/*-----------------------------------------------------------------------
 * File: XPANDSEQ.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "ber_der.h"
#include "ber_internal.h"

/* 
 * expand one level in parsing a BER sequence; return count of items 
 * sucessfully parsed 
 */
sint32 
BER_ExpandSequence(                 /* return count of items sucessfully parsed */
    const uint8 *Octets,                      /* input BER to parse */
    uint32 InputLength,                 /* length of BER to parse (may not all be in memory) */
    sint32 Count,                       /* input number of components */
    const DER_EXPLICIT_CHOICES_PTR *Choices,/* explicit tag choices for each field, if any */
    const uint8 **ContextTags,                /* implicit context-specific tags for each component */
    const uint8 **OriginalTags,               /* original tags for each component */
    const BER_PARSED_ITEM *Defaults,   /* default values */
          BER_PARSED_ITEM *Output)     /* output: array of parsed items */
{
    uint32 remainingInput;
    const uint8  *currentPosition;
    sint16 i, j, k;
    uint32 extTagLength;

    remainingInput = InputLength;
    currentPosition = Octets;
    for (i = 0; i < Count; i++) {
      if (remainingInput > 0) {
        currentPosition = BER_ExpandItem(currentPosition, 
            remainingInput, &Output[i]);
        if (!currentPosition) return i;
        remainingInput = InputLength - (currentPosition - Octets);
        /* search for tag that matches input */
        if (!ContextTags && (!Choices || !Choices[i])) {
            /* if we expect a certain structure, make sure it matches */
            if (OriginalTags && OriginalTags[i] && BER_TagCompare(Output[i].Tag, OriginalTags[i]) != 0)
                return i;   /* doesn't match */
            continue;  /* it matches, or we are accepting anything */
        }
        for (j = i; j < Count; j++) {
            k = -1;
            /* check context-specific tag */
            if (ContextTags && ContextTags[j] &&
                BER_TagCompare(Output[i].Tag, ContextTags[j]) == 0) {
                /* set the tag to the underlying universal tag */
                if (OriginalTags && OriginalTags[j] && 
                    (Output[i].Tag[0] & BER_CLASS_MASK) == BER_CLASS_CONTEXTSPECIFIC)
                    Output[i].Tag = OriginalTags[j];
                break;
            }
            /* check tag choices */
            if (Choices && Choices[j]) {
                if (Choices[j]->ParentField == j /* implicit tag */) {
                    for (k = 0; k < Choices[j]->NumberOfChoices; k++) {
                        if (BER_TagCompare(Output[i].Tag, Choices[j]->Oids[k]->Tag) == 0) { 
                            if (Choices[j]->Subtypes && Choices[j]->Subtypes[k]) 
                                Output[i].Tag = Choices[j]->Subtypes[k]->TypeTag;
                            break;
                        }
                    }
                } else {    /* explicit tag choices */
                    extTagLength = Output[Choices[j]->ParentField].ContentLength;
                    for (k = 0; k < Choices[j]->NumberOfChoices; k++) {
                        if (extTagLength == Choices[j]->Oids[k]->ContentLength &&
                            !cssm_memcmp(Output[Choices[j]->ParentField].Content,
                            Choices[j]->Oids[k]->Content, extTagLength))
                            break;
                    }
                }
                if (k < Choices[j]->NumberOfChoices)
                    break;
            }
        }
        if (j < Count) {
            Output[j] = Output[i];
            for (; i < j; i++) {    
                /* fill in default values for missing components */
                if (Defaults && Defaults[i].Tag) Output[i] = Defaults[i];
                else {
                    Output[i].Tag = 0;
                    Output[i].ContentLength = 0;
                    Output[i].Content = 0;
                }
            }
        } else return i;    /* no match */
      } else {   /* fill in default values */
        if (Defaults && Defaults[i].Tag) Output[i] = Defaults[i];
        else {
            Output[i].Tag = 0;
            Output[i].ContentLength = 0;
            Output[i].Content = 0;
        }
      }
    }
    return i;
}

