/*-----------------------------------------------------------------------
 * File: AUTHCOMP.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "maf_include.h"
/* Windows & CSSM includes */
#include "cssm.h"
#include "cssmspi.h"
#include "guids.h"
#include "internal.h"
#include "free.h"


#define CSSM_AC_BASE_FOR_SPKI_ERRORS (CSSM_AC_PRIVATE_ERROR + 20)

#include "sha.h"                /* SHA-1 structures and calls */
#include "md5.h"                /* MD5 structures and calls */
#include "spkiutil.h"

/***************************************************************************/
/* AC_AuthCompute                                                          */
/*                                                                         */
/* Do the AuthCompute function.                                            */
/*                                                                         */
/* Input is:                                                               */
/* acl: the current ACL (acl ...)                                          */
/* cred: certificate bodies and public keys  (sequence ...)                */
/* lim: test limitations                                                   */
/*    subject: a single subject, presumably the one asking for service     */
/*    tag: the thing that subject is asking to do                          */
/*    now: the current date and time                                       */
/* Any or all of the limitations can be missing, in case we want general   */
/*    results rather than a specific answer to one request.                */
/*                                                                         */
/* The result of this process is a list of S-expressions, possibly         */
/* empty.  These will look like an ACL.  Tuples make it into the result    */
/* list only if they  are ACL-type (have no issuer), are active and have   */
/* the desired subj (if any) as their subject.                             */
/*                                                                         */
/***************************************************************************/


CSSM_RETURN CSSMACI AC_AuthCompute(
              CSSM_AC_HANDLE ach,
              const CSSM_TUPLEGROUP *acl, /* BaseAuthorizations */
              const CSSM_TUPLEGROUP *cred, /* Credentials */
              uint32 ns,        /* NumberOfRequestors */
              const CSSM_LIST subj[], /* Requestors */
              const CSSM_LIST *now_in, /* RequestedAuthorizationPeriod */
              const CSSM_LIST *tag_in, /* RequestedAuthorization */
              CSSM_TUPLEGROUP_PTR aout /* AuthorizationResult */
            )
{
  CSSM_RETURN r  ;              /* error return value */
  pSexp tag = NULL ;            /* the tag constraint */
  pSexp now = NULL ;            /* the date constraint */
  pSexp sbj = NULL ;            /* if there's a single subject, this is it */
  pSexp subjsexp = NULL ;       /* the full subject S-expression */

  if (acl == NULL)
    return ( CSSMERR_AC_INVALID_INPUT_POINTER ) ;
  if (cred == NULL)
    return ( CSSMERR_AC_INVALID_INPUT_POINTER ) ;
  if (ns > 0)
    if (subj == NULL)
      return ( CSSMERR_AC_INVALID_INPUT_POINTER ) ;
  /* now_in and tag_in are allowed to be NULL */
  if (aout == NULL)
    return ( CSSMERR_AC_INVALID_OUTPUT_POINTER ) ;
  aout->NumberOfTuples = 0 ;    /* no output tuples yet */
  hashtab_init() ;              /* init the hash tables */

  /* 0. Pull  keys from the inputs */

  lpass0( ns, subj ) ;          /* from the list of subjects */
  tgpass0( acl ) ;              /* and from the ACL */
  tgpass0( cred ) ;             /* and from the certificates */

  if (ns > 1)
    lactivate( ns, subj ) ;     /* active the subjects */
  if (ns == 1)
    subjsexp = cl_to_sexp( &(subj[0]) ) ;
  if (subjsexp != NULL) {
	sbj = subjsexp->term ;
	if (sbj != NULL) {
	  if (sbj->typ == S_SEXP)
		store_sexp( &sbj, sbj->u.s ) ;
	  else sbj = NULL ;
	} /* if */
  } /* if */

  tag = cl_to_sexp( tag_in ) ;  /* convert the incoming Sexps */
  now = cl_to_sexp( now_in ) ;  /* ... */
  /* 1. Walk the ACL, entering things in the appropriate hash list. */
  r = tg_reduce1( acl, (ns==0), sbj, tag, now ) ; /* walk the ACL structure */
                                /* and enter */
                                /* items from it into the appropriate hash */
                                /* tables, for reduction. */

  /* 2. Walk the cred, entering those. */
  if (r == 0)                   /* if no errors so far */
    r = tg_reduce1( cred, (ns==0), sbj, tag, now ) ; /* walk cred */
                                /* and enter those into the hash tables. */

  /* 3. Gather results into aout. */
  if (r == 0)                   /* if no errors in the process */
    r = tg_report_results( ach, aout ) ; /* get the results into aout */

  hashtab_clear() ;             /* empty the hash tables */
  drop_sexp( tag ) ;            /* clean up temp Sexps */
  drop_sexp( now ) ;            /* ... */
  dump_free_cache( ) ;          /* return pooled free storage */
  if (r < 0)                    /* SPKI errors are negative; CDSA positive */
    r = CSSM_AC_BASE_FOR_SPKI_ERRORS - r ;
  return ( r ) ;                /* return any error code */
} /* AC_AuthCompute */
