/*-----------------------------------------------------------------------
 * File: MODINSTALL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include <sys/types.h>
#include <sys/stat.h>

#include "cssm.h"
#include "cssmerr.h"
#include "cssmport.h"
#include "cssmlock.h"
#include "maf_config.h"
#include "maf_collectn.h"
#include "maf_util.h"
#include "mds.h"
#include "mds_schema.h"
#include "mds_util.h"
#include "intel/installdefs.h"

/* Number of MDS update operations that will be remembered by mds_util for
 * error clean-up during installation.
 */
#define AC_MDS_COMMIT_SIZE  150

static CSSM_RETURN __CleanMDS( MDSU_CONTEXT *pContext )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_CDSA_COMMON_SCHEMA CommonTemplate;
    MDSU_CDSA_COMMON_SCHEMA CommonData;
    MDSU_CDSA_AC_PRIMARY_SCHEMA ACPrimaryTemplate;
    MDSU_CDSA_AC_PRIMARY_SCHEMA ACPrimaryData;
    
    /* Execute a search on every AC record type to get a handle of all
     * records and then delete them.
     */
    /* Common Schema */
    CommonTemplate.ModuleID = MODULE_GUID;
    rv = MDSU_FindFirst( pContext,
                         &IfiMdsuCdsadirCommonSchema,
                         &CommonTemplate,
                         MDSU_COM_MODULE_ID,
                         &CommonData,
                         NULL );
    while ( rv == CSSM_OK )
    {
        if ( CommonData.Manifest.Data )
        {
            MDSU_free( pContext, CommonData.Manifest.Data );
        }

        rv = MDSU_FindNext( pContext,
                            &CommonData,
                            NULL );
    }
    if ( rv == CSSMERR_DL_ENDOFDATA )
    {
        /* AC Primary Schema */
        ACPrimaryTemplate.ModuleID = MODULE_GUID;
        rv = MDSU_FindFirst( pContext,
                             &IfiMdsuCdsadirACPrimarySchema,
                             &ACPrimaryTemplate,
                             MDSU_ACP_MODULE_ID,
                             &ACPrimaryData,
                             NULL );
        while ( rv == CSSM_OK )
        {
            if ( ACPrimaryData.Manifest.Data )
            {
                MDSU_free( pContext, ACPrimaryData.Manifest.Data );
            }

            rv = MDSU_FindNext( pContext,
                                &ACPrimaryData,
                                NULL );
        }
        if ( rv == CSSMERR_DL_ENDOFDATA )
        {
            /* Flush everything from the MDS */
            rv = MDSU_Revert( pContext );
        } /* AC Primary removed */
    } /* Common removed */

    /* Make sure nothing gets removed if there is an error */
    if ( ( rv != CSSMERR_DL_ENDOFDATA ) &&
         ( rv != CSSM_OK ) ) 
    {
        MDSU_Commit( pContext );
    }

    return rv;
}

CSSM_RETURN CSSMAPI INSTALL_ENTRY_NAME(
                        const char *szModuleName,
                        const char *szSearchPath,
                        CSSM_DATA *pCredential,
                        INSTALL_ACTION Action,
                        INSTALL_ERROR *pError )
{
    CSSM_RETURN rv = CSSM_OK;
    MDSU_CONTEXT MDSUContext;
    MDSU_CDSA_COMMON_SCHEMA Common;
    MDSU_CDSA_AC_PRIMARY_SCHEMA ACPrimary;


    /* Initialize the MDS update */
    if ( ( rv = MDSU_Init( &MDSUContext, 
                           &Addin_APIMemFuncs,
                           &MODULE_GUID,
                           MDS_CDSA_DIRECTORY,
                           ( CSSM_DB_ACCESS_READ | 
                             CSSM_DB_ACCESS_WRITE ),
                           AC_MDS_COMMIT_SIZE ) ) != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (Init): %X\n", rv );
        pError->ErrorCode = rv;
        return rv;
    }

    /* Place code to remove existing entries here */
    if ( ( rv = __CleanMDS( &MDSUContext ) ) != CSSM_OK )
    {
        sprintf( pError->ErrorString, "MDS Error (Clean): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Term( &MDSUContext );
        return rv;
    }
    if ( Action == INSTALL_ACTION_UNINSTALL )
    {
        MDSU_Term( &MDSUContext );
        return rv;
    }

    /* Fill in the "easy" common schema structure */
    memcpy( &Common.ModuleID, &MODULE_GUID, sizeof(MODULE_GUID) );
    Common.Manifest = *pCredential;
    strcpy( Common.ModuleName, szModuleName );
    strcpy( Common.Path, szSearchPath );
    Common.CDSAVersion.Major = CSSM_MAJOR;
    Common.CDSAVersion.Minor = CSSM_MINOR;
    strcpy( Common.Desc, ADDIN_NAME );
    Common.DynamicFlag = CSSM_FALSE;
    Common.MultiThreadFlag = CSSM_FALSE;
    Common.ServiceMask = CSSM_SERVICE_AC;

    /* Insert the common schema record into the MDS */
    rv = MDSU_UpdateSchema( &MDSUContext,
                            &IfiMdsuCdsadirCommonSchema,
                            &Common,
                            NULL );
    if ( rv != CSSM_OK ) 
    {
        sprintf( pError->ErrorString, "MDS Error (common-AC): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Term( &MDSUContext );
        return rv;
    }

    /* Fill in the AC primary schema data */
    memcpy( &ACPrimary.ModuleID, &MODULE_GUID, sizeof(MODULE_GUID) );
    ACPrimary.SSID = ADDIN_AC_SUBSERVICE_ID;
    ACPrimary.Manifest.Data = NULL;
    ACPrimary.Manifest.Length = 0;
    strcpy( ACPrimary.ModuleName, szModuleName );
    ACPrimary.ProductVersion.Major = INTEL_ACM_MAJOR_VER;
    ACPrimary.ProductVersion.Minor = INTEL_ACM_MINOR_VER;
    strcpy( ACPrimary.Vendor, ADDIN_VENDOR_NAME );
    
    /* Write the data to MDS */
    rv = MDSU_UpdateSchema( &MDSUContext,
                            &IfiMdsuCdsadirACPrimarySchema,
                            &ACPrimary,
                            NULL );
    if ( rv != CSSM_OK ) 
    {
        sprintf( pError->ErrorString, "MDS Error (AC Primary): %X\n", rv );
        pError->ErrorCode = rv;
        MDSU_Revert( &MDSUContext );
        MDSU_Term( &MDSUContext );
        return rv;
    }


    /* Commit the records */
    MDSU_Commit( &MDSUContext );

    /* Shut down the MDS connection */
    MDSU_Term( &MDSUContext );

    return CSSM_OK;
}
