/*-----------------------------------------------------------------------
 * File: CLDERUTL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * This file contains the routines which directly wrap the BER_DER library
 * and allow the translation between C data structures and DER data structures
 */ 


#include "clutil.h"
#include "clderll.h"
#include "internal.h"

/*-----------------------------------------------------------------------------
 * Name: wrapped_malloc_func, wrapped_calloc_func, wrapped_free_func
 *
 * Description:
 * Several functions in this file are sometimes using memory from the app's
 * memory space and sometimes using local memory.  These routines allocate/free
 * from the appropriate space as identified by the input boolean
 *
 * Parameters: 
 * ToAppMem (input) : A boolean indicator.  If true, use the app's memory space,
 *                    otherwise, use local memory.
 *
 *---------------------------------------------------------------------------*/
#ifndef CSSM_MEMTRACK_ON

#define CSSM_MALLOC_WRAPPER(s, t, u) (s)((t), (u))
#define CSSM_CALLOC_WRAPPER(s, t, u, v) (s)((t) ,(u), (v))
#define CSSM_REALLOC_WRAPPER(s, t, u, v) (s)((t), (u), (v))
#define CSSM_FREE_WRAPPER(s, t, u) (s)((t), (u))

void *wrapped_malloc_func(const CSSM_CLUTIL_CONTEXT *pContext,
                          CSSM_BOOL ToAppMem,
                          CSSM_HANDLE AppHandle,
                          uint32 Size)
{
    if (ToAppMem)
    {
        return CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func,
                            AppHandle,
                            Size);
    }                       
    else
    {
        return CSSM_MALLOC_WRAPPER(internal_malloc, 
                    Size,
                    NULL);
    }
}

void *wrapped_calloc_func(const CSSM_CLUTIL_CONTEXT *pContext,
                          CSSM_BOOL ToAppMem,
                          CSSM_HANDLE AppHandle,
                          uint32 Num,
                          uint32 Size)
{
    if (ToAppMem)
    {
        return CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func,
                            AppHandle,
                            Num,
                            Size);
                            
    }
    else
    {
        return CSSM_CALLOC_WRAPPER(internal_calloc, 
                    Num,
                    Size,
                    NULL
                    );
    }
}

void wrapped_free_func(const CSSM_CLUTIL_CONTEXT *pContext,
                       CSSM_BOOL ToAppMem,
                       CSSM_HANDLE AppHandle,
                       void* MemPtr)
{
    if (MemPtr == NULL)
    {
        return;
    }

    if (ToAppMem)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func,
                                    AppHandle,
                                    MemPtr);
                                    
    }
    else
    {
        CSSM_FREE_WRAPPER(internal_free, MemPtr, NULL);
    }
}

#else /* of CSSM_MEMTRACK_ON*/

#define CSSM_MALLOC_WRAPPER(s, t, u) (s)((t), (u), __FILE__, __LINE__)
#define CSSM_CALLOC_WRAPPER(s, t, u, v) (s)((t) ,(u), (v), __FILE__, __LINE__)
#define CSSM_REALLOC_WRAPPER(s, t, u, v) (s)((t), (u), (v), __FILE__, __LINE__)
#define CSSM_FREE_WRAPPER(s, t, u) (s)((t), (u), __FILE__, __LINE__ )

void *wrapped_malloc_func(const CSSM_CLUTIL_CONTEXT *pContext,
                          CSSM_BOOL ToAppMem,
                          CSSM_HANDLE AppHandle,
                          uint32 Size,
                          const char * szFilename,
                          uint32 u32LineNumber)
{
    if (ToAppMem)
    {
        return CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func,
                            AppHandle,
                            Size);
    }                       
    else
    {
        return CSSM_MALLOC_WRAPPER(internal_malloc, 
                    Size,
                    NULL
                    );
    }
}

void *wrapped_calloc_func(const CSSM_CLUTIL_CONTEXT *pContext,
                          CSSM_BOOL ToAppMem,
                          CSSM_HANDLE AppHandle,
                          uint32 Num,
                          uint32 Size,
                          const char * szFilename,
                          uint32 u32LineNumber)
{
    if (ToAppMem)
    {
        return CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func,
                            AppHandle,
                            Num,
                            Size);
    }                       
    else
    {
        return CSSM_CALLOC_WRAPPER(internal_calloc, 
                    Num,
                    Size,
                    NULL
                    );
    }
}

void wrapped_free_func(const CSSM_CLUTIL_CONTEXT *pContext,
                       CSSM_BOOL ToAppMem,
                       CSSM_HANDLE AppHandle,
                       void* MemPtr,
                       const char * szFilename,
                       uint32 u32LineNumber)
{
    if (MemPtr == NULL)
    {
        return;
    }

    if (ToAppMem)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func,
                                    AppHandle,
                                    MemPtr);
    }
    else
    {
        CSSM_FREE_WRAPPER(internal_free, MemPtr, NULL);
    }
}
#endif

#undef CSSM_MALLOC_WRAPPER
#undef  CSSM_CALLOC_WRAPPER 
#undef  CSSM_REALLOC_WRAPPER 
#undef  CSSM_FREE_WRAPPER 

#include <cssmMemTrack.h>
/*-----------------------------------------------------------------------------
 * Name: clutil_GetHandle
 *
 * Description:
 * This function returns the next handle which can be used to identify a 
 * node in the linked list
 *
 * Parameters: 
 * None
 *---------------------------------------------------------------------------*/
uint32 clutil_GetHandle(CSSM_CLUTIL_CONTEXT *pContext)
{
    return (pContext->CurrentHandle++);
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_IsDerLeafTag
 *
 * Description:
 * This functions checks whether the input tag is one of the supported universal tags
 *
 * Parameters: 
 * Tag (input) : The tag to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether the input tag is one of the supported universal tags
 *---------------------------------------------------------------------------*/
CSSM_BOOL cl_der_IsDerLeafTag(CL_DER_TAG_TYPE Tag)
{
    if ((Tag >= DER_TAG_BOOLEAN && Tag <= DER_TAG_OID) ||
        Tag == DER_TAG_SEQUENCE || Tag == DER_TAG_SET ||
        Tag == DER_TAG_PRINTABLE_STRING || Tag == DER_TAG_T61_STRING ||
        Tag == DER_TAG_IA5_STRING || Tag == DER_TAG_UTC_TIME ||
        Tag == DER_TAG_GENERALIZED_TIME)
    {
        return CSSM_TRUE;
    }
    
    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_IsDerParentTag
 *
 * Description:
 * This functions checks whether the input tag is one of the supported constructed tags
 *
 * Parameters: 
 * Tag (input) : The tag to be checked
 *
 * Return value:
 * A TRUE/FALSE indicator of whether the input tag is one of the supported constructed tags
 * 
 *---------------------------------------------------------------------------*/
CSSM_BOOL cl_der_IsDerParentTag(CL_DER_TAG_TYPE Tag)
{
    if (Tag == DER_TAG_SEQUENCE || Tag == DER_TAG_SET )
    {
        return CSSM_TRUE;
    }

    return CSSM_FALSE;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DerBool2CSSMBool
 *
 * Description:
 * A DER boolean is 1 byte whose value is either 0x00 or 0xFF (typ)
 * A CSSM_BOOL is 4 bytes whose value is either 0x00000000 or 0x00000001
 * This routine translates a DER boolean into a CSSM_BOOL
 *
 * Parameters: 
 * DerBool (input) : The boolean byte to be converted
 *
 * return_data_ptr (output): A pointer to a buffer containing
 *                  the input's CSSM_BOOL equivalent
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_DerBool2CSSMBool(const CSSM_CLUTIL_CONTEXT *pContext,
                                   CSSM_HANDLE AppHandle, uint8 * DerBool,
                                   CSSM_BOOL* *return_data_ptr)
{
    CSSM_RETURN ret = CSSM_OK;
    CSSM_BOOL  bRtn = CSSM_FALSE;
    uint8      der_bool_cmp[1];
    if (!return_data_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *return_data_ptr = NULL;
    *return_data_ptr = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle,sizeof(CSSM_BOOL));
    if (!*return_data_ptr) 
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    port_memset(der_bool_cmp, 0, 1);
    if ( port_memcmp(DerBool, der_bool_cmp, 1) == 0)
    {
        bRtn = CSSM_FALSE;
    }
    else
    {
        bRtn = CSSM_TRUE;
    }

    port_memcpy(*return_data_ptr, &bRtn, sizeof(CSSM_BOOL));
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_CSSMBool2DerBool
 *
 * Description:
 * A CSSM_BOOL is 4 bytes whose value is either 0x00000000 or 0x00000001
 * A DER boolean is 1 byte whose value is either 0x00 or 0xFF (typ)
 * This routine translates a CSSM_BOOL into a DER boolean
 *
 * Parameters: 
 * CSSMBool (input) : The boolean value to be converted
 *
 * return_data_ptr (output) : A pointer to a buffer containing the 
 *                            input's DER boolean equivalent
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_CSSMBool2DerBool(const CSSM_CLUTIL_CONTEXT *pContext,
                                    CSSM_BOOL* CSSMBool,
                                    uint8* *return_data_ptr)
{
    CSSM_RETURN ret = CSSM_OK;
    
    if (!return_data_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }    
    *return_data_ptr = CSSM_MALLOC_WRAPPER(internal_malloc, 1, NULL);
    if (!*return_data_ptr) 
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    if (*CSSMBool)
    {
        port_memset(*return_data_ptr, 0xFF, 1);
    }
    else
    {
        port_memset(*return_data_ptr, 0, 1);
    }

    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_Uint8TagToCLDERTag
 *
 * Description:
 * A DER tag is an arbitrary length buffer, which is typically 1 byte in length 
 *   and ty0ically contains a universal tag
 * A CL DER Tag is a 4 byte representation of the universal type
 * This routine translates a Der tag into a CL Der tag
 *
 * Parameters: 
 * TagIn (input) : The tag to be converted
 *
 * Return value:
 * The input's CL DER equivalent
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
CL_DER_TAG_TYPE cl_der_Uint8TagToCLDERTag(uint8 *TagIn)
{
    CL_DER_TAG_TYPE tag_out = 0;

    if (!TagIn)
    {
        return DER_TAG_UNKNOWN;
    }

    if (BER_LengthOfTag(TagIn) != 1) /* This is not a simple universal tag */
    {
        return DER_TAG_UNKNOWN;
    }

    if ((*TagIn & BER_CLASS_APPLICATION) != 0 ||
        (*TagIn & BER_CLASS_CONTEXTSPECIFIC) != 0 ||
        (*TagIn & BER_CLASS_PRIVATE) != 0 ) /* Application, Ctx Specific, or Private tag */
    {
        return DER_TAG_UNKNOWN;
    }

    tag_out = (CL_DER_TAG_TYPE) *TagIn;
    tag_out = (uint8) (tag_out & ~((uint8) BER_CONSTRUCTED));  /* Remove constructed indicator */

    return tag_out;
}




/*-----------------------------------------------------------------------------
 * Name: cl_der_CLDerTagToDerTag
 *
 * Description:
 * A DER tag is an arbitrary length buffer, which is typically 1 byte in length 
 *   and ty0ically contains a universal tag
 * A CL DER Tag is a 4 byte representation of the universal type
 * This routine translates a CL Der tag into a Der tag
 *
 * Parameters: 
 * TagIn (input) : The tag to be converted
 *
 * tag_out (output) : The input's DER equivalent
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_CLDerTagToDerTag(const CSSM_CLUTIL_CONTEXT *pContext,
                                CL_DER_TAG_TYPE TagIn,
                                uint8* *tag_out)
{
    CSSM_RETURN ret = CSSM_OK;
    if (!tag_out)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *tag_out = NULL;

    if (!cl_der_IsDerLeafTag( (CL_DER_TAG_TYPE) (TagIn & ((uint8)0xDF)) ) )
    {
        return CSSMERR_CL_INVALID_DATA;
      
    }
    *tag_out = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(uint8), NULL);
    if (!*tag_out)
    {
        return CSSMERR_CL_MEMORY_ERROR;
       
    }

    **tag_out = (uint8) (TagIn & 0xFF);
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_InternalDeleteType
 *
 * Description:
 * This functions frees all of the memory associated with a DER_COMPLEX_TYPE
 * structure, including all of its subtypes.
 *
 * Parameters: 
 * Type (input) :   The structure to be freed
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_InternalDeleteType(const CSSM_CLUTIL_CONTEXT *pContext,
                                      CSSM_BOOL ToAppMem,
                                      CSSM_HANDLE AppHandle, 
                                      DER_COMPLEX_TYPE_PTR Type)
{
    sint16 j=0, i=0;
    CSSM_RETURN ret = CSSM_OK;
    /* Check input */
    if (!Type || IsBadReadPtr(Type, sizeof(DER_COMPLEX_TYPE)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }

    /* Free Type data structure */
    wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->TypeTag);
    Type->TypeTag = NULL;
    
    if (Type->ContextTags)
    {
        for (j=0; j<Type->NumberOfDefaults; j++)
        {
            if ( Type->ContextTags[j] )
            {
                wrapped_free_func(pContext, ToAppMem, AppHandle, (void*)Type->ContextTags[j]);
                Type->ContextTags[j] = NULL;
            }
        }
        wrapped_free_func(pContext, ToAppMem, AppHandle, (void*)Type->ContextTags);
        Type->ContextTags = NULL;
    }

    if (Type->OriginalTags)
    {
        for (j=0; j<Type->NumberOfDefaults; j++) 
        {
            if ( Type->OriginalTags[j] )    
            {
                wrapped_free_func(pContext, ToAppMem, AppHandle, (void*)Type->OriginalTags[j]);
                Type->OriginalTags[j] = NULL;
            }
        }
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->OriginalTags);
        Type->OriginalTags = NULL;
    }

    if (Type->Defaults)
    {
        for (j=0; j<Type->NumberOfDefaults; j++)
        {
            wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->Defaults[j].Tag);
            Type->Defaults[j].Tag = NULL;

            wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->Defaults[j].Content);
            Type->Defaults[j].Content = NULL;
        }
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->Defaults);
        Type->Defaults = NULL;
    }

    if (Type->Choices)
    {
        for (j=0; j<Type->NumberOfDefaults; j++)
        {
            if (Type->Choices[j])
            {
                if (Type->Choices[j]->Oids)
                {
                    for (i=0; i<Type->Choices[j]->NumberOfChoices; i++)
                    {
                        if (Type->Choices[j]->Oids[i])
                        {
                            wrapped_free_func(pContext,ToAppMem, AppHandle, 
                                (void*)Type->Choices[j]->Oids[i]->Tag);
                            Type->Choices[j]->Oids[i]->Tag = NULL;
                            wrapped_free_func(pContext,ToAppMem, AppHandle, 
                                (void*)Type->Choices[j]->Oids[i]->Content);
                            Type->Choices[j]->Oids[i]->Content = NULL;
                            wrapped_free_func(pContext,ToAppMem, AppHandle, 
                                (void*)Type->Choices[j]->Oids[i]);
                            Type->Choices[j]->Oids[i] = NULL;
                        }
                    }
                    wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->Choices[j]->Oids);
                    Type->Choices[j]->Oids = NULL;
                }
                if (Type->Choices[j]->Subtypes)
                {
                    for (i=0; i<Type->Choices[j]->NumberOfChoices; i++)
                    {
                        if (Type->Choices[j]->Subtypes[i])
                        {
                            cl_der_InternalDeleteType(pContext,ToAppMem, AppHandle, Type->Choices[j]->Subtypes[i]);
                            Type->Choices[j]->Subtypes[i] = NULL;
                        }
                    }
                    wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->Choices[j]->Subtypes);
                    Type->Choices[j]->Subtypes = NULL;
                }
                wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->Choices[j]);
                Type->Choices[j] = NULL;
            }
        }
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->Choices);
        Type->Choices = NULL;
    }

    if (Type->Subtypes)
    {
        for (j=0; j<Type->NumberOfDefaults; j++)
        {
            if (Type->Subtypes[j]) 
            {
                ret = cl_der_InternalDeleteType(pContext,ToAppMem, AppHandle, Type->Subtypes[j]);
                Type->Subtypes[j] = NULL;
            }
        }
        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type->Subtypes);
        Type->Subtypes = NULL;
    }

    wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)Type);
    Type = NULL;

    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DeleteType
 *
 * Description:
 * This functions frees all of the memory associated with a DER_COMPLEX_TYPE
 * structure, including all of its subtypes.  This assumes that the Type was 
 * allocated in the app's memory space
 *
 * Parameters: 
 * Type (input) :   The structure to be freed
 *
 *---------------------------------------------------------------------------*/
void cl_der_DeleteType(const CSSM_CLUTIL_CONTEXT *pContext,
                       CSSM_HANDLE AppHandle,
                       DER_COMPLEX_TYPE_PTR Type)
{
    cl_der_InternalDeleteType(pContext,CSSM_TRUE, AppHandle, Type);
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_CopyType
 *
 * Description:
 * This functions copies a DER_COMPLEX_TYPE data structure,
 * including all of its subtypes into the memory space identified by the 
 * ToAppMem flag.
 *
 * Parameters: 
 * ToAppMem (input) : An identifier of what memory space to use for memory allocation
 * TypePtr  (input) : The structure to be copied
 *
 * result_val (output): A copy of the TypePtr data structure in either the application's memory space
 *                      or the local memory space.
 * 
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_CopyType(const CSSM_CLUTIL_CONTEXT *pContext,
                            CSSM_BOOL ToAppMem,
                            CSSM_HANDLE AppHandle,
                            DER_COMPLEX_TYPE_PTR TypePtr,
                            DER_COMPLEX_TYPE_PTR *result_val)
{
    DER_COMPLEX_TYPE_PTR results_ptr = NULL;
    CSSM_RETURN ret = CSSM_OK;
    sint16 i=0, j=0;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *result_val = NULL;
    
    /* Check inputs */
    if (!TypePtr)
    {
        return CSSM_OK; /* NULL ptr passed - just return w/ OK. */
    }
    if (IsBadReadPtr(TypePtr, sizeof(DER_COMPLEX_TYPE)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }

    /* Allocate new type structure, initializing all fields to NULL */
    results_ptr = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 1, sizeof(DER_COMPLEX_TYPE));
    if (!results_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Copy data structures internal to the Type structure */
    /* Copy the tag structure and the NumberOfDefaults field */
    results_ptr->TypeTag = wrapped_malloc_func(pContext,ToAppMem, AppHandle, BER_LengthOfTag(TypePtr->TypeTag));
    if (!results_ptr->TypeTag )
    {
        wrapped_free_func(pContext,ToAppMem, AppHandle, results_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    port_memcpy((void*)results_ptr->TypeTag, TypePtr->TypeTag, BER_LengthOfTag(TypePtr->TypeTag));
    results_ptr->NumberOfDefaults = TypePtr->NumberOfDefaults;
    /* The choices structure remains NULL because variants are not supported */

    /* Copy the Context-specific Tags */
    /* If the input type structure has no context-specific tags, */
    /* this field remains NULL */
    if (TypePtr->ContextTags)
    {
        results_ptr->ContextTags = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 
            TypePtr->NumberOfDefaults, sizeof(CL_DER_TAG_TYPE *));
        if (!results_ptr->ContextTags)
        {
            wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)results_ptr->TypeTag);
            wrapped_free_func(pContext,ToAppMem, AppHandle, results_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
        for (i=0; i<TypePtr->NumberOfDefaults; i++)
        {
            if (TypePtr->ContextTags[i])
            {
                results_ptr->ContextTags[i] = wrapped_malloc_func(pContext,ToAppMem, AppHandle, BER_LengthOfTag(TypePtr->ContextTags[i]));
                if (!results_ptr->ContextTags[i])
                {
                    for (j=0; j<i; j++)
                        wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)results_ptr->ContextTags[j]);
                    wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)results_ptr->ContextTags);
                    wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)results_ptr->TypeTag);
                    wrapped_free_func(pContext,ToAppMem, AppHandle, (void*)results_ptr);
                    return CSSMERR_CL_MEMORY_ERROR;
                    
                }

                port_memcpy((void*)results_ptr->ContextTags[i], TypePtr->ContextTags[i], BER_LengthOfTag(TypePtr->ContextTags[i]));
            }
        }
    }

    /* Copy Original Tags */
    /* Allocate memory for the Original Tags */
    results_ptr->OriginalTags = wrapped_calloc_func(pContext,ToAppMem, AppHandle, TypePtr->NumberOfDefaults, sizeof(CL_DER_TAG_TYPE *));
    if (!results_ptr->OriginalTags)
    {
        cl_der_InternalDeleteType(pContext,ToAppMem, AppHandle, results_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    for (i=0; i<TypePtr->NumberOfDefaults; i++)
    {
        if (TypePtr->OriginalTags[i])
        {
            results_ptr->OriginalTags[i] = wrapped_malloc_func(pContext,ToAppMem, AppHandle, 
                BER_LengthOfTag(TypePtr->OriginalTags[i]));
            if (!results_ptr->OriginalTags[i])
            {
                cl_der_InternalDeleteType(pContext,ToAppMem, AppHandle, results_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
                
            }

            port_memcpy((void*)results_ptr->OriginalTags[i], TypePtr->OriginalTags[i], 
                BER_LengthOfTag(TypePtr->OriginalTags[i]));
        }
    }

    /* Copy Defaults */
    /* If the input type structure has no defaults, this field remains NULL */
    if (TypePtr->Defaults)
    {
        results_ptr->Defaults = wrapped_calloc_func(pContext,ToAppMem, AppHandle, 
            TypePtr->NumberOfDefaults, sizeof(BER_PARSED_ITEM));
        if (!results_ptr->Defaults)
        {
            cl_der_InternalDeleteType(pContext,ToAppMem, AppHandle, results_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
        }

        for (i=0; i<TypePtr->NumberOfDefaults; i++)
        {
            if (TypePtr->Defaults[i].Tag)
            {
                results_ptr->Defaults[i].Tag = wrapped_malloc_func(pContext,ToAppMem, AppHandle, 
                    BER_LengthOfTag(TypePtr->Defaults[i].Tag));
                results_ptr->Defaults[i].Content = wrapped_malloc_func(pContext,ToAppMem, AppHandle, 
                    TypePtr->Defaults[i].ContentLength);
                if (!results_ptr->Defaults[i].Tag || !results_ptr->Defaults[i].Content)
                {
                    cl_der_InternalDeleteType(pContext,ToAppMem, AppHandle, results_ptr);
                    return CSSMERR_CL_MEMORY_ERROR;
                    
                }

                port_memcpy((void*)results_ptr->Defaults[i].Tag, 
                    TypePtr->Defaults[i].Tag, BER_LengthOfTag(TypePtr->Defaults[i].Tag));
                results_ptr->Defaults[i].ContentLength = TypePtr->Defaults[i].ContentLength;
                port_memcpy((void*)results_ptr->Defaults[i].Content, 
                    TypePtr->Defaults[i].Content, results_ptr->Defaults[i].ContentLength);
            }
        }
    }


    /* Copy the types of any children */
    /* If the input type structure has no subtypes, this field remains NULL */
    if (TypePtr->Subtypes)
    {
        /* Allocate array of Subtype pointers */
        results_ptr->Subtypes = wrapped_calloc_func(pContext,ToAppMem, AppHandle,
                                                    TypePtr->NumberOfDefaults,
                                                    sizeof(DER_COMPLEX_TYPE_PTR));
        if (!results_ptr->Subtypes)
        {
            cl_der_InternalDeleteType(pContext,ToAppMem, AppHandle, results_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
        }

        /* Copy subtypes */
        for (i=0; i<TypePtr->NumberOfDefaults; i++)
        {
            if (TypePtr->Subtypes[i])
            {
                ret = cl_der_CopyType(pContext,ToAppMem, AppHandle, 
                                      TypePtr->Subtypes[i], 
                                      &results_ptr->Subtypes[i]);
            }
        }
    }


    /* Copy the name of the type data structure */
    /* If the input type structure has no name, this field remains NULL */
    if (TypePtr->Name)
    {
        wrapped_malloc_func(pContext,ToAppMem, AppHandle, strlen(TypePtr->Name) +1);
        if (!results_ptr->Name)
        {
            cl_der_InternalDeleteType(pContext,ToAppMem, AppHandle, results_ptr);
            return CSSMERR_CL_MEMORY_ERROR;
            
        }

        port_memcpy(results_ptr->Name, TypePtr->Name, strlen(TypePtr->Name)+1);
    }

    /* Return the copy */
    
    *result_val = results_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: CssmDataToBerItemPtr
 *
 * Description:
 * This functions copies the input data into a BER_PARSED_ITEM format
 *
 * Parameters: 
 * Tag  (input) : The universal tag for the data
 * Data (input) : The structure to be copied
 *
 * ber_item_ptr (output): A BER_PARSED_ITEM containing the equivalent
 *                        of the input tag and data
 *---------------------------------------------------------------------------*/
CSSM_RETURN  CssmDataToBerItemPtr   (const CSSM_CLUTIL_CONTEXT *pContext,
                                     CL_DER_TAG_TYPE Tag, 
                                     const CSSM_DATA *Data,
                                     BER_PARSED_ITEM_PTR *ber_item_ptr)
{
    CSSM_RETURN ret = CSSM_OK;
    
    if (!ber_item_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *ber_item_ptr = NULL;

    /* Allocate a parsed ber item */
    *ber_item_ptr = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(BER_PARSED_ITEM), NULL);
    if (!(*ber_item_ptr))
    {
        return CSSMERR_CL_MEMORY_ERROR;
       
    }

    ret = cl_der_CLDerTagToDerTag(pContext,Tag, (uint8**)(&(*ber_item_ptr)->Tag));
    if (!(*ber_item_ptr)->Tag)
    {
        CSSM_FREE_WRAPPER(internal_free, (*ber_item_ptr), NULL);
        return CSSMERR_CL_MEMORY_ERROR;
       
    }

    (*ber_item_ptr)->ContentLength = Data->Length;
    if ((*ber_item_ptr)->ContentLength)
    {
        (*ber_item_ptr)->Content = CSSM_MALLOC_WRAPPER(internal_malloc, Data->Length, NULL);
        if (!(*ber_item_ptr)->Content)
        {
            CSSM_FREE_WRAPPER(internal_free, (void*)(*ber_item_ptr)->Tag, NULL);
            CSSM_FREE_WRAPPER(internal_free, (*ber_item_ptr), NULL);
            return CSSMERR_CL_MEMORY_ERROR;
           
        }
        port_memcpy((void*)(*ber_item_ptr)->Content, Data->Data, Data->Length);
    }
    else
    {
        (*ber_item_ptr)->Content = NULL;
    }

    return ret;
}



/*-----------------------------------------------------------------------------
 * Name: SeqOrSetToBerItemPtr
 *
 * Description:
 * This function modifies the input tag to make it constructed and 
 * then calls the routine which translates a CSSM_DATA to a
 * BER_PARSED_ITEM
 *
 * Parameters: 
 * Tag  (input) : The universal tag for the data
 * Data (input) : The structure to be copied
 *
 * results_ptr (output): A BER_PARSED_ITEM containing the equivalent 
 *                       of the input tag and data
 *---------------------------------------------------------------------------*/
CSSM_RETURN  SeqOrSetToBerItemPtr   (const CSSM_CLUTIL_CONTEXT *pContext,
                                     CL_DER_TAG_TYPE Tag, 
                                     const CSSM_DATA *Data,
                                     BER_PARSED_ITEM_PTR *results_ptr)
{
    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *results_ptr = NULL;
    
    
    return CssmDataToBerItemPtr(pContext, (CL_DER_TAG_TYPE)(Tag | BER_CONSTRUCTED), Data, results_ptr);
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_PreprocessTagDataPair
 *
 * Description:
 * This function translates the input C data into Der data and
 * returns it in a BER_PARSED_ITEM structure
 *
 * Parameters: 
 * Tag  (input) : The universal tag for the data
 * Data (input) : The structure to be copied
 *
 * ber_item_ptr (output): A BER_PARSED_ITEM containing the DER equivalent
 *                        of the input tag and data
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN  cl_der_PreprocessTagDataPair(const CSSM_CLUTIL_CONTEXT *pContext,
                                          CL_DER_TAG_TYPE Tag, 
                                          const CSSM_DATA *Data,
                                          BER_PARSED_ITEM_PTR *ber_item_ptr)
{
    CSSM_RETURN         ret = CSSM_OK;
    BER_PARSED_ITEM      temp_ber_item;
    CSSM_DATA            temp_data = {0, NULL};
    CSSM_DATA            null_data = {0, NULL};
    *ber_item_ptr = NULL;
    /* Check inputs */
    if (!cl_der_IsDerLeafTag(Tag))
    {
        return CSSMERR_CL_INVALID_DATA;
      
    }
    if ((Data == NULL) || port_IsBadReadPtr(Data, sizeof(CSSM_DATA)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
     
    }
    if ((Tag != DER_TAG_BIT_STRING) && 
        ((Data->Data != NULL) && port_IsBadReadPtr(Data->Data, Data->Length)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }
    if (Tag == DER_TAG_BIT_STRING &&
        ((Data->Data != NULL) && port_IsBadReadPtr(Data->Data, Data->Length/8)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }

    switch ( Tag ) 
    {
    case DER_TAG_SEQUENCE :
    case DER_TAG_SET :
        ret = SeqOrSetToBerItemPtr(pContext,Tag, Data, ber_item_ptr);
        if (!(*ber_item_ptr))
            return ret;
        break;

    case DER_TAG_BOOLEAN :
        temp_data.Length = 1;
        ret = cl_der_CSSMBool2DerBool(pContext,(CSSM_BOOL *) Data->Data, &temp_data.Data );
        if (!temp_data.Data)
            return ret;

        ret = CssmDataToBerItemPtr(pContext,Tag, &temp_data, ber_item_ptr);
        CSSM_FREE_WRAPPER(internal_free, temp_data.Data, NULL);
        if (!(*ber_item_ptr))
            return ret;
        break;

    case DER_TAG_PRINTABLE_STRING :
    case DER_TAG_IA5_STRING :
    case DER_TAG_UTC_TIME :
    case DER_TAG_GENERALIZED_TIME :
        if (Data->Length)
        {
            temp_data = *Data;
            /* Remove trailing NULLs */
            while (temp_data.Length != 0 && !temp_data.Data[temp_data.Length - 1])  
                temp_data.Length--;
            if (temp_data.Length == 0)
                temp_data = null_data;
            ret = CssmDataToBerItemPtr(pContext,Tag, &temp_data, ber_item_ptr);
        }
        else
        {
            temp_data = null_data;
            ret = CssmDataToBerItemPtr(pContext,Tag, &temp_data, ber_item_ptr);
        }
        if (!(*ber_item_ptr))
            return ret;
        break;

    case DER_TAG_BIT_STRING :
        if (Data->Length)
        {
            if (Data->Length % 8) 
                temp_data.Length = Data->Length/8 +2;
            else
                temp_data.Length = Data->Length/8 +1;
            temp_data.Length = 1 + BER_LengthOfLength(temp_data.Length) + temp_data.Length;
            temp_data.Data = CSSM_MALLOC_WRAPPER(internal_malloc, temp_data.Length, NULL);
            if (!temp_data.Data)
            {
                return CSSMERR_CL_MEMORY_ERROR;
               
            }

            DER_EncodeBitString(Data->Data, Data->Length, temp_data.Data, temp_data.Length);
            BER_ExpandItem(temp_data.Data, temp_data.Length, &temp_ber_item);
            ret = CssmDataToBerItemPtr(pContext,Tag, (CSSM_DATA_PTR) &temp_ber_item.ContentLength, ber_item_ptr);
            CSSM_FREE_WRAPPER(internal_free, temp_data.Data, NULL);
        }
        else
        {
            temp_data = null_data;
            ret = CssmDataToBerItemPtr(pContext,Tag, &temp_data, ber_item_ptr);
        }
        if (!(*ber_item_ptr))
        {
            return ret;
        }
        break;

    default:
        ret = CssmDataToBerItemPtr(pContext,Tag, Data, ber_item_ptr);
        if (!(*ber_item_ptr))
        {
            return ret;
        }
    }

    return ret;
}

/*-----------------------------------------------------------------------------
 * Name: cl_der_CreateLeafNode
 *
 * Description:
 * This function copies the input into an internal structure which can
 * be used to create a DER encoded version of the data.
 *
 * Parameters: 
 * Tag (input)  : Specifies the basic data type
 * Data (input) : The data 
 * 
 * results_handle (output): A handle which can be used by the calling app to specify 
 *                          this node as part of a parent node or to DER encode this data.
 *---------------------------------------------------------------------------*/
CSSM_RETURN  cl_der_CreateLeafNode   (CSSM_CLUTIL_CONTEXT *pContext,
                                      CL_DER_TAG_TYPE Tag, 
                                      uint32 DataLength,
                                      uint8 *Data,
                                      CL_DER_NODE_HANDLE *results_handle)
{
    BER_PARSED_ITEM_PTR  ber_item_ptr = NULL;
    CL_DER_NODE_HANDLE   return_handle = 0;
    CSSM_DATA            temp_data;
    CSSM_RETURN          ret = CSSM_OK;


    if (!results_handle)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    temp_data.Length = DataLength;
    temp_data.Data   = Data;
    *results_handle  = 0;
    /* Translate the input data into a BER_PARSED_ITEM */
    ret = cl_der_PreprocessTagDataPair(pContext,Tag, &temp_data, &ber_item_ptr);

    if(ret != CSSM_OK)
    {
        return ret;
    }

    /* Add parsed ber item as a leaf node */
    ret = cl_der_AddLeafNode(pContext,ber_item_ptr, &return_handle);

    if (ret != CSSM_OK)
    {
        if (ber_item_ptr != NULL)
        {
            if (ber_item_ptr->Tag != NULL)
            {
                CSSM_FREE_WRAPPER(internal_free, (void*)ber_item_ptr->Tag, NULL);
            }
            if (ber_item_ptr->Content != NULL)
            {
                CSSM_FREE_WRAPPER(internal_free, (void*)ber_item_ptr->Content, NULL);
            }
            CSSM_FREE_WRAPPER(internal_free, ber_item_ptr, NULL);
        }
    }

    *results_handle = return_handle;
    return ret;
}



/*-----------------------------------------------------------------------------
 * Name: BER_CreateContextSpecificTag
 *
 * Description:
 * This function creates a context specific tag given the Original Tag and
 * the new tag number
 *
 * Parameters: 
 * OriginalTag   (input) : the base tag
 * ContextNumber (input) : The number (last 5 bits) of the new tag
 * IsExplicit (input)    : indicates whether this is an explicit or implicit 
 *                         context-specific tag.  Explicit tags are always constructed
 * 
 * result_val: A buffer containing the context-specific tag
 *---------------------------------------------------------------------------*/
CSSM_RETURN BER_CreateContextSpecificTag(const CSSM_CLUTIL_CONTEXT *pContext,
                                         const uint8 * OriginalTag, 
                                         uint8 ContextNumber, 
                                         CSSM_BOOL IsExplicit,
                                         uint8 ** result_val)
{
    uint8 *context_tag = NULL;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    if (!OriginalTag)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }
    /* Long context specific tags are not currently supported */
    if (ContextNumber > 0x1F)
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }

    /* Allocate 1 byte for the context-specific tag */
    context_tag = CSSM_CALLOC_WRAPPER(internal_calloc, 1, sizeof(uint8), NULL);
    if (!context_tag)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* The first 2 bits indicate that this is context specific */
    *context_tag = BER_CLASS_CONTEXTSPECIFIC;

    /* If the OriginalTag is constructed or this field is explicit, 
       the 3rd bit indicates that this is constructed */
    if ((*OriginalTag & BER_CONSTRUCTED) || IsExplicit)
    {
        *context_tag += BER_CONSTRUCTED;
    }
    /* The last 5 bits indicate the actual context-specific number */
    *context_tag |= ContextNumber;

    
    *result_val = context_tag;
    return CSSM_OK;

}


/*-----------------------------------------------------------------------------
 * Name: cl_der_CreateParentType
 *
 * Description:
 * This function uses information about the children of this node to build 
 * the parse tree description (type) of this node
 *
 * Parameters: 
 * Tag (input)              : Specifies whether this is a SET or a SEQUENCE
 * NumberOfChildren (input) : Specifies the length of the ChildrenArray
 * ChildrenArray (input)    : Specifies the handles of the children of this
 *                            node.  It also contains information about the 
 *                            SET or SEQ which is used to build the type
 *                            structure (optional parms, defaults, subtypes)
 * NodePtr (input)          : The node that this type will be associated with.
 *                            It is provided as a performance enhancement, 
 *                            because it si useful to have the Original Tags 
 *                            of the children
 * 
 * result_val (output)      : The type for this parent node 
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN  cl_der_CreateParentType (const CSSM_CLUTIL_CONTEXT *pContext,
                                      CL_DER_TAG_TYPE Tag, 
                                      uint32 NumberOfChildren,  
                                      CL_DER_CHILD_INFO_PTR ChildrenArray,
                                      DER_NODE_PTR NodePtr,
                                      DER_COMPLEX_TYPE_PTR *result_val)
{
    DER_COMPLEX_TYPE_PTR    type_ptr = NULL;
    CL_DER_TAG_TYPE         complete_tag = 0;
    uint8                   ctx_tag_number = 0;
    BER_PARSED_ITEM         ber_item;
    CSSM_BOOL               has_subtype = CSSM_FALSE;
    uint32                  i=0, j=0;
    uint32                  length = 0;
    CSSM_RETURN             ret = CSSM_OK;
    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *result_val = NULL;
    /* Check inputs */
    if (!cl_der_IsDerParentTag(Tag))
    {
        return CSSMERR_CL_INVALID_DATA;
    }
    if (port_IsBadReadPtr(ChildrenArray, NumberOfChildren * sizeof(CL_DER_CHILD_INFO)) ||
        !ChildrenArray || NumberOfChildren > 0x7FFF ||
        port_IsBadReadPtr(NodePtr, sizeof(DER_NODE)) )
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }

    /* Allocate a DER node type structure */
    type_ptr = CSSM_CALLOC_WRAPPER(internal_calloc, 1, sizeof(DER_COMPLEX_TYPE), NULL);
    if (!type_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    /* Allocate and fill in the TypeTag = Tag */
    type_ptr->TypeTag = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(CL_DER_TAG_TYPE), NULL);
    if (!type_ptr->TypeTag)
    {
        CSSM_FREE_WRAPPER(internal_free, type_ptr, NULL);
        return CSSMERR_CL_MEMORY_ERROR;
    }
    complete_tag = (CL_DER_TAG_TYPE) (Tag | BER_CONSTRUCTED);
    port_memcpy((void*)type_ptr->TypeTag, &complete_tag, sizeof(CL_DER_TAG_TYPE));
    type_ptr->NumberOfDefaults = (sint16) NumberOfChildren;
    /* The choices structure remains NULL because variants are not supported */

    /* Fill in the OriginalTags */
    /* Allocate an array for the OriginalTags */
    type_ptr->OriginalTags = CSSM_CALLOC_WRAPPER(internal_calloc, NumberOfChildren, sizeof(CL_DER_TAG_TYPE *), NULL);
    if (!type_ptr->OriginalTags)
    {
        CSSM_FREE_WRAPPER(internal_free, (void*)type_ptr->TypeTag, NULL);
        CSSM_FREE_WRAPPER(internal_free, type_ptr, NULL);
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    /* Fill in the individual OriginalTag array elements */
    for (i=0; i<NumberOfChildren; i++)
    {
        if (NodePtr->Child[i].IsLeaf)
        {
            length = BER_LengthOfTag(NodePtr->Child[i].X.Input.Tag);
            if (length != 0)
            {
                /* Allocate space for a new tag */
                type_ptr->OriginalTags[i] = CSSM_MALLOC_WRAPPER(internal_malloc, length, NULL);
                if (!type_ptr->OriginalTags[i])
                {
                    CSSM_FREE_WRAPPER(internal_free, (void*)type_ptr->TypeTag, NULL);
                    for (j=0; j<i; j++)
                    {
                        CSSM_FREE_WRAPPER(internal_free, (void*)type_ptr->OriginalTags[j], NULL);
                    }
                    CSSM_FREE_WRAPPER(internal_free, (void*)type_ptr->OriginalTags, NULL);
                    return CSSMERR_CL_MEMORY_ERROR;
                    
                }

                /* Copy the tag into this array */
                port_memcpy((void*)type_ptr->OriginalTags[i], NodePtr->Child[i].X.Input.Tag, length);
            }
        } 
        else
        {
            if (NodePtr->Child[i].X.Node != NULL)
            {
                length = BER_LengthOfTag(NodePtr->Child[i].X.Node->Tag);
                if (length != 0)
                {
                    /* Allocate space for a new tag */
                    type_ptr->OriginalTags[i] = CSSM_MALLOC_WRAPPER(internal_malloc, length, NULL);
                    if (!type_ptr->OriginalTags[i])
                    {
                        CSSM_FREE_WRAPPER(internal_free, (void*)type_ptr->TypeTag, NULL);
                        for (j=0; j<i; j++)
                        {
                            CSSM_FREE_WRAPPER(internal_free, (void*)type_ptr->OriginalTags[j], NULL);
                        }
                        CSSM_FREE_WRAPPER(internal_free, (void*)type_ptr->OriginalTags, NULL);
                        return CSSMERR_CL_MEMORY_ERROR;
                    }

                    /* Copy the tag into this array */
                    port_memcpy((void*)type_ptr->OriginalTags[i], NodePtr->Child[i].X.Node->Tag, length);
                }
            }
        }
    }

    /* Fill in the Context-specific Tags, if applicable */
    /* Allocate an array for the Context-specific Tags */
    type_ptr->ContextTags = CSSM_CALLOC_WRAPPER(internal_calloc, NumberOfChildren, sizeof(CL_DER_TAG_TYPE *), NULL);
    if (!type_ptr->ContextTags)
    {
        cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    
    /* Fill in the individual Context-specific Tags array elements */
    ctx_tag_number = 0;
    for (i=0; i < NumberOfChildren; i++)
    {
        /* Assign a context specific tag, if appropriate */
        if (ChildrenArray[i].HasCtxSpecificTag)                  
        {
            if (ctx_tag_number == 0xFF) /* No more context-specific tags can be assigned */
            {
                cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);
                return CSSMERR_CL_INTERNAL_ERROR;
            }

            /* Assign the context specific tag to this field */
            BER_CreateContextSpecificTag(pContext,type_ptr->OriginalTags[i], ctx_tag_number++, ChildrenArray[i].IsExplicit, (uint8 **)&(type_ptr->ContextTags[i]));

            /* If this field is also explicit and non-NULL, modify the Input value contents to contain the DER-encoded blob */
            if (ChildrenArray[i].IsExplicit && NodePtr->Child[i].X.Input.ContentLength)
            {
                /* Create a DER object from the current BER item */
                ber_item = NodePtr->Child[i].X.Input;

                /* Tag stays the same */
                /* Get the length of the Der object */
                NodePtr->Child[i].X.Input.ContentLength = BER_LengthOfTag(ber_item.Tag) + BER_LengthOfLength(ber_item.ContentLength) + ber_item.ContentLength;

                /* Encode the ber item into a DER object */
                NodePtr->Child[i].X.Input.Content = CSSM_MALLOC_WRAPPER(internal_malloc, NodePtr->Child[i].X.Input.ContentLength,
                                                                NULL);
                if (!NodePtr->Child[i].X.Input.Content)
                {
                    cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);
                    return CSSMERR_CL_MEMORY_ERROR;
                    
                }
                BER_OutputTag(ber_item.Tag, (void*)&NodePtr->Child[i].X.Input.Content[0]);
                BER_OutputLength(ber_item.ContentLength, (void*)&NodePtr->Child[i].X.Input.Content[BER_LengthOfTag(ber_item.Tag)]);
                port_memcpy((void*)&NodePtr->Child[i].X.Input.Content[ BER_LengthOfTag(ber_item.Tag) + BER_LengthOfLength(ber_item.ContentLength)], 
                    ber_item.Content, ber_item.ContentLength);

                CSSM_FREE_WRAPPER(internal_free, (void*)ber_item.Content, NULL);
            }
        }
    }

    /* If there were no context-specific tags, remove the Context-specific Tags array */
    if (ctx_tag_number == 0)
    {
        CSSM_FREE_WRAPPER(internal_free, (void*)type_ptr->ContextTags, NULL);
        type_ptr->ContextTags = NULL;
    }

    /* Fill in the Defaults, if applicable */  
    /* Due to a bug in the BER/DER library, defaults must be left out of the child array. 
       They are handled in CreateParentNode, so this section has been removed */

    /* Fill in the subtypes array */
    /* Allocate the subtypes array */
    type_ptr->Subtypes = CSSM_CALLOC_WRAPPER(internal_calloc, NumberOfChildren, sizeof(DER_COMPLEX_TYPE_PTR), NULL);
    if (!type_ptr->Subtypes)
    {
        cl_der_InternalDeleteType(pContext,CSSM_FALSE, 0, type_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    for (i=0; i<NumberOfChildren; i++)
    {
        if (ChildrenArray[i].Subtype)
        {
            /* This identifies that there is at least one subtype */
            has_subtype = CSSM_TRUE;

            /* Copy the type structure */
            ret = cl_der_CopyType(pContext,CSSM_FALSE, 0, ChildrenArray[i].Subtype, &type_ptr->Subtypes[i]);
        }
    }

    /* If there are no subtypes, delete the array */
    if (has_subtype == CSSM_FALSE)
    {
        CSSM_FREE_WRAPPER(internal_free, type_ptr->Subtypes, NULL);
        type_ptr->Subtypes = NULL;
    }

    /* By default, the type has no name */

    /* Return */
    
    *result_val = type_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_FillInBaseBerTagCountLength
 *
 * Description:
 * This function fills in the Tag, child count, and length into the input node
 *
 * Parameters: 
 * node_ptr (input) : The node to be filled in
 * Tag (input) : The tag for the node
 * NumberOfChildren (input) : the number of children of this node
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_FillInBaseBerTagCountLength(const CSSM_CLUTIL_CONTEXT *pContext,
                                               DER_NODE_PTR node_ptr, 
                                               CL_DER_TAG_TYPE Tag, 
                                               uint32 NumberOfChildren)
{
    CL_DER_TAG_TYPE complete_tag = 0;

    if (!node_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    if (NumberOfChildren > 0x7FFF)
    {
        return CSSMERR_CL_INVALID_DATA; /* cannot be handled by underlying BER/DER library */
    }
    /* The Length field will be filled in later by the BER/DER static library */
    node_ptr->Tag = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(CL_DER_TAG_TYPE), NULL);
    node_ptr->OriginalTag = CSSM_MALLOC_WRAPPER(internal_malloc, sizeof(CL_DER_TAG_TYPE), NULL);
    if (!node_ptr->OriginalTag || !node_ptr->Tag)
    {
        CSSM_FREE_WRAPPER(internal_free, (void*)node_ptr->Tag, NULL);
        CSSM_FREE_WRAPPER(internal_free, (void*)node_ptr->OriginalTag, NULL);
        CSSM_FREE_WRAPPER(internal_free, (void*)node_ptr, NULL);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }
    complete_tag = (CL_DER_TAG_TYPE) (Tag | BER_CONSTRUCTED);
    port_memcpy((void*)node_ptr->Tag, &complete_tag, sizeof(CL_DER_TAG_TYPE));
    port_memcpy((void*)node_ptr->OriginalTag, &complete_tag, sizeof(CL_DER_TAG_TYPE));
    node_ptr->Count = (sint16) NumberOfChildren;
    node_ptr->Length = 0;

    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_CreateParentNode
 *
 * Description:
 * This function copies the input into an internal structure which can
 * be used to create a DER encoded version of the data.
 *
 * Parameters: 
 * Tag (input)              : Specifies whether this is a SET or a SEQUENCE
 * NumberOfChildren (input) : Specifies the length of the ChildrenArray
 * ChildrenArray (input)    : Specifies the handles of the children of this
 *                            node.  It also contains information about the 
 *                            SET or SEQ which is used to build the type
 *                            structure (optional parms, defaults, subtypes)
 *
 * return_handle_pointer(output) : A handle which can be used by the calling app to specify 
 *                            this node as part of a parent node or to DER encode this data.
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN  cl_der_CreateParentNode (CSSM_CLUTIL_CONTEXT *pContext,
                                      CL_DER_TAG_TYPE Tag, 
                                      uint32 NumberOfChildren,  
                                      CL_DER_CHILD_INFO_PTR ChildrenArray,
                                      CL_DER_NODE_HANDLE *return_handle)
{
    CSSM_RETURN ret = CSSM_OK, ret1 = CSSM_OK; 
                
    DER_NODE_PTR node_ptr = NULL;
    BER_PARSED_ITEM_PTR leaf_node_ptr = NULL;
    uint32 i=0;
    
    if (!return_handle)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *return_handle = 0;
    /* Check inputs */
    if (!cl_der_IsDerParentTag(Tag))
    {
        return CSSMERR_CL_INVALID_DATA;
        
    }
    if ((ChildrenArray && port_IsBadReadPtr(ChildrenArray, NumberOfChildren * sizeof(CL_DER_CHILD_INFO))) || 
        (NumberOfChildren && !ChildrenArray))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }
 

    /* Allocate the DER node structure */
    node_ptr = CSSM_CALLOC_WRAPPER(internal_calloc, 1, sizeof(DER_NODE), NULL);
  
    if (!node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;       
    }

    /* Fill in the Tag, Original Tag, Count, Length fields */ 
    cl_der_FillInBaseBerTagCountLength(pContext,node_ptr, Tag, NumberOfChildren);

    /* Fill in the Child array of the DER Node */
    /* Allocate the Child array of the DER Node */
    node_ptr->Child = CSSM_CALLOC_WRAPPER(internal_calloc, NumberOfChildren, sizeof(struct der_node_child_struct), NULL);
    if (!node_ptr->Child)
    {
        CSSM_FREE_WRAPPER(internal_free, node_ptr, NULL);
        return CSSMERR_CL_MEMORY_ERROR;
    }
    /* Fill in each element of the Child array, 
     * based on whether it is a leaf or parent node */
    for (i=0; i < NumberOfChildren; i++)
    {
        ret = cl_der_GetLeafNode(pContext,ChildrenArray[i].ChildHandle, &leaf_node_ptr);

        if (leaf_node_ptr) /* This is a leaf node */
        {
            node_ptr->Child[i].IsLeaf = CSSM_TRUE;

            /* If this field is optional and its value is NULL and its tag != DER_TAG_NULL
               or if it equals its default value, leave it out of the encoding */
            if ((ChildrenArray[i].IsOptional == CSSM_TRUE && leaf_node_ptr->ContentLength == 0 && cl_der_Uint8TagToCLDERTag((void*)leaf_node_ptr->Tag) != DER_TAG_NULL) ||
                (ChildrenArray[i].HasDefault == CSSM_TRUE && leaf_node_ptr->ContentLength == ChildrenArray[i].DefaultValue.Length &&
                 port_memcmp(ChildrenArray[i].DefaultValue.Data, leaf_node_ptr->Content, ChildrenArray[i].DefaultValue.Length)==0))
            {
                CSSM_FREE_WRAPPER(internal_free, (void*)leaf_node_ptr->Tag, NULL);
                CSSM_FREE_WRAPPER(internal_free, (void*)leaf_node_ptr->Content, NULL);
            } 
            else
            {
                node_ptr->Child[i].X.Input.Tag = leaf_node_ptr->Tag;
                node_ptr->Child[i].X.Input.ContentLength = leaf_node_ptr->ContentLength;
                node_ptr->Child[i].X.Input.Content = leaf_node_ptr->Content;
                
            }
            /* The parent node pointer remains NULL */

            /* Remove the child node from the linked list */
            ret = cl_der_RemoveLeafNode(pContext,ChildrenArray[i].ChildHandle);

        }
        else /* This is either a parent node or an invalid node handle */
        {
            node_ptr->Child[i].IsLeaf = CSSM_FALSE;
            ret = cl_der_GetParentNode(pContext,ChildrenArray[i].ChildHandle, &(node_ptr->Child[i].X.Node));
 
            
            /* The leaf node data remains NULL */

            if (node_ptr->Child[i].X.Node) /* This is a parent node */
            {
                /* Remove this parent node from the linked list */
                ret1 = cl_der_RemoveParentNode(pContext,ChildrenArray[i].ChildHandle);
            
            } 
            else 
            {  /* This is an invalid handle */
                /* Set an error, but continue */
                /* ret = CSSMERR_CL_INVALID_DATA; Error gets set by the function call */
            }
        }
    }

    /* Fill in the Type of the DER Node */
    /* Check if a Type structure is required.  
       If not, set Type to NULL.
       If so,  create a Type structure
     */
    for (i=0; i< NumberOfChildren; i++)
    {
        if (ChildrenArray[i].IsOptional == CSSM_TRUE || 
            ChildrenArray[i].HasDefault == CSSM_TRUE
            || ChildrenArray[i].Subtype)
        {
            break;
        }
    }

    /* If there are no optional fields, default fields, or subtypes, */
    /* the type remains NULL.  Otherwise, build the type structure */
    /* for this parent node */
    if (i != NumberOfChildren)  
    {
        ret = cl_der_CreateParentType(pContext,Tag, NumberOfChildren, ChildrenArray, node_ptr, &node_ptr->Type);
 
        if (CSSM_OK != ret)
        {
            cl_der_FreeDerNodeTree(pContext,node_ptr);
            return ret;
        }
    }

    /* Add parent node to linked list */
    ret = cl_der_AddParentNode(pContext,node_ptr, return_handle);


    if(ret != CSSM_OK)
    {
        cl_der_FreeDerNodeTree(pContext,node_ptr);
        return ret;
    }

    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_PackTree
 *
 * Description:
 * This function packs a parse tree structure 
 * (identified by the input Root node handle) into a DER encoded blob
 *
 * Parameters: 
 * Root (input)             : The handle of the root of the tree to be packed
 * OutputDerType (output)   : The description of the parse tree 
 *
 * return_results_ptr (output): The DER encoded blob
 * 
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN     cl_der_PackTree (CSSM_CLUTIL_CONTEXT *pContext,
                            CSSM_HANDLE AppHandle, 
                            CL_DER_NODE_HANDLE Root, 
                            DER_COMPLEX_TYPE_PTR *OutputDerType,
                            CSSM_DATA_PTR *results_ptr)
{
    CSSM_RETURN     ret = CSSM_OK;
    DER_NODE_PTR    node_ptr = NULL;
    
    *results_ptr = NULL;
    /* Obtain a pointer to the root node */
    ret = cl_der_GetParentNode(pContext,Root, &node_ptr);
    if(ret != CSSM_OK)
    {
        return ret;
    }

    /* Allocate result structure */
    *results_ptr = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CSSM_DATA));
    if (!*results_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;      
    }

    /* Obtain final size of DER encoded object using SizeofDerTree */
    (*results_ptr)->Length = DER_SizeofTree_copyctxtag(node_ptr);

    /* Allocate buffer of correct length */
    (*results_ptr)->Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, (*results_ptr)->Length);
    if (!(*results_ptr)->Data)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr));
        return CSSMERR_CL_MEMORY_ERROR;       
    }
    
    /* Use CopyDerTree to create DER encoded object from tree */
    DER_CopyTree( node_ptr, (*results_ptr)->Data );

    /* If requested (by a positive value of OutputDerType), */
    /*   Copy the type into the app's memory space */
    if (OutputDerType)
    {
        ret = cl_der_CopyType(pContext,CSSM_TRUE, AppHandle, node_ptr->Type, OutputDerType);
    }
 
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_TranslateToBaseBerParseTree
 *
 * Description:
 * Create a BER/DER static library parse tree from the CL DER parse tree
 *
 * Parameters: 
 * NodePtr (input) : The root of the tree to be translated
 *
 * results_ptr (output): The CL DER parse tree version of the BER/DER static library parse tree
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_TranslateToBaseBerParseTree(const CSSM_CLUTIL_CONTEXT *pContext,
                                               CL_DER_PARENT_NODE_PTR NodePtr,
                                               DER_COMPLEX_TYPE_PTR DerType,
                                               DER_NODE_PTR *results_ptr)
{
    CSSM_RETURN ret = CSSM_OK;
    BER_PARSED_ITEM_PTR  ber_item_ptr = NULL;
    uint32 i=0;
    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *results_ptr = NULL;
    /* Check inputs */
    if (!NodePtr)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
       
    }

    /* Allocate output Der node */
    *results_ptr = CSSM_CALLOC_WRAPPER(internal_calloc, sizeof(DER_NODE),1, NULL);
    if (!*results_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
       
    }

    /* Fill in the DER node */
    if ((cl_der_FillInBaseBerTagCountLength(pContext, *results_ptr, NodePtr->Tag, 
                                            NodePtr->NumberOfChildren)) != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(internal_free, *results_ptr, NULL);
        return CSSMERR_CL_MEMORY_ERROR;
      
    }

    /* Fill in the Child array of the DER Node */
    /* Allocate the Child array of the DER Node */
    (*results_ptr)->Child = CSSM_CALLOC_WRAPPER(internal_calloc, NodePtr->NumberOfChildren, 
                                     sizeof(struct der_node_child_struct), NULL);
    if (!(*results_ptr)->Child)
    {
        CSSM_FREE_WRAPPER(internal_free, (void*)(*results_ptr)->Tag, NULL);
        CSSM_FREE_WRAPPER(internal_free, (void*)(*results_ptr)->OriginalTag, NULL);

        CSSM_FREE_WRAPPER(internal_free, *results_ptr, NULL);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Fill in each element of the Child array, 
     * based on whether it is a leaf or parent node */
    for (i=0; i < NodePtr->NumberOfChildren; i++)
    {
        if (NodePtr->ChildrenArray[i].IsLeaf) /* This is a leaf node */
        {
            (*results_ptr)->Child[i].IsLeaf = CSSM_TRUE;

            /* Translate the data into a BER_PARSED_ITEM */
            ret = cl_der_PreprocessTagDataPair
                (pContext,NodePtr->ChildrenArray[i].DerNodeType.LeafNode->Tag, &NodePtr->ChildrenArray[i].DerNodeType.LeafNode->Data, &ber_item_ptr);
            if (ret != CSSM_OK)
            {
                cl_der_FreeDerNodeTree(pContext,*results_ptr);
                CSSM_FREE_WRAPPER(internal_free, *results_ptr, NULL);
                return ret;
            }

            (*results_ptr)->Child[i].X.Input = *ber_item_ptr;
            CSSM_FREE_WRAPPER(internal_free, ber_item_ptr, NULL);

            /* The parent node pointer remains NULL */
        }
        else /* This is either a parent node or an invalid node handle */
        {
            (*results_ptr)->Child[i].IsLeaf = CSSM_FALSE;
            ret = cl_der_TranslateToBaseBerParseTree(pContext,NodePtr->ChildrenArray[i].DerNodeType.ParentNode, NULL, &(*results_ptr)->Child[i].X.Node);
            /* The leaf node data remains NULL */
        }
    }

    /* Fill in the Type */
    ret = cl_der_CopyType(pContext,CSSM_FALSE, 0, DerType, &((*results_ptr)->Type));

    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_PackParseTree
 *
 * Description:
 * This function packs a parse tree structure into a DER encoded blob
 *
 * Parameters: 
 * Root (input)      : The root of the tree to be packed
 * DerType (input)   : The description of the parse tree 
 * results_ptr (output): The DER encoded blob
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_PackParseTree(const CSSM_CLUTIL_CONTEXT *pContext,
                                 CSSM_HANDLE AppHandle, 
                                 CL_DER_PARENT_NODE_PTR Root, 
                                 DER_COMPLEX_TYPE_PTR DerType,
                                 CSSM_DATA_PTR *results_ptr)
{
    CSSM_RETURN     ret = CSSM_OK;
    DER_NODE_PTR  node_ptr = NULL;
    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *results_ptr = NULL;

    /* Check inputs */
    if (!Root)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
       
    }

    /* Translate the parse tree into one of Gary's nodes */
    ret = cl_der_TranslateToBaseBerParseTree(pContext,Root, DerType, &node_ptr);
    if (!node_ptr)
    {
        return ret;
    }

    /* Allocate result structure */
    *results_ptr = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CSSM_DATA));
    if (!*results_ptr)
    {
        cl_der_FreeDerNodeTree(pContext,node_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
    }

    /* Obtain final size of DER encoded object using SizeofDerTree */
    (*results_ptr)->Length = DER_SizeofTree_copyctxtag(node_ptr);

    /* Allocate buffer of correct length */
    (*results_ptr)->Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, (*results_ptr)->Length);
    if (!(*results_ptr)->Data)
    {
        cl_der_FreeDerNodeTree(pContext,node_ptr);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, *results_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
       
    }
    
    /* Use CopyDerTree to create DER encoded object from tree */
    DER_CopyTree( node_ptr, (*results_ptr)->Data );

    /* Cleanup */
    cl_der_FreeDerNodeTree(pContext,node_ptr);

    
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DeleteTree
 *
 * Description:
 * Delete all internal state information associated with this tree structure
 *
 * Parameters: 
 * Root (input) : A handle indicating the root of the tree to be deleted
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_DeleteTree(CSSM_CLUTIL_CONTEXT *pContext,
                              CL_DER_NODE_HANDLE Root)
{
    CSSM_RETURN ret = CSSM_OK;
    
    /* Delete this Node */
    cl_der_DeleteLeafNode(pContext,Root);
    cl_der_DeleteParentNode(pContext,Root);

    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_BerItemToCItem
 *
 * Description:
 * This function translates the input Der data into C data and
 * returns it in a CL_DER_LEAF_NODE structure
 *
 * Parameters: 
 * BerItemPtr  (input) : The tag, length, content in its DER format
 *
 * results_ptr (output) : A CL_DER_LEAF_NODE containing the C equivalent of the input data
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_BerItemToCItem   (const CSSM_CLUTIL_CONTEXT *pContext,
                                     CSSM_HANDLE AppHandle, 
                                     BER_PARSED_ITEM_PTR BerItemPtr,
                                     CL_DER_LEAF_NODE_PTR *results_ptr)
{
    CSSM_RETURN     ret = CSSM_OK;
    CSSM_DATA       null_data = {0, NULL};
    uint32          bit_str_length =0;
    uint32          numBytesInBitString = 0;
    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *results_ptr = NULL;
    /* Allocate leaf node */
    *results_ptr = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CL_DER_LEAF_NODE));
    if (!*results_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
      
    }

    /* Copy into Leaf Node */
    (*results_ptr)->Tag = cl_der_Uint8TagToCLDERTag((void*)BerItemPtr->Tag);

    switch ((*results_ptr)->Tag) {
    case DER_TAG_UNKNOWN:
        (*results_ptr)->Data = null_data;
        break;

    case DER_TAG_BOOLEAN :
        (*results_ptr)->Data.Length = sizeof(CSSM_BOOL);
        ret = cl_der_DerBool2CSSMBool(pContext,AppHandle, (uint8*)BerItemPtr->Content, (CSSM_BOOL**)(&(*results_ptr)->Data.Data));
        if (!(*results_ptr)->Data.Data)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr));
            return ret;
        }
        break;

    case DER_TAG_BIT_STRING :
        if (BerItemPtr->ContentLength)
        {
            /* Decode the bit string, removing the leading byte */
            bit_str_length = 8*(BerItemPtr->ContentLength -1) - BerItemPtr->Content[0];
            numBytesInBitString = BerItemPtr->ContentLength -1;

            (*results_ptr)->Data.Length = bit_str_length;
            if (numBytesInBitString)
            {
                (*results_ptr)->Data.Data = 
                    CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, numBytesInBitString);
                if (!(*results_ptr)->Data.Data)
                {
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr));
                    return CSSMERR_CL_MEMORY_ERROR;
                  
                }
                port_memcpy((*results_ptr)->Data.Data, &BerItemPtr->Content[1], numBytesInBitString);
            } 
            else
            {
                (*results_ptr)->Data.Data = NULL; 
            }
        } 
        else
        {
            (*results_ptr)->Data = null_data; 
        }

        break;

    case DER_TAG_PRINTABLE_STRING :
    case DER_TAG_IA5_STRING :
    case DER_TAG_UTC_TIME :
    case DER_TAG_GENERALIZED_TIME :
        (*results_ptr)->Data.Length = BerItemPtr->ContentLength +1;
        (*results_ptr)->Data.Data = 
            CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, (*results_ptr)->Data.Length);
        if (!(*results_ptr)->Data.Data)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr));
            return CSSMERR_CL_MEMORY_ERROR;
           
        }
        port_memcpy((*results_ptr)->Data.Data, BerItemPtr->Content, BerItemPtr->ContentLength);
        break;

    default:
        (*results_ptr)->Data.Length = BerItemPtr->ContentLength;
        if (BerItemPtr->ContentLength) 
        {
            (*results_ptr)->Data.Data = 
                CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, BerItemPtr->ContentLength);
            if (!(*results_ptr)->Data.Data)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr));
                return CSSMERR_CL_MEMORY_ERROR;
              
            }
            port_memcpy((*results_ptr)->Data.Data, BerItemPtr->Content, (*results_ptr)->Data.Length);
        }
        else
        {
            (*results_ptr)->Data.Data = NULL;
        }
    }

    /* Return */
    return ret;
}

/*-----------------------------------------------------------------------------
 * Name: cl_der_TranslateToCLDerParseTree
 *
 * Description:
 * Create a CL DER parse tree from the BER/DER static library parse tree
 *
 * Parameters: 
 * NodePtr (input) : The root of the tree to be translated
 * 
 * results_ptr (output): The CL DER parse tree version of the BER/DER static
 *                       library parse tree
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_TranslateToCLDerParseTree(const CSSM_CLUTIL_CONTEXT *pContext,
                                                 CSSM_HANDLE AppHandle, 
                                                 DER_NODE_PTR NodePtr,
                                                 CL_DER_NODE_PTR *results_ptr)
{
    CSSM_RETURN     ret = CSSM_OK;
    CL_DER_NODE_PTR temp_pnode = NULL;
    CSSM_DATA       null_data = {0, NULL};
    uint32          i=0;
    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *results_ptr = NULL;
    /* Check inputs */
    if (!NodePtr || IsBadReadPtr(NodePtr, sizeof(DER_NODE)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
     
    }

    /* Allocate and fill in the given parent node */
    *results_ptr = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CL_DER_NODE));
    if (!*results_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
      
    }
    (*results_ptr)->DerNodeType.ParentNode = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CL_DER_PARENT_NODE));
    if (!(*results_ptr)->DerNodeType.ParentNode)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr));
        return CSSMERR_CL_MEMORY_ERROR;
       
    }

    (*results_ptr)->IsLeaf = CSSM_FALSE;
    (*results_ptr)->DerNodeType.ParentNode->Tag = (CL_DER_TAG_TYPE) *NodePtr->OriginalTag;
    (*results_ptr)->DerNodeType.ParentNode->Tag = (CL_DER_TAG_TYPE) ((*results_ptr)->DerNodeType.ParentNode->Tag & ~BER_CONSTRUCTED);
    (*results_ptr)->DerNodeType.ParentNode->NumberOfChildren = (uint32) NodePtr->Count;
    (*results_ptr)->DerNodeType.ParentNode->ChildrenArray = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 
        (*results_ptr)->DerNodeType.ParentNode->NumberOfChildren, sizeof(CL_DER_NODE));
    if (!(*results_ptr)->DerNodeType.ParentNode->ChildrenArray)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr)->DerNodeType.ParentNode);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr));
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* For each child, fill in either the LeafNode or ParentNode */
    for (i=0; i<(*results_ptr)->DerNodeType.ParentNode->NumberOfChildren; i++)
    {
        (*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].IsLeaf = NodePtr->Child[i].IsLeaf;
        if ((*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].IsLeaf == CSSM_TRUE)
        {
            if (NodePtr->Child[i].X.Input.Tag && cl_der_IsDerLeafTag((CL_DER_TAG_TYPE) *NodePtr->Child[i].X.Input.Tag))
            {
                ret =
                    cl_der_BerItemToCItem (pContext,AppHandle, &NodePtr->Child[i].X.Input, &(*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].DerNodeType.LeafNode);
                if (!(*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].DerNodeType.LeafNode)
                {
                    ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, (*results_ptr));
                    return ret;
                }
            }
            else 
            {
                /* Allocate leaf node */
                (*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].DerNodeType.LeafNode = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CL_DER_LEAF_NODE));
                if (!(*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].DerNodeType.LeafNode)
                {
                    ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, (*results_ptr));
                    return CSSMERR_CL_MEMORY_ERROR;
                   
                }
                (*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].DerNodeType.LeafNode->Tag = 0;
                (*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].DerNodeType.LeafNode->Data = null_data;
            }
        }
        else /* This is a parent node */
        {
            ret = cl_der_TranslateToCLDerParseTree(pContext,AppHandle, NodePtr->Child[i].X.Node, &temp_pnode);
            if (!temp_pnode)
            {
                ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, (*results_ptr));
                return ret;
            }

            (*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i] = *temp_pnode;
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, temp_pnode);
            if (!(*results_ptr)->DerNodeType.ParentNode->ChildrenArray[i].DerNodeType.ParentNode)
            {
                ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, (*results_ptr));
                return ret;
            }
        }
    }

   
    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_ParseDerObject
 *
 * Description:
 * This functions translates a DER encoded blob into a parse tree
 *
 * Parameters: 
 * DerObject (input)    : The DER encoded object to be parsed
 * DerType (opt input)  : The parse tree description.  
 *                        If NULL is passed, the library will make its
 *                        best attempt.  Optional & default parameters
 *                        may be incorrect.
 *
 * results_ptr (output) :The parse tree corresponding to the DER encoded blob
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_ParseDerObject(const CSSM_CLUTIL_CONTEXT *pContext,
                                  CSSM_HANDLE AppHandle, 
                                  const CSSM_DATA *DerObject, 
                                  DER_COMPLEX_TYPE_PTR DerType,
                                  CL_DER_NODE_PTR *results_ptr)
{
    CSSM_RETURN ret = CSSM_OK;
    DER_NODE_PTR node_ptr = NULL;
    BER_PARSED_ITEM der_object;
    *results_ptr = NULL;

    /* Check inputs */
    if (IsBadReadPtr(DerObject, sizeof(CSSM_DATA)) || !DerObject)
    {
       return CSSMERR_CL_INVALID_INPUT_POINTER;
    }
    if (!DerObject->Data || !DerObject->Length ||
        IsBadReadPtr(DerObject->Data, DerObject->Length))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
 
    }
    if (DerType && IsBadReadPtr(DerType, sizeof(DER_COMPLEX_TYPE)))
    {
       return CSSMERR_CL_INVALID_INPUT_POINTER;
    }

    /* Initialize DER object as BER parsed item */
    if (BER_ExpandItem(DerObject->Data, DerObject->Length, &der_object) == NULL)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    
    }  

    /* ParseComplexBerObject */
    node_ptr = BER_ParseComplexObject(AppHandle, (BER_MEMORY_FUNCS_PTR)&pContext->MemFuncs, DerType, &der_object);
    if (node_ptr == NULL)
    {
        return CSSMERR_CL_INVALID_DATA; 
    }  

    /* Creating a parse tree of CL_DER_NODEs */
    ret = cl_der_TranslateToCLDerParseTree(pContext,AppHandle, node_ptr, results_ptr);
    
    /* call RecycleDerTree */   
    DER_RecycleTree(AppHandle, (BER_MEMORY_FUNCS_PTR)&pContext->MemFuncs, node_ptr);

    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DeleteParsedDerObject
 *
 * Description:
 * Delete the input parse tree
 *
 * Parameters: 
 * Root (input) : A handle indicating the root of the tree to be deleted
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_DeleteParsedDerObject (const CSSM_CLUTIL_CONTEXT *pContext,
                                          CSSM_HANDLE AppHandle,
                                          CL_DER_NODE_PTR Root)
{
    CSSM_RETURN ret = CSSM_OK;
    uint32 i=0;

    /* Check inputs */
    if (!Root)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    /* If this is a leaf node */
    if (Root->IsLeaf)
    {
        if (Root->DerNodeType.LeafNode->Data.Data != NULL)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, Root->DerNodeType.LeafNode->Data.Data);
        }
        if (Root->DerNodeType.LeafNode != NULL)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, Root->DerNodeType.LeafNode);
        }
        return ret;
    }

    /* If this is not a leaf node, then it is a parent node */
    for (i=0; i<Root->DerNodeType.ParentNode->NumberOfChildren; i++)
    {
        ret = cl_der_DeleteParsedDerObject(pContext,AppHandle, &Root->DerNodeType.ParentNode->ChildrenArray[i]);
    }

    if (Root->DerNodeType.ParentNode->ChildrenArray != NULL)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, Root->DerNodeType.ParentNode->ChildrenArray);
    }
    if (Root->DerNodeType.ParentNode != NULL)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, Root->DerNodeType.ParentNode);
    }
    return ret;
}



/*-----------------------------------------------------------------------------
 * Name: cl_der_EncodeObject
 *
 * Description:
 * This function accepts the Tag and data of an object to be encoded
 * and then encodes that object.
 *
 * Parameters: 
 * Tag (input)   : The DER Tag for the input object
 * Data (input)  : The data to be DER encoded
 *
 * results_ptr (output): The DER encoding for this Tag/value pair
 * Notes:
 *   The returned data is allocated on the API caller's heap.
 *   All intermediate data is allocated on the local (porting library) heap.
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_EncodeObject(const CSSM_CLUTIL_CONTEXT *pContext,
                                CSSM_HANDLE AppHandle, 
                                CL_DER_TAG_TYPE Tag, 
                                const CSSM_DATA *Data,
                                CSSM_DATA_PTR *results_ptr)
{
    CSSM_RETURN         ret = CSSM_OK;
    BER_PARSED_ITEM_PTR  ber_item_ptr = NULL;
    if (!results_ptr)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    *results_ptr = NULL;

    ret = cl_der_PreprocessTagDataPair(pContext,Tag, Data, &ber_item_ptr);
    if (ret != CSSM_OK)
    {
        return ret;
    }

    /* Construct the DER object */
    *results_ptr = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, sizeof(CSSM_DATA));
    if (*results_ptr == NULL)
    {
        CSSM_FREE_WRAPPER(internal_free, (void*)ber_item_ptr->Tag, NULL);
        CSSM_FREE_WRAPPER(internal_free, (void*)ber_item_ptr->Content, NULL);
        CSSM_FREE_WRAPPER(internal_free, ber_item_ptr, NULL);
        return CSSMERR_CL_MEMORY_ERROR;
    }

    (*results_ptr)->Length = BER_LengthOfTag(ber_item_ptr->Tag) + BER_LengthOfLength(ber_item_ptr->ContentLength) + ber_item_ptr->ContentLength;
    (*results_ptr)->Data = CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, (*results_ptr)->Length);
    if ((*results_ptr)->Data == NULL)
    {
        CSSM_FREE_WRAPPER(internal_free, (void*)ber_item_ptr->Tag, NULL);
        CSSM_FREE_WRAPPER(internal_free, (void*)ber_item_ptr->Content, NULL);
        CSSM_FREE_WRAPPER(internal_free, ber_item_ptr, NULL);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, (*results_ptr));
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* Encode the ber item into a DER object */
    BER_OutputTag(ber_item_ptr->Tag, &(*results_ptr)->Data[0]);
    BER_OutputLength(ber_item_ptr->ContentLength, &(*results_ptr)->Data[BER_LengthOfTag(ber_item_ptr->Tag)]);
    port_memcpy(&(*results_ptr)->Data[ BER_LengthOfTag(ber_item_ptr->Tag) + BER_LengthOfLength(ber_item_ptr->ContentLength)], ber_item_ptr->Content, ber_item_ptr->ContentLength);

    CSSM_FREE_WRAPPER(internal_free, (void*)ber_item_ptr->Tag, NULL);
    CSSM_FREE_WRAPPER(internal_free, (void*)ber_item_ptr->Content, NULL);
    CSSM_FREE_WRAPPER(internal_free, ber_item_ptr, NULL);

    return ret;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_DecodeObject
 *
 * Description:
 * This function accepts a DER encoded blob and peels off one layer
 * of DER wrapping
 *
 * Parameters: 
 * Data (input)       : The DER-encoded blob
 *
 * results_ptr (output): A node pointer containing the encoded blob's Tag and data
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_DecodeObject   (const CSSM_CLUTIL_CONTEXT *pContext,
                                   CSSM_HANDLE AppHandle,
                                   CSSM_DATA_PTR Data,
                                   CL_DER_LEAF_NODE_PTR *results_ptr)
{
    BER_PARSED_ITEM      ber_item = {NULL, 0, NULL};
    *results_ptr = NULL;
    /* Expand Data object */
    BER_ExpandItem(Data->Data, Data->Length, &ber_item);

    /* Translate the BER object into a C object */
    return cl_der_BerItemToCItem (pContext,AppHandle, &ber_item, results_ptr);
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_LengthOfTag
 *
 * Description:
 * This function calcuulates the number of bytes in the DER Tag
 *
 * Parameters: 
 * Data (input)       : The DER-encoded blob
 *
 * Return value:
 * the number of bytes in the DER Tag
 *---------------------------------------------------------------------------*/
uint32 cl_der_LengthOfTag(const CSSM_DATA *Data)
{
    BER_PARSED_ITEM      ber_item = {NULL, 0, NULL};

    /* Expand Data object */
    BER_ExpandItem(Data->Data, Data->Length, &ber_item);

    return BER_LengthOfTag(ber_item.Tag);
}

/*-----------------------------------------------------------------------------
 * Name: cl_der_LengthOfLength
 *
 * Description:
 * This function calcuulates the number of bytes in the DER Length
 *
 * Parameters: 
 * Data (input)       : The DER-encoded blob
 *
 * Return value:
 * the number of bytes in the DER Length
 * 
 *---------------------------------------------------------------------------*/
uint32 cl_der_LengthOfLength(const CSSM_DATA *Data)
{
    BER_PARSED_ITEM      ber_item = {NULL, 0, NULL};

    /* Expand Data object */
    BER_ExpandItem(Data->Data, Data->Length, &ber_item);

    return BER_LengthOfLength(ber_item.ContentLength);
}

/*-----------------------------------------------------------------------------
 * Name: cl_der_LengthOfContent
 *
 * Description:
 * This function calculates the number of bytes in the DER Content
 *
 * Parameters: 
 * Data (input)       : The DER-encoded blob
 *
 * Return value:
 * the number of bytes in the DER Content
 * 
 *---------------------------------------------------------------------------*/
uint32 cl_der_LengthOfContent(const CSSM_DATA *Data)
{
    BER_PARSED_ITEM      ber_item = {NULL, 0, NULL};

    /* Expand Data object */
    BER_ExpandItem(Data->Data, Data->Length, &ber_item);

    return ber_item.ContentLength;
}


/*-----------------------------------------------------------------------------
 * Name: cl_der_IsConstructedIndefiniteLength
 *
 * Description:
 * This function uses the first length byte to determine whether this is 
 * of constructed indefinite length
 *
 * Parameters: 
 * Data (input)       : The DER-encoded blob
 *
 * Answer (output)    : A Boolean indicating whether or not this DER-encoded blob has a 
 *                      constructed indefinite length
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_der_IsConstructedIndefiniteLength(const CSSM_CLUTIL_CONTEXT *pContext,
                                                 const CSSM_DATA *Data, CSSM_BOOL *Answer)
{
    uint8* Octets;
    uint32 i=0;

    if (!Data || 
        port_IsBadReadPtr(Data, sizeof(CSSM_DATA)))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }
    if (!Data->Data || !Data->Length ||
        port_IsBadReadPtr(Data->Data, Data->Length))
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }

    i = BER_LengthOfTag(Data->Data);
    if (i==0) return CSSMERR_CL_INVALID_DATA;

    /* bypass tag */
    Octets = &Data->Data[i];

    /* use this first length octet to determine how the length is encoded */
    if ((*Octets & 128) == 0 || (*Octets++ & 127) > 0)  /* short definite or long definite */
    {
        *Answer = CSSM_FALSE; 
        return CSSM_OK;
    }

    /* constructed indefinite length */
    *Answer = CSSM_TRUE;
    return CSSM_OK;
}
