/*-----------------------------------------------------------------------
 * File: CRLENCOD.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

/* This file defines functions for use in Encoding a Certificate Revocation
 * List.
 */

#include "x509defs.h"
#include "X509Time.h"
#include "ber_der.h"
#include "clutil.h"
#include "internal.h"
#include "crllist.h"
#include <cssmMemTrack.h>

/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_DerEncodeRevokedCertEntry
*
* Description:
* This function packs an entry from the revokedCertificates list from a CRL into a
* der encoded memory blob.
*
* If the RCEntry input is NULL, it is treated as a valid entry with no data.  Similarly,
* any of the fields of the entry may be null or empty, and they will be treated as
* valid fields with no data.
*
* Parameters: 
* RCEntry:  The revoked Certificates list entry that is to be encoded.  Note that
*           this function does not require that all non-optional fields of the
*           entry be filled.
*
* 
* *result_val (output CSSM_DATA_PTR):    The der encoded entry.
*
* 
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeRevokedCertEntry(CSSM_CLUTIL_CONTEXT *pContext,
                                                       CSSM_HANDLE AppHandle, 
                                                       CSSM_X509_REVOKED_CERT_ENTRY_PTR RCEntry,
                                                       CSSM_DATA_PTR *return_val)
{
    uint32                  num_types = 3;
    CL_DER_CHILD_INFO       revoked_entry_array[3];
    CL_DER_NODE_HANDLE      root = 0;
    CSSM_DATA_PTR           enc_extensions = NULL;
    uint32                  revocationDate_length = 0;
    
    CSSM_RETURN             retVal = CSSM_OK, ret = CSSM_OK;
    *return_val = NULL;
    /* RCEntry may be NULL, signifying a null placeholder (entry with no data) */
    if (RCEntry)
    {
        /* certificateSerialNumber may have Data==NULL, Length==0, signifying a null placeholder 
           if field hasn't been filled yet. */
        if ( (!RCEntry->certificateSerialNumber.Data && RCEntry->certificateSerialNumber.Length) ||
             (RCEntry->certificateSerialNumber.Data && CSSM_CLUTIL_X509_IsBadReadPtr(RCEntry->certificateSerialNumber.Data, 
                                                                         RCEntry->certificateSerialNumber.Length)))
        {
            
            return CSSMERR_CL_INVALID_DATA;
            
        }
        /* revocationDate may be NULL if field hasn't been filled yet. */
        if (RCEntry->revocationDate.time.Data)
        {
            CSSM_DATA temp_time;
            temp_time.Data = (uint8*) RCEntry->revocationDate.time.Data;
            temp_time.Length = strlen((char *) RCEntry->revocationDate.time.Data);

            if ((IsUtcTime(&temp_time) == CSSM_FALSE) &&
                (IsGeneralizedTime(&temp_time) == CSSM_FALSE))
            {
            
                return CSSMERR_CL_INVALID_DATA;
                
            }
            revocationDate_length = temp_time.Length;
        }
        /* extensions may be NULL if field hasn't been filled in */
        if (RCEntry->extensions.extensions && 
            CSSM_CLUTIL_X509_IsBadExtensionPtr(RCEntry->extensions.extensions, 
                                                RCEntry->extensions.numberOfExtensions)) 
        {
            
            return CSSMERR_CL_INVALID_DATA;
            
        }    
    } /*if RCEntry*/
    else
    {   /* Encode an empty structure */
        RCEntry = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CSSM_X509_REVOKED_CERT_ENTRY));
        if (!RCEntry)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }    
    }
    

    /* initialize children array */
    port_memset(revoked_entry_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));

    /* Encode extensions */
    ret = CSSM_CLUTIL_X509_DerEncodeExtensions(pContext,AppHandle, 
                                                 RCEntry->extensions.extensions, 
                                                 RCEntry->extensions.numberOfExtensions,
                                                 &enc_extensions);
    if (CSSM_OK != ret)
    {
        return ret;
        
    }

    /* Create child nodes */
    /* certificateSerialNumber */
    ret = (RCEntry->certificateSerialNumber.Length != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_INTEGER, 
                              RCEntry->certificateSerialNumber.Length, 
                              RCEntry->certificateSerialNumber.Data, &(revoked_entry_array[0].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_INTEGER, 0, NULL, &(revoked_entry_array[0].ChildHandle));

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
        return ret;
    }


    /* revocationDate */
    {
        CSSM_DATA temp_time;
        temp_time.Data = (uint8*)RCEntry->revocationDate.time.Data;
        temp_time.Length = revocationDate_length;
        if (IsUtcTime(&temp_time) == CSSM_FALSE)
        {
            ret = (revocationDate_length != 0) ?
                cl_der_CreateLeafNode(pContext,DER_TAG_GENERALIZED_TIME, 
                                      revocationDate_length, 
                                      (uint8*) RCEntry->revocationDate.time.Data, &(revoked_entry_array[1].ChildHandle)) :
                cl_der_CreateLeafNode(pContext,DER_TAG_GENERALIZED_TIME, 0, NULL, &(revoked_entry_array[1].ChildHandle));
    
            if(ret != CSSM_OK)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
                retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[0].ChildHandle);
                if(retVal != CSSM_OK) return retVal;
                return ret;
            }
        
        }
        else
        {
            ret = (revocationDate_length != 0) ?
                cl_der_CreateLeafNode(pContext,DER_TAG_UTC_TIME, 
                                      revocationDate_length, 
                                      (uint8*) RCEntry->revocationDate.time.Data, &(revoked_entry_array[1].ChildHandle)) :
                cl_der_CreateLeafNode(pContext,DER_TAG_UTC_TIME, 0, NULL, &(revoked_entry_array[1].ChildHandle));
            
            if(ret != CSSM_OK)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
                retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[0].ChildHandle);
                if(retVal != CSSM_OK) return retVal;
                return ret;
            }
            
        }
    }

    /* extensions */
    revoked_entry_array[2].IsOptional = CSSM_TRUE;
    ret = (cl_der_LengthOfContent(enc_extensions) != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
                              cl_der_LengthOfContent(enc_extensions), 
                              &enc_extensions->Data[cl_der_LengthOfTag(enc_extensions) 
                                + cl_der_LengthOfLength(enc_extensions)], &(revoked_entry_array[2].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &(revoked_entry_array[2].ChildHandle));
 
    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
        retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[0].ChildHandle);
        if(retVal != CSSM_OK) return retVal;
        retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[1].ChildHandle);
        if(retVal != CSSM_OK) return retVal;
        return ret;
    }

    /* Cleanup */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);

    /* Check Handles */
    if (!revoked_entry_array[0].ChildHandle || !revoked_entry_array[1].ChildHandle || 
        !revoked_entry_array[2].ChildHandle)
    {
        retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[0].ChildHandle);
        if(retVal != CSSM_OK) return retVal;
        retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[1].ChildHandle);
        if(retVal != CSSM_OK) return retVal;
        retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[2].ChildHandle);
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Create parent node */
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, revoked_entry_array, &root);

    if (CSSM_OK != ret)
    {   
        retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[0].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[1].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,revoked_entry_array[2].ChildHandle);
        if(retVal != CSSM_OK) return retVal;
        return ret;
    }

    /* Pack the entry struct into a der encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, return_val);

    /* Delete the parse tree */
    cl_der_DeleteTree(pContext,root);

    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* return */

    return ret;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_DerEncodeRevokedCertList
*
* Description:
* This function will pack a revoked certificate list into a der encoded memory
* blob.
*
* If the input RCList is NULL, it will be treated as a valid list with no data. 
*
* Parameters: 
* RCList:   The revoked certificates list that is to be encoded.
*
* return_val (output): The der encoded memory blob.
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeRevokedCertList(CSSM_CLUTIL_CONTEXT *pContext,
                                                      CSSM_HANDLE AppHandle, 
                                                      X509_REVOKED_CERT_LIST_PTR RCList,
                                                      CSSM_DATA_PTR *return_val)
{
    CSSM_X509_REVOKED_CERT_ENTRY_PTR list_entry = NULL;
    uint32                      num_nodes = 0;
    CL_DER_CHILD_INFO           *list_children = NULL;
    CL_DER_NODE_HANDLE          root = 0;
    CSSM_DATA_PTR               *encoded_entries = NULL;
    CSSM_HANDLE                 traverse_handle = 0;
    
    uint32                      i=0, j=0;
    CSSM_RETURN                 ret = CSSM_OK, retVal = CSSM_OK;
    if (!return_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *return_val = NULL;
    /* If RCList is NULL or empty, encode a null placeholder */
    num_nodes = CSSM_CLUTIL_X509_RevokedListNumNodes(pContext,RCList);
    if (!RCList || !num_nodes)
    {
       ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &root);

       if(ret != CSSM_OK)
       {
            cl_der_DeleteTree(pContext,root);
            return ret;
       }
       ret = cl_der_PackTree(pContext,AppHandle, root, NULL, return_val);
       cl_der_DeleteTree(pContext,root);
       return ret;
    }

    /** Otherwise, for a non-empty list, encode each entry **/

    /* Allocate children array and encode entries array */
    list_children = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_nodes, sizeof(CL_DER_CHILD_INFO));
    encoded_entries = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, num_nodes, sizeof(CSSM_DATA));
    if (!list_children || !encoded_entries)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, list_children);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, encoded_entries);
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /** Encode each node **/
    /* Initialize List Traversal */
    ret = CSSM_CLUTIL_X509_RevokedListTraverseInitialize(pContext,RCList, &traverse_handle);
 
    if (CSSM_OK != ret)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, list_children);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, encoded_entries);
        return ret;
    }
    /* traverse list, encoding each entry */
    for (i=0; i < num_nodes; i++)
    {
        /* Get next node */
        ret = CSSM_CLUTIL_X509_RevokedListTraverseNext(pContext,AppHandle, RCList, traverse_handle, &list_entry);
        if (CSSM_OK != ret)
        {
            break;
        }
        /* Encode the revoked cert entry */
        ret = CSSM_CLUTIL_X509_DerEncodeRevokedCertEntry(pContext,AppHandle, list_entry, &encoded_entries[i]);
        CSSM_CLUTIL_X509_FreeRevokedCertEntry(pContext,AppHandle, list_entry);
        if (!encoded_entries[i])
        {
            break;
        }
    }
    /* Abort list traversal */
    CSSM_CLUTIL_X509_RevokedListTraverseAbort(pContext,traverse_handle);
    /* Check for early exit of loop due to error */
    if (i != num_nodes)
    {
        for (j=0; j<i; j++)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, encoded_entries[j]->Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, encoded_entries[j]);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, encoded_entries);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, list_children);
        }
        if (CSSM_OK != ret)
        {
            return ret;
        }
        else
        {
            return CSSMERR_CL_INTERNAL_ERROR;
        }
        
    }

    /* Add each entry to the children array */
    for (i=0; i < num_nodes; i++)
    {
        /* Add to the children array */
        ret = 
            cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
                                  cl_der_LengthOfContent(encoded_entries[i]), 
                                  &encoded_entries[i]->Data[cl_der_LengthOfTag(encoded_entries[i]) + cl_der_LengthOfLength(encoded_entries[i])], &(list_children[i].ChildHandle));
        if (ret != CSSM_OK)
        {
            break;
        }
    }
    /* Clean up and set error if loop was exited early */
    for (j=0; j < num_nodes; j++)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, encoded_entries[j]->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, encoded_entries[j]);
    }
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, encoded_entries);
    if (i != num_nodes)
    {
        for (j=0; j < i; j++)
        {
            retVal = cl_der_DeleteLeafNode(pContext,list_children[j].ChildHandle);
        }
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, list_children);
        if (CSSM_OK != ret)
        {
            return ret;
        }
        else
        {
            return CSSMERR_CL_INTERNAL_ERROR;
        }
        
    }

    /* Create parent node for revoked certificates list */
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_nodes, list_children, &root);

    if(ret != CSSM_OK)
    {

        cl_der_DeleteTree(pContext,root);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, list_children);
        return ret;
    }
    /* Pack the revoked certs struct */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, return_val);

    /* Delete the parse tree */
    cl_der_DeleteTree(pContext,root);

    /* cleanup */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, list_children);

    if (!(*return_val))
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    return ret;
}



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_DerEncodeTbsCertList
*
* Description:
* This function packs a tbsCertList structure into a der encoded memory blob.
*
* If the input tbsCertList is NULL, it will be treated as a valid input with no
* data.  Similarly, if any of the fields of the tbsCertList are null or empty, they
* will be treated as valid fields with no data.
* 
* Parameters: 
* tbsCertList:  The tbsCertList structure to be packed.
*
* return_val (output):The der encoded tbsCertList.
*
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeTbsCertList(CSSM_CLUTIL_CONTEXT *pContext,
                                                  CSSM_HANDLE AppHandle, 
                                                  X509_TBS_CERTLIST_PTR tbsCertList,
                                                  CSSM_DATA_PTR *return_val)
{
    uint32              i = 0;
    uint32              num_types = 7;
    CL_DER_CHILD_INFO   child_array[7];
    CL_DER_NODE_HANDLE  root = 0;
  
    CSSM_RETURN         ret = CSSM_OK, retVal = CSSM_OK;
    CSSM_DATA_PTR       enc_signature = NULL,
                        enc_issuer = NULL,
                        enc_revokedCerts = NULL, 
                        enc_extensions = NULL;
    uint32              thisUpdate_length = 0, 
                        nextUpdate_length = 0;

    if (!return_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *return_val = NULL; 
    /* Check tbsCertList */
    if (!tbsCertList)
    {
        /* tbsCertList may be NULL, signifying a null placeholder */
        /* Replace with an empty structure to be encoded */
        tbsCertList = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(X509_TBS_CERTLIST));
        if (!tbsCertList)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
    }
    

    /* initialize children array */
    port_memset(child_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));

    /** Encode the tbsCertificate **/
    
    /* Call helper encode functions for fields that need them. */
    ret = CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &tbsCertList->signature, &enc_signature);

    if(ret != CSSM_OK)
    {
     if (enc_signature)
     {
         CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);
         CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
     }
     return ret;
    }
    
    ret = CSSM_CLUTIL_X509_DerEncodeName(pContext,AppHandle, &tbsCertList->issuer, &enc_issuer);

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
        if (enc_issuer)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);
        }
        return ret;
    }
    ret = CSSM_CLUTIL_X509_DerEncodeRevokedCertList(pContext,AppHandle, tbsCertList->revokedCertificates, &enc_revokedCerts);

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);
        if(enc_revokedCerts)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);
        }
        return ret;
    }
    
    ret = CSSM_CLUTIL_X509_DerEncodeExtensions(pContext,AppHandle, tbsCertList->extensions.extensions, tbsCertList->extensions.numberOfExtensions, &enc_extensions);


    /* Error Check */
    if (ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
        if(enc_extensions)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
        }
        return ret;
    }

    /** Create child nodes **/

    /* version */
    child_array[0].IsOptional = CSSM_TRUE;
    ret = (tbsCertList->version.Length != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_INTEGER, 
                              tbsCertList->version.Length, tbsCertList->version.Data, &(child_array[0].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_INTEGER, 0, NULL, &(child_array[0].ChildHandle));

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
        return ret;
    }

    /* signature */
    ret = (cl_der_LengthOfContent(enc_signature) != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
                              cl_der_LengthOfContent(enc_signature), 
                              &enc_signature->Data[cl_der_LengthOfTag(enc_signature) 
                                + cl_der_LengthOfLength(enc_signature)], &(child_array[1].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &(child_array[1].ChildHandle));

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
        retVal = cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
        return ret;
    }

    /* issuer */
    ret = (cl_der_LengthOfContent(enc_issuer) != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
                              cl_der_LengthOfContent(enc_issuer), 
                              &enc_issuer->Data[cl_der_LengthOfTag(enc_issuer) 
                                + cl_der_LengthOfLength(enc_issuer)], &(child_array[2].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &(child_array[2].ChildHandle));


    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
        retVal = cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
        if (retVal != CSSM_OK)
        {
            return retVal;
        }
        retVal = cl_der_DeleteLeafNode(pContext,child_array[1].ChildHandle);
        if (retVal != CSSM_OK)
        {
            return retVal;
        }
        return ret;
    }


    /* thisUpdate */
    if (tbsCertList->thisUpdate.time.Data)
    {
        thisUpdate_length = strlen((char *)tbsCertList->thisUpdate.time.Data) + 1;
    }

    {
        CSSM_DATA temp_time;
        temp_time.Data = (uint8*)tbsCertList->thisUpdate.time.Data;
        temp_time.Length = thisUpdate_length;
        if (IsUtcTime(&temp_time) == CSSM_FALSE)
        {
            ret = 
                cl_der_CreateLeafNode(pContext,DER_TAG_GENERALIZED_TIME, 
                                      thisUpdate_length, (uint8*) tbsCertList->thisUpdate.time.Data, &(child_array[3].ChildHandle));
            
            if(ret != CSSM_OK)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[1].ChildHandle);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[2].ChildHandle);
                if(retVal != CSSM_OK) return retVal;
                return ret;
            }
        
        }
        else
        {
            ret =
                cl_der_CreateLeafNode(pContext,DER_TAG_UTC_TIME, 
                                      thisUpdate_length, (uint8*) tbsCertList->thisUpdate.time.Data, &(child_array[3].ChildHandle));
    
            if(ret != CSSM_OK)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[1].ChildHandle);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[2].ChildHandle);
                if(retVal != CSSM_OK)
                {
                    return retVal;
                }
                return ret;
            }
        
        
        }
    }

    /* nextUpdate */
    if (tbsCertList->nextUpdate.time.Data)
    {
        nextUpdate_length = strlen((char *)tbsCertList->nextUpdate.time.Data);
    }

    {
        CSSM_DATA temp_time;
        temp_time.Data = (uint8*)tbsCertList->nextUpdate.time.Data;
        temp_time.Length = nextUpdate_length;
        if (IsUtcTime(&temp_time) == CSSM_FALSE)
        {
            ret = 
                cl_der_CreateLeafNode(pContext,DER_TAG_GENERALIZED_TIME, 
                                      nextUpdate_length, (uint8*) tbsCertList->nextUpdate.time.Data, &(child_array[4].ChildHandle));

            if(ret != CSSM_OK)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
                retVal =cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
                if(retVal != CSSM_OK)
                {
                    return retVal;
                }
                retVal =cl_der_DeleteLeafNode(pContext,child_array[1].ChildHandle);
                if(retVal != CSSM_OK) 
                {
                    return retVal;
                }
                retVal =cl_der_DeleteLeafNode(pContext,child_array[2].ChildHandle);
                if(retVal != CSSM_OK) 
                {
                    return retVal;
                }
                retVal =cl_der_DeleteLeafNode(pContext,child_array[3].ChildHandle);

                return ret;
            }
        
        
        }
        else
        {
            ret =
                cl_der_CreateLeafNode(pContext,DER_TAG_UTC_TIME, 
                                      nextUpdate_length, (uint8*) tbsCertList->nextUpdate.time.Data, &(child_array[4].ChildHandle));
        
        
            if(ret != CSSM_OK)
            {
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
                CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[1].ChildHandle);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[2].ChildHandle);
                retVal = cl_der_DeleteLeafNode(pContext,child_array[3].ChildHandle);
                return ret;
            }
        
        
        }
    }

    /* revokedCertificates */
    child_array[5].IsOptional = CSSM_TRUE;
    ret = (cl_der_LengthOfContent(enc_revokedCerts) != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
                              cl_der_LengthOfContent(enc_revokedCerts), 
                              &enc_revokedCerts->Data[cl_der_LengthOfTag(enc_revokedCerts) 
                                + cl_der_LengthOfLength(enc_revokedCerts)], &(child_array[5].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &(child_array[5].ChildHandle));

    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[1].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[2].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[3].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[4].ChildHandle);
        return ret;
    }


    /* extensions */
    child_array[6].IsOptional = CSSM_TRUE;
    child_array[6].HasCtxSpecificTag = CSSM_TRUE;
    child_array[6].IsExplicit = CSSM_TRUE;
    ret = (cl_der_LengthOfContent(enc_extensions) != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 
                              cl_der_LengthOfContent(enc_extensions), 
                              &enc_extensions->Data[cl_der_LengthOfTag(enc_extensions) 
                                + cl_der_LengthOfLength(enc_extensions)], &(child_array[6].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &(child_array[6].ChildHandle));
 
    if(ret != CSSM_OK)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle,enc_signature->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);      
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);     
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);       
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);        
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[1].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[2].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[3].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[4].ChildHandle);
        retVal =cl_der_DeleteLeafNode(pContext,child_array[5].ChildHandle);
        return ret;
    }   
    



    /* Cleanup */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_signature);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_issuer);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_revokedCerts);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_extensions);

    if (!child_array[0].ChildHandle || !child_array[1].ChildHandle || 
        !child_array[2].ChildHandle || !child_array[3].ChildHandle || 
        !child_array[4].ChildHandle || !child_array[5].ChildHandle || 
        !child_array[6].ChildHandle)
    {
        for (i=0; i < num_types; i++)
        {
            retVal = cl_der_DeleteLeafNode(pContext,child_array[i].ChildHandle);
        }
        return CSSMERR_CL_INTERNAL_ERROR;
        
    }

    /* Create parent node */
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, child_array, &root);

    if(ret != CSSM_OK)
    {
         cl_der_DeleteTree(pContext,root);
         return ret;
    }

    /* Pack the tbsCertList struct into a der encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, return_val);

    /* Delete the parse tree */
    cl_der_DeleteTree(pContext,root);

    if (!(*return_val))
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    /* return */
    return CSSM_OK;
}   



/*-----------------------------------------------------------------------------
* Name: CSSM_CLUTIL_X509_DerEncodeSignedCrl
*
* Description:
* This function packs a signed crl structure into a der encoded memory blob.
*
* If the input crl is NULL, it will be treated as a valid input with no data.
* Similarly, if any of the fields of the crl are null or empty, they will be
* treated as valid fields with no data.
*
* Parameters: 
* crl:      The crl structure to be packed.
* result_val (output): CSSM_DATA_PTR:    The der encoded crl.
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_X509_DerEncodeSignedCrl(CSSM_CLUTIL_CONTEXT *pContext,
                                                CSSM_HANDLE AppHandle, 
                                                X509_SIGNED_CRL_PTR crl, 
                                                CSSM_DATA_PTR *return_val)
{
    
    uint32              num_types = 3;
    CL_DER_CHILD_INFO   child_array[3];
    CL_DER_NODE_HANDLE  root = 0;
    CSSM_DATA_PTR       enc_tbsCertList = NULL,
                        enc_sigalg = NULL;
    CSSM_RETURN         ret = CSSM_OK;
    CSSM_RETURN         ret1 = CSSM_OK,
                        ret2 = CSSM_OK,
                        ret3 = CSSM_OK, retVal = CSSM_OK;
    if (!return_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    *return_val = NULL;
    if (!crl)
    {
        /* crl may be NULL --> just encode a empty structure */
        crl = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(X509_SIGNED_CRL));
        if (!crl)
        {
            return CSSMERR_CL_MEMORY_ERROR;
            
        }
    }
   
    /* Initializations */
    
    port_memset(child_array, 0, num_types*sizeof(CL_DER_CHILD_INFO));

    /*** Encode CRL ***/
    
    /** encode fields **/

    /* tbsCertList */
    ret = CSSM_CLUTIL_X509_DerEncodeTbsCertList(pContext,AppHandle, &crl->tbsCertList, &enc_tbsCertList);
    if (CSSM_OK != ret)
    {
        /* Preserve error code from lower level function */
        return ret;
    }

    /* signature */
    ret = 
        CSSM_CLUTIL_X509_DerEncodeAlgorithmId(pContext,AppHandle, &crl->signature.algorithmIdentifier, &enc_sigalg);
    if (CSSM_OK != ret)
    {
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_tbsCertList->Data);
        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_tbsCertList);
        /* Preserve error code from lower level function */
        return ret;
    }

    /** Create child nodes **/

    /* tbsCertList */
    ret1 = (cl_der_LengthOfContent(enc_tbsCertList) != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE,
                              cl_der_LengthOfContent(enc_tbsCertList),
                              &enc_tbsCertList->Data[cl_der_LengthOfTag(enc_tbsCertList) 
                                + cl_der_LengthOfLength(enc_tbsCertList)], &(child_array[0].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &(child_array[0].ChildHandle));

    /* signature algorithm */
    ret2 = (cl_der_LengthOfContent(enc_sigalg) != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE,
                              cl_der_LengthOfContent(enc_sigalg),
                              &enc_sigalg->Data[cl_der_LengthOfTag(enc_sigalg) 
                                + cl_der_LengthOfLength(enc_sigalg)], &(child_array[1].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_SEQUENCE, 0, NULL, &(child_array[1].ChildHandle));

    /* signature data */
    ret3 = (8*crl->signature.encrypted.Length != 0) ?
        cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING, 
                              8*crl->signature.encrypted.Length,
                              crl->signature.encrypted.Data, &(child_array[2].ChildHandle)) :
        cl_der_CreateLeafNode(pContext,DER_TAG_BIT_STRING, 0, NULL, &(child_array[2].ChildHandle));
    
    /* cleanup */
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_tbsCertList->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_tbsCertList);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_sigalg->Data);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, enc_sigalg);

    if (!child_array[0].ChildHandle || !child_array[1].ChildHandle || 
        !child_array[2].ChildHandle)
    {
        retVal = cl_der_DeleteLeafNode(pContext,child_array[0].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,child_array[1].ChildHandle);
        retVal = cl_der_DeleteLeafNode(pContext,child_array[2].ChildHandle);
        /* Preserve error code from lower level functions */
        if (CSSM_OK != ret1)
        {
            return ret1;
        }
        else if (CSSM_OK != ret2)
        {
            return ret2;
        }
        else
        {
            return ret3;
        }
    }

    /* Create parent node */
    ret = cl_der_CreateParentNode(pContext,DER_TAG_SEQUENCE, num_types, child_array, &root);

    if(ret != CSSM_OK)
    {
          cl_der_DeleteTree(pContext,root);
          return ret;
    }

    /* Pack the crl struct into a der encoded blob */
    ret = cl_der_PackTree(pContext,AppHandle, root, NULL, return_val);

    /* Delete the parse tree */
    cl_der_DeleteTree(pContext,root);

    /* Return */
  
    return ret;
}
                              

/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_DerEncodeCrlGroup
 *
 * Description:
 * This function packs a CSSM CrlGroup structure 
 * into a single DER encoded memory blob.
 *
 * Parameters: 
 * CrlGroupPtr (input) : the CRL group to be DER encoded
 *
 * result_val (output): The DER encoded CRL group
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSM_CLUTIL_DerEncodeCrlGroup        (CSSM_CLUTIL_CONTEXT *pContext,
                                                       CSSM_HANDLE AppHandle, 
                                                    const CSSM_CRLGROUP *CrlGroupPtr,
                                                    CSSM_DATA_PTR *result_val)
{
    CSSM_DATA_PTR           results_ptr = NULL;
    CL_DER_CHILD_INFO_PTR   crl_array = NULL;
    CSSM_DATA_PTR           enc_crl_ptr = NULL;
    CL_DER_NODE_HANDLE      root = 0;
    uint32 i=0, j=0;
    CSSM_RETURN             ret = CSSM_OK, retVal = CSSM_OK;

    if (!result_val)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
  
    /* If there are no crls, return the encoding for a NULL SET */
    if ( !CrlGroupPtr || (!CrlGroupPtr->GroupList.CrlList && !CrlGroupPtr->NumberOfCrls) )
    {
        
        ret = cssm_clutil_EncodeNullSet(pContext,AppHandle, result_val);
        return ret;
    }

    /* Initializations */
    crl_array = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, CrlGroupPtr->NumberOfCrls, sizeof(CL_DER_CHILD_INFO) );
    if (!crl_array)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

    /* For each crl */
    for (i=0; i < CrlGroupPtr->NumberOfCrls; i++)
    {
        enc_crl_ptr = &CrlGroupPtr->GroupList.CrlList[i];

        /* Create a leaf node */
        ret = cl_der_LengthOfContent(enc_crl_ptr) ?
            cl_der_CreateLeafNode  (pContext,DER_TAG_SEQUENCE, 
                                    cl_der_LengthOfContent(enc_crl_ptr),
                                    &enc_crl_ptr->Data[cl_der_LengthOfTag(enc_crl_ptr) 
                                        + cl_der_LengthOfLength(enc_crl_ptr)], &(crl_array[i].ChildHandle)) :
            cl_der_CreateLeafNode  (pContext,DER_TAG_SEQUENCE, 0, NULL, &(crl_array[i].ChildHandle));
                                    
        if (ret != CSSM_OK)
        {
            for(j=0; j<i; j++)
            {
                retVal = cl_der_DeleteLeafNode(pContext,crl_array[j].ChildHandle);
            }
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, crl_array);
            return ret;
            
        }
    }

    ret = cl_der_CreateParentNode(pContext,DER_TAG_SET, CrlGroupPtr->NumberOfCrls, crl_array, &root);
    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, crl_array);

    /* Pack the signatures into a DER encoded blob */
    ret = cl_der_PackTree (pContext,AppHandle, root, NULL, &results_ptr);

    /* Delete the parse tree */
    cl_der_DeleteTree (pContext,root);

    /* Return the result */
    if (CSSM_OK != ret)
    {
        return ret;
    }
    *result_val = results_ptr;
    return CSSM_OK;

}



/*-----------------------------------------------------------------------------
 * Name: CSSM_CLUTIL_X509_RevokedEntryCopy
 *
 * Description:
 * Allocates  a new CRL revoked cert list entry and copies the input data into it. 
 *
 * Parameters: 
 * oldEntry (input) : the input data to be copied
 *
 * Return value:
 * A copy of the revoked entry
 * 
 *---------------------------------------------------------------------------*/
CSSM_X509_REVOKED_CERT_ENTRY_PTR CSSM_CLUTIL_X509_RevokedEntryCopy(const CSSM_CLUTIL_CONTEXT *pContext,
                                                              CSSM_HANDLE AppHandle, 
                                                              CSSM_X509_REVOKED_CERT_ENTRY_PTR oldEntry)
{
    CSSM_X509_REVOKED_CERT_ENTRY_PTR newEntry;
    uint32                      i=0;

    if (!oldEntry)
    {
        return NULL;
    }

   
    newEntry = CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CSSM_X509_REVOKED_CERT_ENTRY));
    if (!newEntry)
    {
        return NULL;
    }

    /* Copy Serial Number */
    newEntry->certificateSerialNumber.Length = oldEntry->certificateSerialNumber.Length;
    if (newEntry->certificateSerialNumber.Length)
    {
        newEntry->certificateSerialNumber.Data = 
            CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, newEntry->certificateSerialNumber.Length);
        if (!newEntry->certificateSerialNumber.Data)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, newEntry);
            return NULL;
        }
        port_memcpy(newEntry->certificateSerialNumber.Data,
                    oldEntry->certificateSerialNumber.Data,
                    newEntry->certificateSerialNumber.Length);
    }
    else
    {
        newEntry->certificateSerialNumber.Data = NULL;
    }

    /* Copy revocationDate */
    if (oldEntry->revocationDate.time.Data)
    {
        newEntry->revocationDate.time.Data = 
            CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, strlen((char *)oldEntry->revocationDate.time.Data) + 1);
        if (!newEntry->revocationDate.time.Data)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, newEntry->certificateSerialNumber.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, newEntry);
            return NULL;
        }
        strcpy((char *)newEntry->revocationDate.time.Data, (char *)oldEntry->revocationDate.time.Data);
        newEntry->revocationDate.time.Length = oldEntry->revocationDate.time.Length;
    }
    else
    {
        newEntry->revocationDate.time.Data = NULL;
        newEntry->revocationDate.time.Length = 0;
    }

    /* Copy extensions */
    newEntry->extensions.numberOfExtensions = oldEntry->extensions.numberOfExtensions;
    if (newEntry->extensions.numberOfExtensions)
    {
        /* allocate new extension list */
        newEntry->extensions.extensions = 
            CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1,
                                         newEntry->extensions.numberOfExtensions * sizeof(CSSM_X509_EXTENSION));
        if (!newEntry->extensions.extensions)
        {
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, newEntry->revocationDate.time.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, newEntry->certificateSerialNumber.Data);
            CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, newEntry);
            return NULL;
        }

        /* copy each extension */
        for (i=0; i < newEntry->extensions.numberOfExtensions; i++)
        {
            /* copy extension ID */
            newEntry->extensions.extensions[i].extnId.Length = 
                oldEntry->extensions.extensions[i].extnId.Length;
            if (newEntry->extensions.extensions[i].extnId.Length)
            {
                newEntry->extensions.extensions[i].extnId.Data = 
                    CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, 
                                                 newEntry->extensions.extensions[i].extnId.Length);
                if (!newEntry->extensions.extensions[i].extnId.Data)
                {
                    CSSM_CLUTIL_X509_FreeExtensions(pContext, 
                                                    AppHandle, 
                                                    newEntry->extensions.extensions, 
                                                    i);
                    return NULL;
                }
                port_memcpy(newEntry->extensions.extensions[i].extnId.Data,
                            oldEntry->extensions.extensions[i].extnId.Data,
                            newEntry->extensions.extensions[i].extnId.Length);
            }
            else
            {
                newEntry->extensions.extensions[i].extnId.Data = NULL;
            }

            /* copy extension critical field */
            newEntry->extensions.extensions[i].critical = 
                oldEntry->extensions.extensions[i].critical;
            /* copy extension type field */
            
            if (oldEntry->extensions.extensions[i].value.tagAndValue)
            {
                newEntry->extensions.extensions[i].value.tagAndValue = 
                CSSM_CALLOC_WRAPPER(pContext->MemFuncs.calloc_func, AppHandle, 1, sizeof(CSSM_X509EXT_TAGandVALUE));
                if (!newEntry->extensions.extensions[i].value.tagAndValue)
                {
                    CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, 
                        newEntry->extensions.extensions[i].extnId.Data);
                    CSSM_CLUTIL_X509_FreeExtensions(pContext,
                        AppHandle, 
                        newEntry->extensions.extensions, 
                        i);
                    return NULL;
                }
                newEntry->extensions.extensions[i].value.tagAndValue->type = 
                    oldEntry->extensions.extensions[i].value.tagAndValue->type;

                /* copy extension value field */
                newEntry->extensions.extensions[i].value.tagAndValue->value.Length = 
                    oldEntry->extensions.extensions[i].value.tagAndValue->value.Length;
                if (newEntry->extensions.extensions[i].value.tagAndValue->value.Length)
                {
                    newEntry->extensions.extensions[i].value.tagAndValue->value.Data =
                        CSSM_MALLOC_WRAPPER(pContext->MemFuncs.malloc_func, AppHandle, 
                            newEntry->extensions.extensions[i].value.tagAndValue->value.Length);
                    if (!newEntry->extensions.extensions[i].value.tagAndValue->value.Data)
                    {
                        CSSM_FREE_WRAPPER(pContext->MemFuncs.free_func, AppHandle, 
                            newEntry->extensions.extensions[i].extnId.Data);
                        CSSM_CLUTIL_X509_FreeExtensions(pContext,
                            AppHandle, 
                            newEntry->extensions.extensions, 
                            i);
                        return NULL;
                    }
                    port_memcpy(newEntry->extensions.extensions[i].value.tagAndValue->value.Data,
                                oldEntry->extensions.extensions[i].value.tagAndValue->value.Data,
                                newEntry->extensions.extensions[i].value.tagAndValue->value.Length);
                }
                else
                {
                    newEntry->extensions.extensions[i].value.tagAndValue->value.Data = NULL;
                }
            }
        }
    }
    else
    {
        newEntry->extensions.extensions = NULL;
    }

    return newEntry;
}



