/*-----------------------------------------------------------------------
 * File: TPCGTOTG.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "maf_include.h"
#include "cssm.h"

#include "spkiutil.h"
#include "addlwords.h"          /* note words not defined by cssmtype.h */
#include "internal.h"

#include "mempool.h"


extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern MEMORY_POOL_PTR poolptr;

static uint32 EstimateNumTuples(CSSM_LIST blob_lists[], uint32 NumLists);

static CSSM_TUPLE_PTR build_tuple_array(CSSM_TP_HANDLE TPHandle,
                                        CSSM_LIST blob_lists[], 
                                        uint32 NumLists, 
                                        uint32 NumTuples);

static CSSM_RETURN cl_hash_lookup(CSSM_TP_HANDLE TPHandle, 
                                  CSSM_LIST *hash, 
                                  CSSM_LIST *public_key);

static void HashToPK(CSSM_TP_HANDLE TPHandle, CSSM_LIST cur_list);

static void create_tuple(CSSM_TP_HANDLE TPHandle, CSSM_TUPLE *t, CSSM_LIST_ELEMENT *x );


static void cplist(CSSM_TP_HANDLE TPHandle, 
                   CSSM_LIST_PTR Dest,
                   CSSM_LIST_PTR Src);

static void free_list(CSSM_TP_HANDLE TPHandle, CSSM_LIST *cl );
/***************************************************************************/
/*  Map a certificate group (sequence...) to a tuplegroup.                 */
/***************************************************************************/

CSSM_RETURN CSSMTPI TP_CertGroupToTupleGroup(
                    CSSM_TP_HANDLE TPHandle,
                    CSSM_CL_HANDLE CLHandle,
                    const CSSM_CERTGROUP *CertGroup,
                    CSSM_TUPLEGROUP_PTR *TupleGroup )
{
    CSSM_TUPLE_PTR tuple_array;
    CSSM_TUPLEGROUP_PTR tuple_group;
    uint32 NumTuples = 0;
    CSSM_LIST *blob_lists;
    Sreader str;
    uint32 i,j;
    CSSM_ENCODED_CERT_PTR cur_cert_ptr;
    CSSM_RETURN ret = CSSM_OK;

    /* Validate Inputs */
    if (!CertGroup)
    {
        return CSSMERR_TP_INVALID_INPUT_POINTER;
    }
    if (!TupleGroup)
    {
        return CSSMERR_TP_INVALID_OUTPUT_POINTER;
    }

    if (CertGroup->CertType != CSSM_CERT_SPKI ||
        CertGroup->CertEncoding != CSSM_CERT_ENCODING_SEXPR ||
        CertGroup->CertGroupType != CSSM_CERTGROUP_ENCODED_CERT ||
        !CertGroup->NumCerts) /* Zero certs in certgroup is deemed an error */
    {
        return CSSMERR_TP_INVALID_CERTGROUP;
    }

    for(i=0;i<CertGroup->NumCerts;i++)
    {
        cur_cert_ptr = &(CertGroup->GroupList.EncodedCertList[i]);
        if (cur_cert_ptr->CertType != CSSM_CERT_SPKI ||
            cur_cert_ptr->CertEncoding != CSSM_CERT_ENCODING_SEXPR)
        {
            /* Error */
            return CSSMERR_TP_INVALID_CERTGROUP;
        }
    }

    hashtab_init();
    /* Create an array of CSSM_LIST structs to store the parsed sexps in */
    blob_lists = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func,
                    sizeof(CSSM_LIST), 
                    CertGroup->NumCerts, 
                    poolptr);

    if (!blob_lists)
    {
        return CSSMERR_TP_MEMORY_ERROR;
    }
    /* Parse each certblob into a CSSM_LIST struct */
    for(i=0;i<CertGroup->NumCerts;i++)
    {
        cur_cert_ptr = &(CertGroup->GroupList.EncodedCertList[i]);
        init_list(&(blob_lists[i]));
        STRR_INIT(str, cur_cert_ptr->CertBlob.Data, cur_cert_ptr->CertBlob.Length);
        ret = cp_sexp(&(blob_lists[i]), &str);
        if (CSSM_OK != ret)
        {
            /* Error */
            for(j=0;j<i;j++)
            {
                drop_list(&(blob_lists[j]));
            }
            CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, blob_lists, poolptr);
            return ret;
        }
    }
    /* Walk all the CSSM_LIST stucts to hash any public key found */
    lpass0(CertGroup->NumCerts, blob_lists);


    /* Modify blob_lists to replace the (hash ...) sublists with 
       (pub-key ...) */
    for(i=0;i<CertGroup->NumCerts;i++)
    {
        HashToPK(TPHandle, blob_lists[i]);
    }
    /* Count the number of tuples to be in the resulting tg */
    NumTuples = EstimateNumTuples(blob_lists, CertGroup->NumCerts);

    if (!NumTuples)
    {
        /* Error */

        for(j=0;j<CertGroup->NumCerts;j++)
        {
            drop_list(&(blob_lists[j]));
        }
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, blob_lists, poolptr);
        return CSSMERR_TP_INVALID_CERTGROUP;
    }
    /* Now we know the number of tuples in resulting tg*/
    /* Build the tuples array to go into the tuplegroup*/
    tuple_array = build_tuple_array(TPHandle, blob_lists, CertGroup->NumCerts, NumTuples);
    if (!tuple_array)
    {
        for(j=0;j<CertGroup->NumCerts;j++)
        {
            drop_list(&(blob_lists[j]));
        }
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, blob_lists, poolptr);
        return CSSMERR_TP_INTERNAL_ERROR;
    }
    tuple_group = App_Calloc(TPHandle, sizeof(CSSM_TUPLEGROUP), 1);
    if (!tuple_group)
    {
        for(j=0;j<CertGroup->NumCerts;j++)
        {
            drop_list(&(blob_lists[j]));
        }
        CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, blob_lists, poolptr);

        for(j=0;j<NumTuples;j++)
        {
            drop_cssm_tuple(&(tuple_array[j]));
        }
        App_Free(TPHandle, tuple_array);

        return CSSMERR_TP_MEMORY_ERROR;
    }
    tuple_group->NumberOfTuples = NumTuples;
    tuple_group->Tuples = tuple_array;
    
    *TupleGroup = tuple_group;
    

    /*Free Local memory used */
    for(j=0;j<CertGroup->NumCerts;j++)
    {
        drop_list(&(blob_lists[j]));
    }
    CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, blob_lists, poolptr);
     
    hashtab_clear();
    dump_free_cache();
    return CSSM_OK;
}

static uint32 EstimateNumTuples(CSSM_LIST blob_lists[], uint32 NumLists)
{
    uint32 i=0;
    uint32 NumTuples = 0;
    for(i=0;i<NumLists;i++)
    {
        CSSM_LIST cur_list = blob_lists[i];
        CSSM_LIST_ELEMENT *w = cur_list.Head;
        CSSM_LIST_ELEMENT *y;
        if (!w )
        {
            return 0;
        }
        switch(w->WordID)
        {
            case CSSM_WORDID_ACL:
                y = w->NextElement;
                while (y != NULL) 
                {          
                    /* count the number of tuples to build */
                    if (y->ElementType == CSSM_LIST_ELEMENT_SUBLIST 
                        && y->WordID == CSSM_WORDID_ENTRY)
                    {
                        NumTuples++ ;
                    }
                    y = y->NextElement ;        /* try the next one */
                } /* while (y) */
                break;
            case CSSM_WORDID_SEQUENCE:
                y = w->Element.Sublist.Head;
                while (y != NULL) 
                {
                    /* count the number of tuples to build */
                    if (y->ElementType == CSSM_LIST_ELEMENT_SUBLIST)
                    {
                        switch (y->WordID) 
                        {
                            case CSSM_WORDID_PUBLIC_KEY:
                            case CSSM_WORDID_ENTRY:/* Should entry be here? FIXME */
                            case CSSM_WORDID_CERT:    
                                NumTuples++ ;   
                            default: 
                                break ;
                        } /* switch */
                    }
                    y = y->NextElement ;        /* try the next one */    
                } /* while (y) */

                break;
            case CSSM_WORDID_CERT:
            case CSSM_WORDID_ENTRY:
                NumTuples++;
                break;

            default:
                break;
        }/* switch (w->WordID)*/
    
    }/*for loop*/
    return NumTuples;
}

static CSSM_TUPLE_PTR build_tuple_array(CSSM_TP_HANDLE TPHandle,
                                        CSSM_LIST blob_lists[], 
                                        uint32 NumLists, 
                                        uint32 NumTuples)
{

    CSSM_TUPLE_PTR tuple_array;
    uint32 i=0;
    uint32 FilledTuples = 0;

    tuple_array = App_Calloc(TPHandle, sizeof(CSSM_TUPLE),NumTuples);
    if (!tuple_array)
    {
        return NULL;
    }

    for(i=0;(i<NumLists && FilledTuples < NumTuples);i++)
    {
        CSSM_LIST cur_list = blob_lists[i];
        CSSM_LIST_ELEMENT *w = cur_list.Head;
        CSSM_LIST_ELEMENT *y;
        if (!w)
        {
            break;
        }
        
        switch(w->WordID)
        {
            case CSSM_WORDID_ACL:

                y = w->NextElement;
                while (y != NULL && FilledTuples<NumTuples) 
                {
                    if (y->ElementType == CSSM_LIST_ELEMENT_SUBLIST
                        && y->WordID == CSSM_WORDID_ENTRY)
                    {
                        
                            create_tuple(TPHandle, &(tuple_array[FilledTuples]), y ) ;
                            FilledTuples++ ;
                    }
                    y = y->NextElement ;        /* try the next one */
                } /* while (y) */
                break;
            case CSSM_WORDID_SEQUENCE:
                y = w->NextElement;
                while (y != NULL  && FilledTuples<NumTuples) 
                {
                    if (y->ElementType == CSSM_LIST_ELEMENT_SUBLIST)
                    {
                        switch (y->WordID) 
                        {
                        case CSSM_WORDID_PUBLIC_KEY: 
                            /* naked public key -- plant as subject */
                            cplist(TPHandle, &(tuple_array[FilledTuples].Subject), &(y->Element.Sublist) ) ;
                            FilledTuples++ ;
                            break ;
                        case CSSM_WORDID_ENTRY:
                        case CSSM_WORDID_CERT:
                            create_tuple(TPHandle,  &(tuple_array[FilledTuples]), y ) ;
                            FilledTuples++ ;
                            break ;
                            
                        default: 
                            break ;
                        } /* switch */
                    }
                    y = y->NextElement ;        /* try the next one */
                } /* while (y) */


                
                break;
            case CSSM_WORDID_CERT:
            case CSSM_WORDID_ENTRY:
                create_tuple(TPHandle, &(tuple_array[FilledTuples]), w ) ;
                FilledTuples++ ;
                break;

            default:
                break;
        }/* switch (w->WordID)*/

    }/*for loop*/
    
    
    if  (i < NumLists)
    {
        /* We broke out of the for loop in error */
        /* Tuple Overflow*/
        uint32 j = 0;
        for(j=0;j<FilledTuples;j++)
        {
            drop_cssm_tuple(&(tuple_array[j]));
        }
        App_Free(TPHandle, tuple_array);
        return NULL;
    }

    return tuple_array;
}


static CSSM_RETURN cl_hash_lookup(CSSM_TP_HANDLE TPHandle, 
                                  CSSM_LIST *hash, 
                                  CSSM_LIST *public_key)
{
    pSexp hash_s;
    pRenlists pRl;
    pSexp public_key_s;

    hash_s = cl_to_sexp(hash);
    if (hash_lookup(hash_s, &pRl) || !pRl)
        /* Get back either PK if hash exists in table*/
        /* or get back hash if hash does not exist in table */
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }
    public_key_s = pRl->obj;
    return sexp_to_cl(TPHandle, public_key, public_key_s);

}

static void HashToPK(CSSM_TP_HANDLE TPHandle, CSSM_LIST cur_list)
{
    
    CSSM_LIST_PTR pub_key_list_ptr;
    CSSM_LIST_ELEMENT *w = cur_list.Head;


    if (!w)
    {
        return;
    }

    if (w->ElementType == CSSM_LIST_ELEMENT_SUBLIST &&
        w->WordID == CSSM_WORDID_HASH)
    {
        /* Termination case for recursion */
        pub_key_list_ptr = CSSM_CALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.calloc_func,
                            sizeof(CSSM_LIST),
                            1,
                            poolptr);
        if (!pub_key_list_ptr)
        {
            return;
        }

        cl_hash_lookup(TPHandle, &cur_list, pub_key_list_ptr);
        /* Update NextElement */
        if (!pub_key_list_ptr->Head)
        {
            CSSM_FREE_WRAPPER(memory_pool_funcs.cssm_memory_funcs.free_func, pub_key_list_ptr, poolptr);
            return ; /* error */
        }
        pub_key_list_ptr->Head->NextElement = w->NextElement;
        w->NextElement = NULL;

        if (cur_list.Tail == w)
        {
            /* cur_list has only 1 element. Update Tail also */
            cur_list.Tail = pub_key_list_ptr->Head;
        }
        cur_list.Head = pub_key_list_ptr->Head;

        /* make pub_key_list point to w for freeing purposes */
        pub_key_list_ptr->Head = w;
        pub_key_list_ptr->Tail = w;
        drop_list(pub_key_list_ptr);
        return;
    }

    while (w)
    {
        if (w->ElementType == CSSM_LIST_ELEMENT_SUBLIST)
        {
            HashToPK(TPHandle, w->Element.Sublist);
     
        }/* if (w->ElementType == CSSM_LIST_ELEMENT_SUBLIST)*/
        w = w->NextElement;
    }
    return ;
}


/***************************************************************************/
/* create_tuple                                                            */
/*                                                                         */
/* Fill a CSSM_TUPLE from the parsed S-expression x.                       */
/* Move lists from there to the tuple, for the list entries.               */
/***************************************************************************/
/* Similar to fill_tuple in clutil.c, but creates copies instead of moving */
static void create_tuple(CSSM_TP_HANDLE TPHandle, CSSM_TUPLE *t, CSSM_LIST_ELEMENT *x )
{
  while (x != NULL) {
    if (x->ElementType == CSSM_LIST_ELEMENT_SUBLIST)
      switch (x->WordID) {
      case CSSM_WORDID_ENTRY:
      case CSSM_WORDID_CERT:
        /* walk into these structures to find the tuple */
        create_tuple(TPHandle, t, x->Element.Sublist.Head ) ;
        return ;

      case CSSM_WORDID_ISSUER:
        cplist(TPHandle, &(t->Issuer), &(x->Element.Sublist) ) ;
        break ;

      case CSSM_WORDID_SUBJECT:
        cplist(TPHandle, &(t->Subject), &(x->Element.Sublist) ) ;
        break ;

      case CSSM_WORDID_PROPAGATE:
        t->Delegate = CSSM_TRUE ;
        break ;

      case CSSM_WORDID_TAG:
        cplist(TPHandle, &(t->AuthorizationTag), &(x->Element.Sublist) ) ;
        break ;

      case CSSM_WORDID_VALID:
        cplist(TPHandle, &(t->ValidityPeriod), &(x->Element.Sublist) ) ;
        break ;

      default: break ;
      } /* switch (x) */
    x = x->NextElement ;
  } /* while (x) */
} /* create_tuple */

static void cplist(CSSM_TP_HANDLE TPHandle, 
                   CSSM_LIST_PTR Dest,
                   CSSM_LIST_PTR Src)
{
    CSSM_LIST_ELEMENT_PTR dw, sw;

    if (!Dest || !Src)
    {
        return;
    }

    

    Dest->ListType = Src->ListType;


    sw = Src->Head;

    /* Set up first element in the list */
    dw = App_Malloc(TPHandle, sizeof( CSSM_LIST_ELEMENT ));
    if (!dw)
    {
        return;
    }
    dw->NextElement = NULL;
    dw->WordID = sw->WordID;
    dw->ElementType = sw->ElementType;
    if (dw->ElementType == CSSM_LIST_ELEMENT_SUBLIST)
    {
        cplist(TPHandle, &(dw->Element.Sublist), &(sw->Element.Sublist));
    }
    else
    {
        if (sw->Element.Word.Length)
        {
            dw->Element.Word.Length = sw->Element.Word.Length;
            dw->Element.Word.Data = App_Malloc(TPHandle, dw->Element.Word.Length);
            if (!dw->Element.Word.Data)
            {
                free_list(TPHandle, Dest);
                App_Free(TPHandle, dw);
            }
            port_memcpy(dw->Element.Word.Data, 
                        sw->Element.Word.Data,
                        dw->Element.Word.Length);
        }
    }

    Dest->Head = dw;
    Dest->Tail = dw;
    sw = sw->NextElement;

    while (sw)
    {
        /* copy list element */
        dw = App_Malloc(TPHandle, sizeof( CSSM_LIST_ELEMENT ));;
        if (!dw)
        {
            free_list(TPHandle, Dest);
        }
        dw->NextElement = NULL;
        dw->WordID = sw->WordID;
        dw->ElementType = sw->ElementType;
        if (dw->ElementType == CSSM_LIST_ELEMENT_SUBLIST)
        {
            cplist(TPHandle, &(dw->Element.Sublist), &(sw->Element.Sublist));
        }
        else
        {
            if (sw->Element.Word.Length)
            {
                dw->Element.Word.Length = sw->Element.Word.Length;
                dw->Element.Word.Data = App_Malloc(TPHandle, dw->Element.Word.Length);
                if (!dw->Element.Word.Data)
                {
                    free_list(TPHandle, Dest);
                    App_Free(TPHandle, dw);
                }
                port_memcpy(dw->Element.Word.Data, 
                            sw->Element.Word.Data,
                            dw->Element.Word.Length);
            }
        }

        Dest->Tail->NextElement = dw;
        Dest->Tail = dw;
        sw = sw->NextElement;
    }
    return;
}


/***************************************************************************/
/* free_list                                                               */
/*                                                                         */
/* Drop a CSSM_LIST  from the App_* space                                  */
/***************************************************************************/

static void free_list(CSSM_TP_HANDLE TPHandle, CSSM_LIST *cl ) 
{
  CSSM_LIST_ELEMENT *le, *nxt ; /* walkers of the list */

  if (cl == NULL) return ;      /* no list, no work to do */
  if (cl->ListType != CSSM_LIST_TYPE_SEXPR) return ; 
  nxt = cl->Head ;              /* the next node to drop */
  cl->Head = NULL ; 
  cl->Tail = NULL ; 
  while ((le = nxt) != NULL) {  /* while there are more nodes to drop */
    nxt = le->NextElement ;     /* note the next element before dropping le */
    switch (le->ElementType) {
    case CSSM_LIST_ELEMENT_DATUM: 
      App_Free(TPHandle, le->Element.Word.Data ) ; 
      break ; 

    case CSSM_LIST_ELEMENT_SUBLIST: 
      free_list(TPHandle, &(le->Element.Sublist) ) ; 
      break ; 

    default: break ;            /* no work to do */
    } /* switch */
    App_Free(TPHandle, le) ;                /* free this node */
    le = NULL ;                 /* and set it to NULL */
  } /* while (le) */
  return ;
} /* free_list */
