/*-----------------------------------------------------------------------
 * File: CRTGRPFR.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "maf_include.h"

/* Windows & CSSM includes */
#ifndef VMS
#pragma warning (disable:4201 4214 4115)
#include <stdio.h>
#pragma warning (default:4201 4214 4115)
#else
#include <stdio.h>
#endif
#include "cssm.h"
#include "cssmspi.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"

/* Certificate-specific includes */
#include "x509defs.h"
#include "x_fndefs.h"

#include "oidsx509.h"
#include "oidsalg.h"
#include "x509hlpr.h"
#include "clutil.h"
#include "x_qtable.h"
#include "internal.h"

#include "api_shim.h"
#include "maf_include.h"

#include <cssmMemTrack.h>

#include "ber_der.h"
#include "pkcs7oid.h"
#include "mempool.h"
#ifndef VMS
#pragma warning(disable: 4090) /* This disables the warnings that are generated
                                when CSSM_DATA::Data <- BER_PARSED_ITEM::Content
                                assignment takes place */
#endif
#define X509_FORMAT_VERSION 2

extern const CSSM_API_MEMORY_FUNCS Addin_APIMemFuncs;
extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern BER_MEMORY_FUNCS berder_funcs_wrapper;
extern PORT_TLS_INDEX TLS_MemoryPool;

/* Types for BER_DER typed parsing */
extern DER_COMPLEX_TYPE BER_SignedPKCS7Type;
extern DER_COMPLEX_TYPE BER_PKCS7Type;
extern DER_COMPLEX_TYPE BER_SignedDataTypeWrapper;
extern DER_COMPLEX_TYPE BER_SignedDataType;
extern DER_COMPLEX_TYPE BER_SetOfCertificatesType;
extern DER_COMPLEX_TYPE BER_PKCS7CertNamesOnlyType;
extern DER_COMPLEX_TYPE BER_PKCS7CertNamesOnlySignedPartType;
extern DER_COMPLEX_TYPE setOfSignerInfoType;
extern DER_COMPLEX_TYPE signerInfoType;
extern DER_COMPLEX_TYPE DER_IssuerandSerialNumberType;

static CSSM_RETURN VerifySignatureAndCreatePKCS7ParseTree(MEMORY_POOL_PTR poolptr, 
                                                          CSSM_CL_HANDLE CLHandle,
                                                          CSSM_CC_HANDLE CCHandle,
                                                          const CSSM_DATA *SignerCert,
                                                          BER_PARSED_ITEM *itemptr, 
                                                          DER_NODE_PTR *result_val);

static CSSM_RETURN ExtractCertGroup(CSSM_CL_HANDLE CLHandle, 
                                    DER_NODE_PTR SetOfCerts_node, 
                                    CSSM_CERTGROUP_PTR *result_val);

static uint32 CertIndexSignerInfo(DER_NODE_PTR SignedData_node, DER_NODE_PTR SetOfCerts_node);

static CSSM_BOOL IsEquivBerItem (BER_PARSED_ITEM item1,BER_PARSED_ITEM item2);

static void ReorderCertBySignerInfo(CSSM_CERTGROUP_PTR certgroup_ptr, uint32 swap_index);

CSSM_ALGORITHMS CSSMAPI X509_AlgorithmOidToAlgId (CSSM_OID_PTR Oid);

CSSM_RETURN 
GetCCHandle(MEMORY_POOL_PTR poolptr,
            CSSM_CL_HANDLE CLHandle, 
            CSSM_CC_HANDLE CCHandle, 
            const CSSM_DATA *SignerCert,
            CSSM_ALGORITHMS verifyAlg,
            CSSM_CSP_HANDLE *NewCspHandle,
            CSSM_SUBSERVICE_UID_PTR *SsUidp,
            CSSM_CC_HANDLE *result_val);

void
FreeCCHandle(CSSM_CL_HANDLE CLHandle,
                CSSM_CC_HANDLE CCHandle, 
                CSSM_CC_HANDLE NewCCHandle, 
                CSSM_CSP_HANDLE NewCspHandle,
                CSSM_SUBSERVICE_UID *SsUidp);

/*-----------------------------------------------------------------------------
 * Name: CL_CertGroupFromVerifiedBundle
 *
 * Description:
 * This function accepts as input a certificate bundle
 * (a codified and signed aggregation of
 * the certificates in the group), verifies the signature
 * of the bundle (if a signature is present)
 * and returns a certificate group (as an array of
 * individual certificates) including every
 * certificate contained in the bundle. The signature
 * on the certificate aggregate is verified using the
 * cryptographic context and possibly using the input 
 * signer certificate. The CL module embeds the knowledge
 * of the verification scope for the bundle types that it
 * supports.  A CL module's supported bundle types and
 * encodings are available to applications by querying
 * the CSSM registry.  The type and encoding of the
 * certificate bundle must be specified with the input
 * bundle. If signature verification is successful, the
 * certificate aggregate will be parsed into a certificate
 * group whose order corresponds to the certificate
 * aggregate ordering.  This certificate group will then
 * be returned to the calling application.
 * 
 * Parameters: 
 * CCHandle (input)          : Handle identifying the Cryptographic Context 
 *                             to use for verification
 * CertBundle (input) : The certificate bundle to be verified and from which the returned certificate group is extracted
 * SignerCert (input)        : The cert which was used for signing
 * CertGroup (input)        : The cert group extracted from the input bundle
 *
 * Return value:
 * An indicator of whether the function (including verification) was successful
 *

Signed PKCS7 Tree Structure:
===========================
                        Signed_PKCS7_Sturcture
                        ___________|_________________
                        |             |             |
                  PKCS7_Structure   AlgOID      Signature
                ________|________
                |               |
                OID           Signed_Data_Wrapper
           (Signed_Data OID)            |
                                    Signed_Data
    ____________________________________|__________________________________________
      |             |                |                |               |           |
    Version    DigestAlgIDs     ContentInfo       CertGroup       CrlGroup    SignerInfo                   
  (INTEGER:1)  (NULL SET)    (NULL: 06 00 05 00)  (Enc Certs)       (NULL)      (NULL)
 *---------------------------------------------------------------------------*/
CSSM_RETURN CSSMCLI CL_CertGroupFromVerifiedBundle
                        (CSSM_CL_HANDLE CLHandle,
                         CSSM_CC_HANDLE CCHandle,
                         const CSSM_CERT_BUNDLE *CertBundle,
                         const CSSM_DATA *SignerCert,
                         CSSM_CERTGROUP_PTR *CertGroup)
{
    
    
    CSSM_CERTGROUP_PTR    certgroup_ptr = NULL;
    CSSM_RETURN     ret = CSSM_OK;
    BER_PARSED_ITEM item;
    const uint8 *tail;
    DER_NODE_PTR tree, pkcs7node;
    DER_NODE_PTR SignedDataWrapper_node, SignedData_node, SetOfCerts_node;


    MEMORY_POOL pool;
    

    /* Check inputs */
    /* CLHandle and CCHandle are checked by CSSM */
    if (!CertBundle || 
        CSSM_CLUTIL_IsBadCssmDataPtr(&CertBundle->Bundle) ||
        !CertBundle->Bundle.Length || 
        !CertBundle->Bundle.Data)
    {
        return CSSMERR_CL_INVALID_BUNDLE_POINTER ;
        
    }
    if (CertBundle->BundleHeader.BundleType != CSSM_CERT_BUNDLE_PKCS7_SIGNED_DATA )
    {
        return CSSMERR_CL_INVALID_BUNDLE_INFO;
        
    }
    if (CertBundle->BundleHeader.BundleEncoding != CSSM_CERT_BUNDLE_ENCODING_DER )
    {
        return CSSMERR_CL_INVALID_BUNDLE_INFO;
        
    }
    if (SignerCert && 
        CSSM_CLUTIL_IsBadCssmDataPtr(SignerCert) )
    {
        return CSSMERR_CL_INVALID_CERT_POINTER;
        
    }
    if (!CertGroup || 
        port_IsBadWritePtr(CertGroup, sizeof(CSSM_CERTGROUP_PTR)) )
    {
        return CSSMERR_CL_INVALID_CERTGROUP_POINTER;
        
    }
    
    /* Init result value */
    *CertGroup = NULL;

    /* Setup the Memory pool */
    ret = (*memory_pool_funcs.init)((CSSM_MEMORY_FUNCS_PTR)&Addin_APIMemFuncs, NULL, &pool);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, &pool);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* parse the PKCS 7 bundle and obtain necessary parts */


    /* First construct a BER_PARSED_ITEM for further use */
    tail = BER_ExpandItem((uint8*) CertBundle->Bundle.Data, CertBundle->Bundle.Length, &item);
    if (!tail)
    {
        (*memory_pool_funcs.recycle)(&pool);
		/* CertBundle Data or Length (format as a whole) is invalid. */
		return CSSMERR_CL_UNKNOWN_FORMAT;    
    }


    /* Now, Generate Parse tree for PKCS7 object */
    /* This depends on whether we have a signature or not. If we have a signature, we use the BER_SignedPKCS7Type to
    do the parsing. If not, we use the BER_PKCS7Type to do the parsing */

    if (CCHandle || SignerCert)
    {
        /* Signature present and must be verified. Use BER_SignedPKCS7Type.*/
        
        ret = VerifySignatureAndCreatePKCS7ParseTree(&pool, 
                                                     CLHandle,
                                                     CCHandle, 
                                                     SignerCert, 
                                                     &item, 
                                                     &pkcs7node);
        if (CSSM_OK != ret)
        {
            (*memory_pool_funcs.recycle)(&pool);
            return ret;
        }

    }
    else
    {
        /* No signature verification to be done */
        tree = BER_ParseTypedObject(TLS_MemoryPool,
            &berder_funcs_wrapper, 
            &BER_PKCS7Type, &item);
        if (!tree)
        {
            tree = BER_ParseTypedObject(TLS_MemoryPool,
            &berder_funcs_wrapper, 
            &BER_SignedPKCS7Type, &item);
            if (tree)
            {
                /* we got a signed bundle with no means of verifying the signature */
                /* lets claim its an invalid signer cert although what we're really 
                saying is that both CCHandle and SignerCert are zero */
                (*memory_pool_funcs.recycle)(&pool);
                return CSSMERR_CL_INVALID_CERT_POINTER;
            }
            else
            {
                /* we have a format that we cant handle */
                (*memory_pool_funcs.recycle)(&pool);
                return CSSMERR_CL_UNKNOWN_FORMAT;
            }
        }
        pkcs7node = tree;
    }

    /* Extract Enc PKCS7 data */
    if  (!pkcs7node || pkcs7node->Type != &BER_PKCS7Type || !pkcs7node->Child) 
    {
        (*memory_pool_funcs.recycle)(&pool);
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }

    /* PKCS7 node has 2 children: [OID, Signed_data_wrapper] */
    SignedDataWrapper_node = pkcs7node->Child[1].X.Node;
    if (!SignedDataWrapper_node || SignedDataWrapper_node->Type != &BER_SignedDataTypeWrapper 
        || !SignedDataWrapper_node->Child || !SignedDataWrapper_node->Count)
    {
        (*memory_pool_funcs.recycle)(&pool);
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    
    /* Signed_data_wrapper has one child - [signed_data] */
    SignedData_node = SignedDataWrapper_node->Child[0].X.Node;
    if (!SignedData_node || SignedData_node->Type != &BER_SignedDataType
        || !SignedData_node->Child || !SignedData_node->Count)
    {
        (*memory_pool_funcs.recycle)(&pool);
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    
    
    /* this signed_data node has 6 children: 
            [version, algid, contentinfo, certs, crls, signerinfo] */
    SetOfCerts_node = SignedData_node->Child[3].X.Node;
    if (!SetOfCerts_node || SetOfCerts_node->Type != &BER_SetOfCertificatesType 
        || !SetOfCerts_node->Child || !SetOfCerts_node->Count)
    { 
        (*memory_pool_funcs.recycle)(&pool);
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }

    ret = ExtractCertGroup(CLHandle, SetOfCerts_node, &certgroup_ptr);
    if (CSSM_OK != ret)
    {
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }
    
    /* Extract the serial number of the First Signer Info. We need to reorder the 
    certs (if necessary) so that the first cert has the same serial number of the 
    first signer info  */
    ReorderCertBySignerInfo(certgroup_ptr,CertIndexSignerInfo(SignedData_node,SetOfCerts_node));
    
    /* Set the outputs and return */

    (*memory_pool_funcs.recycle)(&pool);
    *CertGroup = certgroup_ptr;
    return CSSM_OK;
}

static CSSM_RETURN VerifySignatureAndCreatePKCS7ParseTree(MEMORY_POOL_PTR poolptr, 
                                                          CSSM_CL_HANDLE CLHandle,
                                                          CSSM_CC_HANDLE CCHandle,
                                                          const CSSM_DATA *SignerCert,
                                                          BER_PARSED_ITEM *itemptr, 
                                                          DER_NODE_PTR *result_val)
{
    DER_NODE_PTR pkcs7node, algid_node, signaturenode, tree;
    CSSM_DATA_PTR enc_pkcs7_ptr, signature;
    CSSM_OID algoid;
    CSSM_RETURN ret=CSSM_OK;
    BER_PARSED_ITEM pkcs7item, sig_item, algoid_item;
    CSSM_ALGORITHMS DigestAlgorithm;
    CSSM_CC_HANDLE          useCCHandle = 0;
    CSSM_CSP_HANDLE         newCspHandle = 0;
    CSSM_SUBSERVICE_UID_PTR ssUidp = NULL;

    tree = BER_ParseTypedObject(TLS_MemoryPool,
    &berder_funcs_wrapper, 
    &BER_SignedPKCS7Type, itemptr);

    if (!tree || !tree->Child || tree->Count != 3)
    {
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    /* BER_SignedPKCS7Type has 3 children: [BER_PKCS7Type, AlgID, Signature]. So the tree is in child[0] */
    pkcs7node = tree->Child[0].X.Node;
    /* Input is the simple representation of the subordinate BER component: In this case PKCS7Type */
    pkcs7item = tree->Child[0].X.Input;

    if (!pkcs7node || pkcs7node->Type != &BER_PKCS7Type)
    {
        
        return CSSMERR_CL_UNKNOWN_FORMAT;
    }
    
    /* Build a CSSM_DATA structure with the encoded PKCS7 structure to verify signature */
    enc_pkcs7_ptr = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
            sizeof(CSSM_DATA), poolptr);
    
    if (!enc_pkcs7_ptr)
    {
        
        return CSSMERR_CL_MEMORY_ERROR;
    }

    enc_pkcs7_ptr->Length = BER_LengthOfTag(pkcs7item.Tag) 
                        + BER_LengthOfLength(pkcs7item.ContentLength) 
                        + pkcs7item.ContentLength;
    enc_pkcs7_ptr->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
            enc_pkcs7_ptr->Length, poolptr);
    if (!enc_pkcs7_ptr->Data)
    {
        
        return CSSMERR_CL_MEMORY_ERROR;
    }
    port_memcpy(enc_pkcs7_ptr->Data, pkcs7item.Tag, BER_LengthOfTag(pkcs7item.Tag));
    BER_OutputLength(pkcs7item.ContentLength, enc_pkcs7_ptr->Data + BER_LengthOfTag(pkcs7item.Tag));
    port_memcpy(enc_pkcs7_ptr->Data + BER_LengthOfTag(pkcs7item.Tag) + BER_LengthOfLength(pkcs7item.ContentLength), 
        pkcs7item.Content, pkcs7item.ContentLength);


    /* Now get the Algid.
       To do this we need the Algid which is the 2nd element of the type */
    /* BER_SignedPKCS7Type = [BER_PKCS7Type, AlgID, Signature]. So the ALGID is in child[1] */
    algid_node = tree->Child[1].X.Node;
    /* ALGID had algoid and parameteres as children */
    algoid_item = algid_node->Child[0].X.Input;
    algoid.Length = algoid_item.ContentLength;
    algoid.Data = algoid_item.Content;

    DigestAlgorithm = X509_AlgorithmOidToAlgId(&algoid);
    /* Now get the signature into a CSSM_DATA */
    /* BER_SignedPKCS7Type = [BER_PKCS7Type, AlgID, Signature]. So the signature is in child[2] */
    signaturenode = tree->Child[2].X.Node;
    /* Input is the simple representation of the subordinate BER component: In this case the signature */
    sig_item = tree->Child[2].X.Input;

    signature = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
            sizeof(CSSM_DATA), poolptr);
    if (!signature )
    {
     
        return CSSMERR_CL_MEMORY_ERROR;
    }
    signature->Length = sig_item.ContentLength - 1;
    signature->Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
            signature->Length, poolptr);

    if (!signature->Data )
    {
     
        return CSSMERR_CL_MEMORY_ERROR;
    }
    port_memcpy(signature->Data, sig_item.Content+1, signature->Length); /* first byte of encoding is ignored.
                                                                            See the ASN.1 rules. */

    /* Use SignerCert for verification if specified. */
    if (SignerCert)
    {
        ret = GetCCHandle(poolptr, CLHandle, CCHandle, SignerCert, DigestAlgorithm, &newCspHandle, &ssUidp, &useCCHandle);
        if (CSSM_OK != ret)
        {
            FreeCCHandle(CLHandle, CCHandle, useCCHandle, newCspHandle, ssUidp);
            return ret;
        }
        
    }
    else /* else use CCHandle */
    {
        useCCHandle = CCHandle;
    }



    /* Check the signature */
    ret = addin_VerifyData (useCCHandle, enc_pkcs7_ptr, 1, CSSM_ALGID_NONE, signature);
    *result_val = pkcs7node;
    FreeCCHandle(CLHandle, CCHandle, useCCHandle, newCspHandle, ssUidp);
    return ret;
}

static CSSM_RETURN ExtractCertGroup(CSSM_CL_HANDLE CLHandle, DER_NODE_PTR SetOfCerts_node, CSSM_CERTGROUP_PTR *result_val)
{
    CSSM_CERTGROUP_PTR    certgroup_ptr = NULL;
    DER_NODE_PTR current_cert_node;
    BER_PARSED_ITEM current_cert_item;

    uint32 i,j;

    /* Allocate a CertGroup structure */
    certgroup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_CERTGROUP));
    if (!certgroup_ptr)
    {
        
        return CSSMERR_CL_MEMORY_ERROR;
    }

    /* SetOfCerts_node->Count contains the number of certificates */
    certgroup_ptr->NumCerts = SetOfCerts_node->Count;
    certgroup_ptr->GroupList.CertList = 
        App_Calloc(CLHandle, certgroup_ptr->NumCerts, sizeof(CSSM_DATA));
    if (!certgroup_ptr->GroupList.CertList)
    {
        
        App_Free(CLHandle, certgroup_ptr);
        return CSSMERR_CL_MEMORY_ERROR;
    }

    /* 
        Copy each of the certificates from the SetOfCerts_node->Child array.
        Remember that SetOfCerts_node->Child[i].X.Input contains the BER_PARSED_ITEM
        structure for the certificate #i
    */

    for (i = 0; i < certgroup_ptr->NumCerts; i++)
    {
        current_cert_item = SetOfCerts_node->Child[i].X.Input;
        current_cert_node = SetOfCerts_node->Child[i].X.Node;
        certgroup_ptr->GroupList.CertList[i].Length = BER_LengthOfTag(current_cert_item.Tag) 
            + BER_LengthOfLength(current_cert_item.ContentLength) 
            + current_cert_item.ContentLength;
        
        if (certgroup_ptr->GroupList.CertList[i].Length && current_cert_node->Type == &BER_PKCS7CertNamesOnlyType)
        {
            certgroup_ptr->GroupList.CertList[i].Data = 
                App_Calloc(CLHandle, 1, certgroup_ptr->GroupList.CertList[i].Length);

            if (!certgroup_ptr->GroupList.CertList[i].Data)
            {
                for (j=0; j < i; j++)
                    App_Free(CLHandle, certgroup_ptr->GroupList.CertList[j].Data);
                App_Free(CLHandle, certgroup_ptr->GroupList.CertList);
                App_Free(CLHandle, certgroup_ptr);
                return CSSMERR_CL_MEMORY_ERROR;
            
            }
            port_memcpy(certgroup_ptr->GroupList.CertList[i].Data, current_cert_item.Tag, BER_LengthOfTag(current_cert_item.Tag));
            BER_OutputLength(current_cert_item.ContentLength, certgroup_ptr->GroupList.CertList[i].Data + BER_LengthOfTag(current_cert_item.Tag));
            port_memcpy(certgroup_ptr->GroupList.CertList[i].Data + BER_LengthOfTag(current_cert_item.Tag) + BER_LengthOfLength(current_cert_item.ContentLength), current_cert_item.Content, current_cert_item.ContentLength);
        }
    }
    *result_val = certgroup_ptr;
    return CSSM_OK;

}

static CSSM_BOOL IsEquivBerItem (BER_PARSED_ITEM item1,BER_PARSED_ITEM item2)
{
    
    if (*(item1.Tag) != *(item2.Tag))
        return CSSM_FALSE;
    if (item1.ContentLength != item2.ContentLength)
        return CSSM_FALSE;
    if (port_memcmp(item1.Content,item2.Content,item1.ContentLength))
        return CSSM_FALSE;
    return CSSM_TRUE;
}
/* Returns the index of the certificate that has the same issuer and serial num
as the first signer info */
static uint32 CertIndexSignerInfo(DER_NODE_PTR SignedData_node, DER_NODE_PTR SetOfCerts_node)
{
    DER_NODE_PTR SetOfSignerInfo_node, SignerInfo_node, IssuerAndSerialNumber_node;
    BER_PARSED_ITEM IssuerName, SerialNumber;
    BER_PARSED_ITEM Cert_IssuerName, Cert_SerialNumber;
    DER_NODE_PTR current_cert_node, current_tbs_cert_node;
    uint32 i;

    SetOfSignerInfo_node = SignedData_node->Child[5].X.Node;
    if (!SetOfSignerInfo_node || SetOfSignerInfo_node->Type != &setOfSignerInfoType 
        || !SetOfSignerInfo_node->Child || !SetOfSignerInfo_node->Count)
    {
        return 0;
    }
    
    /* Take first signer info for processing */
    SignerInfo_node = SetOfSignerInfo_node->Child[0].X.Node;
    if (!SignerInfo_node || SignerInfo_node->Type != &signerInfoType 
        || !SignerInfo_node->Child || !SignerInfo_node->Count)
    {
        return 0;
    }
    
    IssuerAndSerialNumber_node = SignerInfo_node->Child[1].X.Node;
    if (!IssuerAndSerialNumber_node || IssuerAndSerialNumber_node->Type != &DER_IssuerandSerialNumberType
        || !IssuerAndSerialNumber_node->Child || !IssuerAndSerialNumber_node->Count)
    {
        return 0;
    }
    IssuerName = IssuerAndSerialNumber_node->Child[0].X.Input;
    SerialNumber = IssuerAndSerialNumber_node->Child[1].X.Input;

    for(i=0;i<(uint32)SetOfCerts_node->Count;i++)
    {
        current_cert_node = SetOfCerts_node->Child[i].X.Node;
        if (!current_cert_node  || current_cert_node ->Type != &BER_PKCS7CertNamesOnlyType 
        || !current_cert_node->Child || !current_cert_node->Count)
        {
            return 0;
        }
        current_tbs_cert_node = current_cert_node->Child[0].X.Node;
        if (!current_tbs_cert_node  || current_tbs_cert_node ->Type != &BER_PKCS7CertNamesOnlySignedPartType
        || !current_tbs_cert_node->Child || !current_tbs_cert_node->Count)
        {
            return 0;
        }
        Cert_IssuerName = current_tbs_cert_node->Child[3].X.Input;
        Cert_SerialNumber = current_tbs_cert_node->Child[1].X.Input;
        if (IsEquivBerItem(IssuerName, Cert_IssuerName) &&
            IsEquivBerItem(SerialNumber, Cert_SerialNumber))
            return i;
    }
    return 0;
}

/* Swap Certs in 0 and swap_index positions in certgroup_ptr*/
static void ReorderCertBySignerInfo(CSSM_CERTGROUP_PTR certgroup_ptr, uint32 swap_index)
{
    CSSM_DATA tmpcert;

    if (!swap_index || swap_index >= certgroup_ptr->NumCerts)
        return;
    tmpcert = certgroup_ptr->GroupList.CertList[swap_index];
    certgroup_ptr->GroupList.CertList[swap_index] = certgroup_ptr->GroupList.CertList[0];
    certgroup_ptr->GroupList.CertList[0] = tmpcert;
    return;
}
