/*-----------------------------------------------------------------------
 * File: CRTGRPTO.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

/* Windows & CSSM includes */
#ifndef VMS
#pragma warning (disable:4201 4214 4115)
#include <stdio.h>
#pragma warning (default:4201 4214 4115)
#else
#include <stdio.h>
#endif
#include "cssm.h"
#include "cssmspi.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"

/* Certificate-specific includes */
#include "x509defs.h"
#include "x_fndefs.h"

#include "oidsx509.h"
#include "oidsalg.h"
#include "x509hlpr.h"
#include "clutil.h"
#include "x_qtable.h"
#include "internal.h"

#include "api_shim.h"
#include "maf_include.h"

#include "algoid.h"
#include <cssmMemTrack.h>
#include "pkcs7oid.h"

#include "clberder.h"

extern const CSSM_API_MEMORY_FUNCS Addin_APIMemFuncs;
extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern PORT_TLS_INDEX TLS_MemoryPool;

#define X509_FORMAT_VERSION 2
extern BER_MEMORY_FUNCS berder_funcs_wrapper;

/* Types for BER/DER parsing */
extern DER_COMPLEX_STRUCT BER_PKCS7Type;
extern DER_COMPLEX_STRUCT BER_SignedDataTypeWrapper;
extern DER_COMPLEX_STRUCT BER_SignedDataType;
extern DER_COMPLEX_STRUCT BER_SetOfCertificatesType; 
extern DER_COMPLEX_STRUCT BER_SignedPKCS7Type;
extern DER_COMPLEX_STRUCT DER_AlgorithmIdentifierType;
/* BER DER Tags */
extern unsigned char DER_constructedSequence[];
extern unsigned char DER_constructedSet[];
extern unsigned char DER_objectIDTag[];
extern unsigned char DER_signedDataObjectID[];
extern unsigned char DER_integerTag[];
extern unsigned char DER_bitStringTag[];
extern unsigned char DER_nullTag[];


static CSSM_RETURN Create_Unsigned_PKCS7_ParseTree(MEMORY_POOL_PTR poolptr, 
                                                   const CSSM_CERTGROUP *CertGroupToBundle,
                                                   DER_NODE_PTR *result_val);

static CSSM_RETURN Create_Signed_PKCS7_ParseTree(MEMORY_POOL_PTR poolptr, 
                                                 CSSM_CC_HANDLE CCHandle,
                                                 const CSSM_DATA_PTR tbs_pkcs7_ptr,
                                                 CSSM_DATA_PTR signed_data_ptr,
                                                 DER_NODE_PTR pkcs7_node_ptr,
                                                 DER_NODE_PTR *result_val);


/*-----------------------------------------------------------------------------
 * Name: CL_CertGroupToSignedBundle
 *
 * Description:
 * This function accepts as input a certificate group
 * (as an array of individual certificates) and returns
 * a certificate bundle (a codified and signed aggregation
 * of the certificates in the group). The certificate
 * group will first be encoded according to the BundleInfo
 * input by the user. If BundleInfo is NULL, the library
 * will perform a default encoding for its default bundle
 * type. If possible, the certificate group ordering will
 * be maintained in this certificate aggregate encoding. 
 * After encoding, the certificate aggregate will be signed
 * using the input context and signer certificate. The CL
 * module embeds knowledge of the signing scope for the
 * bundle types it supports.  The signature is then
 * associated with the certificate aggregate according to
 * the bundle type and encoding rules and is returned as
 * a bundle to the calling application. 
 * 
 * Parameters: 
 * CCHandle (input)          : Handle identifying the Cryptographic Context 
 *                             to use for signing
 * CertGroupToBundle (input) : The certificate group to be aggregated into
 *                             a bundle
 * BundleInfo (input)        : The type and encoding format requested for the bundle
 *
 * result_val (output)       : The bundled and signed certificate bundle
 *
 Signed PKCS7 Tree Structure:
 ============================
                        Signed_PKCS7_Sturcture
                        ___________|_________________
                        |             |             |
                  PKCS7_Structure   AlgOID      Signature
                ________|________
                |               |
                OID           Signed_Data_Wrapper
           (Signed_Data OID)            |
                                    Signed_Data
    ____________________________________|__________________________________________
      |             |                |                |               |           |
    Version    DigestAlgIDs     ContentInfo       CertGroup       CrlGroup    SignerInfo                   
  (INTEGER:1)  (NULL SET)    (NULL: 06 00 05 00)  (Enc Certs)       (NULL)      (NULL)
*---------------------------------------------------------------------------*/
CSSM_RETURN CSSMCLI CL_CertGroupToSignedBundle
                        (CSSM_CL_HANDLE CLHandle,
                         CSSM_CC_HANDLE CCHandle,
                         const CSSM_CERTGROUP *CertGroupToBundle,
                         const CSSM_CERT_BUNDLE_HEADER *BundleInfo,
                         CSSM_DATA_PTR result_val)
{
    
    CSSM_RETURN       ret = CSSM_OK;
    MEMORY_POOL pool;
    CSSM_DATA       signed_data = {0, NULL};
    CSSM_DATA       tbs_pkcs7, signed_pkcs7 ;
    DER_NODE_PTR    signed_pkcs7_node_ptr, pkcs7_node_ptr;
    

    if (port_IsBadWritePtr(result_val,sizeof(CSSM_DATA)))
    {
		return CSSMERR_CL_INVALID_BUNDLE_POINTER;  
    }


    /* Clear result */
    result_val->Length = 0;
    result_val->Data = NULL;
    

    /* Check inputs */
    /* CLHandle and CCHandle are checked by CSSM */
    if (!CertGroupToBundle || 
        CSSM_CLUTIL_IsBadCertGroup(CertGroupToBundle) )
    {
        return CSSMERR_CL_INVALID_CERTGROUP_POINTER;
        
    }
    if (CL_IsBadReadPtr(BundleInfo, sizeof(CSSM_CERT_BUNDLE_HEADER)) )
    {
        return CSSMERR_CL_INVALID_BUNDLE_POINTER ;
        
    }
    if (BundleInfo && BundleInfo->BundleType != CSSM_CERT_BUNDLE_PKCS7_SIGNED_DATA )
    {
        return CSSMERR_CL_INVALID_BUNDLE_INFO;
        
    }
    if (BundleInfo && 
        (BundleInfo->BundleEncoding != CSSM_CERT_BUNDLE_ENCODING_BER) &&
        (BundleInfo->BundleEncoding != CSSM_CERT_BUNDLE_ENCODING_DER) )
    {
        return CSSMERR_CL_INVALID_BUNDLE_INFO;
        
    }
    
    /* Setup the Memory pool */
    ret = (*memory_pool_funcs.init)((CSSM_MEMORY_FUNCS_PTR)&Addin_APIMemFuncs, NULL, &pool);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    
    /* Store poolptr in TLS: so we can send inTLS_MemoryPool as BER_HANDLE */
    ret = port_SetTlsValue(TLS_MemoryPool, &pool);
    if (CSSM_OK != ret)
    {
        return ret;
    }

    /* Create a PKCS7 ParseTree with the fields filled in correctly */
    ret = Create_Unsigned_PKCS7_ParseTree(&pool, CertGroupToBundle, &pkcs7_node_ptr);
    if (CSSM_OK != ret)
    {
        (*memory_pool_funcs.recycle)(&pool);
        return ret;
    }

    /* Now we have a typed parse tree in pkcs7_node_ptr */
    /* Lets generate a DER blob for it */
    tbs_pkcs7.Data = DER_TypedGenerate(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        pkcs7_node_ptr);
    if (!tbs_pkcs7.Data)
    {
        (*memory_pool_funcs.recycle)(&pool);
        return CSSMERR_CL_FUNCTION_FAILED;
    }
    tbs_pkcs7.Length = BER_SizeofObject(tbs_pkcs7.Data);
    
    
    if (!CCHandle) /* No need to sign the PKCS7 structure */
    {
        /* this is *not* an error condition */
        result_val->Length = tbs_pkcs7.Length;
        result_val->Data = App_Malloc(CLHandle, tbs_pkcs7.Length);
        if (!result_val->Data)
        {
           (*memory_pool_funcs.recycle)(&pool);
           return CSSMERR_CL_MEMORY_ERROR;
        }
        port_memcpy(result_val->Data, tbs_pkcs7.Data, tbs_pkcs7.Length);   
        
        (*memory_pool_funcs.recycle)(&pool);
        return CSSM_OK;
    }

    
    /* Build a signed PKCS7 Parse Tree:
        Signature generated inside this function*/
    ret = Create_Signed_PKCS7_ParseTree(&pool,
                                        CCHandle,
                                        &tbs_pkcs7,
                                        &signed_data,
                                        pkcs7_node_ptr,
                                        &signed_pkcs7_node_ptr);
    if (CSSM_OK != ret)
    {
        (*memory_pool_funcs.recycle)(&pool);
        if (signed_data.Data)
            App_Free(CLHandle, signed_data.Data);
        return ret;
    }

    /* Now build the signed blob */
    signed_pkcs7.Data = DER_TypedGenerate(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        signed_pkcs7_node_ptr);

    /* Can free the signature now, regardless of success/failure */
    App_Free(CLHandle, signed_data.Data);

    if (!signed_pkcs7.Data)
    {
        /* DER_TypedGenerate failed */
        (*memory_pool_funcs.recycle)(&pool);
        return CSSMERR_CL_FUNCTION_FAILED ;
    }
    signed_pkcs7.Length = BER_SizeofObject(signed_pkcs7.Data);
    
    /* send the result back to the caller. We're done now! */
    result_val->Length = signed_pkcs7.Length;
    result_val->Data = App_Malloc(CLHandle, signed_pkcs7.Length);
    if (!result_val->Data)
    {
       (*memory_pool_funcs.recycle)(&pool);
       return CSSMERR_CL_MEMORY_ERROR;
    }
    port_memcpy(result_val->Data, signed_pkcs7.Data, signed_pkcs7.Length);   
    
    /* Clean up and return */
    (*memory_pool_funcs.recycle)(&pool);
    return CSSM_OK;

}

/*-----------------------------------------------------------------------------
 * Name: Create_Unsigned_PKCS7_ParseTree
 *
 * Description: This function creates a typed parse tree of DER_NODE structures
 * that can be used as input for the DER_TypedGenerate function to produce a DER
 * blob. The tree is allocated from the MEMORY_POOL and returned in result_val.
 * The structure of the tree is found above.
 *
 * Parameters:
 * poolptr (Input): Memory Pool
 * CertGroupToBundle (Input): Certgroup structure
 * result_val (Output): Pointer to parse tree of Unsigned PKCS7 structure.
 * 
 * Return value: CSSM_OK if no error.
 *               Appropriate error code if error.
 *
 ------------------------------------------------------------------------------*/
static CSSM_RETURN Create_Unsigned_PKCS7_ParseTree(MEMORY_POOL_PTR poolptr, 
                                                   const CSSM_CERTGROUP *CertGroupToBundle,
                                                   DER_NODE_PTR *result_val)
{
    
    DER_NODE_PTR    pkcs7_node_ptr, signed_data_wrapper_node_ptr, 
        signed_data_node_ptr, certgroup_node_ptr;
    CSSM_DATA_PTR enc_cert_ptr;
    sint32          i;
    static uint8           pkcs7_version_data = 1; /* PKCS7 Version */
    static uint8           content_info[] = {0x06, 0x00, 0x05, 0x00}; /* Hard coded content info structure */
    static unsigned char DER_signedDataObjectID[] =     { PKCSOID, 7, 2};    

    
    /* Do all the Node allocations to build a PKCS7 parse tree (unsigned) */
    /* PKCS7 Node */
    pkcs7_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        2); /* PKCS7 type has OID and signeddatawrapper as children */
    if (!pkcs7_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    

    /* signed_data_wrapper Node */
    signed_data_wrapper_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        1); /* signeddatawrapper has signed_data as child */

    if (!signed_data_wrapper_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    

    /* signed_data Node */
    signed_data_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        6); /* this signed_data node has the 6 children: 
            version, algid, contentinfo, certs, crls, signerinfo */

    if (!signed_data_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    /* certgroup Node */
    certgroup_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        CertGroupToBundle->NumCerts); 
    if (!certgroup_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
 
    /* All nodes alloced. Time to fill in data */
    /* Set up the PKCS7 node */
    Init_DER_NODE_Header(pkcs7_node_ptr, DER_constructedSequence, DER_constructedSequence, 2, BER_PKCS7Type);
    
    /* child[0] is the signed_data OID */
    pkcs7_node_ptr->Child[0].IsLeaf = 1; /* OID is a leaf*/
    pkcs7_node_ptr->Child[0].X.Node = NULL;
    pkcs7_node_ptr->Child[0].X.Input.Tag = DER_objectIDTag;
    pkcs7_node_ptr->Child[0].X.Input.ContentLength = sizeof(DER_signedDataObjectID);
    pkcs7_node_ptr->Child[0].X.Input.Content = DER_signedDataObjectID;
    /* child[1] is the signeddatawrapper */
    pkcs7_node_ptr->Child[1].IsLeaf = 0;
    pkcs7_node_ptr->Child[1].X.Node = signed_data_wrapper_node_ptr;



    /* Set up the signed_data_wrapper_node_ptr */
    Init_DER_NODE_Header(signed_data_wrapper_node_ptr, DER_constructedSequence,
        DER_constructedSequence,
        1, BER_SignedDataTypeWrapper);

    /*child[0] is the signed_data type */
    signed_data_wrapper_node_ptr->Child[0].IsLeaf = 0;
    signed_data_wrapper_node_ptr->Child[0].X.Node = signed_data_node_ptr;

    /* Set up the signed_data_node_ptr */
    /* this signed_data node has the 6 children: 
            version, algid, contentinfo, certs, crls, signerinfo */
    Init_DER_NODE_Header(signed_data_node_ptr, DER_constructedSequence,
        DER_constructedSequence, 6, BER_SignedDataType);
    
    /* child[0] is the version */
    signed_data_node_ptr->Child[0].IsLeaf = 1;
    signed_data_node_ptr->Child[0].X.Node = 0;
    signed_data_node_ptr->Child[0].X.Input.Tag = DER_integerTag;
    signed_data_node_ptr->Child[0].X.Input.ContentLength = sizeof(uint8);
    signed_data_node_ptr->Child[0].X.Input.Content = &pkcs7_version_data;    
    /*Child[1] is the Digest Alg Ids - NULL set */
    signed_data_node_ptr->Child[1].IsLeaf = 1;
    signed_data_node_ptr->Child[1].X.Node=0;
    signed_data_node_ptr->Child[1].X.Input.Tag = DER_constructedSet;
    signed_data_node_ptr->Child[1].X.Input.ContentLength = 0;
    signed_data_node_ptr->Child[1].X.Input.Content = NULL;
    /* Child[2] is the Content Info */
    /* Content Info is basically NULL. A null content info corresponds to a sequence that contains 2 elements: 
        - an OID with null data (06 00) and
        - a NULL type (05 00) */
    signed_data_node_ptr->Child[2].IsLeaf = 1;
    signed_data_node_ptr->Child[2].X.Node = 0;
    signed_data_node_ptr->Child[2].X.Input.Tag = DER_constructedSequence;
    signed_data_node_ptr->Child[2].X.Input.ContentLength = 4;
    signed_data_node_ptr->Child[2].X.Input.Content = content_info; /* hard coded 06 00 05 00 */ 
    /* Child[3] is the ExtendedCertificates */
    signed_data_node_ptr->Child[3].IsLeaf = 0;
    signed_data_node_ptr->Child[3].X.Node = certgroup_node_ptr;
    /* Child[4] is the CRLs optional. For us, its NULL */
    signed_data_node_ptr->Child[4].IsLeaf = 1;
    signed_data_node_ptr->Child[4].X.Node = 0;
    signed_data_node_ptr->Child[4].X.Input.Tag = DER_constructedSet;
    signed_data_node_ptr->Child[4].X.Input.ContentLength = 0;
    signed_data_node_ptr->Child[4].X.Input.Content = NULL;
    /* Child[5] is the Signer Info. Again its a NULL set for us */
    signed_data_node_ptr->Child[5].IsLeaf = 1;  
    signed_data_node_ptr->Child[5].X.Node = 0;
    signed_data_node_ptr->Child[5].X.Input.Tag = DER_constructedSet;
    signed_data_node_ptr->Child[5].X.Input.ContentLength = 0;
    signed_data_node_ptr->Child[5].X.Input.Content = NULL;

    /* Now to set up the certgroup_node_ptr */
    Init_DER_NODE_Header(certgroup_node_ptr, DER_constructedSet, DER_constructedSet,
        CertGroupToBundle->NumCerts, BER_SetOfCertificatesType);/* extended certificates */
    for(i=0;i<certgroup_node_ptr->Count;i++)
    {
        /* Set up the certificates as its children */
        enc_cert_ptr = &CertGroupToBundle->GroupList.CertList[i];
        certgroup_node_ptr->Child[i].IsLeaf = 1;
        certgroup_node_ptr->Child[i].X.Node = 0;
        certgroup_node_ptr->Child[i].X.Input.Tag = DER_constructedSequence;
        certgroup_node_ptr->Child[i].X.Input.ContentLength = cl_der_LengthOfContent(enc_cert_ptr);
        certgroup_node_ptr->Child[i].X.Input.Content = 
            &enc_cert_ptr->Data[cl_der_LengthOfTag(enc_cert_ptr) + cl_der_LengthOfLength(enc_cert_ptr)];
    }
 
    *result_val = pkcs7_node_ptr;
    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: Create_Signed_PKCS7_ParseTree
 *
 * Description: This function creates a typed parse tree of DER_NODE structures
 * that can be used as input for the DER_TypedGenerate function to produce a DER
 * blob. The tree is allocated from the MEMORY_POOL and returned in result_val.
 * The structure of the tree is found above. It takes as input the Unsigned PKCS7
 * tree created earlier in the process.
 *
 * Parameters:
 * poolptr (Input): Memory Pool
 * CCHandle (Input): Crytographic Context Handle
 * tbs_pkcs7_ptr (Input): DER blob of Unsigned PKCS7 
 * signed_data_ptr(Output): Signature generated by addin_signdata
 * pkcs7_node_ptr(Input): Pointer to parse tree of Unsigned PKCS7 structure.
 * result_val (Output): Pointer to parse tree of Signed PKCS7 structure.
 * 
 * Return value: CSSM_OK if no error.
 *               Appropriate error code if error.
 *
 ------------------------------------------------------------------------------*/

static CSSM_RETURN Create_Signed_PKCS7_ParseTree(MEMORY_POOL_PTR poolptr, 
                                                 CSSM_CC_HANDLE CCHandle,
                                                 const CSSM_DATA_PTR tbs_pkcs7_ptr,
                                                 CSSM_DATA_PTR signed_data_ptr,
                                                 DER_NODE_PTR pkcs7_node_ptr,
                                                 DER_NODE_PTR *result_val)
{
    DER_NODE_PTR    signed_pkcs7_node_ptr, sig_algid_node_ptr;
    CSSM_DATA       temp_data = {0, NULL};
    static CSSM_OID_PTR          algoid;
    CSSM_CONTEXT_PTR  context = NULL;
    uint32            alg_id = 0;
    CSSM_RETURN        ret = CSSM_OK;


    /* Get the algorithm id */
    ret = addin_GetContext(CCHandle, &context);
    if (ret != CSSM_OK)
    {
        /* error set by addin_GetContext */
        return ret; 
    }
    alg_id = context->AlgorithmType;
    addin_FreeContext(context);

    /* Obtain the OID for the algorithm */
    algoid = X509_AlgIdToAlgorithmOid(alg_id);
    if (algoid == NULL)
    {
        return CSSMERR_CL_INVALID_DATA;
    }

    /* Get the signature */
    ret = addin_SignData (CCHandle, tbs_pkcs7_ptr, 1, CSSM_ALGID_NONE, signed_data_ptr);
    if (ret != CSSM_OK)
    {
        /* error set by addin_SignData */
        return ret; 
    }
    /* Allocate a signed_pkcs7_node_ptr */
    signed_pkcs7_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        3); 
    if (!signed_pkcs7_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    
    /* Allocate a sig_algid_node_ptr */
    sig_algid_node_ptr = DER_AllocateNode(TLS_MemoryPool, 
        &berder_funcs_wrapper,
        2); 
    if (!sig_algid_node_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }
    

    /* Fill in the fields of the signed_pkcs7_node */
    /* 3 children: child[0] is the PKCS7 node from above. 
        child[1] is the algoid. child[2] is the signature */
    Init_DER_NODE_Header(signed_pkcs7_node_ptr, DER_constructedSequence,
        DER_constructedSequence, 
        3, BER_SignedPKCS7Type);

    signed_pkcs7_node_ptr->Child[0].IsLeaf = 0;
    signed_pkcs7_node_ptr->Child[0].X.Node = pkcs7_node_ptr;
    signed_pkcs7_node_ptr->Child[1].IsLeaf = 0;
    signed_pkcs7_node_ptr->Child[1].X.Node = sig_algid_node_ptr;
    signed_pkcs7_node_ptr->Child[2].IsLeaf = 1;
    signed_pkcs7_node_ptr->Child[2].X.Node = 0;
    if (signed_data_ptr->Length)
    {
        temp_data.Length = signed_data_ptr->Length +1;
        temp_data.Length = 1 + BER_LengthOfLength(temp_data.Length) + temp_data.Length;
        temp_data.Data = CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func, 
            temp_data.Length, poolptr);
        if (!temp_data.Data)
        {
            return CSSMERR_CL_MEMORY_ERROR;
           
        }

        DER_EncodeBitString(signed_data_ptr->Data, 8*(signed_data_ptr->Length), temp_data.Data, temp_data.Length);
        BER_ExpandItem(temp_data.Data, temp_data.Length, &signed_pkcs7_node_ptr->Child[2].X.Input);
        
    } else 
    {
        signed_pkcs7_node_ptr->Child[2].X.Input.Tag = DER_bitStringTag;
        signed_pkcs7_node_ptr->Child[2].X.Input.ContentLength = 0;
        signed_pkcs7_node_ptr->Child[2].X.Input.Content = 0;        
    }

    
    
    /* Fill in the fields of sig_algid_node_ptr */
    /* Two children: child[0] is the algoid pbtained above. child[1] is NULL */
    Init_DER_NODE_Header(sig_algid_node_ptr, DER_constructedSequence,
        DER_constructedSequence,
        2, DER_AlgorithmIdentifierType);
    
    sig_algid_node_ptr->Child[0].IsLeaf = 1;
    sig_algid_node_ptr->Child[0].X.Node = 0;
    sig_algid_node_ptr->Child[0].X.Input.Tag = DER_objectIDTag;
    sig_algid_node_ptr->Child[0].X.Input.ContentLength = algoid->Length;
    sig_algid_node_ptr->Child[0].X.Input.Content = algoid->Data;
    sig_algid_node_ptr->Child[1].IsLeaf = 1;
    sig_algid_node_ptr->Child[1].X.Node = 0;
    sig_algid_node_ptr->Child[1].X.Input.Tag = DER_nullTag;
    sig_algid_node_ptr->Child[1].X.Input.ContentLength = 0;
    sig_algid_node_ptr->Child[1].X.Input.Content = NULL;
    
    *result_val = signed_pkcs7_node_ptr;
    return CSSM_OK;
}
