/*-----------------------------------------------------------------------
 * File: INTERNAL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * This file implements certificate functions which are used only by this library.
 */
#include "maf_include.h"
/* CSSM includes */
#include "cssm.h"
#include "guids.h"
#include "cssmport.h"
#include "cssmutil.h"

/* Certificate-specific includes */
#include "x509defs.h"
#include "x_fndefs.h"
#include "internal.h"
#include "clutil.h"

#include "api_shim.h"
#include "maf_util.h"

#include <cssmMemTrack.h>

extern CSSM_UPCALLS MemoryFunctions;


/*-----------------------------------------------------------------------------
 * Name: cl_SetSignerInfoFromSignerCert
 *
 * Description:
 * This function gets the value or values from SignerCert to fill in the
 * SignerInfo in the Pkcs7SignedData.
 * Currently, this is always set to NULL because there is no example in industry
 * to set an expectation for the contents of this structure.  VeriSign leaves it NULL.
 *
 * Parameters: 
 * SignerCert  (input)       : The certificate to be used for signerInfo
 * Pkcs7SignedData   (input) : The Pkcs7 structure whose SignerInfo should be filled in
 *
 * Return value:
 * A CSSM_RETURN indicator of whether or not the signerInfo was set
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN cl_SetSignerInfoFromSignerCert(CSSM_CL_HANDLE CLHandle, 
                                           const CSSM_DATA *SignerCert,
                                           cssm_PKCS7_SIGNED_DATA_PTR Pkcs7SignedData)
{
        /* unused parameters */
        CLHandle;
        SignerCert;

        Pkcs7SignedData->numberOfSignerInfos = 0;
        Pkcs7SignedData->signerInfos = NULL;
        return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: x509_cl_CertAddExtensions
 *
 * Description:
 * This function adds the input extensions to the list of extensions in the input cert.
 * This routine assumes that it can keep all of the pointers internal to the extensions. 
 * Therefore, DO NOT free those pointers in the calling routine 
 *
 * Parameters: 
 * Cert (input/output) : The cert with the original extensions 
 * NewExtensions (input) : The extensions to be added to the cert
 * NumNewExtensions (input) : The length of the NewExtensions array
 *
 *---------------------------------------------------------------------------*/
CSSM_RETURN x509_cl_CertAddExtensions(CSSM_CL_HANDLE AppHandle, 
                               CSSM_X509_SIGNED_CERTIFICATE_PTR Cert, 
                               CSSM_X509_EXTENSION_PTR NewExtensions, 
                               uint32 NumNewExtensions)
{
    CSSM_X509_EXTENSION_PTR temp_ext_array_ptr = NULL;

    /* Check inputs */
    if ( CSSM_CLUTIL_X509_IsBadCertPtr(Cert) ) 
    {
        return CSSMERR_CL_INVALID_CERT_POINTER;
        
    }
    if ( CSSM_CLUTIL_X509_IsBadExtensionPtr (NewExtensions, NumNewExtensions) ) 
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
        
    }

    /* Allocate memory for the old + new extensions */
    temp_ext_array_ptr = (CSSM_X509_EXTENSION_PTR) App_Calloc(AppHandle, 
        (Cert->certificate.extensions.numberOfExtensions + NumNewExtensions), sizeof(CSSM_X509_EXTENSION) );
    if (!temp_ext_array_ptr)
    {
        return CSSMERR_CL_MEMORY_ERROR;
        
    }

        /* Copy the old & new extensions into the new memory */
    port_memcpy(temp_ext_array_ptr, Cert->certificate.extensions.extensions, 
        Cert->certificate.extensions.numberOfExtensions*sizeof(CSSM_X509_EXTENSION));
    port_memcpy(&temp_ext_array_ptr[Cert->certificate.extensions.numberOfExtensions], NewExtensions, 
        NumNewExtensions*sizeof(CSSM_X509_EXTENSION) );

        /* Free the old certificate extensions */
    App_Free(AppHandle, Cert->certificate.extensions.extensions );

    /* Update the extensions pointer */
    Cert->certificate.extensions.extensions = temp_ext_array_ptr;
    Cert->certificate.extensions.numberOfExtensions = Cert->certificate.extensions.numberOfExtensions + NumNewExtensions;

    return CSSM_OK;
}



/*-----------------------------------------------------------------------------
 * Name: get_signedcertificate_field
 *
 * Description:
 * Makes a copy of the SignedCert C structure in memory from the
 * application's heap.  
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * SignedCert (input) :       The SignedCert to be copied. 
 *
 * Return value:
 * A non-NULL pointer to a CSSM_DATA_PTR whose Data field points to the 
 *  SignedCert C structure. 
 * or NULL if there was an error.
 * 
 *---------------------------------------------------------------------------*/
CSSM_DATA_PTR 
get_signedcertificate_field(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_SIGNED_CERTIFICATE_PTR SignedCert)
{
    CSSM_DATA_PTR ret_ptr = NULL;

    /* allocate the return structure */
    ret_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_DATA));
    if (ret_ptr == NULL)
    {
        return NULL;
    }
    /* set the length */
    ret_ptr->Length = CSSM_FIELDVALUE_COMPLEX_DATA_TYPE;

    /* duplicate the signed certificate */
    ret_ptr->Data = (uint8 *) dup_signedcert(CLHandle, SignedCert);
    if (ret_ptr->Data == NULL)  /* if unsuccessful, then free the return structure */
    {
        App_Free(CLHandle, ret_ptr);
        return NULL;
    }

    return ret_ptr;
}

/*-----------------------------------------------------------------------------
 * Name: get_certificate_field
 *
 * Description:
 * Makes a copy of the Certificate C structure in memory from the
 * application's heap.  
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * SignedCert (input) :       The SignedCert to copy the Certificate from.
 *
 * Return value:
 * A non-NULL pointer to a CSSM_DATA_PTR whose Data field points to the 
 *  Cert C structure copy.
 * or NULL if there was an error.
 * 
 *---------------------------------------------------------------------------*/
CSSM_DATA_PTR 
get_certificate_field(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_SIGNED_CERTIFICATE_PTR SignedCert)
{
    CSSM_DATA_PTR ret_ptr = NULL;

    /* allocate the return structure */
    ret_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_DATA));
    if (ret_ptr == NULL)
    {
        return NULL;
    }

    /* set the length */
    ret_ptr->Length = CSSM_FIELDVALUE_COMPLEX_DATA_TYPE;

    /* duplicate the certificate */
    ret_ptr->Data = (uint8 *) dup_cert(CLHandle, &SignedCert->certificate);
    if (ret_ptr->Data == NULL)  /* if unsuccessful, then free the return structure */
    {
        App_Free(CLHandle, ret_ptr);
        return NULL;
    }

    return ret_ptr;
}

/*-----------------------------------------------------------------------------
 * Name: get_certificate_extensions_field
 *
 * Description:
 * Makes a copy of the Extensions C structure in memory from the
 * application's heap.  
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * SignedCert (input) :       The SignedCert to copy the Extension structures from.
 *
 * Return value:
 * A non-NULL pointer to a CSSM_DATA_PTR whose Data field points to an array of
 *  Cert extension structures
 * or NULL if there was an error.
 *---------------------------------------------------------------------------*/
CSSM_DATA_PTR 
get_certificate_extensions_field(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_SIGNED_CERTIFICATE_PTR SignedCert)
{
    CSSM_DATA_PTR ret_ptr = NULL;

    /* allocate the return structure */
    ret_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_DATA));
    if (ret_ptr == NULL)
    {
        return NULL;
    }

    /* set the length */
    ret_ptr->Length = CSSM_FIELDVALUE_COMPLEX_DATA_TYPE;

    /* duplicate the extensions */

    ret_ptr->Data = (uint8 *) dup_extensions(CLHandle, &SignedCert->certificate.extensions);
    if (ret_ptr->Data == NULL)  /* if unsuccessful, then free the return structure */
    {
        App_Free(CLHandle, ret_ptr);
        return NULL;
    }

    return ret_ptr;
}

/*-----------------------------------------------------------------------------
 * Name: dup_signedcert
 *
 * Description:
 * Duplicate a SignedCert C structure in memory from the Application's heap.
 * This call allocates the CSSM_X509_SIGNED_CERTIFICATE structure.  
 * The copy_* functions allocate the sub-structures.
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * SignedCert (input) :       The SignedCert to be copied. 
 *
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN 
copy_signedcert(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_SIGNED_CERTIFICATE_PTR toSignedCert,
    CSSM_X509_SIGNED_CERTIFICATE_PTR SignedCert)
{
    CSSM_RETURN retval = CSSM_OK;

    if (toSignedCert == NULL)
    {
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }

    /* copy the certificate */
    retval = copy_certvalues(CLHandle, &toSignedCert->certificate, &SignedCert->certificate);
    if (retval != CSSM_OK)
    {
        return retval;
    }

    /* copy the signature */
    retval = copy_signaturevalues(CLHandle, &toSignedCert->signature, &SignedCert->signature);
    if (retval != CSSM_OK)
    {
        free_signedcert(CLHandle, toSignedCert);
        return retval;
    }

    return CSSM_OK;
}


CSSM_X509_SIGNED_CERTIFICATE_PTR
dup_signedcert(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_SIGNED_CERTIFICATE_PTR SignedCert)
{
    CSSM_X509_SIGNED_CERTIFICATE_PTR dup_ptr = NULL;
    CSSM_RETURN retval = CSSM_OK;

    /* allocate the signed certificate structure */
    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_X509_SIGNED_CERTIFICATE));
    if (dup_ptr == NULL)
        return NULL;

    /* copy the certificate */
    retval = copy_signedcert(CLHandle, dup_ptr, SignedCert);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, dup_ptr);
        return NULL;
    }

    return dup_ptr;
}

/*-----------------------------------------------------------------------------
 * Name: dup_cert
 *
 * Description:
 * Duplicate a Cert C structure in memory from the Application's heap.
 * This call allocates the CSSM_X509_TBS_CERTIFICATE structure.  
 * The copy_* functions allocate the sub-structures.
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * Cert (input) :             The Cert to be copied. 
 *
 * Return value:
 * A non-NULL pointer to the Cert C structure. 
 * or NULL if there was an error.
 * 
 *---------------------------------------------------------------------------*/
CSSM_X509_TBS_CERTIFICATE_PTR
dup_cert(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_TBS_CERTIFICATE_PTR Cert)
{
    CSSM_X509_TBS_CERTIFICATE_PTR dup_ptr = NULL;
    CSSM_RETURN retval = CSSM_OK;

    /* allocate the signed certificate structure */
    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_X509_TBS_CERTIFICATE));
    if (dup_ptr == NULL)
        return NULL;

    /* copy the certificate */
    retval = copy_certvalues(CLHandle, dup_ptr, Cert);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, dup_ptr);
        return NULL;
    }

    return dup_ptr;
}

/*-----------------------------------------------------------------------------
 * Name: free_signedcert
 *
 * Description:
 * Free a CSSM_X509_SIGNED_CERTIFICATE C structure in memory from the Application's heap.
 * The copy_* functions free the sub-structures.
 * This call frees the CSSM_X509_SIGNED_CERTIFICATE structure.  
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * free_ptr (input) :         The structure to be freed. 
 *
 * Return value:
 * CSSM_OK if successful.
 * Error code if there was an error.
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN
free_signedcert(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_SIGNED_CERTIFICATE_PTR free_ptr)
{
    CSSM_RETURN retval1 = CSSM_OK, retval2 = CSSM_OK;
    

    if (free_ptr == NULL)
    {
        return CSSM_OK;
    }

    /* free the certificate structure */
    retval1 = free_certvalues(CLHandle, &free_ptr->certificate);
    
    /* free the signature structure */
    retval2 = free_signaturevalues(CLHandle, &free_ptr->signature);
    
    /* free the signed certificate structure */
    App_Free(CLHandle, free_ptr);

    if (CSSM_OK != retval1)
    {
        return retval1;
    }
    if (CSSM_OK != retval2)
    {
        return retval2;
    }
    return CSSM_OK;
}

/*-----------------------------------------------------------------------------
 * Name: copy_cert
 *
 * Description:
 * Copy from the Cert C structure to toCert.  
 * The copy_* functions allocate substructures as necessary.
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * toCert (input) :           The structure to copy to.
 * Cert (input) :             The structure to copy from.
 *
 * Return value:
 * CSSM_OK on success.
 * Error code if there was an error. 
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN
copy_certvalues(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_TBS_CERTIFICATE_PTR toCert, 
    CSSM_X509_TBS_CERTIFICATE_PTR Cert)
{
    CSSM_RETURN retval = CSSM_OK;

    retval = copy_CSSM_DATAvalue(CLHandle, &toCert->version, &Cert->version);
    if (retval != CSSM_OK)
    {
        return retval;
    }
    
    retval = copy_CSSM_DATAvalue(CLHandle, &toCert->serialNumber, &Cert->serialNumber);
    if (retval != CSSM_OK)
    {
        free_certvalues(CLHandle, toCert);
        return retval;
    }

    retval = copy_algorithm(CLHandle, &toCert->signature, &Cert->signature);
    if (retval != CSSM_OK)
    {
        free_certvalues(CLHandle, toCert);
        return retval;
    }

    retval = copy_name(CLHandle, &toCert->issuer, &Cert->issuer);
    if (retval != CSSM_OK)
    {
        free_certvalues(CLHandle, toCert);
        return retval;
    }
    
    retval = copy_validity(CLHandle, &toCert->validity, &Cert->validity);
    if (retval != CSSM_OK)
    {
        free_certvalues(CLHandle, toCert);
        return retval;
    }

    retval = copy_name(CLHandle, &toCert->subject, &Cert->subject);
    if (retval != CSSM_OK)
    {
        free_certvalues(CLHandle, toCert);
        return retval;
    }

    retval = copy_subjectPublicKeyInfo(CLHandle, &toCert->subjectPublicKeyInfo, 
                &Cert->subjectPublicKeyInfo);
    if (retval != CSSM_OK)
    {
        free_certvalues(CLHandle, toCert);
        return retval;
    }
    
    retval = copy_CSSM_DATAvalue(CLHandle, &toCert->issuerUniqueIdentifier, 
                &Cert->issuerUniqueIdentifier);
    if (retval != CSSM_OK)
    {
        free_certvalues(CLHandle, toCert);
        return retval;
    }

    retval = copy_CSSM_DATAvalue(CLHandle, &toCert->subjectUniqueIdentifier, 
                &Cert->subjectUniqueIdentifier);
    if (retval != CSSM_OK)
    {
        free_certvalues(CLHandle, toCert);
        return retval;
    }

    if (Cert->extensions.numberOfExtensions != 0)
    {
        retval = copy_extensions(CLHandle, &toCert->extensions, &Cert->extensions);
        if (retval != CSSM_OK)
        {
            free_certvalues(CLHandle, toCert);
            return retval;
        }
    }

    return CSSM_OK;
}


/*-----------------------------------------------------------------------------
 * Name: free_certvalues
 *
 * Description:
 * Free CSSM_X509_TBS_CERTIFICATE C structure values.
 * The free_* functions free the sub-structures.
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * free_ptr (input) :         The structure to be freed. 
 *
 * Return value:
 * CSSM_OK if successful.
 * Error code if there was an error.
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN
free_certvalues(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_TBS_CERTIFICATE_PTR free_ptr)
{
    CSSM_RETURN retval = CSSM_OK, err_retval = CSSM_OK;
    int errcount = 0;

    if (free_ptr == NULL)
    {
        return CSSM_OK;
    }

    retval = free_CSSM_DATAvalue(CLHandle, &free_ptr->version);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }
    
    retval = free_CSSM_DATAvalue(CLHandle, &free_ptr->serialNumber);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }

    retval = free_algorithm(CLHandle, &free_ptr->signature);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }

    retval = free_name(CLHandle, &free_ptr->issuer);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }
    
    retval = free_validity(CLHandle, &free_ptr->validity);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }

    retval = free_name(CLHandle, &free_ptr->subject);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }

    retval = free_subjectPublicKeyInfo(CLHandle, &free_ptr->subjectPublicKeyInfo);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }
    
    retval = free_CSSM_DATAvalue(CLHandle, &free_ptr->issuerUniqueIdentifier);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }

    retval = free_CSSM_DATAvalue(CLHandle, &free_ptr->subjectUniqueIdentifier);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }

    retval = free_extensions(CLHandle, &free_ptr->extensions);
    if (retval != CSSM_OK)
    {
        err_retval = retval;
        errcount++;
    }

    return ((errcount == 0) ? CSSM_OK : err_retval);
}


/*-----------------------------------------------------------------------------
 * Name: copy_signature
 *
 * Description:
 * Copy from the Signature C structure to Signature.  
 * The copy_* functions allocate substructures as necessary.
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * toSignature (input) :      The structure to copy to.
 * Signature (input) :        The structure to copy from.
 *
 * Return value:
 * CSSM_OK on success.
 * Error Code if there was an error. 
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN
copy_signaturevalues(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_SIGNATURE_PTR toSignature, 
    CSSM_X509_SIGNATURE_PTR Signature)
{
    CSSM_RETURN retval = CSSM_OK;

    retval = copy_CSSM_DATAvalue(CLHandle, &toSignature->algorithmIdentifier.algorithm, 
                &Signature->algorithmIdentifier.algorithm);
    if (retval != CSSM_OK)
    {
        return retval;
    }

    retval = copy_CSSM_DATAvalue(CLHandle, &toSignature->algorithmIdentifier.parameters, 
                &toSignature->algorithmIdentifier.parameters);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, toSignature->algorithmIdentifier.algorithm.Data);
        return retval;
    }

    retval = copy_CSSM_DATAvalue(CLHandle, &toSignature->encrypted, &Signature->encrypted);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, toSignature->algorithmIdentifier.parameters.Data);
        App_Free(CLHandle, toSignature->algorithmIdentifier.algorithm.Data);
        return retval;
    }

    return CSSM_OK;
}


CSSM_RETURN 
free_signaturevalues(CSSM_HANDLE CLHandle, 
    CSSM_X509_SIGNATURE_PTR ptr)
{
    if (ptr == NULL)
    {
        return CSSM_OK;
    }

    free_algorithm(CLHandle, &ptr->algorithmIdentifier);

    free_CSSM_DATAvalue(CLHandle, &ptr->encrypted);

    return CSSM_OK;
}


CSSM_RETURN
free_CSSM_DATAvalue(CSSM_CL_HANDLE CLHandle,
    CSSM_DATA_PTR ptr)
{
    if (ptr == NULL)
    {
        return CSSM_OK;
    }

    if (ptr->Data != NULL)
    {
        App_Free(CLHandle, ptr->Data);
        ptr->Data = NULL;
        ptr->Length = 0;
    }

    return CSSM_OK;
}

CSSM_RETURN
copy_CSSM_DATAvalue(CSSM_CL_HANDLE CLHandle, 
    CSSM_DATA_PTR toData,
    CSSM_DATA_PTR Data)
{
    if (Data->Length == 0)
        return CSSM_OK;

    if (Data->Data == NULL)
        return CSSMERR_CL_INVALID_DATA;

    toData->Data = App_Malloc(CLHandle, Data->Length);
    if (toData->Data == NULL)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    memcpy(toData->Data, Data->Data, Data->Length);
    toData->Length = Data->Length;

    return CSSM_OK;
}

CSSM_DATA_PTR
dup_CSSM_DATA(CSSM_CL_HANDLE CLHandle, CSSM_DATA_PTR Data)
{
    CSSM_DATA_PTR dup_ptr = NULL;
    CSSM_RETURN retval = CSSM_OK;

    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_DATA));
    if (dup_ptr == NULL)
        return NULL;

    retval = copy_CSSM_DATAvalue(CLHandle, dup_ptr, Data);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, dup_ptr);
        return NULL;
    }

    return dup_ptr;
}

CSSM_RETURN 
free_algorithm(CSSM_HANDLE CLHandle, 
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR ptr)
{
    CSSM_RETURN retval1=CSSM_OK, retval2=CSSM_OK;

    if (ptr != NULL) 
    {
        retval1 = free_CSSM_DATAvalue(CLHandle, &ptr->algorithm);
        retval2 = free_CSSM_DATAvalue(CLHandle, &ptr->parameters);
    }
    if (CSSM_OK != retval1)
    {
        return retval1;
    }
    if (CSSM_OK != retval2)
    {
        return retval2;
    }
    return CSSM_OK;

}


CSSM_RETURN
copy_algorithm(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR toAlg,
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR Alg)
{
    CSSM_RETURN retval = CSSM_OK;

    retval = copy_CSSM_DATAvalue(CLHandle, &toAlg->algorithm, &Alg->algorithm);
    if (retval != CSSM_OK)
    {
        return retval;
    }

    retval = copy_CSSM_DATAvalue(CLHandle, &toAlg->parameters, &Alg->parameters);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, toAlg->algorithm.Data);
        return retval;
    }

    return CSSM_OK;
}


CSSM_X509_ALGORITHM_IDENTIFIER_PTR
dup_algorithm(CSSM_CL_HANDLE CLHandle, CSSM_X509_ALGORITHM_IDENTIFIER_PTR Alg)
{
    CSSM_X509_ALGORITHM_IDENTIFIER_PTR dup_ptr = NULL;
    CSSM_RETURN retval = CSSM_OK;

    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_X509_ALGORITHM_IDENTIFIER));
    if (dup_ptr == NULL)
        return NULL;

    retval = copy_algorithm(CLHandle, dup_ptr, Alg);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, dup_ptr);
        return NULL;
    }

    return dup_ptr;
}

CSSM_RETURN
free_Pair(CSSM_CL_HANDLE CLHandle, CSSM_X509_TYPE_VALUE_PAIR_PTR ptr)
{
    CSSM_RETURN retval1 = CSSM_OK, retval2 = CSSM_OK;

    if (ptr != NULL)
    {
        retval1 = free_CSSM_DATAvalue(CLHandle, &ptr->type);
        retval2 = free_CSSM_DATAvalue(CLHandle, &ptr->value);
    }

    if (CSSM_OK != retval1)
        return retval1;
    if (CSSM_OK != retval2)
        return retval2;
    return CSSM_OK;
}


CSSM_RETURN
free_RDN(CSSM_HANDLE CLHandle,
    CSSM_X509_RDN_PTR ptr)
{
    uint32 i=0;
    int errorcount = 0;

    if (ptr == NULL)
        return CSSM_OK;

    if (ptr->AttributeTypeAndValue == NULL)
        return CSSM_OK;

    /* For each attribute/value pair */
    for (i=0; i < ptr->numberOfPairs; i++)
    {
        errorcount += (int) free_Pair(CLHandle, &ptr->AttributeTypeAndValue[i]);
    }
    App_Free(CLHandle, ptr->AttributeTypeAndValue);
    ptr->AttributeTypeAndValue;
    
    return (errorcount == 0) ? CSSM_OK : CSSMERR_CL_INTERNAL_ERROR;
}

CSSM_RETURN
free_RDNs(CSSM_HANDLE CLHandle,
    uint32 NumRDNs,
    CSSM_X509_RDN_PTR ptr)
{
    uint32 i=0;
    int errorcount = 0;

    if (ptr == NULL)
    {
        return CSSM_OK;
    }

    if (ptr->AttributeTypeAndValue == NULL)
    {
        return CSSM_OK;
    }

    for (i=0; i<NumRDNs; i++)
    {
        errorcount += (int) free_RDN(CLHandle, &ptr[i]);
    }

    return (errorcount == 0) ? CSSM_OK : CSSMERR_CL_INTERNAL_ERROR;
}

CSSM_RETURN 
free_name(CSSM_HANDLE CLHandle, 
    CSSM_X509_NAME_PTR ptr)
{
    uint32 i=0;
    int errorcount = 0;

    if (ptr == NULL)
    {
        return CSSM_OK;
    }

    if (ptr->RelativeDistinguishedName == NULL)
    {
        return CSSM_OK;
    }

    /* For each RDN */
    for (i=0; i < ptr->numberOfRDNs; i++)
    {
        errorcount += (int) free_RDN(CLHandle, &ptr->RelativeDistinguishedName[i]);
    }

    App_Free(CLHandle, ptr->RelativeDistinguishedName);
    ptr->RelativeDistinguishedName = NULL;

    return CSSM_OK;
}


CSSM_RETURN
copy_name(CSSM_CL_HANDLE CLHandle, CSSM_X509_NAME_PTR toName, CSSM_X509_NAME_PTR Name)
{
    

    toName->numberOfRDNs = Name->numberOfRDNs;

    toName->RelativeDistinguishedName = dup_RDNs(CLHandle, Name->numberOfRDNs, 
                                            Name->RelativeDistinguishedName);
    if (toName->RelativeDistinguishedName == NULL)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    return CSSM_OK;
}

CSSM_X509_RDN_PTR 
dup_RDNs(CSSM_CL_HANDLE CLHandle, 
    uint32 NumRDNs, 
    CSSM_X509_RDN_PTR RDN)
{
    CSSM_X509_RDN_PTR dup_ptr = NULL;
    uint32 i=0;
    CSSM_RETURN retval = CSSM_OK;

    dup_ptr = App_Calloc(CLHandle, NumRDNs, sizeof(CSSM_X509_RDN));
    if (dup_ptr == NULL)
        return NULL;

    for (i=0; i<NumRDNs; i++)
    {
        retval = copy_RDN(CLHandle, &dup_ptr[i], &RDN[i]);
        if (retval != CSSM_OK)
        {
            free_RDNs(CLHandle, NumRDNs, dup_ptr);
            return NULL;
        }
    }

    return (CSSM_X509_RDN_PTR) dup_ptr;
}

CSSM_RETURN
copy_RDN(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_RDN_PTR toRDN, 
    CSSM_X509_RDN_PTR RDN)
{
    toRDN->numberOfPairs = RDN->numberOfPairs;

    toRDN->AttributeTypeAndValue = dup_Pairs(CLHandle, RDN->numberOfPairs, 
                                                RDN->AttributeTypeAndValue);
    if (toRDN->AttributeTypeAndValue == NULL)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    return CSSM_OK;
}


CSSM_RETURN
free_Pairs(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_TYPE_VALUE_PAIR_PTR ptr)
{
    CSSM_RETURN retval1 = CSSM_OK, retval2= CSSM_OK;
    
    if (ptr == NULL)
    {
        return CSSM_OK;
    }

    retval1 = free_CSSM_DATAvalue(CLHandle, &ptr->type);
    retval2 = free_CSSM_DATAvalue(CLHandle, &ptr->value);
    
    if (CSSM_OK != retval1)
        return retval1;
    if (CSSM_OK != retval2)
        return retval2;
    return CSSM_OK;

    
}

CSSM_X509_TYPE_VALUE_PAIR_PTR
dup_Pairs(CSSM_CL_HANDLE CLHandle, uint32 NumPairs, CSSM_X509_TYPE_VALUE_PAIR_PTR Pair)
{
    CSSM_X509_TYPE_VALUE_PAIR_PTR dup_ptr = NULL;
    uint32 i=0;
    CSSM_RETURN retval = CSSM_OK;

    dup_ptr = App_Calloc(CLHandle, NumPairs, sizeof(CSSM_X509_TYPE_VALUE_PAIR));
    if (dup_ptr == NULL)
        return NULL;

    for (i=0; i<NumPairs; i++)
    {
        retval = copy_Pair(CLHandle, &dup_ptr[i], &Pair[i]);
        if (retval != CSSM_OK)
        {
            free_Pairs(CLHandle, dup_ptr);
            return NULL;
        }
    }

    return dup_ptr;
}

CSSM_RETURN
copy_Pair(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_TYPE_VALUE_PAIR_PTR toPair,
    CSSM_X509_TYPE_VALUE_PAIR_PTR Pair)
{
    CSSM_RETURN retval = CSSM_OK;

    retval = copy_CSSM_DATAvalue(CLHandle, &toPair->type, &Pair->type);
    if (retval != CSSM_OK)
    {
        free_Pair(CLHandle, toPair);
        return retval;
    }

    toPair->valueType = Pair->valueType;
    
    retval = copy_CSSM_DATAvalue(CLHandle, &toPair->value, &Pair->value);
    if (retval != CSSM_OK)
    {
        free_Pair(CLHandle, toPair);
        return retval;
    }

    return CSSM_OK;
}



CSSM_X509_TYPE_VALUE_PAIR_PTR
dup_Pair(CSSM_CL_HANDLE CLHandle, CSSM_X509_TYPE_VALUE_PAIR_PTR Pair)
{
    CSSM_X509_TYPE_VALUE_PAIR_PTR dup_ptr = NULL;
    CSSM_RETURN retval = CSSM_OK;

    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_X509_TYPE_VALUE_PAIR));
    if (dup_ptr == NULL)
        return NULL;

    retval = copy_CSSM_DATAvalue(CLHandle, &dup_ptr->type, &Pair->type);
    if (retval != CSSM_OK)
    {
        free_Pair(CLHandle, Pair);
        return NULL;
    }

    dup_ptr->valueType = Pair->valueType;
    
    retval = copy_CSSM_DATAvalue(CLHandle, &dup_ptr->value, &Pair->value);
    if (retval != CSSM_OK)
    {
        free_Pair(CLHandle, Pair);
        return NULL;
    }

    return dup_ptr;
}


CSSM_RETURN
free_validity(CSSM_CL_HANDLE CLHandle, 
CSSM_X509_VALIDITY_PTR Validity)
{
    if (Validity == NULL)
    {
        return CSSM_OK;
    }

    if (Validity->notBefore.time.Data != NULL)
    {
        App_Free(CLHandle, Validity->notBefore.time.Data);
        Validity->notBefore.time.Data = NULL;
        Validity->notBefore.time.Length = 0;
    }

    if (Validity->notAfter.time.Data != NULL);
    {
        App_Free(CLHandle, Validity->notAfter.time.Data);
        Validity->notAfter.time.Data = NULL;
        Validity->notBefore.time.Length = 0;
    }

    return CSSM_OK;
}


CSSM_RETURN
copy_validity(CSSM_CL_HANDLE CLHandle, CSSM_X509_VALIDITY_PTR toValidity, CSSM_X509_VALIDITY_PTR Validity)
{
    



    toValidity->notBefore = dup_time(CLHandle, Validity->notBefore);
    if (toValidity->notBefore.time.Data == NULL)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    
    
    toValidity->notAfter = dup_time(CLHandle, Validity->notAfter);
    
    if (toValidity->notAfter.time.Data == NULL)
    {
        App_Free(CLHandle, toValidity->notBefore.time.Data);
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    return CSSM_OK;
}

CSSM_X509_TIME
dup_time(CSSM_CL_HANDLE CLHandle,CSSM_X509_TIME Time)
{
    CSSM_X509_TIME dup_time;

    dup_time.time.Data = NULL;
    dup_time.time.Length = 0;
    
    dup_time.time.Data = App_Calloc(CLHandle, 1, strlen((char *)Time.time.Data)+1);
    if (dup_time.time.Data == NULL)
        return dup_time;
    dup_time.time.Length = Time.time.Length;
    memcpy(dup_time.time.Data, Time.time.Data, strlen((char *)Time.time.Data)+1);

    return dup_time;
}

CSSM_RETURN
copy_subjectPublicKeyInfo(CSSM_CL_HANDLE CLHandle, 
CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR toInfo, 
CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR Info)
{
    CSSM_RETURN retval = CSSM_OK;

    retval = copy_CSSM_DATAvalue(CLHandle, &toInfo->algorithm.algorithm, &Info->algorithm.algorithm);
    if (retval != CSSM_OK)
    {
        return retval;
    }

    retval = copy_CSSM_DATAvalue(CLHandle, &toInfo->algorithm.parameters, &Info->algorithm.parameters);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, toInfo->algorithm.algorithm.Data);
        return retval;
    }

    retval = copy_CSSM_DATAvalue(CLHandle, &toInfo->subjectPublicKey, &Info->subjectPublicKey);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, toInfo->algorithm.parameters.Data);
        App_Free(CLHandle, toInfo->algorithm.algorithm.Data);
        return retval;
    }

    return CSSM_OK;
}

CSSM_RETURN 
free_subjectPublicKeyInfo(CSSM_HANDLE CLHandle, 
CSSM_X509_SUBJECT_PUBLIC_KEY_INFO_PTR ptr)
{
    int errorcount = 0;

    if (ptr == NULL)
    {
        return CSSM_OK;
    }

    errorcount += (int) free_algorithm(CLHandle, &ptr->algorithm);

    errorcount += (int) free_CSSM_DATAvalue(CLHandle, &ptr->subjectPublicKey);

    return (errorcount == 0) ? CSSM_OK : CSSM_FALSE;
}


CSSM_RETURN
free_extensions(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_EXTENSIONS_PTR ptr)
{
    uint32 i=0;
    int errorcount = 0;

    if (ptr == NULL)
    {
        return CSSM_OK;
    }

    if (ptr->numberOfExtensions == 0)
    {
        return CSSM_OK;
    }

    for (i=0; i<ptr->numberOfExtensions; i++)
    {
        errorcount += free_extension(CLHandle, &ptr->extensions[i]);
    }
    if (ptr->extensions != NULL)
    {
        App_Free(CLHandle, ptr->extensions);
        ptr->extensions = NULL;
    }
    ptr->numberOfExtensions = 0;        

    return (errorcount == 0) ? CSSM_OK : CSSMERR_CL_INTERNAL_ERROR;
}

CSSM_RETURN
free_extension(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_EXTENSION_PTR ptr)
{
    CSSM_RETURN retval1=CSSM_OK, retval2=CSSM_OK;

    if (ptr == NULL)
    {
        return CSSM_OK;
    }
    retval1 = free_CSSM_DATAvalue(CLHandle, &ptr->extnId);
    if (ptr->value.tagAndValue)
    {
        retval2 = free_CSSM_DATAvalue(CLHandle, &ptr->value.tagAndValue->value);
        App_Free(CLHandle, ptr->value.tagAndValue);    
    }

    if (CSSM_OK != retval1)
        return retval1;
    if (CSSM_OK != retval2)
        return retval2;
    return CSSM_OK;

}

CSSM_RETURN
copy_extension_value(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_EXTENSION_PTR toExtension,
    CSSM_X509_EXTENSION_PTR Extension)  
{
    CSSM_RETURN retval = CSSM_OK;

    switch(Extension->format)
    {
        
        default:
            toExtension->value.tagAndValue = App_Calloc(CLHandle,1,sizeof(CSSM_X509EXT_TAGandVALUE));
            if (!toExtension->value.tagAndValue)
                return CSSMERR_CL_MEMORY_ERROR;
            toExtension->value.tagAndValue->type = Extension->value.tagAndValue->type;
            retval = copy_CSSM_DATAvalue(CLHandle, 
                        &toExtension->value.tagAndValue->value,
                        &Extension->value.tagAndValue->value);
            if (retval != CSSM_OK)
            {
                return retval;
            }           
            break;
    }
    return CSSM_OK;
}


CSSM_RETURN
copy_extension(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_EXTENSION_PTR toExtension,
    CSSM_X509_EXTENSION_PTR Extension)  
{
    CSSM_RETURN retval = CSSM_OK;

    retval = copy_CSSM_DATAvalue(CLHandle, &toExtension->extnId, &Extension->extnId);
    if (retval != CSSM_OK)
    {
        free_extension(CLHandle, toExtension);
        return retval;
    }
    
    toExtension->critical = Extension->critical;

    toExtension->format = Extension->format;
    retval = copy_extension_value(CLHandle, toExtension, Extension);
    if (retval != CSSM_OK)
    {
        free_extension(CLHandle, toExtension);
        return retval;
    }

    retval = copy_CSSM_DATAvalue(CLHandle, &toExtension->BERvalue, &Extension->BERvalue);
    if (retval != CSSM_OK)
    {
        free_extension(CLHandle, toExtension);
        return retval;
    }

    return CSSM_OK;
}

CSSM_RETURN
copy_extensions(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_EXTENSIONS_PTR toExtensions,
    CSSM_X509_EXTENSIONS_PTR Extensions)    
{
    CSSM_RETURN retval = CSSM_OK;
    uint32 i;

    toExtensions->numberOfExtensions = Extensions->numberOfExtensions;

    toExtensions->extensions = App_Calloc(CLHandle, 
                                                Extensions->numberOfExtensions, 
                                                sizeof(CSSM_X509_EXTENSION));
    if (toExtensions->extensions == NULL)
        return CSSMERR_CL_INTERNAL_ERROR;

    for (i=0; i<Extensions->numberOfExtensions; i++)
    {
        retval = copy_extension(CLHandle, &toExtensions->extensions[i], &Extensions->extensions[i]);
        if (retval != CSSM_OK)
        {
            free_extensions(CLHandle, toExtensions);
            return retval;
        }
    }

    return CSSM_OK;
}


CSSM_X509_EXTENSIONS_PTR
dup_extensions(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_EXTENSIONS_PTR Extensions)
{
    CSSM_X509_EXTENSIONS_PTR dup_ptr=NULL;
    
    CSSM_RETURN retval = CSSM_OK;

    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_X509_EXTENSIONS));
    if (dup_ptr == NULL)
        return NULL;

    if (Extensions->numberOfExtensions != 0)
    {
        retval = copy_extensions(CLHandle, dup_ptr, Extensions);
        if (retval != CSSM_OK)
            return NULL;
    }

    return dup_ptr;
}

/*-----------------------------------------------------------------------------
 * Name: get_signedCRL_field
 *
 * Description:
 * Makes a copy of the SignedCRL C structure in memory from the
 * application's heap.  
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * SignedCRL (input) :       The SignedCRL to be copied. 
 * ret_ptr (output):       A non-NULL pointer to a CSSM_DATA_PTR whose Data 
 *                         field points to the SignedCRL C structure or NULL 
 *                         if there was an error.
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN
get_signedCRL_field(CSSM_CL_HANDLE CLHandle, 
X509_SIGNED_CRL_PTR SignedCRL, 
    CSSM_DATA_PTR ret_ptr)
{
    if (ret_ptr == NULL)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    /* set the length */
    ret_ptr->Length = CSSM_FIELDVALUE_COMPLEX_DATA_TYPE;

    /* duplicate the signed certificate */
    ret_ptr->Data = (uint8 *) dup_signedCRL(CLHandle, SignedCRL);
    if (ret_ptr->Data == NULL)  /* if unsuccessful, then free the return structure */
    {
        App_Free(CLHandle, ret_ptr);
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    return CSSM_OK;
}

CSSM_RETURN 
free_certListvalues(CSSM_CL_HANDLE CLHandle, 
X509_TBS_CERTLIST_PTR CertList)
{
    
    int errorcount = 0;

    if (CertList == NULL)
    {
        return CSSM_OK;
    }

    /* version */
    if (free_CSSM_DATAvalue(CLHandle, &CertList->version) != CSSM_OK)
        errorcount++;

    /* signature */
    if (free_algorithm(CLHandle, &CertList->signature) != CSSM_OK)
        errorcount++;

    /* issuer */
    if (free_name(CLHandle, &CertList->issuer) != CSSM_OK)
        errorcount++;

    /* thisUpdate */
    if (CertList->thisUpdate.time.Data)
    {
        App_Free(CLHandle, CertList->thisUpdate.time.Data);
        CertList->thisUpdate.time.Length = 0;
    }

    /* nextUpdate */
    if (CertList->nextUpdate.time.Data)
    {
        App_Free(CLHandle, CertList->nextUpdate.time.Data);
        CertList->nextUpdate.time.Length = 0;
    }

    /* revokedCertificates */
    if (free_revokedCertificates(CLHandle, CertList->revokedCertificates) != CSSM_OK)
        errorcount++;
    App_Free(CLHandle, CertList->revokedCertificates);
    CertList->revokedCertificates = NULL;

    /* extensions */
    if (free_extensions(CLHandle, &CertList->extensions) != CSSM_OK)
        errorcount++;

    return ((errorcount==0) ? CSSM_OK : CSSMERR_CL_INTERNAL_ERROR);
}


CSSM_RETURN 
free_certListvalues_array(CSSM_CL_HANDLE CLHandle, 
CSSM_X509_TBS_CERTLIST_PTR CertList)
{
    
    int errorcount = 0;

    if (CertList == NULL)
    {
        return CSSM_OK;
    }

    /* version */
    if (free_CSSM_DATAvalue(CLHandle, &CertList->version) != CSSM_OK)
        errorcount++;

    /* signature */
    if (free_algorithm(CLHandle, &CertList->signature) != CSSM_OK)
        errorcount++;

    /* issuer */
    if (free_name(CLHandle, &CertList->issuer) != CSSM_OK)
        errorcount++;

    /* thisUpdate */
    if (CertList->thisUpdate.time.Data)
    {
        App_Free(CLHandle, CertList->thisUpdate.time.Data);
        CertList->thisUpdate.time.Length = 0;
    }

    /* nextUpdate */
    if (CertList->nextUpdate.time.Data)
    {
        App_Free(CLHandle, CertList->nextUpdate.time.Data);
        CertList->nextUpdate.time.Length = 0;
    }

    /* revokedCertificates */
    if (free_revokedCertificates_array(CLHandle, CertList->revokedCertificates) != CSSM_OK)
        errorcount++;
    CertList->revokedCertificates = NULL;

    /* extensions */
    if (free_extensions(CLHandle, &CertList->extensions) != CSSM_OK)
        errorcount++;

    return ((errorcount==0) ? CSSM_OK : CSSMERR_CL_INTERNAL_ERROR);
}

CSSM_RETURN 
copy_certListvalues(CSSM_CL_HANDLE CLHandle, 
CSSM_X509_TBS_CERTLIST_PTR toCertList,
X509_TBS_CERTLIST_PTR CertList)
{
    CSSM_RETURN retval = CSSM_OK;

    /* version */
    retval = copy_CSSM_DATAvalue(CLHandle, &toCertList->version, &CertList->version);
    if (retval != CSSM_OK)
    {
        return retval;
    }

    /* signature */
    retval = copy_algorithm(CLHandle, &toCertList->signature, &CertList->signature);
    if (retval != CSSM_OK)
    {
        free_certListvalues_array(CLHandle, toCertList);
        return retval;
    }

    /* issuer */
    retval = copy_name(CLHandle, &toCertList->issuer, &CertList->issuer);
    if (retval != CSSM_OK)
    {
        free_certListvalues_array(CLHandle, toCertList);
        return retval;
    }

    /* thisUpdate */
    toCertList->thisUpdate = dup_time(CLHandle, CertList->thisUpdate);
    if (toCertList->thisUpdate.time.Data == NULL)
    {
        free_certListvalues_array(CLHandle, toCertList);
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    /* nextUpdate */
    toCertList->nextUpdate = dup_time(CLHandle, CertList->nextUpdate);
    if (toCertList->nextUpdate.time.Data == NULL)
    {
        free_certListvalues_array(CLHandle, toCertList);
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    /* revokedCertificates */
    if (CertList->revokedCertificates != NULL)
    {
        toCertList->revokedCertificates = dup_revokedCertificates(CLHandle, 
                                                CertList->revokedCertificates);
        if (toCertList->revokedCertificates == NULL)
        {
            free_certListvalues_array(CLHandle, toCertList);
            return CSSMERR_CL_INTERNAL_ERROR;
        }
    }

    /* extensions */
    if (CertList->extensions.numberOfExtensions != 0)
    {
        retval = copy_extensions(CLHandle, &toCertList->extensions, &CertList->extensions);
        if (retval != CSSM_OK)
        {
            free_certListvalues_array(CLHandle, toCertList);
            return retval;
        }
    }

    return CSSM_OK;
}

CSSM_RETURN 
copy_signedCRL(CSSM_CL_HANDLE CLHandle, 
CSSM_X509_SIGNED_CRL_PTR toSignedCRL,
X509_SIGNED_CRL_PTR SignedCRL)
{
    CSSM_RETURN retval = CSSM_OK;

    if (toSignedCRL == NULL)
    { 
        return CSSMERR_CL_INVALID_INPUT_POINTER;
    }

    /* copy the certificate */
    retval = copy_certListvalues(CLHandle, &toSignedCRL->tbsCertList, &SignedCRL->tbsCertList);
    if (retval != CSSM_OK)
    {
        return retval;
    }

    /* copy the signature */
    retval = copy_signaturevalues(CLHandle, &toSignedCRL->signature, &SignedCRL->signature);
    if (retval != CSSM_OK)
    {
        return retval;
    }

    return CSSM_OK;
}

CSSM_RETURN 
free_signedCRL(CSSM_CL_HANDLE CLHandle,
X509_SIGNED_CRL_PTR SignedCRL)
{
    CSSM_RETURN retval1, retval2;

    if (SignedCRL == NULL)
    {
        return CSSM_OK;
    }
    
    retval1 = free_certListvalues(CLHandle, &SignedCRL->tbsCertList);
        

    retval2= free_signaturevalues(CLHandle, &SignedCRL->signature);
        

    App_Free(CLHandle, SignedCRL);

    
    if (CSSM_OK != retval1)
        return retval1;
    if (CSSM_OK != retval2)
        return retval2;
    return CSSM_OK;

}

CSSM_RETURN 
free_signedCRL_array(CSSM_CL_HANDLE CLHandle,
CSSM_X509_SIGNED_CRL_PTR SignedCRL)
{
    CSSM_RETURN retval1, retval2;

    if (SignedCRL == NULL)
    {
        return CSSM_OK;
    }
    
    retval1 = free_certListvalues_array(CLHandle, &SignedCRL->tbsCertList);
        

    retval2= free_signaturevalues(CLHandle, &SignedCRL->signature);
        

    App_Free(CLHandle, SignedCRL);

    
    if (CSSM_OK != retval1)
        return retval1;
    if (CSSM_OK != retval2)
        return retval2;
    return CSSM_OK;

}

CSSM_RETURN 
free_certlist(CSSM_CL_HANDLE CLHandle,
X509_TBS_CERTLIST_PTR list)
{
    CSSM_RETURN retval;

    if (list == NULL)
    {
        return CSSM_OK;
    }
    
    retval = free_certListvalues(CLHandle, list);
        

    App_Free(CLHandle, list);

    return retval;
}


CSSM_X509_SIGNED_CRL_PTR
dup_signedCRL(CSSM_CL_HANDLE CLHandle, 
 X509_SIGNED_CRL_PTR SignedCRL)
{
    CSSM_X509_SIGNED_CRL_PTR dup_ptr = NULL;
    CSSM_RETURN retval = CSSM_OK;

    /* allocate the signed CRL structure */
    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_X509_SIGNED_CRL));
    if (dup_ptr == NULL)
        return NULL;

    /* copy the CRL */
    retval = copy_signedCRL(CLHandle, dup_ptr, SignedCRL);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, dup_ptr);
        return NULL;
    }

    return dup_ptr;
}

CSSM_X509_REVOKED_CERT_LIST_PTR
dup_revokedCertificates(CSSM_CL_HANDLE CLHandle, 
 X509_REVOKED_CERT_LIST_PTR CertList)
{
    CSSM_X509_REVOKED_CERT_LIST_PTR dup_ptr = NULL;
    CSSM_RETURN retval = CSSM_OK;

    /* allocate the revoked certificate list structure */
    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_X509_REVOKED_CERT_LIST));
    if (dup_ptr == NULL)
        return NULL;

    /* copy the revoked certificate list */
    retval = copy_revokedCertificates(CLHandle, dup_ptr, CertList);
    if (retval != CSSM_OK)
    {
        App_Free(CLHandle, dup_ptr);
        return NULL;
    }

    return dup_ptr;
}


CSSM_RETURN 
copy_revokedCertificates(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_REVOKED_CERT_LIST_PTR toRevokedCertList,
    X509_REVOKED_CERT_LIST_PTR RevokedCertList)
{
    CSSM_RETURN retval = CSSM_OK;
    uint32 i = 0;
    uint32 j = 0;
    uint32 numNodes = 0;
    
    X509_REVOKED_LIST_NODE_PTR node_ptr = NULL;
    CSSM_X509_REVOKED_CERT_ENTRY_PTR toEntry_ptr = NULL;

    toRevokedCertList->numberOfRevokedCertEntries = RevokedCertList->numNodes;
    toRevokedCertList->revokedCertEntry = NULL;

    numNodes = RevokedCertList->numNodes;
    node_ptr = RevokedCertList->head;

    if (numNodes == 0)
        return CSSM_OK;

    /* allocate contiguous array for entries */
    toEntry_ptr = App_Calloc(CLHandle, numNodes, sizeof(CSSM_X509_REVOKED_CERT_ENTRY));
    if (toEntry_ptr == NULL)
    {
        return CSSMERR_CL_MEMORY_ERROR;
    }

    


    for (i=0; i < numNodes; i++)
    {
        /* fill in cert entry */
        retval = copy_revokedCertEntry(CLHandle, &toEntry_ptr[i], node_ptr->revokedCertEntry);

        if (retval != CSSM_OK)
        {
            for(j=0;j<i;j++)
            {
                free_revokedCertEntry(CLHandle, &toEntry_ptr[j]);
            }
            App_Free(CLHandle, toEntry_ptr);
            return retval;
        }
        
        node_ptr = node_ptr->nextNode;
    }

    toRevokedCertList->revokedCertEntry = toEntry_ptr;
    return CSSM_OK;
}

CSSM_RETURN 
copy_revokedCertEntry(CSSM_CL_HANDLE CLHandle, 
 CSSM_X509_REVOKED_CERT_ENTRY_PTR toRevokedCertEntry,
 CSSM_X509_REVOKED_CERT_ENTRY_PTR RevokedCertEntry)
{
    CSSM_RETURN retval = CSSM_OK;
    

    /* certificateSerialNumber */
    retval = copy_CSSM_DATAvalue(CLHandle, 
            &toRevokedCertEntry->certificateSerialNumber,
            &RevokedCertEntry->certificateSerialNumber);
    if (CSSM_OK != retval)      
    {
        return retval;
    }

    /* revocationDate */
    toRevokedCertEntry->revocationDate = dup_time(CLHandle, RevokedCertEntry->revocationDate);
    if ((RevokedCertEntry->revocationDate.time.Data != NULL) 
        && (toRevokedCertEntry->revocationDate.time.Data == NULL))
    {
        free_revokedCertEntry(CLHandle, toRevokedCertEntry);
    }

    /* extensions */
    if (RevokedCertEntry->extensions.numberOfExtensions != 0)
    {
        retval = copy_extensions(CLHandle, 
                    &toRevokedCertEntry->extensions, 
                    &RevokedCertEntry->extensions);
        if (retval != CSSM_OK)
        {
            free_revokedCertEntry(CLHandle, toRevokedCertEntry);
            return retval;
        }
    }

    return CSSM_OK;
}

CSSM_RETURN 
free_revokedCertEntry(CSSM_CL_HANDLE CLHandle, 
 CSSM_X509_REVOKED_CERT_ENTRY_PTR RevokedCertEntry)
{
    
    
    CSSM_RETURN retval1=CSSM_OK, retval2=CSSM_OK;

    if (RevokedCertEntry == NULL)
    {
        return CSSM_OK;
    }

    /* certificateSerialNumber */
    retval1 = free_CSSM_DATAvalue(CLHandle, &RevokedCertEntry->certificateSerialNumber);
    
    /* revocationDate */
    if (RevokedCertEntry->revocationDate.time.Data != NULL)
    {
        App_Free(CLHandle, RevokedCertEntry->revocationDate.time.Data);
    }

    /* extensions */
    if (RevokedCertEntry->extensions.numberOfExtensions != 0)
    {       
        retval2 = free_extensions(CLHandle, &RevokedCertEntry->extensions);
            
    }

    
    if (CSSM_OK != retval1)
        return retval1;
    if (CSSM_OK != retval2)
        return retval2;
    return CSSM_OK;

}

CSSM_RETURN 
free_revokedCertificatevalues(CSSM_CL_HANDLE CLHandle, 
    X509_REVOKED_CERT_LIST_PTR RevokedCertList)
{
    
    uint32 i = 0;
    uint32 numNodes = 0;
    X509_REVOKED_LIST_NODE_PTR node_ptr = NULL;
    uint32 errorcount = 0;

    if (RevokedCertList == NULL)
    {
        return CSSM_OK;
    }

    numNodes = RevokedCertList->numNodes;

    if (numNodes == 0)
        return CSSM_OK;

    node_ptr = RevokedCertList->head;

    for (i=0; i < numNodes; i++)
    {
        if (free_revokedCertEntry(CLHandle, node_ptr->revokedCertEntry) != CSSM_OK)


            errorcount++;
            
        node_ptr = node_ptr->nextNode;
    }


    return CSSM_OK;
}

CSSM_RETURN 
free_revokedCertificates(CSSM_CL_HANDLE CLHandle, 
    X509_REVOKED_CERT_LIST_PTR RevokedCertList)
{
    CSSM_RETURN retval = CSSM_OK;
    uint32 i = 0;
    uint32 numNodes = 0;
    X509_REVOKED_LIST_NODE_PTR node_ptr = NULL;
    uint32 errorcount = 0;

    if (RevokedCertList == NULL)
    {
        return CSSM_OK;
    }

    numNodes = RevokedCertList->numNodes;

    if (numNodes == 0)


        return CSSM_OK;


    node_ptr = RevokedCertList->head;
    if (node_ptr == NULL)
        return CSSMERR_CL_INTERNAL_ERROR;

    for (i=0; i < numNodes && node_ptr != NULL; i++)
    {
        /* delete cert entry */
        retval = free_revokedCertEntry(CLHandle, node_ptr->revokedCertEntry);
        if (retval != CSSM_OK)
        {
            errorcount++;
        }
        if (node_ptr->revokedCertEntry != NULL)
        {
            App_Free(CLHandle, node_ptr->revokedCertEntry);
        }
        
        /* delete node */
        node_ptr->revokedCertEntry = NULL;
        if ((i+1) < numNodes) {
            RevokedCertList->head = node_ptr->nextNode;
            App_Free(CLHandle, node_ptr);
            node_ptr = RevokedCertList->head;
            RevokedCertList->numNodes--;
        }
    }
    if (i != numNodes)
    {
        return CSSMERR_CL_INTERNAL_ERROR;
    }
    if (node_ptr)
        App_Free(CLHandle, node_ptr);
    

    
    return CSSM_OK;

}

CSSM_RETURN 
free_revokedCertificates_array(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_REVOKED_CERT_LIST_PTR RevokedCertList)
{
    CSSM_RETURN retval = CSSM_OK;
    uint32 i = 0;
    uint32 numNodes = 0;
    
    uint32 errorcount = 0;

    if (RevokedCertList == NULL)
    {
        return CSSM_OK;
    }

    numNodes = RevokedCertList->numberOfRevokedCertEntries;

    if (numNodes == 0)
        return CSSM_OK;



    for (i=0; i < numNodes; i++)
    {
        /* delete cert entry */
        retval = free_revokedCertEntry(CLHandle, &RevokedCertList->revokedCertEntry[i]);
        if (retval != CSSM_OK)
        {
            errorcount++;
        }
        
    }
    App_Free(CLHandle, RevokedCertList->revokedCertEntry);
    App_Free(CLHandle, RevokedCertList);
    return CSSM_OK;

}


/*-----------------------------------------------------------------------------
 * Name: get_TBSCertList_field
 *
 * Description:
 * Makes a copy of the TBSCertList in memory from the
 * application's heap.  
 *
 * Parameters: 
 * CLHandle (input) :         The caller's CL handle.
 * TBSCertList (input) :      The TBSCertList to be copied. 
 *
 * Return value:
 * A non-NULL pointer to a CSSM_DATA_PTR whose Data field points to the 
 *  TBSCertList C structure. 
 * or NULL if there was an error.
 * 
 *---------------------------------------------------------------------------*/
CSSM_RETURN
get_TBSCertList_field(CSSM_CL_HANDLE CLHandle, 
    X509_TBS_CERTLIST_PTR from_ptr, 
    CSSM_DATA_PTR ret_ptr)
{
    if (ret_ptr == NULL)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    /* set the length */
    ret_ptr->Length = CSSM_FIELDVALUE_COMPLEX_DATA_TYPE;

    /* duplicate the TBS cert list */
    ret_ptr->Data = (uint8 *) dup_TBSCertList(CLHandle, from_ptr);
    if (ret_ptr->Data == NULL)  /* if unsuccessful, then free the return structure */
    {
        App_Free(CLHandle, ret_ptr);
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    return CSSM_OK;
}

CSSM_X509_TBS_CERTLIST_PTR
dup_TBSCertList(CSSM_CL_HANDLE CLHandle, 
    X509_TBS_CERTLIST_PTR from_ptr)
{
    CSSM_X509_TBS_CERTLIST_PTR dup_ptr = NULL;

    /* allocate the cert list structure */
    dup_ptr = App_Calloc(CLHandle, 1, sizeof(CSSM_X509_TBS_CERTLIST));
    if (dup_ptr == NULL)
        return NULL;

    /* copy the TBS cert list */
    if (copy_certListvalues(CLHandle, dup_ptr, from_ptr) != CSSM_OK)
    {
        App_Free(CLHandle, dup_ptr);
        return NULL;
    }

    return dup_ptr;
}

CSSM_RETURN
get_RevokedCertificates_field(CSSM_CL_HANDLE CLHandle, 
    X509_REVOKED_CERT_LIST_PTR from_ptr, 
    CSSM_DATA_PTR ret_ptr)
{
    if (ret_ptr == NULL)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }

    /* set the length */
    ret_ptr->Length = CSSM_FIELDVALUE_COMPLEX_DATA_TYPE;

    /* duplicate */
    ret_ptr->Data = (uint8 *) dup_revokedCertificates(CLHandle, from_ptr);
    if (ret_ptr->Data == NULL)  /* if unsuccessful, then free the return structure */
    {
        App_Free(CLHandle, ret_ptr);
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    return CSSM_OK;
}

CSSM_RETURN
get_Extensions_field(CSSM_CL_HANDLE CLHandle, 
    CSSM_X509_EXTENSIONS_PTR from_ptr,
    CSSM_DATA_PTR ret_ptr)
{
    if (ret_ptr == NULL)
    {
        return CSSMERR_CL_INVALID_OUTPUT_POINTER;
    }
    
    /* set the length */
    ret_ptr->Length = CSSM_FIELDVALUE_COMPLEX_DATA_TYPE;

    /* duplicate */
    ret_ptr->Data = (uint8 *) dup_extensions(CLHandle, from_ptr);
    if (ret_ptr->Data == NULL)  /* if unsuccessful, then free the return structure */
    {
        App_Free(CLHandle, ret_ptr);
        return CSSMERR_CL_INTERNAL_ERROR;
    }

    return CSSM_OK;
}


