/*-----------------------------------------------------------------------
 * File: DERIVE.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

extern CSSM_RETURN EAYCSP_CheckSupportedAttrs( uint32 KeyAttr );

static CSSM_RETURN EAYCSP_ValidatePBKDF1Params(
                        CSSM_PKCS5_PBKDF1_PARAMS *pParams )
{
    CSSM_RETURN rv = CSSM_OK;

    if ( ( ( rv = CSM_ValidateInputBuffers( &pParams->Passphrase,
                                            1,
                                            CSSM_FALSE ) ) == CSSM_OK ) &&
         ( ( rv = CSM_ValidateOutputBuffers( &pParams->InitVector,
                                             1,
                                             CSSM_FALSE ) ) == CSSM_OK ) )
    {
        if ( pParams->Passphrase.Length == 0 )
        {
            ERR( rv = CSSMERR_CSP_INVALID_INPUT_POINTER );
        }
    }

    return rv;
}

static CSSM_RETURN EAYCSP_ValidatePBKDF2Params(
                        CSSM_PKCS5_PBKDF2_PARAMS *pParams )
{
    CSSM_RETURN rv = CSSM_OK;

    if ( ( rv = CSM_ValidateInputBuffers( &pParams->Passphrase,
                                          1,
                                          CSSM_FALSE ) ) == CSSM_OK )
    {
        if ( pParams->PseudoRandomFunction != 
                                        CSSM_PKCS5_PBKDF2_PRF_HMAC_SHA1 )
        {
            ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
        }
        else if ( pParams->Passphrase.Length == 0 )
        {
            ERR( rv = CSSMERR_CSP_INVALID_INPUT_POINTER );
        }
    }

    return rv;
}

static CSSM_RETURN __PerformPBKDF1(
                        CSSM_CSP_HANDLE CSPHandle,
                        uint32 KeySize,
                        const CSSM_DATA *pPassPhrase,
                        const CSSM_DATA *pSalt,
                        uint32 Iterations,
                        EVP_MD *pDigest,
                        CSSM_KEY *pKey,
                        CSSM_DATA *pIV )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 KeyBytes;
    CSSM_BOOL bAllocatedIV = CSSM_FALSE;
    uint8 TempBuf[64];
    uint32 DataLen;
    EVP_MD_CTX MdCtx;
    uint32 i;

    /* Can the key size be accomodated? */
    KeyBytes = KeySize / 8;
    if ( KeyBytes > 8 ) /* PKCS #5 specified limitation */
    {
        ERR( rv = CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH );
    }
    else
    {
        /* Allocate space for the IV */
        if ( ( rv = MAF_CreateOutputBuffer( CSPHandle,
                                            8,
                                            pIV,
                                            &bAllocatedIV ) ) == CSSM_OK )
        {
            /* Allocate space for the key buffer */
            pKey->KeyData.Data = Addin_malloc( KeyBytes, NULL );
            if ( pKey->KeyData.Data == NULL )
            {
                ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
            }
            else
            {   
                pKey->KeyData.Length = KeyBytes;

                /* Do the initial iteration of the operation */
                memset( &MdCtx, 0, sizeof(MdCtx) );
                EVP_DigestInit( &MdCtx, pDigest );
                EVP_DigestUpdate( &MdCtx,
                                  pPassPhrase->Data,
                                  pPassPhrase->Length );
                EVP_DigestUpdate( &MdCtx,
                                  pSalt->Data, 
                                  pSalt->Length );
                EVP_DigestFinal( &MdCtx, TempBuf, (unsigned int *)&DataLen ); /* ifdef VMS apply cast */

                /* Do the rest of the iterations */
                for ( i = 1; i < Iterations; i++ )
                {
                    EVP_DigestInit( &MdCtx, pDigest );
                    EVP_DigestUpdate( &MdCtx, TempBuf, DataLen );
                    EVP_DigestFinal( &MdCtx, TempBuf, (unsigned int *)&DataLen ); /* ifdef VMS apply cast */
                }

                /* Split up the result into the key and IV */
                memcpy( pKey->KeyData.Data, TempBuf, KeyBytes );
                memcpy( pIV->Data, &TempBuf[8], 8 );

                /* Clear the state info */
                memset( &MdCtx, 0, sizeof(&MdCtx) );
                memset( TempBuf, 0, sizeof(TempBuf) );
            } /* Allocate key buffer */

            if ( rv != CSSM_OK )
            {
                MAF_DeleteOutputBuffer( CSPHandle,
                                        pIV,
                                        bAllocatedIV );
            }
        } /* Allocate IV */
    } /* Check key length */

    return rv;
}

static CSSM_RETURN __DeriveKeyPBKDF1(
                        CSM_SESSION *pSession,
                        CSSM_ALGORITHMS KeyType,
                        uint32 KeySize,
                        CSSM_DATA_PTR pAlgParam,
                        EVP_MD *pDigest,
                        CSSM_KEY *pKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_PKCS5_PBKDF1_PARAMS *pParams = NULL;
    uint32 Iterations = 1000;
    CSSM_DATA *pSalt = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pIterationAttr = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pSaltAttr = NULL;

    /* Make sure the parameter structure is probably valid */
    if ( port_IsBadWritePtr( pAlgParam, sizeof(*pAlgParam) ) ||
         port_IsBadReadPtr( pAlgParam->Data, pAlgParam->Length ) ||
         ( pAlgParam->Length != sizeof(*pParams) ) )
    {
        ERR( rv = CSSMERR_CSP_INVALID_INPUT_POINTER );
    }
    else
    {
        /* Cast the parameter pointer */
        pParams = (CSSM_PKCS5_PBKDF1_PARAMS*)pAlgParam->Data;

        /* Validate the parameter structure contents */
        if ( ( rv = EAYCSP_ValidatePBKDF1Params( pParams ) ) == CSSM_OK )
        {
            /* Fetch the params from the context */
            pIterationAttr = CSM_GetContextAttribute(
                                            pSession->pContext,
                                            CSSM_ATTRIBUTE_ITERATION_COUNT );
            pSaltAttr = CSM_GetContextAttribute( pSession->pContext,
                                                 CSSM_ATTRIBUTE_SALT );
            if ( pIterationAttr == NULL )
            {
                ERR( rv = CSSMERR_CSP_MISSING_ATTR_ITERATION_COUNT );
            }
            else if ( pSaltAttr == NULL )
            {
                ERR( rv = CSSMERR_CSP_MISSING_ATTR_SALT );
            }
            else if ( pSaltAttr->Attribute.Data->Length != 8 )
            {
                ERR( rv = CSSMERR_CSP_INVALID_ATTR_SALT );
            }
            else
            {
                pSalt = pSaltAttr->Attribute.Data;
                Iterations = pIterationAttr->Attribute.Uint32;

                /* Check if we can derive the requested key type */
                if ( KeyType == CSSM_ALGID_DES )
                {
                    /* Validate the key length */
                    if ( ( rv = EAYCSP_CheckKeyLimits(
                                                KeyType,
                                                KeySize ) ) == CSSM_OK )
                    {
                        /* Derive the values */
                        if ( ( rv = __PerformPBKDF1(
                                            pSession->hCsp,
                                            KeySize,
                                            &pParams->Passphrase,
                                            pSalt,
                                            Iterations,
                                            pDigest,
                                            pKey,
                                            &pParams->InitVector ) )
                                                            == CSSM_OK )
                        {
                            /* Adjust the key header */
                            pKey->KeyHeader.AlgorithmId = KeyType;
                            pKey->KeyHeader.LogicalKeySizeInBits = KeySize;
                        }
                    } /* Check key limits */
                } /* Valid derived key type */
            } /* Fetch and validate context attributes */
        } /* Validate the PBKDF1 params */
    } /* Convert param struct to PBKDF1 params */

    return rv;
}

static void __PKCS5_PBKD2_F(
                        const CSSM_DATA *Password,
                        const CSSM_DATA *Salt,
                        uint32 Iterations,
                        uint32 Index,
                        uint8 *Result )
{
    uint8 IndexBytes[4];
    uint32 i, j;
    uint8 TempF[SHA_DIGEST_LENGTH];
    uint8 IterationResult[SHA_DIGEST_LENGTH];
    uint32 ResultSize = sizeof(IterationResult);
    HMAC_CTX HmacCtx;

    /* Convert the index value into bytes */
    for ( i = 0; i < sizeof(IndexBytes); i++ )
    {
        IndexBytes[i] = (uint8)( ( Iterations >> ( 8 * i ) ) & 0xFF );
    }

    /* Perform the first iteration */
    HMAC_Init( &HmacCtx, Password->Data, Password->Length, EVP_sha1() );
    HMAC_Update( &HmacCtx, Salt->Data, Salt->Length );
    HMAC_Update( &HmacCtx, IndexBytes, sizeof(IndexBytes) );
    HMAC_Final( &HmacCtx, IterationResult, (unsigned int *)&ResultSize );    /* ifdef VMS apply cast */
    HMAC_cleanup( &HmacCtx );
    memcpy( TempF, IterationResult, sizeof(IterationResult) );

    /* Perform the other iterations */
    for ( i = 1; i < Iterations; i++ )
    {
        /* HMAC the previous iteration with the password */
        HMAC( EVP_sha1(),
              Password->Data, Password->Length,
              IterationResult, ResultSize,
              IterationResult, (unsigned int *)&ResultSize );  /* ifdef VMS apply cast */

        /* XOR this iteration with the temporary state */
        for ( j = 0; j < sizeof(TempF); j++ )
        {
            TempF[j] ^= IterationResult[j];
        }
    }

    /* Copy the temporary buffer to the result */
    memcpy( Result, TempF, sizeof(TempF) );
    memset( IterationResult, 0, sizeof(IterationResult) );
    memset( TempF, 0, sizeof(TempF) );

}

static CSSM_RETURN __PerformPBKDF2(
                        CSSM_CSP_HANDLE CSPHandle,
                        uint32 KeySize,
                        const CSSM_DATA *pPassPhrase,
                        const CSSM_DATA *pSalt,
                        uint32 Iterations,
                        CSSM_KEY *pKey )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 KeyBytes;
    uint32 TopIndex;
    uint32 Index;
    uint8 *DataPtr = NULL;
    uint8 TempBuf[SHA_DIGEST_LENGTH];

    /* Calculate algorithm parameters */
    KeyBytes = KeySize / 8;

    /* Allocate space for the key buffer */
    pKey->KeyData.Data = Addin_malloc( KeyBytes, NULL );
    if ( pKey->KeyData.Data == NULL )
    {
        ERR( rv = CSSMERR_CSP_MEMORY_ERROR );
    }
    else
    {   
        pKey->KeyData.Length = KeyBytes;

        /* Calculate the number of blocks to calculate */
        TopIndex = ( KeyBytes / SHA_DIGEST_LENGTH ) +
                    ( ( ( KeyBytes % SHA_DIGEST_LENGTH ) > 0 ) ? 1 : 0 );

        /* Calculate TopIndex blocks to produce DataLength bytes */
        DataPtr = pKey->KeyData.Data;
        for ( Index = 1; Index <= TopIndex; Index++ )
        {
            __PKCS5_PBKD2_F( pPassPhrase,
                             pSalt,
                             Iterations,
                             Index,
                             TempBuf );
            memcpy( DataPtr,
                    TempBuf,
                    min( SHA_DIGEST_LENGTH, KeyBytes ) );
            KeyBytes -= SHA_DIGEST_LENGTH;
            DataPtr += SHA_DIGEST_LENGTH;
        }

        /* Clear the state info */
        memset( TempBuf, 0, sizeof(TempBuf) );
    } /* Allocate key buffer */

    return rv;
}

static CSSM_RETURN __DeriveKeyPBKDF2(
                        CSM_SESSION *pSession,
                        CSSM_ALGORITHMS KeyType,
                        uint32 KeySize,
                        CSSM_DATA_PTR pAlgParam,
                        CSSM_KEY *pKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_PKCS5_PBKDF2_PARAMS *pParams = NULL;
    uint32 Iterations = 1000;
    CSSM_DATA *pSalt = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pIterationAttr = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pSaltAttr = NULL;

    /* Make sure the parameter structure is probably valid */
    if ( port_IsBadWritePtr( pAlgParam, sizeof(*pAlgParam) ) ||
         port_IsBadReadPtr( pAlgParam->Data, pAlgParam->Length ) ||
         ( pAlgParam->Length != sizeof(*pParams) ) )
    {
        ERR( rv = CSSMERR_CSP_INVALID_INPUT_POINTER );
    }
    else
    {
        /* Cast the parameter pointer */
        pParams = (CSSM_PKCS5_PBKDF2_PARAMS*)pAlgParam->Data;

        /* Validate the parameter structure contents */
        if ( ( rv = EAYCSP_ValidatePBKDF2Params( pParams ) ) == CSSM_OK )
        {
            /* Fetch the params from the context */
            pIterationAttr = CSM_GetContextAttribute(
                                            pSession->pContext,
                                            CSSM_ATTRIBUTE_ITERATION_COUNT );
            pSaltAttr = CSM_GetContextAttribute( pSession->pContext,
                                                 CSSM_ATTRIBUTE_SALT );
            if ( pIterationAttr == NULL )
            {
                ERR( rv = CSSMERR_CSP_MISSING_ATTR_ITERATION_COUNT );
            }
            else if ( pSaltAttr == NULL )
            {
                ERR( rv = CSSMERR_CSP_MISSING_ATTR_SALT );
            }
            else
            {
                pSalt = pSaltAttr->Attribute.Data;
                Iterations = pIterationAttr->Attribute.Uint32;

                /* Check if we can derive the requested key type */
                if ( KeyType == CSSM_ALGID_DES )
                {
                    /* Validate the key length */
                    if ( ( rv = EAYCSP_CheckKeyLimits(
                                                KeyType,
                                                KeySize ) ) == CSSM_OK )
                    {
                        /* Derive the values */
                        if ( ( rv = __PerformPBKDF2(
                                            pSession->hCsp,
                                            KeySize,
                                            &pParams->Passphrase,
                                            pSalt,
                                            Iterations,
                                            pKey ) )
                                                            == CSSM_OK )
                        {
                            /* Adjust the key header */
                            pKey->KeyHeader.AlgorithmId = KeyType;
                            pKey->KeyHeader.LogicalKeySizeInBits = KeySize;
                        }
                    } /* Check key limits */
                } /* Valid derived key type */
            } /* Fetch and validate context attributes */
        } /* Validate the PBKDF1 params */
    } /* Convert param struct to PBKDF1 params */

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DeriveKey(
                        CSM_SESSION *pSession,
                        CSSM_DATA_PTR Param,
                        uint32 KeyUsage,
                        uint32 KeyAttr,
                        const CSSM_DATA *KeyLabel,
                        const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                        CSSM_KEY_PTR DerivedKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_KEY TempKey;
    const CSSM_CONTEXT_ATTRIBUTE *pKeySizeAttr = NULL;
    const CSSM_CONTEXT_ATTRIBUTE *pKeyTypeAttr = NULL;
    CSSM_ALGORITHMS KeyType;
    uint32 KeySize;

    /* Fetch the attributes common to all algorithms */
    pKeySizeAttr = CSM_GetContextAttribute( pSession->pContext,
                                            CSSM_ATTRIBUTE_KEY_LENGTH );
    pKeyTypeAttr = CSM_GetContextAttribute( pSession->pContext,
                                            CSSM_ATTRIBUTE_KEY_TYPE );
    if ( pKeySizeAttr == NULL )
    {
        ERR( rv = CSSMERR_CSP_MISSING_ATTR_KEY_LENGTH );
    }
    else if ( pKeyTypeAttr == NULL )
    {
        ERR( rv = CSSMERR_CSP_MISSING_ATTR_KEY_TYPE );
    }
    else
    {
        KeySize = pKeySizeAttr->Attribute.Uint32;
        KeyType = pKeyTypeAttr->Attribute.Uint32;

        /* Do some basic key attribute checks */
		if ( ( ( rv = EAYCSP_CheckSupportedAttrs( KeyAttr ) ) == CSSM_OK ) &&
             ( ( rv = EAYCSP_ValidateReturnAttr(
                                KeyAttr,
                                CSSM_KEYCLASS_SESSION_KEY,
                                KeyLabel ) ) == CSSM_OK ) &&
             ( ( rv = EAYCSP_ValidateReturnFormat(
                                pSession->pContext,
                                KeyAttr,
                                CSSM_ATTRIBUTE_SYMMETRIC_KEY_FORMAT,
                                KeyType,
                                CSSM_KEYCLASS_SESSION_KEY ) ) == CSSM_OK ) )

        {

            /* Generate the key material */
            memset( &TempKey, 0, sizeof(TempKey) );
            TempKey.KeyHeader.HeaderVersion = CSSM_KEYHEADER_VERSION;
            TempKey.KeyHeader.CspId = ADDIN_GUID;
            TempKey.KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
            TempKey.KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING;
            TempKey.KeyHeader.KeyClass = CSSM_KEYCLASS_SESSION_KEY;
            TempKey.KeyHeader.KeyAttr = KeyAttr;
            TempKey.KeyHeader.KeyUsage = KeyUsage;
            switch ( pSession->pContext->AlgorithmType )
            {
            case CSSM_ALGID_PKCS5_PBKDF1_MD5:
                {
                    rv = __DeriveKeyPBKDF1( pSession,
                                            KeyType,
                                            KeySize,
                                            Param,
                                            EVP_md5(),
                                            &TempKey );
                    break;
                }
            case CSSM_ALGID_PKCS5_PBKDF1_SHA1:
                {
                    rv = __DeriveKeyPBKDF1( pSession,
                                            KeyType,
                                            KeySize,
                                            Param,
                                            EVP_sha1(),
                                            &TempKey );
                    break;
                }
            case CSSM_ALGID_PKCS5_PBKDF2:
                {
                    rv = __DeriveKeyPBKDF2( pSession,
                                            KeyType,
                                            KeySize,
                                            Param,
                                            &TempKey );
                    break;
                }
            default:
                {
                    rv = CSSMERR_CSP_INVALID_ALGORITHM;
                }
            }

            /* "cook" the key into a returnable format */
            if ( rv == CSSM_OK )
            {
                rv = EAYCSP_CookKey( pSession,
                                     &TempKey,
                                     NULL,
                                     EAYCSP_GetKeyReturnType( 
                                                KeyAttr,
                                                CSSM_KEYCLASS_SESSION_KEY ),
                                     KeyLabel,
                                     CredAndAclEntry,
                                     DerivedKey );

                /* Wipe and delete the temporary key */
                memset( TempKey.KeyData.Data, 0, TempKey.KeyData.Length );
                Addin_free( TempKey.KeyData.Data, NULL );
            }
            else if ( rv == CSSMERR_CSP_UNSUPPORTED_KEY_SIZE )
            {
                /* Adjust the err code if it indicates invalid key size
                 * since the size came from a context attribute.
                 */
                ERR( rv = CSSMERR_CSP_INVALID_ATTR_KEY_LENGTH );
            }

        } /* Key attribute checks */
    } /* Fetch common attributes */

    return rv;
}

