/*-----------------------------------------------------------------------
 * File: OBTNPRIV.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

CSSM_RETURN CSSMAPI ADDIN_CSM_ObtainPrivateKeyFromPublicKey(
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_KEY *PublicKey,
                        CSSM_KEY *PrivateKey )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_ALGORITHMS AlgId;
    const ADDIN_LOAD_DATA *pLoadData = NULL;
    CSM_KM_KEY_HANDLE hKey = CSSM_INVALID_HANDLE;
    CSM_KM_KEY_HANDLE hPrivKey = CSSM_INVALID_HANDLE;
    CSM_KM_FORMAT_CONTEXT KmFormatCtx;
    const CSSM_KEY *pKey = PublicKey;
    CSSM_KEY StoredKey;
    CSSM_BOOL bAllocatedPrivate = CSSM_FALSE;
    
    AlgId = PublicKey->KeyHeader.AlgorithmId;
    if ( ( PublicKey->KeyHeader.BlobType == CSSM_KEYBLOB_REFERENCE ) &&
         ( PublicKey->KeyHeader.Format == CSSM_KEYBLOB_REF_FORMAT_INTEGER ) )
    {
        /* The key is really in the key manager. Do all of the remaining
         * checks on that one.
         */
        if ( PublicKey->KeyData.Length != sizeof(CSM_KM_KEY_HANDLE) )
        {
            ERR( rv = CSSMERR_CSP_INVALID_KEY_REFERENCE );
        }
        else
        {
            pLoadData = Addin_GetLoadData();
            assert( pLoadData );
            hKey = *(CSM_KM_KEY_HANDLE*)pKey->KeyData.Data;
            if ( ( ( rv = EAYCSP_KMFormatCtxByAlgId( 
                                            AlgId,
                                            CSSM_KEYCLASS_PUBLIC_KEY,
                                            &KmFormatCtx ) ) == CSSM_OK ) &&
                 ( ( rv = CSM_KMFindKeyByHandle( 
                                            pLoadData->hKeyFile,
                                            hKey,
                                            &KmFormatCtx,
                                            NULL,
                                            &StoredKey ) ) == CSSM_OK ) )
            {
                pKey = &StoredKey;
            }
        }
    } /* Reference key */
    else if ( pKey->KeyHeader.BlobType == CSSM_KEYBLOB_RAW )
    {
        /* Nothing needs to be done here. The pKey pointer is already
         * referencing the correct key structure.
         */
    } /* Raw Key */
    else
    {
        ERR( rv = CSSMERR_CSP_INVALID_KEY_FORMAT );
    }

    /* Is there a key to parse? */
    if ( rv == CSSM_OK )
    {
        /* Do algorithm and usage checks */
        if ( pKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PUBLIC_KEY )
        {
            ERR( rv = CSSMERR_CSP_INVALID_KEY_CLASS );
        }
        else
        {
            /* Search the key manager for a matching private key */
            pLoadData = Addin_GetLoadData();
            assert( pLoadData );
            if ( ( ( rv = CSM_KMFindKeyByRefKey(
                                    pLoadData->hKeyFile,
                                    pKey,
                                    NULL,
                                    NULL,
                                    NULL,
                                    &hPrivKey ) ) == CSSM_OK ) &&
                 ( ( rv = CSM_KMGetKeyHeader(
                                    pLoadData->hKeyFile,
                                    hPrivKey,
                                    &PrivateKey->KeyHeader ) ) == CSSM_OK ) )

            {
                /* Format the output key */
                PrivateKey->KeyData.Data = NULL;
                PrivateKey->KeyData.Length = 4;
                if ( ( rv = MAF_CreateOutputBuffer( 
                                        CSPHandle,
                                        sizeof(hPrivKey),
                                        &PrivateKey->KeyData,
                                        &bAllocatedPrivate ) ) == CSSM_OK )
                {
                    /* Make sure the wrapping values from the private key
                     * header are cleared, and the blob type information is
                     * corrected.
                     */
                    PrivateKey->KeyHeader.BlobType = CSSM_KEYBLOB_REFERENCE;
                    PrivateKey->KeyHeader.Format = CSSM_KEYBLOB_REF_FORMAT_INTEGER;
                    PrivateKey->KeyHeader.WrapAlgorithmId = CSSM_ALGID_NONE;
                    PrivateKey->KeyHeader.WrapMode = CSSM_ALGMODE_NONE;

                    /* Copy the key handle */
                    memcpy( PrivateKey->KeyData.Data,
                            (void*)&hPrivKey,
                            sizeof(hPrivKey) );
                    PrivateKey->KeyData.Length = sizeof(hPrivKey);
                }
            }
        }

        /* Free the key buffer from the key manager */
        if ( pKey == &StoredKey )
        {
            memset( pKey->KeyData.Data, 0, pKey->KeyData.Length );
            MAF_DeleteOutputBuffer( CSPHandle,
                                    (CSSM_DATA*)&pKey->KeyData, /* strip const */
                                    CSSM_TRUE );
        }
    } /* Key fetched */

    return rv;
}

