/*-----------------------------------------------------------------------
 * File: CSP_SIGN.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

typedef struct _sig_state
{
    EVP_MD_CTX MdCtx;
    EVP_PKEY *pSigKey;
}
EAYCSP_SIG_STATE;

uint32 EAYCSP_CalcSigLength( 
                        CSSM_ALGORITHMS AlgId,
                        EVP_PKEY *pSigKey )
{
    uint32 rv = 0;

    switch ( AlgId )
    {
    case CSSM_ALGID_DSA:
    case CSSM_ALGID_SHA1WithDSA:
        {
            rv = 48; /* Add 8 bytes for DER encoding that will be stripped */
            break;
        }
    default:
        {
            /* By the time this function get called, the CSP should have
             * already determined that this is an invalid algorithm.
             */
        }
    }

    return rv;
}


static CSSM_RETURN __ConstructSignatureHash(
                        CSSM_ALGORITHMS Algorithm,
                        EVP_MD_CTX *pMdCtx )
{
    CSSM_RETURN rv = CSSM_OK;

    assert( pMdCtx );

    memset( pMdCtx, 0, sizeof(*pMdCtx) );
    switch ( Algorithm )
    {
    case CSSM_ALGID_SHA1WithDSA:
        {
            EVP_SignInit( pMdCtx, EVP_dss1() );
            break;
        }
    default:
        {
            ERR( rv = CSSMERR_CSP_INVALID_ALGORITHM );
        }
    } /* switch Algorithm */

    return rv;
}

static CSSM_RETURN __SignDataInit(
                        CSM_SESSION *pSession )
{
    CSSM_RETURN rv = CSSM_OK;
    EAYCSP_SIG_STATE *pState = NULL;

    /* Allocate a place to store the hash state */
    if ( ( rv = CSM_AllocateSessionState(
                                    pSession,
                                    sizeof(EAYCSP_SIG_STATE) ) ) == CSSM_OK )
    {
        pState = (EAYCSP_SIG_STATE*)(pSession->pCryptoState);

        /* Fetch the signing key */
        if ( ( rv = EAYCSP_FetchPrivateKey( pSession,
                                            CSSM_KEYUSE_SIGN,
                                            CSSM_TRUE,
                                            &pState->pSigKey ) ) == CSSM_OK )
        {
            /* Create the hashing object */
            rv = __ConstructSignatureHash( pSession->pContext->AlgorithmType,
                                           &pState->MdCtx );

			if ( rv != CSSM_OK )
			{
				ADDIN_CSM_SignDataError( pSession );
			}
        }
    }

    return rv;
}

static CSSM_RETURN __SignDataUpdate(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *DataBufs,
                        uint32 DataBufCount )
{
    CSSM_RETURN rv = CSSM_OK;
    EAYCSP_SIG_STATE *pState = NULL;
    uint32 uCount;

    /* Fetch the hash object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EAYCSP_SIG_STATE) ) );
    pState = (EAYCSP_SIG_STATE*)pSession->pCryptoState;
    assert( pState );

    /* Add the data to the hash */
    for ( uCount = 0; uCount < DataBufCount; uCount++ )
    {
        /* Only call update if there is really data */
        if ( DataBufs[uCount].Length != 0 )
        {
            EVP_SignUpdate( &pState->MdCtx,
                            DataBufs[uCount].Data,
                            DataBufs[uCount].Length );
        }
    }

    return rv;
}

static CSSM_RETURN __SignDataFinal(
                        CSM_SESSION *pSession,
                        CSSM_DATA *Signature )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 SignatureLength = 0;
    CSSM_BOOL bAllocatedSig = CSSM_FALSE;
    EAYCSP_SIG_STATE *pState = NULL;

    /* Fetch the hash object from the session */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EAYCSP_SIG_STATE) ) );
    pState = (EAYCSP_SIG_STATE*)pSession->pCryptoState;
    assert( pState );

    /* Get the length of the output */
    SignatureLength = EAYCSP_CalcSigLength( pSession->pContext->AlgorithmType,
                                            pState->pSigKey );

    /* Make sure there is enough space to return the digest */
    if ( ( rv = MAF_CreateOutputBuffer( pSession->hCsp,
                                        SignatureLength, 
                                        Signature,
                                        &bAllocatedSig ) ) == CSSM_OK )
    {
        /* Complete the digest and sign */
        if ( !EVP_SignFinal( &pState->MdCtx,
                             Signature->Data,
                             (unsigned int *)&Signature->Length,
                             pState->pSigKey ) )
        {
            /* Release the signature buffer on error */
            MAF_DeleteOutputBuffer( pSession->hCsp,
                                    Signature,
                                    bAllocatedSig );

            ERR( rv = CSSMERR_CSP_FUNCTION_FAILED );
        }                    
    }

    return rv;
}

void ADDIN_CSM_SignDataError( CSM_SESSION *pSession )
{
    EAYCSP_SIG_STATE *pState;

    /* Just make sure the state is correct */
    assert( pSession->pCryptoState &&
            ( pSession->uCryptoStateSize == sizeof(EAYCSP_SIG_STATE) ) );

    /* Clear the context */
    pState = (EAYCSP_SIG_STATE*)pSession->pCryptoState;
    EVP_PKEY_free( pState->pSigKey );
    memset( pState, 0, sizeof(*pState) );
}

CSSM_RETURN CSSMAPI ADDIN_CSM_SignData(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *DataBufs,
                        uint32 DataBufCount,
                        CSSM_ALGORITHMS DigestAlgorithm,
                        CSSM_DATA_PTR Signature )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Make sure that only combo-ops are requested */
    if ( DigestAlgorithm != CSSM_ALGID_NONE )
    {
        ERR( rv = CSSMERR_CSP_INVALID_DIGEST_ALGORITHM );
    }
    else
    {
        /* Start the digest operation */
        if ( ( rv = __SignDataInit( pSession ) ) == CSSM_OK )
        {
            /* Perform the digesting operation */
            if ( ( rv = __SignDataUpdate( pSession,
                                          DataBufs,
                                          DataBufCount ) ) == CSSM_OK )
            {
                /* Finalize the digest */
                rv = __SignDataFinal( pSession,
                                      Signature );
            } /* Update */

            /* Make sure the session is cleared of the digest operation */
            ADDIN_CSM_SignDataError( pSession );
        } /* Init */
    }

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_SignDataInit(
                        CSM_SESSION *pSession )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Initialize the digest operation */
    rv = __SignDataInit( pSession );

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_SignDataUpdate(
                        CSM_SESSION *pSession,
                        const CSSM_DATA *DataBufs,
                        uint32 DataBufCount )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Feed data buffers into the digest one at a time */
    rv = __SignDataUpdate( pSession,
                           DataBufs,
                           DataBufCount );

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_SignDataFinal( CSM_SESSION *pSession,
                                      CSSM_DATA *Digest )
{
    CSSM_RETURN rv = CSSM_OK;

    /* Finalize the digest operation */
    rv = __SignDataFinal( pSession,
                          Digest );

    /* Clean up the signature state */
    ADDIN_CSM_SignDataError( pSession );

    return rv;
}
