/*-----------------------------------------------------------------------
 * File: UNWRAP.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

CSSM_RETURN EAYCSP_GetRawFormatFromWrapped(
                        CSSM_KEYCLASS KeyClass,
                        CSSM_ALGORITHMS AlgId,
						CSSM_KEYBLOB_TYPE BlobType,
                        uint32 WrappedFormat,						
                        uint32 *RawFormat )
{
    CSSM_RETURN rv = CSSM_OK;

	if ( BlobType == CSSM_KEYBLOB_WRAPPED) 
	{
		if ( KeyClass == CSSM_KEYCLASS_PRIVATE_KEY )
		{
			if ( WrappedFormat == CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS8 )
			{
				*RawFormat = CSSM_KEYBLOB_RAW_FORMAT_PKCS8;
			}
			else
			{
				ERR( rv = CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT );
			}
		}
		else if ( KeyClass == CSSM_KEYCLASS_SESSION_KEY )
		{
			if ( WrappedFormat == CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS7 )
			{
				*RawFormat = CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING;
			}
			else
			{
				ERR( rv = CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT );
			}
		}
		else
		{
			ERR( rv = CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT );
		}
	}
	else {
		ERR( rv = CSSMERR_CSP_KEY_BLOB_TYPE_INCORRECT );
	}

    return rv;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_UnwrapKey(
                        CSM_SESSION *pSession,
                        const CSSM_KEY *PublicKey,
                        const CSSM_WRAP_KEY *WrappedKey,
                        uint32 KeyUsage,
                        uint32 KeyAttr,
                        const CSSM_DATA *KeyLabel,
                        const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                        CSSM_KEY_PTR UnwrappedKey,
                        CSSM_DATA_PTR DescriptiveData,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_RETURN rv = CSSM_OK;
    CSSM_KEY TempPublicKey;
    CSSM_KEY TempPrivateKey;	
    uint32 RawKeyFormat = CSSM_KEYBLOB_RAW_FORMAT_NONE;
    CSSM_CSP_HANDLE SaveCSPHandle = CSSM_INVALID_HANDLE;

    memset( &TempPublicKey, 0, sizeof(TempPublicKey) );
    memset( &TempPrivateKey, 0, sizeof(TempPrivateKey) );

    /* Check to make sure the public key was only passed for a private
     * key unwrap.
     */
    if ( ( ( WrappedKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PRIVATE_KEY ) &&
           ( PublicKey != NULL ) ) ||
         ( ( WrappedKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY ) &&
           ( PublicKey == NULL ) ) )
    {
        ERR( rv = CSSMERR_CSP_INVALID_KEY_POINTER );
    }
    /* Can't unwrap public keys */
    else if ( WrappedKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PUBLIC_KEY )
    {
        ERR( rv = CSSMERR_CSP_INVALID_KEY );
    }
    /* None of the algorithms support descriptive data */
    else if ( ( DescriptiveData != NULL ) &&
              ( DescriptiveData->Data != NULL ) )
    {
        ERR( rv = CSSMERR_CSP_INVALID_OUTPUT_POINTER );
    }
	else if ( ( PublicKey ) && ( PublicKey->KeyHeader.AlgorithmId != WrappedKey->KeyHeader.AlgorithmId ) )		
	{
		rv = CSSMERR_CSP_ALGID_MISMATCH;
	}

    /* Do some basic key attribute checks and format conversions */
    else if ( ( ( rv = EAYCSP_GetRawFormatFromWrapped( 
                            WrappedKey->KeyHeader.KeyClass,
                            WrappedKey->KeyHeader.AlgorithmId,
							WrappedKey->KeyHeader.BlobType,
                            WrappedKey->KeyHeader.Format,
                            &RawKeyFormat ) ) == CSSM_OK ) &&
              ( ( rv = EAYCSP_ValidateReturnAttr(
                            KeyAttr,
                            WrappedKey->KeyHeader.KeyClass,
                            KeyLabel ) ) == CSSM_OK ) &&
              ( ( rv = EAYCSP_ValidateReturnFormat(
                            pSession->pContext,
                            KeyAttr,
                            ( ( WrappedKey->KeyHeader.KeyClass == 
                                                CSSM_KEYCLASS_PRIVATE_KEY )
                                ? CSSM_ATTRIBUTE_PRIVATE_KEY_FORMAT
                                : CSSM_ATTRIBUTE_SYMMETRIC_KEY_FORMAT ),
                            pSession->pContext->AlgorithmType,
                            WrappedKey->KeyHeader.KeyClass ) ) == CSSM_OK ) )

    {
        /* Fetch the key material for the public key */
        if ( PublicKey != NULL )
         {
            rv = EAYCSP_FetchKeyData( pSession,
                                      PublicKey,
                                      NULL,
                                      NULL,
                                      CSSM_KEYCLASS_PUBLIC_KEY,
                                      &TempPublicKey );
            PublicKey = &TempPublicKey;
        }		

        if ( rv == CSSM_OK )
        {			
            /* unwrap the key material */
            memcpy( &TempPrivateKey.KeyHeader,
                    &WrappedKey->KeyHeader,
                    sizeof(WrappedKey->KeyHeader) );

            /* Override the real CSP handle and set it so that the decrypt
             * function will allocate on the local heap.
             */
            SaveCSPHandle = pSession->hCsp;
            pSession->hCsp = MAF_ALLOCATE_LOCAL;

            if ( ( rv = EAYCSP_DecryptBlock( 
                                    pSession,
                                    CSSM_KEYUSE_UNWRAP,
                                    &WrappedKey->KeyData,
                                    &TempPrivateKey.KeyData ) ) == CSSM_OK )
            {
                /* Reset the CSP handle to it's original value */
                pSession->hCsp = SaveCSPHandle;

                /* Update the key header */
                TempPrivateKey.KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
                TempPrivateKey.KeyHeader.Format = RawKeyFormat;
                TempPrivateKey.KeyHeader.WrapAlgorithmId = CSSM_ALGID_NONE;
                TempPrivateKey.KeyHeader.WrapMode = CSSM_ALGMODE_NONE;

                /* "cook" the key into a returnable format */
                rv = EAYCSP_CookKey( 
                                pSession,
                                &TempPrivateKey,
                                PublicKey,
                                EAYCSP_GetKeyReturnType( 
                                            KeyAttr, 
                                            WrappedKey->KeyHeader.KeyClass ),
                                KeyLabel,
                                CredAndAclEntry,
                                UnwrappedKey );

                /* Wipe and delete the temporary private key */
                memset( TempPrivateKey.KeyData.Data,
                        0,
                        TempPrivateKey.KeyData.Length );
                Addin_free( TempPrivateKey.KeyData.Data, NULL );
            }

            /* Wipe and delete the temporary public key */
            if ( TempPublicKey.KeyData.Data != NULL )
            {
                memset( TempPublicKey.KeyData.Data,
                        0,
                        TempPublicKey.KeyData.Length );
                Addin_free( TempPublicKey.KeyData.Data, NULL );
            }
        } /* Fetch optional public key */
    } /* Key attribute checks */

    return rv;
}

