/*-----------------------------------------------------------------------
 * File: WRAP.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "cspinc.h"

CSSM_RETURN CSSMAPI ADDIN_CSM_WrapKey(
                        CSM_SESSION *pSession,
                        const CSSM_ACCESS_CREDENTIALS *Credentials,
                        const CSSM_KEY *Key,
                        const CSSM_DATA *DescriptiveData,
                        CSSM_WRAP_KEY_PTR WrappedKey,
                        CSSM_PRIVILEGE Privilege )
{
    CSSM_RETURN rv = CSSM_OK;
    CSM_KM_FORMAT_CONTEXT KmFormatSpec;
    CSSM_KEY KeyToWrap;
    const CSSM_CONTEXT_ATTRIBUTE *pAttr = NULL;

    /* Descriptive data is not supported for any wrapping algorithm */
    if ( ( DescriptiveData != NULL ) &&
         ( DescriptiveData->Data != NULL ) )
    {
        ERR( rv = CSSMERR_CSP_INVALID_INPUT_POINTER );
    }
    else
    {
        /* Fetch the output format attribute and validate */
        pAttr = CSM_GetContextAttribute( 
                                        pSession->pContext,
                                        CSSM_ATTRIBUTE_WRAPPED_KEY_FORMAT );
        if ( ( pAttr != NULL ) &&
             ( ( ( Key->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY ) &&
                 ( pAttr->Attribute.Uint32 != 
                                    CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS8 ) ) ||
               ( ( Key->KeyHeader.KeyClass == CSSM_KEYCLASS_SESSION_KEY ) &&
                 ( pAttr->Attribute.Uint32 != 
                                    CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS7 ) ) ) )
        {
            ERR( rv = CSSMERR_CSP_INVALID_ATTR_WRAPPED_KEY_FORMAT );
        }
        else
        {
            /* Fetch the key to be wrapped */
            memset( &KeyToWrap, 0, sizeof(KeyToWrap) );
            if ( Key->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY )
            {
                KmFormatSpec.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS8;
                KmFormatSpec.SigAlg = CSSM_ALGID_NONE;
                rv = EAYCSP_FetchKeyData( pSession,
                                          Key,
                                          Credentials,
                                          &KmFormatSpec,
                                          CSSM_KEYCLASS_PRIVATE_KEY,
                                          &KeyToWrap );
            }
            else if ( Key->KeyHeader.KeyClass == CSSM_KEYCLASS_SESSION_KEY )
            {
                KmFormatSpec.Format = CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING;
                KmFormatSpec.SigAlg = CSSM_ALGID_NONE;
                rv = EAYCSP_FetchKeyData( pSession,
                                          Key,
                                          Credentials,
                                          &KmFormatSpec,
                                          CSSM_KEYCLASS_SESSION_KEY,
                                          &KeyToWrap );
            }
            else
            {
                ERR( rv = CSSMERR_CSP_INVALID_KEY_CLASS );
            }
            if ( rv == CSSM_OK )
            {
                /* Make sure the key is allowed to be extracted */
                if ( !CHECKBITS( KeyToWrap.KeyHeader.KeyAttr,
                                 CSSM_KEYATTR_EXTRACTABLE ) )
                {
                    /* Manipulation into a wrapped form is not allowed */
                    ERR( rv = CSSMERR_CSP_OBJECT_MANIP_AUTH_DENIED );
                }
                else
                {
                    /* Encrypt the key buffer */
                    WrappedKey->KeyData.Data = NULL;
                    WrappedKey->KeyData.Length = 0;
                    if ( ( rv = EAYCSP_EncryptBlock( 
                                        pSession,
                                        CSSM_KEYUSE_WRAP,
                                        &KeyToWrap.KeyData,
                                        &WrappedKey->KeyData ) ) == CSSM_OK )
                    {
                        /* Copy the key header */
                        memcpy( &WrappedKey->KeyHeader,
                                &KeyToWrap.KeyHeader,
                                sizeof(KeyToWrap.KeyHeader) );

                        /* Adjust the key header to match the wrap alg */
                        WrappedKey->KeyHeader.BlobType = 
                                                    CSSM_KEYBLOB_WRAPPED;
                        WrappedKey->KeyHeader.Format = 
                                    ( KeyToWrap.KeyHeader.KeyClass == 
                                                CSSM_KEYCLASS_PRIVATE_KEY )
                                        ? CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS8
                                        : CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS7;
                        WrappedKey->KeyHeader.WrapAlgorithmId = 
                                        pSession->pContext->AlgorithmType;
                        pAttr = CSM_GetContextAttribute(
                                                    pSession->pContext,
                                                    CSSM_ATTRIBUTE_MODE );
                        WrappedKey->KeyHeader.WrapMode =
                                ( pAttr != NULL ) ? pAttr->Attribute.Uint32
                                                  : CSSM_ALGMODE_NONE ;

                    } /* Encrypt key */
                } /* Check extractable */

                /* Delete the buffer with the key to wrap */
                memset( KeyToWrap.KeyData.Data, 0, KeyToWrap.KeyData.Length );
                Addin_free( KeyToWrap.KeyData.Data, NULL );
            } /* Fetch formatted key data to wrap */
        } /* Fetch and validate wrapped key format */
    } /* Check descriptive data */

    return rv;
}

