/*-----------------------------------------------------------------------
 * File: CSP_DGST.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * File Description: Contains digest operations for the CSP.
 */

#include "geninc.h"

/* BSafe Digest Choosers */
B_ALGORITHM_METHOD  *DIGEST_MD5_CHOOSER[] = {
    &AM_MD5,
    (B_ALGORITHM_METHOD*)NULL_PTR
};

B_ALGORITHM_METHOD  *DIGEST_MD2_CHOOSER[] = {
    &AM_MD2,
    (B_ALGORITHM_METHOD*)NULL_PTR
};

B_ALGORITHM_METHOD  *DIGEST_SHA_CHOOSER[] = {
    &AM_SHA,
    (B_ALGORITHM_METHOD*)NULL_PTR
};

CSSM_RETURN CSSMAPI ADDIN_CSM_DigestData( CSM_SESSION *pSession,
                                          const CSSM_DATA *DataBufs,
                                          uint32 DataBufCount,
                                          CSSM_DATA_PTR Digest )
{
    CSSM_RETURN  retval = CSSM_OK;

    retval = ADDIN_CSM_DigestDataInit(pSession);

    if (retval == CSSM_OK)
    {
        retval = ADDIN_CSM_DigestDataUpdate(pSession, DataBufs, DataBufCount);
        if (retval == CSSM_OK)
        {
            retval = ADDIN_CSM_DigestDataFinal(pSession, Digest);
        }
    }

    CleanState( (BSAFECSP_STATE_PTR)pSession->pCryptoState );
    
    return retval;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DigestDataInit( CSM_SESSION *pSession )
{
    int                 status;
    BSAFECSP_STATE_PTR  pState = NULL;
    CSSM_RETURN         retval = CSSM_OK;

    /* Allocate space to store the algorithm map pointer */
    retval = CSM_AllocateSessionState( pSession,
                                         sizeof(BSAFECSP_STATE) );

    if ( retval != CSSM_OK )
    {
        retval = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
        memset(pState, 0, sizeof(BSAFECSP_STATE));

        /* Create the digest algorithm object */
        if ((status = B_CreateAlgorithmObject(&pState->algObject)) != 0)
        {
            retval = ConvertBsafeError(status);
        }
        else
        {
            /* Set and initialize the digest algorithm */
            switch (pSession->pContext->AlgorithmType) {
            case CSSM_ALGID_MD2:
                if ((status = B_SetAlgorithmInfo(pState->algObject, AI_MD2,
                                    NULL_PTR)) != 0)
                {
                    CleanState(pState);
                    retval = ConvertBsafeError(status);
                }
                else if ((status = B_DigestInit(pState->algObject, (B_KEY_OBJ)NULL_PTR,
                                    DIGEST_MD2_CHOOSER, (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    CleanState(pState);
                    retval = ConvertBsafeError(status);
                }
                break;
            case CSSM_ALGID_MD5:
                if ((status = B_SetAlgorithmInfo(pState->algObject, AI_MD5,
                                    NULL_PTR)) != 0)
                {
                    CleanState(pState);
                    retval = ConvertBsafeError(status);
                }
                else if ((status = B_DigestInit(pState->algObject, (B_KEY_OBJ)NULL_PTR,
                                    DIGEST_MD5_CHOOSER, (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    CleanState(pState);
                    retval = ConvertBsafeError(status);
                }
                break;
            case CSSM_ALGID_SHA1:
                if ((status = B_SetAlgorithmInfo(pState->algObject, AI_SHA1,
                                    NULL_PTR)) != 0)
                {
                    CleanState(pState);
                    retval = ConvertBsafeError(status);
                }
                else if ((status = B_DigestInit(pState->algObject, (B_KEY_OBJ)NULL_PTR,
                                    DIGEST_SHA_CHOOSER, (A_SURRENDER_CTX*)NULL_PTR)) != 0)
                {
                    CleanState(pState);
                    retval = ConvertBsafeError(status);
                }
                break;
            default:
                retval = CSSMERR_CSP_INVALID_ALGORITHM;
            }
        }
    }

    return retval;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DigestDataUpdate( CSM_SESSION *pSession,
                                                const CSSM_DATA *DataBufs,
                                                uint32 DataBufCount )
{
    int                 status;
    uint32              i = 0;
    BSAFECSP_STATE_PTR  pState = NULL;
    CSSM_RETURN         retval = CSSM_OK;

    /* Get the pState->algObject object from the algorithm state */
    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    if (!pState)
    {
        retval = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        for (i=0; i<DataBufCount; i++) {
            if ((status = B_DigestUpdate(pState->algObject, DataBufs[i].Data, 
                        DataBufs[i].Length, (A_SURRENDER_CTX*)NULL_PTR)) != 0)
            {
                CleanState(pState);
                retval = ConvertBsafeError(status);
                if (retval == CSSMERR_CSP_INTERNAL_ERROR)
                    retval = CSSMERR_CSP_FUNCTION_FAILED;
                return retval;
            }
        }
    }

    return retval;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_DigestDataFinal( CSM_SESSION *pSession,
                                               CSSM_DATA *Digest )
{
    int                 status;
    unsigned char       digestedData[MAX_HASHED_OUTPUT_LEN];
    unsigned int        digestedDataLen;
    BSAFECSP_STATE_PTR  pState = NULL;
    CSSM_RETURN         retval = CSSM_OK;
    CSSM_BOOL           allocate = CSSM_FALSE;

    /* Get the pState->algObject object from the algorithm state */
    pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
    if (!pState)
    {
        retval = CSSMERR_CSP_INTERNAL_ERROR;
    }
    else
    {
        /* Get the final digest */
        if ((status = B_DigestFinal(pState->algObject, digestedData, 
                &digestedDataLen, sizeof(digestedData), (A_SURRENDER_CTX*)NULL_PTR)) != 0)
        {
            CleanState(pState);
            retval = ConvertBsafeError(status);
            if (retval == CSSMERR_CSP_INTERNAL_ERROR)
                retval = CSSMERR_CSP_FUNCTION_FAILED;
        }
        else
        {
            retval = MAF_CreateOutputBuffer(pSession->hCsp, digestedDataLen, 
                                              Digest, &allocate);

            if (retval == CSSM_OK)
            {
                memcpy(Digest->Data, digestedData, digestedDataLen);
                Digest->Length = digestedDataLen;
            }
            CleanState(pState);
        }
    }

    return retval;
}

void ADDIN_CSM_DigestDataError( CSM_SESSION *pSession )
{
    BSAFECSP_STATE_PTR  pState = NULL;

    if (pSession)
    {
        /* Get the pState->algObject object from the algorithm state */
        pState = (BSAFECSP_STATE_PTR)pSession->pCryptoState;
        if (pState)
        {
            CleanState(pState);
        }
    }
}

/* For DigestDataClone */
CSSM_RETURN ADDIN_CSM_DuplicateDigestState( 
                                const CSSM_CONTEXT *pContext,
                                const void *pSrcState,
                                uint32 uSrcStateSize,
                                void *pDestState )
{
    pContext;
    pSrcState;
    uSrcStateSize;
    pDestState;

    return CSSMERR_CSP_FUNCTION_NOT_IMPLEMENTED;
}
