/*-----------------------------------------------------------------------
 * File: CSP_UTIL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */
 
/*
 * File Description: Contains utility functions for the CSP.
 */

#include "geninc.h"

/*---------------------------------------------------------------
 *Name: csp_GetEnvironmentRandom
 *
 *Description:
 * This function get variables from system and generates random material.
 *
 *Parameters: 
 * pRandomBuf(output)  - The output random buffer
 * pnRandomLen(input/output) - The random buffer/random material length
 *
 *Returns:
 * CSSM_OK -   Generated an environment random
 * CSSM_FAIL - Random generation failed
 *-------------------------------------------------------------*/
CSSM_RETURN csp_GetEnvironmentRandom(uint8 *pRandomBuf, uint32 *pnRandomLen)
#if defined (WIN32)
{
    uint8       RanBuf[MAX_BUFFER_LEN];
    uint32      RanCount=0;
    sint32      nLen=0;
    uint32      dwVal;
    uint32      wVal;
    uint8       bVal;
    
    /* RSA Labs and DoD recommendations for seed generation */

    dwVal = GetTickCount();
    memcpy(RanBuf + RanCount, (char *)&dwVal, sizeof(dwVal));
    RanCount += sizeof(dwVal);

    {
        POINT       point;

        GetCursorPos(&point);

        bVal = (uint8)(LOBYTE(point.x) ^ HIBYTE(point.x));
        memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
        RanCount += sizeof(bVal);

        bVal = (uint8)(LOBYTE(point.y) ^ HIBYTE(point.y));
        memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
        RanCount += sizeof(bVal);
    }
    
    {
        SYSTEMTIME  sysTime;

        GetLocalTime(&sysTime);

        memcpy(RanBuf + RanCount, (char *)&sysTime.wMilliseconds, 
            sizeof(sysTime.wMilliseconds));
        RanCount += sizeof(sysTime.wMilliseconds);

        bVal = (uint8)(LOBYTE(sysTime.wSecond) ^ LOBYTE(sysTime.wMinute));
        memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
        RanCount += sizeof(bVal);
    }
        
    {
        LARGE_INTEGER   liPerfCount;

        if (QueryPerformanceCounter(&liPerfCount))
        {
            bVal = (uint8)(LOBYTE(LOWORD(liPerfCount.LowPart)) 
                          ^ LOBYTE(LOWORD(liPerfCount.HighPart)));
            memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
            RanCount += sizeof(bVal);

            bVal = (uint8)(HIBYTE(LOWORD(liPerfCount.LowPart)) 
                          ^ LOBYTE(LOWORD(liPerfCount.HighPart)));
            memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
            RanCount += sizeof(bVal);

            bVal = (uint8)(LOBYTE(HIWORD(liPerfCount.LowPart)) 
                           ^ LOBYTE(LOWORD(liPerfCount.HighPart)));
            memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
            RanCount += sizeof(bVal);

            bVal = (uint8)(HIBYTE(HIWORD(liPerfCount.LowPart)) 
                          ^ LOBYTE(LOWORD(liPerfCount.HighPart)));
            memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
            RanCount += sizeof(bVal);
        }
    }

    {
        MEMORYSTATUS    mstMemStat;

        mstMemStat.dwLength = sizeof(MEMORYSTATUS);
        GlobalMemoryStatus(&mstMemStat);

        wVal = HIWORD(mstMemStat.dwAvailPhys);
        memcpy(RanBuf + RanCount, (char *)&wVal, sizeof(wVal));
        RanCount += sizeof(wVal);

        wVal = HIWORD(mstMemStat.dwAvailPageFile);
        memcpy(RanBuf + RanCount, (char *)&wVal, sizeof(wVal));
        RanCount += sizeof(wVal);

        bVal = LOBYTE(HIWORD(mstMemStat.dwAvailVirtual));
        memcpy(RanBuf + RanCount, (char *)&bVal, sizeof(bVal));
        RanCount += sizeof(bVal);
    }
    /* Copy the environment random number to out buffer */
    nLen = (*pnRandomLen > RanCount)? RanCount: *pnRandomLen;
    memcpy(pRandomBuf, RanBuf, nLen);
    *pnRandomLen = nLen;
    return CSSM_OK;
}
#elif defined (__digital__) || defined (VMS)
{
    static int seeded = 0;
    uint32 i;

    if (!seeded)
    {
        srandom(time(NULL));
        seeded = 1;
    }

    for (i = 0; i < *pnRandomLen; i++)
        pRandomBuf[i] = random();

    return CSSM_OK;
}
#else
{
#error Not implemented.    
	return CSSM_FAIL;
}
#endif

/*----------------------------------------------------------------------
 *  Name: csp_KeyHeader_Fill
 *
 *  Description:
 *  Fill the key header values from the other input parameters.
 *
 *  Parameters:
 *  pKeyHdr (output) - Contains a key header to be filled.
 *  nBlobType (input) - See BlobType #define's
 *  nFormat (input) - Raw or Reference format
 *  nAlgorithmId(input) - Algorithm ID of key
 *  nKeyClass (input) - Public/Private/Secret etc.
 *  nLogicalKeySizeInBits (input) - Logical key size in bits
 *  nKeyAttr (input) - Attribute flags
 *  nKeyUsage(input) - Key use flags
 *  pStartDate (input) - Effective date of key
 *  pEndDate (input) - Expiration date of key
 *  nWrapAlgorithmId (input) - == CSSM_ALGID_NONE if clear key
 *  nWrapMode (input) -   if alg supports multiple wrapping modes
 *  nReserved (input) - 
 *
 *  Returns:
 *    CSSM_OK:  Call is successful.
 *----------------------------------------------------------------------*/
CSSM_RETURN csp_KeyHeader_Fill(CSSM_KEYHEADER_PTR pKeyHdr, uint32 nBlobType,
                             uint32 nFormat, uint32 nAlgorithmId, 
                             uint32 nKeyClass, uint32 nLogicalKeySizeInBits,
                             uint32 nKeyAttr, uint32 nKeyUsage, 
                             CSSM_DATE_PTR pStartDate, CSSM_DATE_PTR pEndDate,
                             uint32 nWrapAlgorithmId, uint32 nWrapMode, 
                             uint32 nReserved)
{
    if (pKeyHdr == NULL)
    {
        return CSSMERR_CSP_INVALID_INPUT_POINTER;
    }

    pKeyHdr->HeaderVersion      = CSSM_KEYHEADER_VERSION;
    memcpy(&pKeyHdr->CspId, &ADDIN_GUID, sizeof(CSSM_GUID));
    pKeyHdr->BlobType           = nBlobType;
    pKeyHdr->Format             = nFormat;
    pKeyHdr->AlgorithmId        = nAlgorithmId;
    pKeyHdr->KeyClass           = nKeyClass;
    pKeyHdr->LogicalKeySizeInBits   = nLogicalKeySizeInBits;
    pKeyHdr->KeyAttr            = nKeyAttr;
    pKeyHdr->KeyUsage           = nKeyUsage;
    pKeyHdr->WrapAlgorithmId    = nWrapAlgorithmId;
    pKeyHdr->WrapMode           = nWrapMode;
    pKeyHdr->Reserved           = nReserved;
    if (pStartDate)
    {
        memcpy(&pKeyHdr->StartDate, pStartDate, sizeof(CSSM_DATE));
    }
    else
    {
        memset(&pKeyHdr->StartDate, 0, sizeof(CSSM_DATE));
    }
    if (pEndDate)
    {
        memcpy(&pKeyHdr->EndDate, pEndDate, sizeof(CSSM_DATE));
    }
    else
    {
        memset(&pKeyHdr->EndDate, 0, sizeof(CSSM_DATE));
    }

    return CSSM_OK;
}

void CleanState(BSAFECSP_STATE_PTR pState)
{
    if (pState)
    {
        if (pState->algObject)
        {
            B_DestroyAlgorithmObject(&pState->algObject);
            pState->algObject = NULL;
        }
        if (pState->keyObject)
        {
            B_DestroyKeyObject(&pState->keyObject);
            pState->keyObject = NULL;
        }
    }
}
