/*-----------------------------------------------------------------------
 * File: CSP_WRAP.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 /*
** OpenVMS History
**
 *    01-Nov-1999	Carol Kelly
 *		o Merge in changes from Unix code base
 *
 */

/*
 * File Description: Contains wrap/unwrap operations for the CSP.
 */

#include "geninc.h"

CSSM_RETURN CSSMAPI ADDIN_CSM_WrapKey( 
                        CSM_SESSION *pSession,
                        const CSSM_ACCESS_CREDENTIALS *AccessCred,
                        const CSSM_KEY *Key,
                        const CSSM_DATA *DescriptiveData,
                        CSSM_WRAP_KEY_PTR WrappedKey,
                        CSSM_PRIVILEGE Privilege )
{
    const CSSM_KEY                  *pKeyToBeWrapped = NULL;
    CSM_KM_FORMAT_CONTEXT           kmFormat;
    CSSM_KEY                        priKey;
    const ADDIN_LOAD_DATA           *loadData;
    const CSSM_DATA                 *pClearBuf = NULL;
    const CSSM_CONTEXT_ATTRIBUTE    *pModeAttr = NULL;
    uint32                          nBytesEncrypted = 0;
    CSSM_RETURN                     rtn = CSSM_OK;
	CSSM_BOOL bAllocatedKeyData = CSSM_FALSE;

    /* Unused parameters */
    DescriptiveData;

    if (pSession->pContext->ContextType != CSSM_ALGCLASS_SYMMETRIC)
    {
        return CSSMERR_CSP_INVALID_CONTEXT;
    }

#ifdef VMS  /* COMPAQ_FIX */
    /*
     * Make sure that they key attributes are correct since these are used
     * later for the wrapped key.
     */
    if ((Key->KeyHeader.KeyAttr & (CSSM_KEYATTR_RETURN_DEFAULT |
				   CSSM_KEYATTR_RETURN_DATA |
				   CSSM_KEYATTR_RETURN_REF |
				   CSSM_KEYATTR_RETURN_NONE)) != 0)
    {
        return CSSMERR_CSP_KEY_HEADER_INCONSISTENT;
    }
#endif
    /* Do the context correctness check in the EncryptData call */

    /* Verify the target key header and get the raw private key */
    switch (Key->KeyHeader.BlobType)
    {
    case CSSM_KEYBLOB_RAW:
//  case CSSM_KEYBLOB_RAW_BERDER:
        {
            pKeyToBeWrapped = Key;
            break;
        }
    case CSSM_KEYBLOB_REFERENCE:
        {
            memset(&priKey, 0, sizeof(CSSM_KEY));

            /* Get the raw private key from the private key reference */
            if (Key->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY)
            {
                /* Get the key from the key manager */
                loadData = Addin_GetLoadData();

                if (loadData == NULL)
                {
                    rtn = CSSMERR_CSP_INTERNAL_ERROR;
                    goto FAIL;
                }

#ifdef VMS  /* COMPAQ_FIX */
		/* If raw data is requested use a different format */
		if (pSession->pContext->AlgorithmType == CSSM_ALGID_NONE)
		{
		  /* Pick raw blob format based on Algorithm ID */
		  switch (Key->KeyHeader.AlgorithmId)
                  {
                    case CSSM_ALGID_RSA:
		      kmFormat.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
                      break;

		    case CSSM_ALGID_DSA:
                      kmFormat.Format = CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
                      break;
 
                    default:
                      rtn = CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
                      goto FAIL;
                  }
                }
                else
                {
                  kmFormat.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS8;
                }
#else
                kmFormat.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS8;
#endif
                kmFormat.SigAlg = CSSM_ALGID_NONE;

                if ( ( Key->KeyData.Length != sizeof(CSM_KM_KEY_HANDLE) ) ||
                                        port_IsBadReadPtr( Key->KeyData.Data,
                                        sizeof(CSM_KM_KEY_HANDLE) ) )
                {
                    rtn = CSSMERR_CSP_INVALID_KEY_REFERENCE;
                    goto FAIL;                      
                }

                rtn = CSM_KMFindKeyByHandle(loadData->hKeyFile, *(CSM_KM_KEY_HANDLE*)Key->KeyData.Data,
                    &kmFormat, AccessCred, &priKey);

                if (rtn != CSSM_OK)
                {
                    goto FAIL;
                }
            }

            pKeyToBeWrapped = &priKey;

			bAllocatedKeyData = CSSM_TRUE;

            break;
        }
    case CSSM_KEYBLOB_WRAPPED:
    default:                            
        {
            return CSSMERR_CSP_KEY_BLOB_TYPE_INCORRECT;
        }
    }

#ifdef VMS /* COMPAQ_FIX */
    pClearBuf       = &pKeyToBeWrapped->KeyData;

    /* If just the raw data is being requested, just need to copy key */
    if (pSession->pContext->AlgorithmType == CSSM_ALGID_NONE)
    {
      WrappedKey->KeyData.Data = App_Malloc(pSession->hCsp,
					    pClearBuf->Length);
      if (WrappedKey->KeyData.Data == NULL)
      {
	rtn = CSSMERR_CSP_MEMORY_ERROR;
      }
      else
      {
	rtn = CSSM_OK;
	WrappedKey->KeyData.Length = pClearBuf->Length;
        memcpy(WrappedKey->KeyData.Data,
               pClearBuf->Data,
               pClearBuf->Length);
      }
    }
    else
    {
    /* Call ADDIN_CSM_EncryptData to encrypt the key */
#else
    /* Call ADDIN_CSM_EncryptData to encrypt the key */
    pClearBuf       = &pKeyToBeWrapped->KeyData;
#endif

    /* Insure that the encrypt call will allocate space for the output */
    WrappedKey->KeyData.Data = NULL;
    WrappedKey->KeyData.Length = 0;

    rtn = ADDIN_CSM_EncryptData(pSession, pClearBuf, 1, &(WrappedKey->KeyData), 1, 
                             &nBytesEncrypted, NULL, Privilege);
#ifdef VMS /* COMPAQ_FIX */
    }
#endif

    /* Clear out the memory of the raw key */
    memset(pClearBuf->Data, 0, pClearBuf->Length);

    if (rtn != CSSM_OK)
    {
        goto FAIL;
    }

    /* Set the key header */
    memcpy(&WrappedKey->KeyHeader, &Key->KeyHeader, sizeof(CSSM_KEYHEADER));
    switch (Key->KeyHeader.BlobType)
    {
    case CSSM_KEYBLOB_RAW:
    case CSSM_KEYBLOB_REFERENCE:
        {
#ifdef VMS /* COMPAQ_FIX */
            /* If requested it returned as raw data */
            if (pSession->pContext->AlgorithmType == CSSM_ALGID_NONE)
            {
              WrappedKey->KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
              /* Indicate raw blob format based on Algorithm ID */
              switch (Key->KeyHeader.AlgorithmId)
              {
                case CSSM_ALGID_RSA:
                  WrappedKey->KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS1;
                  break;

                case CSSM_ALGID_DSA:
                  WrappedKey->KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_FIPS186;
                  break;
              }
            }
            else
            {
              WrappedKey->KeyHeader.BlobType = CSSM_KEYBLOB_WRAPPED;
              WrappedKey->KeyHeader.Format = CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS8;
            }
#else
            WrappedKey->KeyHeader.BlobType = CSSM_KEYBLOB_WRAPPED;
            WrappedKey->KeyHeader.Format = CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS8;
#endif
            break;
        }
    case CSSM_KEYBLOB_WRAPPED:
    default:
        {
            rtn = CSSMERR_CSP_KEY_BLOB_TYPE_INCORRECT;
            goto FAIL;
        }
    }
    WrappedKey->KeyHeader.WrapAlgorithmId = pSession->pContext->AlgorithmType;
    
    /* Get the mode attribute. */
    pModeAttr = CSM_GetContextAttribute(pSession->pContext, CSSM_ATTRIBUTE_MODE);
    WrappedKey->KeyHeader.WrapMode = pModeAttr->Attribute.Uint32;

FAIL:
	MAF_DeleteOutputBuffer (MAF_ALLOCATE_LOCAL, (CSSM_DATA *)pClearBuf, bAllocatedKeyData );
    return rtn;
}

CSSM_RETURN CSSMAPI ADDIN_CSM_UnwrapKey( 
                        CSM_SESSION *pSession,
                        const CSSM_KEY *PublicKey,
                        const CSSM_WRAP_KEY *WrappedKey,
                        uint32 KeyUsage,
                        uint32 KeyAttr,
                        const CSSM_DATA *KeyLabel,
                        const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                        CSSM_KEY_PTR UnwrappedKey,
                        CSSM_DATA_PTR DescriptiveData,
                        CSSM_PRIVILEGE Privilege )
{
    uint32                      nBytesDecrypted = 0, temp;
    const CSSM_DATA             *pCipherBuf = NULL;
    CSSM_KEY                    tempKey;
    const ADDIN_LOAD_DATA       *loadData;
    CSM_KM_KEY_HANDLE           keyHandle;
    CSSM_BOOL                   bAllocatedOutput = CSSM_FALSE;
    CSSM_RETURN                 rtn = CSSM_OK;
	uint32						KeyReturnType = 0;
    const CSSM_CONTEXT_ATTRIBUTE *pAttr = NULL;
	CSSM_ATTRIBUTE_TYPE			AttrType;
    uint32						BlobFormat;
    CSSM_ALGORITHMS				AlgId;
    const ADDIN_LOAD_DATA		*pLoadData = NULL;
    CSM_KM_KEY_HANDLE			hKey = CSSM_INVALID_HANDLE;
    CSM_KM_FORMAT_CONTEXT		KmFormatCtx;
    CSSM_KEY					TempPublicKey;

    /* Unused parameters */
    DescriptiveData;
    KeyLabel;

    /* Initialize */
    memset(&tempKey, 0, sizeof(CSSM_KEY));
    memset(&TempPublicKey, 0, sizeof(TempPublicKey));

    if (pSession->pContext->ContextType != CSSM_ALGCLASS_SYMMETRIC)
    {
        return CSSMERR_CSP_INVALID_CONTEXT;
    }

    /* Check to make sure the public key was only passed for a private
     * key unwrap.
     */
    if (((WrappedKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PRIVATE_KEY) &&
         (PublicKey != NULL)) ||
        ((WrappedKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY) &&
         (PublicKey == NULL)))
    {
        return (CSSMERR_CSP_INVALID_KEY_POINTER);
    }
    /* Can't unwrap public keys */
    else if (WrappedKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PUBLIC_KEY)
    {
        return (CSSMERR_CSP_INVALID_KEY);    
	}
    /* None of the algorithms support descriptive data */
    else if ((DescriptiveData != NULL) &&
             (DescriptiveData->Data != NULL))
    {
        return (CSSMERR_CSP_INVALID_OUTPUT_POINTER);
    }
	else if ((PublicKey != NULL) && 
			 (PublicKey->KeyHeader.AlgorithmId != WrappedKey->KeyHeader.AlgorithmId))
	{
		return (CSSMERR_CSP_ALGID_MISMATCH);
	}

#ifndef VMS /* COMPAQ_FIX */
    /* The symmetric context correctness check is done in the DecryptData. */
    if ( ( WrappedKey->KeyHeader.BlobType != CSSM_KEYBLOB_WRAPPED ) ||
         ( WrappedKey->KeyHeader.Format != CSSM_KEYBLOB_WRAPPED_FORMAT_PKCS8 ) )
    {
        return CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT; 
    }
#endif
	if (WrappedKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY)
	{
		AttrType = CSSM_ATTRIBUTE_PRIVATE_KEY_FORMAT;
	}
	else
	{
		AttrType = CSSM_ATTRIBUTE_SYMMETRIC_KEY_FORMAT;
	}
 
	KeyReturnType = KeyAttr & (CSSM_KEYATTR_RETURN_DATA |
							   CSSM_KEYATTR_RETURN_REF |
							   CSSM_KEYATTR_RETURN_NONE);

    if (KeyReturnType == CSSM_KEYATTR_RETURN_DEFAULT)
    {
        if ((WrappedKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY))
        {
            KeyReturnType = CSSM_KEYATTR_RETURN_REF;
        }
        else
        {
            KeyReturnType = CSSM_KEYATTR_RETURN_DATA;
        }
    }

	AlgId = pSession->pContext->AlgorithmType;

	pAttr = CSM_GetContextAttribute(pSession->pContext, AttrType);


    if (pAttr != NULL)
    {
        BlobFormat = pAttr->Attribute.Uint32;

        /* Switch through the formats trying to find a match to the request */
        switch (KeyReturnType)
		{
		case CSSM_KEYATTR_RETURN_NONE:
			{
				/* Specifying the format attribute is an error */
				if (pAttr)
				{
					switch (AttrType)
					{
					case CSSM_ATTRIBUTE_SYMMETRIC_KEY_FORMAT:
						{
							return (CSSMERR_CSP_INVALID_ATTR_SYMMETRIC_KEY_FORMAT);
							break;
						}
					case CSSM_ATTRIBUTE_PUBLIC_KEY_FORMAT:
						{ 
							return (CSSMERR_CSP_INVALID_ATTR_PUBLIC_KEY_FORMAT);
							break;
						}
                    case CSSM_ATTRIBUTE_PRIVATE_KEY_FORMAT:
                        { 
							return (CSSMERR_CSP_INVALID_ATTR_PRIVATE_KEY_FORMAT);
							break;
						}
                    default:
                        {
							return (CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT);
							break;
						}
                    }
                }

                break;
            }
        case CSSM_KEYATTR_RETURN_REF:
            {
                if (BlobFormat != CSSM_KEYBLOB_REF_FORMAT_INTEGER)
                {
                }
                break;
            }
        case CSSM_KEYATTR_RETURN_DATA:
            {
                switch (AlgId)
                {
                case CSSM_ALGID_DES:
				case CSSM_ALGID_3DES_3KEY:
				case CSSM_ALGID_DESX:
                    {
                        if (BlobFormat != CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING)
                        {
                            return (CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT);
                        }
                        break;
                    }
				case CSSM_ALGID_RSA:
				case CSSM_ALGID_RC2:
				case CSSM_ALGID_RC4:
				case CSSM_ALGID_RC5:
                    {
                        break;
                    }
                default:
                    {
                        return (CSSMERR_CSP_INVALID_ALGORITHM);
                    }
                } /* switch AlgId */
            }
        } /* switch ReturnType */
    }

	if (PublicKey != NULL)
	{
        /* Attempt to resolve the key */
        
        if ((PublicKey->KeyHeader.BlobType == CSSM_KEYBLOB_REFERENCE) &&
            (PublicKey->KeyHeader.Format == CSSM_KEYBLOB_REF_FORMAT_INTEGER))
        {
            /* The key is really in the key manager. Do all of the remaining
             * checks on that one.
             */
            if ((PublicKey->KeyData.Length != sizeof(CSM_KM_KEY_HANDLE)) ||
                port_IsBadReadPtr(PublicKey->KeyData.Data, sizeof(CSM_KM_KEY_HANDLE)))
            {
                return (CSSMERR_CSP_INVALID_KEY_REFERENCE);
            }
            else
            {
                pLoadData = Addin_GetLoadData();
                assert(pLoadData);
                hKey = *(CSM_KM_KEY_HANDLE*)PublicKey->KeyData.Data;

                KmFormatCtx.SigAlg = CSSM_ALGID_NONE;
				switch (AlgId)
				{
                case CSSM_ALGID_DES:
				case CSSM_ALGID_3DES_3KEY:
				case CSSM_ALGID_DESX:
					{
						KmFormatCtx.Format = CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING;
						break;
					}
				case CSSM_ALGID_RSA:
				case CSSM_ALGID_RC2:
				case CSSM_ALGID_RC4:
				case CSSM_ALGID_RC5:
                    {
                        break;
                    }
				default:
					{
						return (CSSMERR_CSP_INVALID_ALGORITHM);
					}
				}

				rtn = CSM_KMFindKeyByHandle(pLoadData->hKeyFile,
											hKey,
											&KmFormatCtx,
											NULL,
											&TempPublicKey);

                /* If the application passes in the wrong type of key,
                 * and the passed key requires access credentials (even if
                 * the correct key doesn't), the key manager may return
                 * CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS instead of the
                 * _REAL_ error, CSSMERR_CSP_INVALID_KEY_CLASS.
                 */
                if ((rtn == CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS) &&
                    (PublicKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PUBLIC_KEY))
                {
                    return (CSSMERR_CSP_INVALID_KEY_CLASS);
                }
                else if (rtn == KMERR_INVALID_HANDLE)
                {
                    return (CSSMERR_CSP_INVALID_KEY_REFERENCE);
                }
                else if (rtn == KMERR_CAN_NOT_INTERPRET_KEY_DATA)
                {
                    return (CSSMERR_CSP_INVALID_KEY);
                }
				else if (rtn != CSSM_OK)
				{
					return (rtn);
				}
            }
        } /* Reference key */
        else if (PublicKey->KeyHeader.BlobType == CSSM_KEYBLOB_RAW)
        {
            /* Reformat the key blob to match the format request. */
            memcpy(&(TempPublicKey.KeyHeader), 
				   &(PublicKey->KeyHeader), 
				   sizeof(PublicKey->KeyHeader));

            TempPublicKey.KeyData.Data = Addin_malloc(PublicKey->KeyData.Length,
													   NULL);
            
			if (TempPublicKey.KeyData.Data != NULL)
            {
                TempPublicKey.KeyData.Length = PublicKey->KeyData.Length;

                memcpy(TempPublicKey.KeyData.Data,
					   PublicKey->KeyData.Data,
					   PublicKey->KeyData.Length);
            }
            else
            {
                rtn = CSSMERR_CSP_MEMORY_ERROR;
				goto FAIL;
            }
        } /* Raw Key */
        else
        {
            return (CSSMERR_CSP_INVALID_KEY_FORMAT);
        }
	}

	PublicKey = &TempPublicKey;

#ifdef VMS /* COMPAQ_FIX */
    /* If just the raw data is being requested, just need to copy key */
    if (pSession->pContext->AlgorithmType == CSSM_ALGID_NONE)
    {
        tempKey.KeyData.Data =
          Addin_malloc(WrappedKey->KeyData.Length, NULL);
        if (tempKey.KeyData.Data == NULL)
        {
          rtn = CSSMERR_CSP_MEMORY_ERROR;
          goto FAIL;
        }
        memcpy(tempKey.KeyData.Data,
               WrappedKey->KeyData.Data,
               WrappedKey->KeyData.Length);
        tempKey.KeyData.Length = WrappedKey->KeyData.Length;
    }
    else
    {
#endif
    /* Change the handle for this call to 0 since it is a local call */
    temp = pSession->hCsp;
    pSession->hCsp = 0;
    pCipherBuf      = &WrappedKey->KeyData;

    if ((rtn = ADDIN_CSM_DecryptData(pSession, pCipherBuf, 1, &(tempKey.KeyData), 1, 
                           &nBytesDecrypted, NULL, Privilege)) != CSSM_OK)
    {
        goto FAIL;
    }

    /* Reset the handle in the session to the appropriate value */
    pSession->hCsp = temp;

#ifdef VMS /* COMPAQ_FIX */
    } /* End of just raw data being requested */
#endif

    /* Reset the key header */
    memcpy(&tempKey.KeyHeader, &WrappedKey->KeyHeader, sizeof(CSSM_KEYHEADER));
    if (WrappedKey->KeyHeader.BlobType == CSSM_KEYBLOB_WRAPPED)
    {
        tempKey.KeyHeader.BlobType = CSSM_KEYBLOB_RAW;
    }
    tempKey.KeyHeader.WrapAlgorithmId = CSSM_ALGID_NONE;
    tempKey.KeyHeader.WrapMode = CSSM_ALGMODE_NONE;

    if ((KeyAttr & CSSM_KEYATTR_PERMANENT) != 0)
    {
        tempKey.KeyHeader.KeyAttr |= CSSM_KEYATTR_PERMANENT;
    }

    if ((KeyUsage != 0) && (tempKey.KeyHeader.KeyUsage != KeyUsage))
    {
        tempKey.KeyHeader.KeyUsage = KeyUsage;
    }

    if ( tempKey.KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY )
    {
#ifdef VMS /* COMPAQ_FIX */
      /* If just the raw data is being requested, use format from key */
      if (pSession->pContext->AlgorithmType == CSSM_ALGID_NONE)
      {
        tempKey.KeyHeader.Format = WrappedKey->KeyHeader.Format;
      }
      else
      {
        tempKey.KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS8;
      }
#else
        tempKey.KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_PKCS8;
#endif
    }
    else
    {
        tempKey.KeyHeader.Format = CSSM_KEYBLOB_RAW_FORMAT_OCTET_STRING;
    }

    if (KeyReturnType == CSSM_KEYATTR_RETURN_DATA)
    {
        /* The key must be extractable to be returned as data.
         * If the key is extractable but sensitive, then the key can
         * not be returned raw. It must be extracted wrapped. Public
         * keys are the exception to the rule. They can never be
         * sensitive, and are always extractable regardless of the
         * attribute value.
         */
        if ((WrappedKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PUBLIC_KEY) &&
            CHECKBITS(KeyAttr, CSSM_KEYATTR_SENSITIVE))
        {
            rtn = CSSMERR_CSP_INVALID_KEYATTR_MASK;
			goto FAIL;
        }
        else if ((WrappedKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PUBLIC_KEY) &&
                 (!CHECKBITS(KeyAttr, CSSM_KEYATTR_EXTRACTABLE) ||
                  CHECKBITS(KeyAttr, CSSM_KEYATTR_SENSITIVE)))
        {
            rtn = CSSMERR_CSP_INVALID_KEYATTR_MASK;
			goto FAIL;
        }

		if (UnwrappedKey)
		{
			/* CSP is required to allocate space for the key, so ignore any
			 * values that were set in the KeyData.
			 */
			UnwrappedKey->KeyData.Length = 0;
			UnwrappedKey->KeyData.Data = NULL;

			if ((rtn = MAF_CreateOutputBuffer(pSession->hCsp, 
											  tempKey.KeyData.Length, 
											  &(UnwrappedKey->KeyData),
											  &bAllocatedOutput)) != CSSM_OK)
			{
				goto FAIL;
			}

			memcpy(&(UnwrappedKey->KeyHeader),
				   &(tempKey.KeyHeader),
				   sizeof(CSSM_KEYHEADER));

			memcpy(UnwrappedKey->KeyData.Data,
				   tempKey.KeyData.Data,
				   tempKey.KeyData.Length);

			UnwrappedKey->KeyData.Length = tempKey.KeyData.Length;
		}
    }
    else if (KeyReturnType == CSSM_KEYATTR_RETURN_REF)
	{
		/* Place the private key into the key manager */
		loadData = Addin_GetLoadData();

		if (loadData == NULL)
		{
			rtn = CSSMERR_CSP_INTERNAL_ERROR;
			goto FAIL;
		}

		if ((rtn = CSM_KMInsertKey(loadData->hKeyFile, &tempKey, PublicKey, NULL,
							  CredAndAclEntry, &keyHandle)) != CSSM_OK)
		{
			goto FAIL;
		}

		/* Only return the private key reference if it was asked for */
		if ((KeyAttr & CSSM_KEYATTR_RETURN_NONE) == 0)
		{
			if (UnwrappedKey)
			{
				/* CSP is required to allocate space for the key, so ignore any
				 * values that were set in the KeyData.
				 */
				UnwrappedKey->KeyData.Length = 0;
				UnwrappedKey->KeyData.Data = NULL;

				if ((rtn = MAF_CreateOutputBuffer(pSession->hCsp, 
							sizeof(CSM_KM_KEY_HANDLE), &(UnwrappedKey->KeyData),
							&bAllocatedOutput)) != CSSM_OK)
				{
					goto FAIL;
				}
				memcpy(&(UnwrappedKey->KeyHeader), &(tempKey.KeyHeader), sizeof(CSSM_KEYHEADER));

				/* Correct the private key header information */
				UnwrappedKey->KeyHeader.BlobType = CSSM_KEYBLOB_REFERENCE;
				UnwrappedKey->KeyHeader.Format = CSSM_KEYBLOB_REF_FORMAT_INTEGER;

				memcpy(UnwrappedKey->KeyData.Data, &keyHandle, sizeof(CSM_KM_KEY_HANDLE));
				UnwrappedKey->KeyData.Length = sizeof(CSM_KM_KEY_HANDLE);
			}
		}
	}
    else if (KeyReturnType == CSSM_KEYATTR_RETURN_NONE)
	{
        /* The key must be permanent and named with a label. */
        if (!CHECKBITS(KeyAttr, CSSM_KEYATTR_PERMANENT))
        {
            rtn = CSSMERR_CSP_INVALID_KEYATTR_MASK;
			goto FAIL;
        }
        else if ((KeyLabel == NULL) && 
				 (WrappedKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PRIVATE_KEY))
        {
            rtn = CSSMERR_CSP_INVALID_KEY_LABEL;
			goto FAIL;
        }
	}

FAIL:
    /* Cleanup */
    if (tempKey.KeyData.Data != NULL)
    {
		memset(tempKey.KeyData.Data,
			   0,
			   tempKey.KeyData.Length);
        Addin_free(tempKey.KeyData.Data, NULL);
    }

    /* Wipe and delete the temporary public key */
    if (TempPublicKey.KeyData.Data != NULL)
    {
        memset(TempPublicKey.KeyData.Data,
               0,
               TempPublicKey.KeyData.Length);
        Addin_free(TempPublicKey.KeyData.Data, NULL);
    }

    return rtn;
}
