/*-----------------------------------------------------------------------
 * File: CSM_KEYMGR_DB.H
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#ifndef __CSM_KEYMGR_DB_H__
#define __CSM_KEYMGR_DB_H__

#ifdef __cplusplus
extern "C" {
#endif

#ifndef _CSSMTYPE_H
#error cssmtype.h must be included before this file
#endif

/*
 * The database record constants.
 */
#define CSM_KM_DB_VERSION                   (2)
#define CSM_KM_DB_MAX_KEY_SIZE              (2048)
#define CSM_KM_DB_HASH_SIZE                 (20)
#define CSM_KM_PBE_SALT_LENGTH              (20)
#define CSM_KM_PBE_MAX_IV_LENGTH            (24)
#define CSM_KM_MAX_ACL_AUTH_TAGS            (16)

/*
 * The database data types.
 */
typedef struct _csm_km_db_header
{
    uint32 dbVer;
    uint32 dbHeaderSize;
    uint32 dbRecordSize;
    /* Perhaps an ACL could be stored in the header. */
} CSM_KM_DB_HEADER;

/*
 * Key ACL information structure. Designed to support password ACLs since
 * an encryption key can be derived from them.
 */
typedef struct _csm_km_acl_info
{
    CSSM_ACL_SUBJECT_TYPE aclType;
    CSSM_STRING prompt;
    CSSM_ACL_AUTHORIZATION_TAG authTags[CSM_KM_MAX_ACL_AUTH_TAGS];
}
CSM_KM_ACL_INFO;




typedef struct _csm_km_db_record
{
    CSSM_BOOL recordIsValid;
    CSM_KM_ACL_INFO acl;
    CSSM_KEYHEADER keyHeader;
    CSSM_STRING keyName;
    uint8 WrapIV[CSM_KM_PBE_MAX_IV_LENGTH];
    uint32 WrapIVLength;
    uint8 WrapPBESalt[CSM_KM_PBE_SALT_LENGTH];
    CSSM_BOOL keyRefHashValid;
    uint8 keyRefHash[CSM_KM_DB_HASH_SIZE];
    uint32 keyLength;
    uint8 key[CSM_KM_DB_MAX_KEY_SIZE];
} CSM_KM_DB_RECORD;

/*
 * A macro to determine if a given file size could be a valid database size.
 * Also used to determine if an index into the database is valid.
 */
#define CSM_KM_DB_SIZE_IS_VALID(size) \
                        (((size) >= sizeof(CSM_KM_DB_HEADER)) && \
                         ((((size) - sizeof(CSM_KM_DB_HEADER)) % \
                           sizeof(CSM_KM_DB_RECORD)) == 0))

/*
 * Database operations.
 */

/* The types are used to specify what sort of data is in a query item. */
typedef uint32 CSM_KM_DB_QUERY_TYPE;
#define CSM_KM_DB_QUERY_HOLE       (0x01)   /* No associated data for holes. */
#define CSM_KM_DB_QUERY_INDEX      (0x02)   /* long *                        */
#define CSM_KM_DB_QUERY_RECORD     (0x04)   /* CSM_KM_DB_RECORD *            */
#define CSM_KM_DB_QUERY_KEY_NAME   (0x08)   /* CSSM_STRING *                 */
#define CSM_KM_DB_QUERY_HASH       (0x10)   /* uint8 *                       */

/*
 * This is the most items that can compose a query.
 * Looking for a hole contradicts any of the other
 * query types.
 */
#define CSM_KM_DB_MAX_QUERY_ITEMS  (4)

typedef struct _csm_kmDbQueryItem
{
    CSM_KM_DB_QUERY_TYPE type;
    void * value;
} CSM_KM_DB_QUERY_ITEM;

CSSM_RETURN csm_kmCreateDb(
                        const char * pFilename );

CSSM_RETURN csm_kmDestroyDb(
                        const char * pFilename );

CSSM_RETURN csm_kmOpenDb(
                        const char * pFilename,
                        int * pFd );

CSSM_RETURN csm_kmCloseDb(
                        int fd );

CSSM_RETURN csm_kmInsertRecord(
                        int fd,
                        const CSM_KM_DB_RECORD * pRecord,
                        long * pDbIndex );

CSSM_RETURN csm_kmDeleteRecord(
                        int fd,
                        unsigned int numQueries,
                        CSM_KM_DB_QUERY_ITEM * pQueries );

CSSM_RETURN csm_kmQueryDb(
                        int fd,
                        unsigned int numQueries,
                        CSM_KM_DB_QUERY_ITEM * pQueries,
                        unsigned int numRequests,
                        CSM_KM_DB_QUERY_ITEM * pRequests,
                        PORT_FLOCK_TYPE LockType,
                        PORT_FLOCK_RANGE *pLockRange,
                        CSSM_BOOL * pFound );

CSSM_RETURN csm_kmRecordsMatch(
                        const CSM_KM_DB_RECORD * pRecord,
                        const CSM_KM_DB_RECORD * pQryRecord,
                        const CSSM_STRING * pQryKeyName,
                        const uint8 * pQryHash,
                        CSSM_BOOL * pMatch );

CSSM_RETURN csm_kmRecordToHostOrder(
                        const CSM_KM_DB_RECORD * pSrcRec,
                        CSM_KM_DB_RECORD * pDstRec );

CSSM_RETURN csm_kmRecordToNetOrder(
                        const CSM_KM_DB_RECORD * pSrcRec,
                        CSM_KM_DB_RECORD * pDstRec );

#ifdef __cplusplus
}
#endif

#endif /* ifndef __CSM_KEYMGR_DB_H__ */
