/*-----------------------------------------------------------------------
 * File: CSM_KEYMGR_INTERNAL.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "maf_include.h"

#ifndef LINUX
#ifndef VMS
#include <io.h>
#endif
#endif

#include "csm_keymgr.h"
#include "csm_keymgr_db.h"
#include "csm_keymgr_internal.h"

#include "eicl.h"

#include "hmacsha1.h"
#include "pkcs5v2.h"

/*
 * Name: csm_kmBuildNewDBRecord
 *
 * Description:
 *     Allocates and fills in a new DB record. The record is allocated using
 *     the Addin_calloc function with a NULL AllocRef parameter. The key
 *     referenced by the pKey parameter is encrypted using the csm_kmWrapKey
 *     function.
 *
 * Parameters:
 *     const CSSM_KEY *                       pKey
 *     const CSSM_KEY *                       pRefKey
 *     const CSSM_DATA *                      pName
 *     const CSSM_RESOURCE_CONTROL_CONTEXT *  pCredAndAclEntry
 *     uint8 *                                pDigest
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 */
CSSM_RETURN csm_kmBuildNewDBRecord(
                        const CSSM_KEY *pKey,
                        const CSSM_KEY *pRefKey,
                        const CSSM_DATA *pName,
                        const CSSM_RESOURCE_CONTROL_CONTEXT *pCredAndAclEntry,
                        CSM_KM_DB_RECORD **ppRecord )
{
    CSSM_RETURN ret = CSSM_OK;
    CSM_KM_DB_RECORD *pRecord = NULL;
    CSSM_KEY wrappedKey;

    /* Allocate memory for the db record. */
    pRecord = (CSM_KM_DB_RECORD *)Addin_calloc( sizeof(CSM_KM_DB_RECORD),
                                                1,
                                                NULL);
    if (!pRecord)
    {
        ERR( ret = CSSMERR_CSP_MEMORY_ERROR );
    }
    else
    {
        /* Set the fields of the record. */

        pRecord->recordIsValid = CSSM_TRUE;

        if ( pName )
        {
            if ( pName->Length > CSSM_MODULE_STRING_SIZE )
            {
                ERR( ret = CSSMERR_CSP_INPUT_LENGTH_ERROR );
            }
            else
            {
                memcpy( pRecord->keyName, pName->Data,  pName->Length );
                pRecord->keyName[pName->Length] = '\0';
            }
        }

        if ( ret == CSSM_OK )
        {
            if ( ( pKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY ) )
            {
                /* Hash the public key and store it in the record. */

                /*
                 * Determine which key to hash.  If pKey is a private key,
                 * then hash pRefKey.  If pKey is a public key, then hash pKey.
                 */
                if ( pKey->KeyHeader.KeyClass == CSSM_KEYCLASS_PRIVATE_KEY )
                {
                    ret = csm_kmHashKey( pRefKey, pRecord->keyRefHash );
                }
                else
                {
                    ret = csm_kmHashKey( pKey, pRecord->keyRefHash );
                }

                if (ret == CSSM_OK)
                {
                    pRecord->keyRefHashValid = CSSM_TRUE;
                }
            }
            else
            {
                pRecord->keyRefHashValid = CSSM_FALSE;
            }
        }

        if (ret == CSSM_OK)
        {
            if ( ( pKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PUBLIC_KEY ) &&
                 ( pCredAndAclEntry != NULL ) )
            { /* Key is protected by an ACL */
                CSSM_ACL_SUBJECT_TYPE SubjectType;
                CSSM_DATA DeriveSrc = { 0, NULL };
                CSSM_DATA PromptData = { 0, NULL };
                CSSM_BOOL bAllocatedDeriveSrc = CSSM_FALSE;
                CSSM_DATA PbeSalt;
                CSSM_DATA PbeIV;

                wrappedKey.KeyData.Data = NULL;
                wrappedKey.KeyData.Length = 0;

                /* Resolve the initial ACL to a value that can be used to derive an
                 * encryption key.
                 */
                if ( ( ret = MAF_GetAclSubjectType( 
                                        &pCredAndAclEntry->
                                            InitialAclEntry.
                                                Prototype.
                                                    TypedSubject,
                                        &SubjectType ) ) == CSSM_OK )
                {
                    switch ( SubjectType )
                    {
                    case CSSM_ACL_SUBJECT_TYPE_PASSWORD:
                    case CSSM_ACL_SUBJECT_TYPE_PROMPTED_PASSWORD:
                        {
                            ret = MAF_ResolvePasswordSubject( 
                                        MAF_ALLOCATE_LOCAL,
                                        &pCredAndAclEntry->InitialAclEntry,
                                        &DeriveSrc,
                                        &PromptData,
                                        &bAllocatedDeriveSrc );
                            break;
                        }
                    default:
                        {
                            ret = CSSMERR_CSP_INVALID_NEW_ACL_ENTRY;
                        }
                    }

                    if ( ret == CSSM_OK )
                    {
                        /* Set the ACL information */
                        if ( PromptData.Length != 0 )
                        {
                            pRecord->acl.aclType = CSSM_ACL_SUBJECT_TYPE_PROMPTED_PASSWORD;
                            if ( PromptData.Length > ( sizeof(pRecord->acl.prompt) - 1 ) )
                            {
                                ERR( ret = CSSMERR_CSP_INPUT_LENGTH_ERROR );
                            }
                            else
                            {
                                memcpy( pRecord->acl.prompt, PromptData.Data, PromptData.Length );
                                pRecord->acl.prompt[PromptData.Length] = '\0';
                            }
                        }
                        else
                        {
                            pRecord->acl.aclType = CSSM_ACL_SUBJECT_TYPE_PASSWORD;
                        }

                        if ( ret == CSSM_OK )
                        {
                            /* Wrap the key. */
                            PbeSalt.Data = pRecord->WrapPBESalt;
                            PbeSalt.Length = sizeof(pRecord->WrapPBESalt);
                            PbeIV.Data = pRecord->WrapIV;
                            PbeIV.Length = sizeof(pRecord->WrapIV);
                            if ( ( ret = csm_kmWrapKey( 
                                                MAF_ALLOCATE_LOCAL,
                                                pKey,
                                                &wrappedKey,
                                                &DeriveSrc,
                                                &PbeSalt,
                                                &PbeIV ) ) == CSSM_OK )
                            {
                                pRecord->WrapIVLength = PbeIV.Length;

                                /* Copy the wrapped key. */
                                if ( wrappedKey.KeyData.Length > 
                                                    CSM_KM_DB_MAX_KEY_SIZE )
                                {
                                    ERR( ret = CSSMERR_CSP_INPUT_LENGTH_ERROR );
                                }
                                else
                                {
                                    /* Copy the key header. */
                                    memcpy( &pRecord->keyHeader,
                                            &wrappedKey.KeyHeader,
                                            sizeof(wrappedKey.KeyHeader) );

                                    pRecord->keyLength = 
                                                wrappedKey.KeyData.Length;

                                    memcpy( &pRecord->key,
                                            wrappedKey.KeyData.Data,
                                            wrappedKey.KeyData.Length );
                                }

                                /* Wipe and free the wrapped key buffer */
                                memset( wrappedKey.KeyData.Data,
                                        0,
                                        wrappedKey.KeyData.Length );
                                Addin_free( wrappedKey.KeyData.Data, NULL );
                                wrappedKey.KeyData.Data = NULL;
                            } /* Encrypt the key */
                        } /* ACL password gathering succeeded */
                        
                        /* Free the password buffer used to derive the
                         * wrapping key.
                         */
                        MAF_DeleteOutputBuffer( MAF_ALLOCATE_LOCAL,
                                                &DeriveSrc,
                                                bAllocatedDeriveSrc );
                    } /* key derivation source gathering succeeded */
                } /* Get ACL subject type */

            } /* Key is protected by an ACL */
            else
            { /* Key is not restricted */
                /* Store it in the clear. */
                
                if (pKey->KeyData.Length > CSM_KM_DB_MAX_KEY_SIZE)
                {
                    ERR( ret = CSSMERR_CSP_INPUT_LENGTH_ERROR );
                }
                else
                {
                    /* Copy the key. */
                    memcpy( &pRecord->keyHeader,
                            &pKey->KeyHeader,
                            sizeof(pKey->KeyHeader) );

                    pRecord->keyLength = pKey->KeyData.Length;

                    memcpy( &pRecord->key,
                            pKey->KeyData.Data,
                            pKey->KeyData.Length );
                }
            } /* Key is not restricted */
        } /* Name copied and reference key hashed */

        /* Wipe and free the DB record if we had an error */
        if ( ret != CSSM_OK )
        {
            memset( pRecord, 0, sizeof(*pRecord) );
            Addin_free( pRecord, NULL );
            pRecord = NULL;
        }
    } /* Allocate the record structure */

    /* Return the record */
    *ppRecord = pRecord;

    return ret;
}

/*
 * Name: csm_kmHashKey
 *
 * Description:
 *     Hashes the key *pKey and places the resulting SHA-1 hash in *pDigest.
 *     *pDigest must be at least CSM_KM_DB_HASH_SIZE bytes.
 *
 * Parameters:
 *     const CSSM_KEY *                       pKey
 *     uint8 *                                pDigest
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     CSSM_CSP_OPERATION_UNSUPPORTED       - Could not hash this key.
 *     CSSM_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 */
CSSM_RETURN csm_kmHashKey(
                        const CSSM_KEY * pKey,
                        uint8 * pDigest )
{
    CSSM_RETURN ret = CSSM_OK;
    CSSM_DATA keyParts[CSM_MAX_PUBLIC_KEY_PARTS];
    uint32 numKeyParts = CSM_MAX_PUBLIC_KEY_PARTS;
    EICLSHAState shaState;
    EICLData shaData;
    EICLSHADigest digest;

    if ( !pKey ||
         ( ( pKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PUBLIC_KEY ) &&
           ( pKey->KeyHeader.KeyClass != CSSM_KEYCLASS_PRIVATE_KEY ) ) ||
         ( pKey->KeyHeader.BlobType != CSSM_KEYBLOB_RAW ) )
    {
        ret = CSSMERR_CSP_INVALID_KEY;
    }
    else if ( !pDigest )
    {
        ret = KMERR_INVALID_PARAMETERS;
    }
    else if (pKey->KeyHeader.KeyClass == CSSM_KEYCLASS_SESSION_KEY)
    {
        /* Don't hash symmetric keys. */
        ret = KMERR_NOT_SUPPORTED;
    }
    else
    {
        /*
         * Only hash specific parts of the key if it is an RSA key
         * or DSA key.  For any other asymmetric algorithm, hash the
         * whole key.
         */

        if ((pKey->KeyHeader.AlgorithmId == CSSM_ALGID_RSA) ||
            (pKey->KeyHeader.AlgorithmId == CSSM_ALGID_DSA))
        {
            if ((ret = CSM_ParseKeyblob(pKey->KeyHeader.AlgorithmId,
                                        pKey->KeyHeader.KeyClass,
                                        pKey->KeyHeader.Format,
                                        keyParts,
                                        &numKeyParts,
                                        &(pKey->KeyData))) != CSSM_OK)
            {
                if (ret == CSSMERR_CSP_FUNCTION_NOT_IMPLEMENTED)
                {
                    ret = CSSMERR_CSP_UNSUPPORTED_KEY_FORMAT;
                }               
            }
            else
            {
                if (pKey->KeyHeader.AlgorithmId == CSSM_ALGID_RSA)
                {
                    /* Use the public modulus to hash. */

                    shaData.length = keyParts[CSM_RSA_MODULUS].Length;
                    shaData.value = (EICLByte *)
                                    (keyParts[CSM_RSA_MODULUS].Data);
                }
                else
                {
                    /* Use the public value to hash. */

                    shaData.length = keyParts[CSM_DSA_PUBLIC_VALUE].Length;
                    shaData.value = (EICLByte *)
                                    (keyParts[CSM_DSA_PUBLIC_VALUE].Data);
                }
            }
        }
        else
        {
            shaData.length = pKey->KeyData.Length;
            shaData.value = (EICLByte *)(pKey->KeyData.Data);
        }

        if (ret == CSSM_OK)
        {
            EICL_SHABegin(&shaState);
            EICL_SHAProcess(&shaData, &shaState);
            EICL_SHAEnd(&shaState, digest);

            memcpy((void *)pDigest,
                   (const void *)&digest,
                   CSM_KM_DB_HASH_SIZE);
        }
    }

    return ret;
}

/*
 * Name: csm_kmWrapKey
 *
 * Description:
 *     Wraps the key *pKey using SHA-1/DES, and places the resuling
 *     wrapped key in *pWrappedKey.  Space for the wrapped key material
 *     will be allocated by the callee, and must include space for the
 *     salt value (the size of an ICLSalt, which is probably two
 *     unsigned longs).
 *
 *     Currently, there is only support for prompted passwords.
 *
 * Parameters:
 *     const CSSM_KEY *                       pKey
 *     CSSM_KEY *                             pWrappedKey
 *     CSSM_CHALLENGE_CALLBACK                callback
 *     void *                                 pCallbackContext
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     CSSMERR_CSP_MEMORY_ERROR                - Couldn't allocate memory.
 *     CSSM_CSP_INPUT_LENGTH_OVERSIZE       - The input was too big.
 *     CSSM_CSP_OPERATION_UNSUPPORTED       - Could not hash this key.
 *     CSSM_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 */
CSSM_RETURN csm_kmWrapKey(
                        CSSM_HANDLE hModule,
                        const CSSM_KEY *pKey,
                        CSSM_KEY *pWrappedKey,
                        const CSSM_DATA *pDeriveSrc,
                        CSSM_DATA_PTR pPbeSalt,
                        CSSM_DATA_PTR pIV )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 KeyDataLength;
    uint8 KeyDataBuf[CSM_KM_PBE_MAX_KEY_LENGTH];
    uint32 IVLength;

    assert( pKey && 
            pKey->KeyData.Data &&
            ( pKey->KeyData.Length > 0 ) &&
            pWrappedKey &&
            ( pWrappedKey->KeyData.Data == NULL ) &&
            ( pWrappedKey->KeyData.Length == 0 ) &&
            pPbeSalt &&
            pPbeSalt->Data &&
            ( pPbeSalt->Length >= CSM_KM_PBE_SALT_LENGTH ) &&
            pIV &&
            pIV->Data &&
            ( pIV->Length >= CSM_KM_PBE_MAX_IV_LENGTH ) 
          );

    /* Copy the key header. */
    memcpy( &pWrappedKey->KeyHeader,
            &pKey->KeyHeader,
            sizeof(CSSM_KEYHEADER) );


    /* Get the encryption key parameters */
    rv = CSM_KMCallout_GetWrapParams( &KeyDataLength,
                                      &IVLength,
                                      &pWrappedKey->KeyHeader.WrapAlgorithmId,
                                      &pWrappedKey->KeyHeader.WrapMode );
    assert( rv == CSSM_OK );
    assert( IVLength <= pIV->Length );
    assert( KeyDataLength <= sizeof(KeyDataBuf) );

    /* Generate random IV and Salt values */
    if  ( ( ( rv = CSM_KMCallout_GetRandomBytes(
                                    IVLength,            
                                    pIV->Data ) ) == CSSM_OK ) &&
          ( ( rv = CSM_KMCallout_GetRandomBytes(
                                    CSM_KM_PBE_SALT_LENGTH,
                                    pPbeSalt->Data ) ) == CSSM_OK ) )
    {
        pIV->Length = IVLength;
        pPbeSalt->Length = CSM_KM_PBE_SALT_LENGTH;

        /* Derive the encryption key from the credential result */
        if ( ( rv = PKCS5_v2_PBKD2_HMAC_SHA1( pDeriveSrc,
                                              pPbeSalt,
                                              KeyDataLength,
                                              CSM_KM_PBE_ITERATIONS,
                                              KeyDataBuf ) ) == CSSM_OK )
        {
            /* Encrypt the data blob */
            rv = CSM_KMCallout_EncryptData( KeyDataBuf,
                                            pIV->Data,
                                            &pKey->KeyData,
                                            &pWrappedKey->KeyData );
        } /* Derive the key value */
    } /* Generate random IV and Salt values */
                                
    return rv;
}

/*
 * Name: csm_kmUnwrapKey
 *
 * Description:
 *     Unwraps the key *pWrappedKey using SHA-1/DES, and places the
 *     resuling key in *pKey.  Space for the key material will be
 *     allocated by the callee, and will be smaller than the wrapped
 *     key due to the salt also being stored in the wrapped key.  The
 *     key will be smaller than the wrapped key by the size of the salt
 *     value (the size of an ICLSalt, which should is probably two
 *     unsigned longs).
 *
 *     Currently, there is only support for prompted passwords.
 *
 * Parameters:
 *     const CSSM_KEY *                       pKey
 *     CSSM_KEY *                             pWrappedKey
 *     CSSM_CHALLENGE_CALLBACK                callback
 *     void *                                 pCallbackContext
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     CSSMERR_CSP_MEMORY_ERROR                - Couldn't allocate memory.
 *     CSSM_CSP_INPUT_LENGTH_OVERSIZE       - The input was too big.
 *     CSSM_CSP_OPERATION_UNSUPPORTED       - Could not hash this key.
 *     CSSM_CSP_INTERNAL_ERROR              - Something failed that shouldn't.
 */
CSSM_RETURN csm_kmUnwrapKey(
                        CSSM_HANDLE hModule,
                        const CSSM_KEY *pWrappedKey,
                        CSSM_KEY *pKey,
                        const CSSM_ACCESS_CREDENTIALS *pAccessCred,
                        CSSM_SAMPLE_TYPE ExpectedSample,
                        const CSSM_DATA *pPrompt,
                        const CSSM_DATA *pPbeSalt,
                        const CSSM_DATA *pIV )
{
    CSSM_RETURN rv = CSSM_OK;
    uint32 KeyDataLength;
    uint8 KeyDataBuf[CSM_KM_PBE_MAX_KEY_LENGTH];
    uint32 IVLength;
    CSSM_ALGORITHMS AlgorithmId;
    CSSM_ENCRYPT_MODE AlgMode;
    CSSM_DATA DeriveSrc = { 0, NULL };
    CSSM_BOOL bAllocatedDeriveSrc = CSSM_FALSE;

    assert( pWrappedKey && 
            pWrappedKey->KeyData.Data &&
            ( pWrappedKey->KeyData.Length > 0 ) &&
            pKey &&
            ( pKey->KeyData.Data == NULL ) &&
            ( pKey->KeyData.Length == 0 ) &&
            pPbeSalt &&
            pPbeSalt->Data &&
            ( pPbeSalt->Length == CSM_KM_PBE_SALT_LENGTH ) &&
            pIV &&
            pIV->Data 
          );

    /* Copy the key header. */
    memcpy( &pKey->KeyHeader,
            &pWrappedKey->KeyHeader,
            sizeof(CSSM_KEYHEADER) );
    pKey->KeyHeader.WrapAlgorithmId = CSSM_ALGID_NONE;
    pKey->KeyHeader.WrapMode = CSSM_ALGMODE_NONE;


    /* Get the encryption key parameters */
    rv = CSM_KMCallout_GetWrapParams( &KeyDataLength,
                                      &IVLength,
                                      &AlgorithmId,
                                      &AlgMode );
    assert( rv == CSSM_OK );
    assert( IVLength == pIV->Length );
    assert( KeyDataLength <= sizeof(KeyDataBuf) );

    /* Check the encryption parameters to see if they will work */
    if ( ( pWrappedKey->KeyHeader.WrapAlgorithmId != AlgorithmId ) ||
         ( pWrappedKey->KeyHeader.WrapMode != AlgMode ) )
    {
        ERR( rv = CSSMERR_CSP_INTERNAL_ERROR );
    }
    else
    {
        /* Fetch the key derivation source value */
        switch ( ExpectedSample )
        {
        case CSSM_SAMPLE_TYPE_PASSWORD:
        case CSSM_SAMPLE_TYPE_PROMPTED_PASSWORD:
            {
                assert( ( ExpectedSample == CSSM_SAMPLE_TYPE_PASSWORD ) ||
                        ( ( ExpectedSample == CSSM_SAMPLE_TYPE_PROMPTED_PASSWORD ) &&
                          ( pPrompt ) ) );

                rv = MAF_ResolvePasswordCredential( hModule,
                                                    pAccessCred,
                                                    pPrompt,
                                                    &DeriveSrc,
                                                    &bAllocatedDeriveSrc );
                break;
            }
        default:
            {
                ERR( rv = CSSMERR_CSP_INVALID_ACCESS_CREDENTIALS );
            }
        }

        if ( rv == CSSM_OK )
        {
            /* Derive the encryption key from the credential result */
            if ( ( rv = PKCS5_v2_PBKD2_HMAC_SHA1( &DeriveSrc,
                                                  pPbeSalt,
                                                  KeyDataLength,
                                                  CSM_KM_PBE_ITERATIONS,
                                                  KeyDataBuf ) ) == CSSM_OK )
            {
                /* Encrypt the data blob */
                rv = CSM_KMCallout_DecryptData( KeyDataBuf,
                                                pIV->Data,
                                                &pWrappedKey->KeyData,
                                                &pKey->KeyData );
            } /* Derive the key value */

            MAF_DeleteOutputBuffer( hModule,
                                    &DeriveSrc,
                                    bAllocatedDeriveSrc );
        }
    } /* Encryption algorithms match */
    
    return rv;
}

/*
 * Name: csm_kmInitHandles
 *
 * Description:
 *     Initializes the handle store.
 *
 * Parameters:
 *     CSM_KM_HANDLES *                       pHandles
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 */
CSSM_RETURN csm_kmInitHandles(
                        CSM_KM_HANDLES * pHandles )
{
    CSSM_RETURN ret = CSSM_OK;

    if (!pHandles)
    {
        ret = KMERR_INVALID_PARAMETERS;
    }
    else
    {
        pHandles->handles = NULL;
    }

    return ret;
}

/*
 * Name: csm_kmNoHandles
 *
 * Description:
 *     Proposition for store emptiness.
 *
 * Parameters:
 *     CSM_KM_HANDLES *                       pHandles
 *     CSSM_BOOL *                            pNoHandles
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 */
CSSM_RETURN csm_kmNoHandles(
                        CSM_KM_HANDLES * pHandles,
                        CSSM_BOOL * pNoHandles )
{
    CSSM_RETURN ret = CSSM_OK;

    if (!pHandles || !pNoHandles)
    {
        ret = KMERR_INVALID_PARAMETERS;
    }
    else
    {
        if (pHandles->handles)
        {
            *pNoHandles = CSSM_FALSE;
        }
        else
        {
            *pNoHandles = CSSM_TRUE;
        }
    }

    return ret;
}

/*
 * Name: csm_kmInsertHandle
 *
 * Description:
 *     Inserts a handle into the store.  It does not check for duplicates.
 *
 * Parameters:
 *     CSM_KM_HANDLES *                       pHandles
 *     CSM_KM_HANDLE                          handle
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 *     CSSMERR_CSP_MEMORY_ERROR                - Couldn't allocate memory.
 */
CSSM_RETURN csm_kmInsertHandle(
                        CSM_KM_HANDLES * pHandles,
                        CSM_KM_HANDLE handle )
{
    CSSM_RETURN ret = CSSM_OK;
    CSM_KM_HANDLES_NODE * pNewHandleNode = NULL;

    if ( !pHandles )
    {
        ret = KMERR_INVALID_PARAMETERS;
    }
    else if ( !handle )
    {
        ret = KMERR_INVALID_HANDLE_POINTER;
    }
    else
    {
        pNewHandleNode = (CSM_KM_HANDLES_NODE *)
                         Addin_malloc(sizeof(CSM_KM_HANDLES_NODE), NULL);
        if (!pNewHandleNode)
        {
            ret = CSSMERR_CSP_MEMORY_ERROR;
        }
        else
        {
            pNewHandleNode->handle = handle;
            pNewHandleNode->next = pHandles->handles;
            pHandles->handles = pNewHandleNode;
        }
    }

    return ret;
}

/*
 * Name: csm_kmGetHandle
 *
 * Description:
 *     Removes a handle from the store.  If there are no handles in the store,
 *     *pHandle is set to 0.
 *
 * Parameters:
 *     CSM_KM_HANDLES *                       pHandles
 *     CSM_KM_HANDLE *                        pHandle
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 */
CSSM_RETURN csm_kmGetHandle(
                        CSM_KM_HANDLES * pHandles,
                        CSM_KM_HANDLE * pHandle )
{
    CSSM_RETURN ret = CSSM_OK;
    CSM_KM_HANDLES_NODE * pHandleNode = NULL;

    if ( !pHandles )
    {
        ret = KMERR_INVALID_PARAMETERS;
    }
    else if ( !pHandle )
    {
        ret = KMERR_INVALID_HANDLE_POINTER;
    }
    else
    {
        if (pHandles->handles)
        {
            *pHandle = pHandles->handles->handle;
            pHandleNode = pHandles->handles;
            pHandles->handles = pHandleNode->next;
            Addin_free((void *)pHandleNode, NULL);
        }
        else
        {
            *pHandle = 0;
        }
    }

    return ret;
}

/*
 * Name: csm_kmDeleteAllHandles
 *
 * Description:
 *     Removes all handles from the store.
 *
 * Parameters:
 *     CSM_KM_HANDLES *                       pHandles
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 */
CSSM_RETURN csm_kmDeleteAllHandles(
                        CSM_KM_HANDLES * pHandles )
{
    CSSM_RETURN ret = CSSM_OK;
    CSM_KM_HANDLES_NODE * pHandleNode = NULL;

    if (!pHandles)
    {
        ret = KMERR_INVALID_PARAMETERS;
    }
    else
    {
        while (pHandles->handles)
        {
            pHandleNode = pHandles->handles;
            pHandles->handles = pHandleNode->next;
            Addin_free((void *)pHandleNode, NULL);
        }
    }

    return ret;
}

/*
 * Name: csm_kmTermHandles
 *
 * Description:
 *     Removes all handles from the store and then
 *     shuts down the handle store.
 *
 * Parameters:
 *     CSM_KM_HANDLES *                       pHandles
 *
 * Return:
 *     CSSM_OK                              - The operation was successful.
 *     CSSM_INVALID_INPUT                   - The input was invalid.
 */
CSSM_RETURN csm_kmTermHandles(
                        CSM_KM_HANDLES * pHandles )
{
    return csm_kmDeleteAllHandles(pHandles);
}

/*
 * Name: csm_kmTearDownKeyFileHandle
 *
 * Description:
 *     Destroys a key file handles node.
 *
 * Parameters:
 *     CSM_KM_KEY_FILE_HANDLES_NODE *         pItem
 *
 * Return:
 *     void
 */
void MLC_FASTCALL csm_kmTearDownKeyFileHandle(
                        void *pItem )
{
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = pItem;

    if (pKeyFileHandlesNode)
    {
        if (pKeyFileHandlesNode->fd != -1)
        {
            csm_kmCloseDb(pKeyFileHandlesNode->fd);
        }

        if (MLC_Term(&(pKeyFileHandlesNode->keys)) != CSSM_OK)
        {
            return;
        }
        else if (csm_kmTermHandles(&(pKeyFileHandlesNode->discardedKeyHandles))
                 != CSSM_OK)
        {
            return;
        }

        Addin_free((void *)pKeyFileHandlesNode, NULL);
    }

    return;
}

/*
 * Name: csm_kmFindKeyFileHandle
 *
 * Description:
 *     Matches key file handles.
 *
 * Parameters:
 *     CSM_KM_KEY_FILE_HANDLES_NODE *         pItem
 *     CSM_KM_HANDLE *                        pKey
 *
 * Return:
 *     0                                     - A match or an error.
 *     1                                     - No match.
 */
int MLC_FASTCALL csm_kmFindKeyFileHandle(
                        void *pItem,
                        void *pKey )
{
    int ret = 0;
    int error = 0;
    CSM_KM_KEY_FILE_HANDLES_NODE * pKeyFileHandlesNode = pItem;
    CSM_KM_HANDLE * pHandle = pKey;

    if (!pKeyFileHandlesNode || !pHandle)
    {
        error = 1;
    }
    else
    {
        ret = !((*pHandle) == (pKeyFileHandlesNode->hKeyFile));
    }

    if (error)
    {
        ret = 0;
    }

    return ret;
}

/*
 * Name: csm_kmTearDownKeys
 *
 * Description:
 *     Destroys a keys node.
 *
 * Parameters:
 *     CSM_KM_KEYS_NODE *                     pItem
 *
 * Return:
 *     void
 */
void MLC_FASTCALL csm_kmTearDownKeys(
                        void *pItem )
{
    CSM_KM_KEYS_NODE * pKeysNode = pItem;

    if (pKeysNode)
    {
        if (pKeysNode->pRecord)
        {
            memset((void *)pKeysNode->pRecord, 0,
                   sizeof(CSM_KM_DB_RECORD));
            Addin_free((void *)(pKeysNode->pRecord), NULL);
        }

        Addin_free((void *)pKeysNode, NULL);
    }

    return;
}

/*
 * Name: csm_kmFindKeyByHash
 *
 * Description:
 *     Matches key hashes.
 *
 * Parameters:
 *     CSM_KM_KEYS_NODE *                     pItem
 *     CSSM_DATA *                            pKey
 *
 * Return:
 *     0                                     - A match or an error.
 *     1                                     - No match.
 */
int MLC_FASTCALL csm_kmFindKeyByHash(
                        void *pItem,
                        void *pKey )
{
    int ret = 0;
    int error = 0;
    CSM_KM_KEYS_NODE * pKeysNode = pItem;
    CSSM_DATA * pHash = pKey;

    if (!pKeysNode || !(pKeysNode->pRecord) || !pHash)
    {
        error = 1;
    }
    else
    {
        if (CSM_KM_DB_HASH_SIZE != (pHash->Length))
        {
            error = 1;
        }
        else
        {
            ret = !(memcmp((const void *)(pKeysNode->pRecord->keyRefHash),
                           (const void *)(pHash->Data),
                           CSM_KM_DB_HASH_SIZE)
                    == 0);
        }
    }

    if (error)
    {
        ret = 0;
    }

    return ret;
}

/*
 * Name: csm_kmFindKeyByName
 *
 * Description:
 *     Matches key names.
 *
 * Parameters:
 *     CSM_KM_KEYS_NODE *                     pItem
 *     CSSM_DATA *                            pKey
 *
 * Return:
 *     0                                     - A match or an error.
 *     1                                     - No match.
 */
int MLC_FASTCALL csm_kmFindKeyByName(
                        void *pItem,
                        void *pKey )
{
    int ret = 0;
    int error = 0;
    CSM_KM_KEYS_NODE * pKeysNode = pItem;
    CSSM_DATA * pName = pKey;

    if (!pKeysNode || !(pKeysNode->pRecord) || !pName)
    {
        error = 1;
    }
    else
    {
        if (CSSM_MODULE_STRING_SIZE < (pName->Length))
        {
            error = 1;
        }
        else
        {
            ret = !(strncmp((const char *)(pKeysNode->pRecord->keyName),
                            (const char *)(pName->Data),
                            CSSM_MODULE_STRING_SIZE)
                    == 0);
        }
    }

    if (error)
    {
        ret = 0;
    }

    return ret;
}

/*
 * Name: csm_kmFindKeyByHandle
 *
 * Description:
 *     Matches key handles.
 *
 * Parameters:
 *     CSM_KM_KEYS_NODE *                     pItem
 *     CSM_KM_KEY_HANDLE *                    pKey
 *
 * Return:
 *     0                                     - A match or an error.
 *     1                                     - No match.
 */
int MLC_FASTCALL csm_kmFindKeyByHandle(
                        void *pItem,
                        void *pKey )
{
    int ret = 0;
    int error = 0;
    CSM_KM_KEYS_NODE * pKeysNode = pItem;
    CSM_KM_KEY_HANDLE * pHandle = pKey;

    if (!pKeysNode || !pHandle)
    {
        error = 1;
    }
    else
    {
        ret = !((*pHandle) == (pKeysNode->hKey));
    }

    if (error)
    {
        ret = 0;
    }

    return ret;
}

/*
 * Name: csm_kmFindKeyDuplicate
 *
 * Description:
 *     If a hash is given, it is matched against the hash in the record.
 *     If a key name is given, it is also matched against the name in
 *     the record.
 *
 * Parameters:
 *     CSM_KM_KEYS_NODE *                     pItem
 *     CSM_KM_KEYS_FIND_DUPLICATE *           pKey
 *
 * Return:
 *     0                                     - A match or an error.
 *     1                                     - No match.
 */
int MLC_FASTCALL csm_kmFindKeyDuplicate(
                        void *pItem,
                        void *pKey )
{
    int ret = 1;
    int error = 0;
    CSM_KM_KEYS_NODE * pKeysNode = pItem;
    CSM_KM_KEYS_FIND_DUPLICATE * pDuplicate = pKey;

    if (!pKeysNode || !pDuplicate ||
        !(pDuplicate->pHash || pDuplicate->pName))
    {
        error = 1;
    }
    else
    {
        if (pDuplicate->pHash)
        {
            ret = !(memcmp((const void *)(pKeysNode->pRecord->keyRefHash),
                           (const void *)(pDuplicate->pHash),
                           CSM_KM_DB_HASH_SIZE)
                    == 0);
        }

        if (ret && pDuplicate->pName)
        {
            if (CSSM_MODULE_STRING_SIZE < (pDuplicate->pName->Length))
            {
                error = 1;
            }
            else
            {
                ret = !(strncmp((const char *)(pKeysNode->pRecord->keyName),
                                (const char *)(pDuplicate->pName->Data),
                                CSSM_MODULE_STRING_SIZE)
                        == 0);
            }
        }
    }

    if (error)
    {
        ret = 0;
    }

    return ret;
}
