/*-----------------------------------------------------------------------
 * File: TPM_CSPI.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/* Implements the SPI shims needed to provide resource protection around
 * existing TP functionality using MAF.
 */

/* Generic CSSM stuff */
#include "cssm.h"
#include "cssmspi.h"
#include "cssmtpi.h"

#ifdef ISL_INCLUDED
/* EISL stuff */
#include "eislapi.h"
#endif

/* MAF stuff */
#include "maf_include.h"

/* Local headers */
#include "tpm_cspi.h"

#include <cssmMemTrack.h>

CSSM_RETURN CSSMAPI 
TPM_SubmitCredRequest(
                        CSSM_TP_HANDLE TPHandle,
                        const CSSM_TP_AUTHORITY_ID *PreferredAuthority,
                        CSSM_TP_AUTHORITY_REQUEST_TYPE RequestType,
                        const CSSM_TP_REQUEST_SET *RequestInput,
                        const CSSM_TP_CALLERAUTH_CONTEXT *CallerAuthContext,
                        sint32 *EstimatedTime,
                        CSSM_DATA_PTR ReferenceIdentifier )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_SubmitCredRequest(TPHandle,
								PreferredAuthority,
								RequestType,
								RequestInput,
								CallerAuthContext,
								EstimatedTime,
								ReferenceIdentifier);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}

CSSM_RETURN CSSMAPI 
TPM_RetrieveCredResult(
                        CSSM_TP_HANDLE TPHandle,
                        const CSSM_DATA *ReferenceIdentifier,
                        const CSSM_TP_CALLERAUTH_CONTEXT *CallerAuthCredentials,
                        sint32 *EstimatedTime,
                        CSSM_BOOL *ConfirmationRequired,
                        CSSM_TP_RESULT_SET_PTR *RetrieveOutput )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_RetrieveCredResult(TPHandle,
								ReferenceIdentifier,
								CallerAuthCredentials,
								EstimatedTime,
								ConfirmationRequired,
								RetrieveOutput);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_ConfirmCredResult(
                        CSSM_TP_HANDLE TPHandle,
                        const CSSM_DATA *ReferenceIdentifier,
                        const CSSM_TP_CALLERAUTH_CONTEXT *CallerAuthCredentials,
                        const CSSM_TP_CONFIRM_RESPONSE *Responses,
                        const CSSM_TP_AUTHORITY_ID *PreferredAuthority )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_ConfirmCredResult(
                        TPHandle,
                        ReferenceIdentifier,
                        CallerAuthCredentials,
                        Responses,
                        PreferredAuthority);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_ReceiveConfirmation(
                        CSSM_TP_HANDLE TPHandle,
                        const CSSM_DATA *ReferenceIdentifier,
                        CSSM_TP_CONFIRM_RESPONSE_PTR *Responses,
                        sint32 *ElapsedTime )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_ReceiveConfirmation(
                        TPHandle,
                        ReferenceIdentifier,
                        Responses,
                        ElapsedTime);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_CertReclaimKey(
                        CSSM_TP_HANDLE TPHandle,
                        const CSSM_CERTGROUP *CertGroup,
                        uint32 CertIndex,
                        CSSM_LONG_HANDLE KeyCacheHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertReclaimKey(
                        TPHandle,
                        CertGroup,
                        CertIndex,
                        KeyCacheHandle,
                        CSPHandle,
                        CredAndAclEntry);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_CertReclaimAbort(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_LONG_HANDLE KeyCacheHandle )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertReclaimAbort(
                        TPHandle,
                        KeyCacheHandle);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_FormRequest(
                        CSSM_TP_HANDLE TPHandle,
                        const CSSM_TP_AUTHORITY_ID *PreferredAuthority,
                        CSSM_TP_FORM_TYPE FormType,
                        CSSM_DATA_PTR BlankForm )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_FormRequest(
                        TPHandle,
                        PreferredAuthority,
                        FormType,
                        BlankForm);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_FormSubmit(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_TP_FORM_TYPE FormType,
                        const CSSM_DATA *Form,
                        const CSSM_TP_AUTHORITY_ID *ClearanceAuthority,
                        const CSSM_TP_AUTHORITY_ID *RepresentedAuthority,
                        CSSM_ACCESS_CREDENTIALS_PTR Credentials )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_FormSubmit(
                        TPHandle,
                        FormType,
                        Form,
                        ClearanceAuthority,
                        RepresentedAuthority,
                        Credentials);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}


CSSM_RETURN CSSMAPI 
TPM_CertGroupVerify(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_CERTGROUP *CertGroupToBeVerified,
                        const CSSM_TP_VERIFY_CONTEXT *VerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR VerifyContextResult )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertGroupVerify( TPHandle,
                                      CLHandle,
                                      CSPHandle,
                                      CertGroupToBeVerified,
                                      VerifyContext,
                                      VerifyContextResult);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}

CSSM_RETURN CSSMAPI 
TPM_CertCreateTemplate(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        uint32 NumberOfFields,
                        const CSSM_FIELD *CertFields,
                        CSSM_DATA_PTR CertTemplate )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertCreateTemplate(
                        TPHandle,
                        CLHandle,
                        NumberOfFields,
                        CertFields,
                        CertTemplate);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_CertGetAllTemplateFields(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        const CSSM_DATA *CertTemplate,
                        uint32 *NumberOfFields,
                        CSSM_FIELD_PTR *CertFields )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertGetAllTemplateFields(
                        TPHandle,
                        CLHandle,
                        CertTemplate,
                        NumberOfFields,
                        CertFields);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_CertSign(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_DATA *CertTemplateToBeSigned,
                        const CSSM_CERTGROUP *SignerCertGroup,
                        const CSSM_TP_VERIFY_CONTEXT *SignerVerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR SignerVerifyResult,
                        CSSM_DATA_PTR SignedCert )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertSign( TPHandle, 
                               CLHandle, 
                               CCHandle, 
                               CertTemplateToBeSigned, 
                               SignerCertGroup,
                               SignerVerifyContext,
                               SignerVerifyResult, 
                               SignedCert);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}


CSSM_RETURN CSSMAPI 
TPM_CrlVerify(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_ENCODED_CRL *CrlToBeVerified,
                        const CSSM_CERTGROUP *SignerCertGroup,
                        const CSSM_TP_VERIFY_CONTEXT *VerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR RevokerVerifyResult )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CrlVerify( TPHandle,
                                CLHandle,
                                CSPHandle,
                                CrlToBeVerified,
                                SignerCertGroup,
                                VerifyContext,
                                RevokerVerifyResult);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}

CSSM_RETURN CSSMAPI 
TPM_CrlCreateTemplate(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        uint32 NumberOfFields,
                        const CSSM_FIELD *CrlFields,
                        CSSM_DATA_PTR NewCrlTemplate )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CrlCreateTemplate(
                        TPHandle,
                        CLHandle,
                        NumberOfFields,
                        CrlFields,
                        NewCrlTemplate);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}


CSSM_RETURN CSSMAPI 
TPM_CertRevoke(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_DATA *OldCrlTemplate,
                        const CSSM_CERTGROUP *CertGroupToBeRevoked,
                        const CSSM_CERTGROUP *RevokerCertGroup,
                        const CSSM_TP_VERIFY_CONTEXT *RevokerVerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR RevokerVerifyResult,
                        CSSM_TP_CERTCHANGE_REASON Reason,
                        CSSM_DATA_PTR NewCrlTemplate )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertRevoke( TPHandle,
                                 CLHandle,
                                 CSPHandle,
                                 OldCrlTemplate,
                                 CertGroupToBeRevoked,
                                 RevokerCertGroup,
                                 RevokerVerifyContext,
                                 RevokerVerifyResult,
                                 Reason,
                                 NewCrlTemplate);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}

CSSM_RETURN CSSMAPI 
TPM_CertRemoveFromCrlTemplate(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_DATA *OldCrlTemplate,
                        const CSSM_CERTGROUP *CertGroupToBeRemoved,
                        const CSSM_CERTGROUP *RevokerCertGroup,
                        const CSSM_TP_VERIFY_CONTEXT *RevokerVerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR RevokerVerifyResult,
                        CSSM_DATA_PTR NewCrlTemplate )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertRemoveFromCrlTemplate(
                        TPHandle,
                        CLHandle,
                        CSPHandle,
                        OldCrlTemplate,
                        CertGroupToBeRemoved,
                        RevokerCertGroup,
                        RevokerVerifyContext,
                        RevokerVerifyResult,
                        NewCrlTemplate);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}


CSSM_RETURN CSSMAPI 
TPM_CrlSign(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_ENCODED_CRL *CrlToBeSigned,
                        const CSSM_CERTGROUP *SignerCertGroup,
                        const CSSM_TP_VERIFY_CONTEXT *SignerVerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR SignerVerifyResult,
                        CSSM_DATA_PTR SignedCrl )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CrlSign( TPHandle,
                              CLHandle,
                              CCHandle,
                              CrlToBeSigned,
                              SignerCertGroup,
                              SignerVerifyContext,
                              SignerVerifyResult,
                              SignedCrl );

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}

CSSM_RETURN CSSMAPI 
TPM_ApplyCrlToDb(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_ENCODED_CRL *CrlToBeApplied,
                        const CSSM_CERTGROUP *SignerCertGroup,
                        const CSSM_TP_VERIFY_CONTEXT *ApplyCrlVerifyContext,
                        CSSM_TP_VERIFY_CONTEXT_RESULT_PTR ApplyCrlVerifyResult )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_ApplyCrlToDb( TPHandle, 
                                     CLHandle,
                                     CSPHandle,
                                     CrlToBeApplied,
                                     SignerCertGroup,
                                     ApplyCrlVerifyContext,
                                     ApplyCrlVerifyResult);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}

CSSM_RETURN CSSMAPI 
TPM_CertGroupConstruct(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CSP_HANDLE CSPHandle,
                        const CSSM_DL_DB_LIST *DBList,
                        const void *ConstructParams,
                        const CSSM_CERTGROUP *CertGroupFrag,
                        CSSM_CERTGROUP_PTR *CertGroup )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertGroupConstruct( TPHandle,
                                         CLHandle,
                                         CSPHandle,
                                         DBList,
                                         ConstructParams,
                                         CertGroupFrag,
                                         CertGroup);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}

CSSM_RETURN CSSMAPI 
TPM_CertGroupPrune(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        const CSSM_DL_DB_LIST *DBList,
                        const CSSM_CERTGROUP *OrderedCertGroup,
                        CSSM_CERTGROUP_PTR *PrunedCertGroup )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertGroupPrune( TPHandle,
                                     CLHandle,
                                     DBList,
                                     OrderedCertGroup,
                                     PrunedCertGroup);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}

CSSM_RETURN CSSMAPI 
TPM_CertGroupToTupleGroup(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        const CSSM_CERTGROUP *CertGroup,
                        CSSM_TUPLEGROUP_PTR *TupleGroup )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_CertGroupToTupleGroup(
                        TPHandle,
                        CLHandle,
                        CertGroup,
                        TupleGroup);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI
TPM_TupleGroupToCertGroup(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        const CSSM_TUPLEGROUP *TupleGroup,
                        CSSM_CERTGROUP_PTR *CertTemplates )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_TupleGroupToCertGroup(
                        TPHandle,
                        CLHandle,
                        TupleGroup,
                        CertTemplates);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;

}

CSSM_RETURN CSSMAPI 
TPM_PassThrough(
                        CSSM_TP_HANDLE TPHandle,
                        CSSM_CL_HANDLE CLHandle,
                        CSSM_CC_HANDLE CCHandle,
                        const CSSM_DL_DB_LIST *DBList,
                        uint32 PassThroughId,
                        const void *InputParams,
                        void **OutputParams )
{
    void *pLockRef = NULL;
    CSSM_RETURN rv = CSSM_OK;

    /* Setup the current execution context */
    ADDIN_SPI_BEGIN( rv, TPHandle, pLockRef );
    if (rv != CSSM_OK)  
    {
        return rv;
    }
    if (pLockRef == NULL)
    {
        return CSSMERR_TP_INTERNAL_ERROR;
    }

    /* Perform the actual service */
    rv = TP_PassThrough( TPHandle, 
                                  CLHandle, 
                                  CCHandle, 
                                  DBList, 
                                  PassThroughId,
                                  InputParams,
                                  OutputParams);

    /* Shut down the SPI handler */
    Addin_SPIEnd( pLockRef );

    /* Return the proper value */
    return rv;
}
