/***************************************************************************/
/* ac_tuple.c                                                              */
/*                                                                         */
/* Process tuples -- doing all reductions.                                 */
/*                                                                         */
/* This version includes the interface code for AuthCompute, in CDSA.      */
/*                                                                         */
/***************************************************************************/

#include "cssm.h"
#include "mempool.h"

#include "sha.h"                /* SHA-1 structures and calls */
#include "md5.h"                /* MD5 structures and calls */

#include "abbrev.h"             /* abbreviations for data types */
#include "errcodes.h"           /* error codes */
#include "tuple.h"              /* structures for tuple reduction */
#include "hashtabs.h"           /* definitions for hash tables */
#include "free.h"               /* free routines */
#include "cwordtab.h"           /* defines spelling of reserved words */
#include "lookup.h"             /* defines integer values of res'd words */
#include "sutil.h"              /* S-expression utility code */
#include "bs_compare.h"         /* byte string comparison */
#include "tagint.h"             /* tag intersection routines */
#include "strread.h"            /* for parsing S-expressions */
#include "psexp.h"              /* for parsing -- for output */
#include "cdsa-util.h"          /* CDSA data structure conversions */

extern MEMORY_POOL_FUNCS memory_pool_funcs;
extern MEMORY_POOL_PTR poolptr;

/***************************************************************************/
/***************************************************************************/
/* To make AC thread-safe, move these hash tables to a per-invocation      */
/* data block.                                                             */
/***************************************************************************/
/***************************************************************************/

/***************************************************************************/
/* Define hash tables                                                      */
/***************************************************************************/

static pPkh pktab[ HASHTABLTH ] ; /* for finding public key definitions */
                                /* when an object refers to the key */
                                /* only by its hash, or for holding hash */
                                /* values that have no matching key. */

/***************************************************************************/
/* These two are a pair.  A tuple is entered in one and scanned against    */
/* the other, looking for a chance to reduce.                              */
/* In each case, the hash is on the hash value of the key/hash and the     */
/* first name in a name string.                                            */
/***************************************************************************/

static pRen5 intab[ HASHTABLTH ] ; /* issuers ready for name reduction  */
static pRen5 sntab[ HASHTABLTH ] ; /* subjects ready for name reduction  */

/***************************************************************************/
/* acl_ren                                                                 */
/*                                                                         */
/* This is a rendezvous list for tuples that are ACLs -- have no issuer.   */
/***************************************************************************/

static pRen5 acl_ren = NULL ; /* list for ACL tuples */

/***************************************************************************/
/***************************************************************************/
/* End of data to be moved to an AC invocation state block.                */
/***************************************************************************/
/***************************************************************************/

/***************************************************************************/
/* Forward references to internal routines.                                */
/***************************************************************************/

static void mpl_hash( SHA_CTX *pSi, MD5_CTX *pMi, UCP b, int l ) ;
static void mpl_bs_hash( SHA_CTX *pSi, MD5_CTX *pMi, UCP b, int l ) ;
static void mpl_bytes_hash( SHA_CTX *pSi, MD5_CTX *pMi, pBytes b ) ;
static void walk_hash( SHA_CTX *pSi, MD5_CTX *pMi, pSexp k ) ;
static void mpl_hash_sexp( SHA_CTX *pSi, MD5_CTX *pMi, pSexp s ) ;
static int hindex( UCP hv ) ;
static int plant_tuple( pTuple t, pSexp subj, int no_subj ) ;

/***************************************************************************/
/* hashtab_init                                                            */
/*                                                                         */
/* Zero the hash tables, on first starting up.  Their contents is          */
/* assumed to be trash.                                                    */
/***************************************************************************/

void hashtab_init( void )
{
  int i ;

  for (i=0; i<HASHTABLTH; i++) {
    pktab[i] = NULL ;
    intab[i] = NULL ;
    sntab[i] = NULL ;
  } /* for */
  acl_ren = NULL ;
} /* hashtab_init */

/***************************************************************************/
/* hashtab_clear                                                           */
/*                                                                         */
/* Remove any contents of the hash tables and return them to free storage. */
/*  This is to be called on exit from the package.                         */
/***************************************************************************/

void hashtab_clear( void )
{
  int i ;

  for (i=0; i<HASHTABLTH; i++ ) {
    drop_pkh_list( pktab[i] ) ;
    drop_ren5_list( intab[i] ) ;
    drop_ren5_list( sntab[i] ) ;
  } /* for */
  drop_ren5_list( acl_ren ) ;
} /* hashtab_clear */

/***************************************************************************/
/* mpl_hash                                                                */
/*                                                                         */
/* Hash the given byte string by all required hashes.                      */
/***************************************************************************/

static void mpl_hash( SHA_CTX *pSi, MD5_CTX *pMi, UCP b, int l )
{
  if ((b == NULL)||(l < 1)) return ; /* no work to do */
  if (pSi != NULL) SHA1_Update( pSi, b, l ) ; /* gather the SHA */
  if (pMi != NULL) MD5_Update( pMi, b, l ) ; /* and/or the MD5 */
} /* mpl_hash */

/***************************************************************************/
/* mpl_bs_hash                                                             */
/*                                                                         */
/* Hash a byte string in canonical form.  That is, convert the length      */
/* to a decimal integer followed by a colon and hash that in.              */
/*                                                                         */
/* For example, "testing" becomes "7:testing" .                            */
/*                                                                         */
/***************************************************************************/

static void mpl_bs_hash( SHA_CTX *pSi, MD5_CTX *pMi, UCP b, int l )
{
  UC lth[32] ;                  /* overkill for a length field - be safe */
  int llth ;                    /* length of the length string */

  if (b == NULL) return ;

  /* don't bother clamping the length.  If it's negative or wrong, */
  /* the hash will be wrong. */

  llth = sprintf( (char *)lth, "%d:", l ) ; /* get the length string and its length */
  mpl_hash( pSi, pMi, lth, llth ) ; /* hash in the length field */
  mpl_hash( pSi, pMi, b, l ) ;  /* and the string */

  return ;                      /* and return */
} /* mpl_bs_hash */

/***************************************************************************/
/* mpl_bytes_hash                                                          */
/*                                                                         */
/* Hash a Bytes block.                                                     */
/*  If there is a display hint, hash that in too.                          */
/***************************************************************************/

static void mpl_bytes_hash( SHA_CTX *pSi, MD5_CTX *pMi, pBytes b )
{
  if (b == NULL) return ;

  if (b->disp != NULL) {
    mpl_hash( pSi, pMi, (UCP)"[", 1 ) ;
    mpl_bs_hash( pSi, pMi, (UCP)b->disp, b->d_lth ) ;
    mpl_hash( pSi, pMi, (UCP)"]", 1 ) ;
  }
  mpl_bs_hash( pSi, pMi, b->val, b->v_lth ) ;
  return ;
} /* mpl_bytes_hash */

/***************************************************************************/
/* walk_hash                                                               */
/*                                                                         */
/* Compute the hash(es) of an S-expression by walking the parsed structure */
/* This is for use when there is no canonical representation available.    */
/*                                                                         */
/***************************************************************************/

static void walk_hash( SHA_CTX *pSi, MD5_CTX *pMi, pSexp k )
{
  if (k == NULL) return ;       /* no action if no expression */

  mpl_hash( pSi, pMi, (UCP)"(", 1 ) ; /* opening paren */

  while (k != NULL) {
    CSSM_DATA *sp ;

    switch( k->typ ) {
    case S_KW:
      sp = get_spelling( k->u.k ) ;
      mpl_bs_hash( pSi, pMi, (UCP)sp->Data, sp->Length ) ;
      break ;
    case S_UKW: case S_BS:
      mpl_bytes_hash( pSi, pMi, k->u.b ) ;
      break ;
    case S_SEXP:
      mpl_hash_sexp( pSi, pMi, k->u.s ) ;
      break ;
    default:
      break ;
    } /* switch */

    k = k->term ;               /* and process the next term */
  } /* while */

  mpl_hash( pSi, pMi, (UCP)")", 1 ) ; /* closing paren */

  return ;

} /* walk_hash */

/***************************************************************************/
/* mpl_hash_sexp                                                           */
/*                                                                         */
/* Compute the hash(es) of an S-expression.  Use the canonical packed      */
/* block, if it is available, or the parse tree, otherwise.                */
/***************************************************************************/

static void mpl_hash_sexp( SHA_CTX *pSi, MD5_CTX *pMi, pSexp s )
{
  if (s == NULL) return ;

  if (s->expb != NULL)
    mpl_hash( pSi, pMi, s->expb, s->explth ) ;
  else
    walk_hash( pSi, pMi, s ) ;

  return ;
} /* mpl_hash_sexp */

/***************************************************************************/
/* hindex                                                                  */
/*                                                                         */
/* pull and return a hash index from the supplied hash value.              */
/***************************************************************************/

static int hindex( UCP hv )
{
#if (HASHTABLTH != 256)
  int val = hv[0] ;
#endif

#if (HASHTABLTH > 256)
  val <<= 8 ;
  val |= hv[1] ;
#endif

#if (HASHTABLTH > 65536)
  val <<= 8 ;
  val |= hv[2] ;
#endif

#if (HASHTABLTH == 256)
  return (hv[0]) ;
#else
  return ( val & HASHMASK ) ;
#endif

} /* hindex */

/***************************************************************************/
/* key_lookup                                                              */
/*                                                                         */
/* Have a public-key S-expression, s.  Look it up in the hash table and    */
/* if it's not there, create a Renlists node for it and enter it in        */
/* the table by all the hashes we know how to compute (two at the moment). */
/*                                                                         */
/*  Return the Renlists block.                                             */
/*                                                                         */
/***************************************************************************/

static int key_lookup( pSexp s, pRenlists *pRl )
{
  SHA_CTX si ;                 /* for the SHA of it */
  MD5_CTX mi ;                  /* for the MD5 */
  int wi ;                      /* walker index into the hash table */
  pPkh w = NULL ;               /* walker of the Pkh blocks (no drop) */
  pPkh sh = NULL ;              /* new Pkh block for this key for SHA */
  pPkh mh = NULL ;              /* new Pkh block for this key for MD5 */
  HashVal hv ;                  /* SHA-1 hash value, to see if the */
                                /* key is already there */
  pRenlists rl = NULL ;         /* Renlists node for a new key (drop) */

  if ((s == NULL)||(pRl == NULL))
    return ( SPKI_ERR_NULL_PARAM ) ; /* have to have the parameter */

  SHA1_Init( &si ) ;              /* init the hash context blocks */
  MD5_Init( &mi ) ;              /* ... */
  mpl_hash_sexp( &si, &mi, s ) ; /* hash the S-expression in canonical form */
  SHA1_Final( hv, &si ) ;         /* get the SHA1 value (our choice of alg) */

  /* now, see if the key has been entered already */
  wi = hindex( hv ) ;           /* compute the hash index */
  w = pktab[ wi ] ;             /* get the list of blocks in that bucket */
  while (w != NULL) {
    if (memcmp( hv, w->hash, SHA_DIGEST_LENGTH ) == 0) { /* if the SHA-1 matches */
      store_renlists( pRl, w->obj ) ;   /* use the key already there */
      return ( 0 ) ;            /* and return, success. */
    } /* if */
    w = w->nxt ;                /* try the next entry in the list */
  } /* while (w) */

  /* The key is not in the list yet.  Place it there. */
  sh = new_pkh() ;              /* get a new Pkh block for SHA-1 */
  rl = new_renlists() ;         /* and a new Renlists block */
  mh = new_pkh() ;              /* and a new Pkh for MD5 */
  if ((sh == NULL) || (rl == NULL) || (mh == NULL)) {
    drop_pkh( sh ) ;            /* drop any that succeeded */
    drop_renlists( rl ) ;       /* ... */
    drop_pkh( mh ) ;            /* ... */
    return ( SPKI_ERR_NO_MEM ) ;
  } /* if failure to carve blocks */

  /** key_lookup  ********  continued **/

  /* copy the SHA result */
  memcpy( sh->hash, hv, SHA_DIGEST_LENGTH ) ;
  memcpy( rl->hash, hv, SHA_DIGEST_LENGTH ) ; /* use SHA as the preferred hash */
  store_sexp( &(rl->obj), s ) ; /* note for later reporting */
  /* and close the MD5 */
  MD5_Final( &(mh->hash[0]), &mi ) ; /* Note: cseq reversed from SHA-1 */
                                /* mh->hash was cleared when carved */
  store_renlists( &(sh->obj), rl ) ; /* point to the Renlists node */
  store_renlists( &(mh->obj), rl ) ; /* from both hashes */
  store_renlists( pRl, rl ) ;   /* return the result also */
  store_renlists( &rl, NULL ) ; /* and drop my temp copy */

  /* now, plant those Pkh blocks in the hash table */
  /* wi is already the right hindex for SHA */
  sh->nxt = pktab[ wi ] ;       /* the SHA result */
  pktab[ wi ] = sh ;            /* ... */

  wi = hindex( mh->hash ) ;     /* get the index from the MD5 value */
  mh->nxt = pktab[ wi ] ;       /* the MD5 hash */
  pktab[ wi ] = mh ;            /* ... */

  return ( 0 ) ;
} /* key_lookup */

/***************************************************************************/
/* pass0                                                                   */
/*                                                                         */
/* Walk an S-expression, looking for public-key sub-expressions.           */
/* For each such sub-expression found, call key_lookup to canonicalize it. */
/***************************************************************************/

static int pass0( pSexp s )
{
  pSexp w = NULL ;              /* walker of the S-expression */
                                /* don't drop this reference */
  int r = 0 ;                   /* return value */
  pRenlists rl = NULL ;         /* dummy but still need to drop */

  if (s == NULL)
    return ( SPKI_ERR_NULL_PARAM ) ; /* can't walk */
  w = s ;                       /* get the top of the expression */
  while ((w != NULL) && (r == 0)) {
    switch (w->typ) {
    case S_KW:                  /* known keyword, must lead the sexp */
      if (w == s)               /* has to be the first term, but check */
        if (w->u.k == O_PUBLIC_KEY) {
          r = key_lookup( s, &rl ) ; /* look it up */
          store_renlists( &rl, NULL ) ; /* and forget the rl for now  */
        } /*  if, if */
      break ;
    case S_SEXP:                /* sub-expression */
      r = pass0( w->u.s ) ;     /* recurse */
      break ;
    default:
        ;
    } /* switch */
    w = w->term ;               /* go to the next term in the expression */
  } /* while */
  return ( r ) ;                /* return */
} /* pass0 */

/***************************************************************************/
/* hash_lookup                                                             */
/*                                                                         */
/* Have a hash S-expression, s.  Look it up in the public key hash table   */
/* and if it's there, return its pRenlists entry.  If it's not there,      */
/* add it to the table and carve a new pRenlists for it.                   */
/*                                                                         */
/***************************************************************************/

int hash_lookup( pSexp s, pRenlists *pRl )
{
  int wi ;                      /* walker index into the hash table */
  pPkh w = NULL ;               /* walker of the Pkh blocks (no drop) */
  pPkh h = NULL ;               /* new Pkh block for this hash (no drop) */
  HashVal hv ;                  /* SHA-1 hash value, to see if the */
                                /* key is already there */
  UCP hvp ;                     /* hash value pointer */
  UI hvl ;                      /* hash value length */
  int h_alg ;                   /* hash algorithm ID */
  int r ;                       /* return value */
  pRenlists rl = NULL ;         /* a new Renlists node, if we need it */

  if ((s == NULL) || (pRl == NULL))
    return ( SPKI_ERR_NULL_PARAM ) ; /* have to have the parameters */
  r = pull_hashval( s, &hvp, &hvl, &h_alg ) ;
  if (r != 0) return ( r ) ;

  switch( h_alg ) {
  case O_SHA1:
    if (hvl != SHA_DIGEST_LENGTH) return ( SPKI_ERR_DIFF_HASH_LTHS ) ; break ;
  case O_MD5:
    if (hvl != MD5_DIGEST_LENGTH) return ( SPKI_ERR_DIFF_HASH_LTHS ) ; break ;
  default: /* some other hash algorithm, not implemented here */
    return ( SPKI_ERR_UNKNOWN_HASH ) ; /* unsuccessful return */
  } /* switch */

  memset( hv, 0, MAXHASHLTH ) ; /* clear hv */
  memcpy( hv, hvp, hvl ) ;      /* copy the hash value into storage here */

  /* now, see if the key has been entered */
  wi = hindex( hv ) ;           /* compute the hash index */
  w = pktab[ wi ] ;             /* get the list of blocks in that bucket */
  while (w != NULL) {
    if (memcmp( hv, w->hash, MAXHASHLTH ) == 0) { /* if the hash matches */
      store_renlists( pRl, w->obj ) ; /* yes, use the one defined already */
      return ( 0 ) ;            /* and return, successfully */
    } /* if */
    w = w->nxt ;                /* try the next entry in the list */
  } /* while (w) */

  /** hash_lookup  ********  continued **/

  /* The hash is not in the list yet.  Place it there. */
  h = new_pkh() ;               /* get a new block */
  rl = new_renlists() ;         /* and a new Renlists block */
  if ((h == NULL) || (rl == NULL)) {
    drop_pkh( h ) ;
    drop_renlists( rl ) ;
    return ( SPKI_ERR_NO_MEM ) ;
  } /* if no mem */

  /* copy the hash result */
  memcpy( h->hash, hv, MAXHASHLTH ) ;
  memcpy( rl->hash, hv, MAXHASHLTH ) ;
  store_sexp( &(rl->obj), s ) ; /* save for later reporting */
  store_renlists( &(h->obj), rl ) ;
  store_renlists( pRl, rl ) ;   /* return the result */
  store_renlists( &rl, NULL ) ; /* drop my storage */

  /* now, plant that block in the hash table */
  /* wi is already the right hindex */
  h->nxt = pktab[ wi ] ;        /* ... */
  pktab[ wi ] = h ;             /* ... */

  return ( 0 ) ;                /* successful return */
} /* hash_lookup */

/***************************************************************************/
/* kh_lookup                                                               */
/*                                                                         */
/* Have either a public-key or a hash S-expression, s.  Look it up in      */
/* the hash table and return its Renlists block.                           */
/*                                                                         */
/***************************************************************************/

static int kh_lookup( pSexp s, pRenlists *pRl )
{
  if ((s == NULL) || (pRl == NULL)) return ( SPKI_ERR_NULL_PARAM ) ;
  switch (pull_resword( s )) {
  case O_PUBLIC_KEY: return ( key_lookup( s, pRl ) ) ;
  case O_HASH: return ( hash_lookup( s, pRl ) ) ;
  default: break ;
  } /* switch */
  return ( SPKI_ERR_NOT_KEY_OR_HASH ) ;
} /* kh_lookup */

/***************************************************************************/
/* subj_lookup                                                             */
/*                                                                         */
/* Have what should be a (subject...), with either a key or a hash as      */
/* the subject.  Look it up.                                               */
/***************************************************************************/

static int subj_lookup( pSexp s, pRenlists *pRl )
{
  if ((s == NULL) || (pRl == NULL)) return ( SPKI_ERR_NULL_PARAM ) ;
  if (s->typ != S_KW) return ( SPKI_ERR_NOT_KEY_OR_HASH ) ;
  if (s->u.k != O_SUBJECT) return ( SPKI_ERR_NOT_KEY_OR_HASH ) ;
  s = s->term ;                 /* the next element */
  if (s == NULL) return ( SPKI_ERR_NOT_KEY_OR_HASH ) ;
  if (s->typ != S_SEXP) return ( SPKI_ERR_NOT_KEY_OR_HASH ) ;
  return ( kh_lookup( s->u.s, pRl ) ) ;
} /* subj_lookup */

/***************************************************************************/
/* pull_date                                                               */
/*                                                                         */
/* Pull and return the date string from this (not-xxx ) structure.         */
/***************************************************************************/

static pBytes pull_date( pSexp d )
{
  if (d == NULL)
    return ( NULL ) ;
  d = d->term ;                 /* skip to the next term */
  if (pull_typ( d ) != S_BS)
    return ( NULL ) ;
  return ( d->u.b ) ;           /* return the date field */
} /* pull_date */

/***************************************************************************/
/* intersect_dates                                                         */
/*                                                                         */
/* Walk d looking for dates to intersect with those in t.                  */
/* Return an error if the result should be NULL.                           */
/***************************************************************************/

static int intersect_dates( pTuple t, pSexp d )
{
  int r = 0 ;                   /* return value */

  if ((d == NULL) || (t == NULL)) return ( SPKI_ERR_NULL_PARAM ) ;
  switch (pull_resword( d )) {
  case O_NOT_BEFORE:            /* pull the date */
    if (t->nb != NULL)
      store_bytes( &(t->nb), bs_max( t->nb, pull_date( d ) ) ) ;
    else
      store_bytes( &(t->nb), pull_date( d ) ) ;
    break ;
  case O_NOT_AFTER:             /* pull the date */
    if (t->na != NULL)
      store_bytes( &(t->na), bs_min( t->na, pull_date( d ) ) ) ;
    else
      store_bytes( &(t->na), pull_date( d ) ) ;
    break ;
  case O_VALID:
    while ((d = d->term) != NULL) /* walk the interior */
      switch (d->typ) {
      case S_SEXP:
        r = intersect_dates( t, d->u.s ) ; /* recurse */
        break ;
      case S_BS:                /* a single date, to be both nb and na */
        if (t->nb != NULL)
          store_bytes( &(t->nb), bs_max( t->nb, d->u.b ) ) ;
        else
          store_bytes( &(t->nb), d->u.b ) ;
        if (t->na != NULL)
          store_bytes( &(t->na), bs_min( t->na, d->u.b ) ) ;
        else
          store_bytes( &(t->na), d->u.b ) ;
        break ;
      default:
        break ;
      } /* switch (d->typ) */
    break ;
  default: break ;
  } /* switch */

  if (r == 0)
    if ((t->na != NULL) && (t->nb != NULL))
      if (bs_compare( t->nb, t->na ) == GREATER)
        r = SPKI_ERR_DATE_FAIL ;

  return ( r ) ;
} /* intersect_dates */

/***************************************************************************/
/* store_dates                                                             */
/*                                                                         */
/* Walk d looking for dates to put away in t.                              */
/***************************************************************************/

static int store_dates( pTuple t, pSexp d )
{
  int r = 0 ;                   /* return value */

  if ((d == NULL) || (t == NULL)) return ( SPKI_ERR_NULL_PARAM ) ;
  switch (pull_resword( d )) {
  case O_NOT_BEFORE:            /* pull the date */
    if (t->nb != NULL) r = SPKI_ERR_MPL_NOT_BEFORE ;
    store_bytes( &(t->nb), pull_date( d ) ) ;
    break ;
  case O_NOT_AFTER:             /* pull the date */
    if (t->na != NULL) r = SPKI_ERR_MPL_NOT_AFTER ;
    store_bytes( &(t->na), pull_date( d ) ) ;
    break ;
  case O_VALID:
    while ((d = d->term) != NULL) /* walk the interior */
      if (d->typ == S_SEXP)     /* dealing only with sub-expressions */
        r = store_dates( t, d->u.s ) ; /* recurse */
    break ;
  default:
      ;
  } /* switch */

  return ( r ) ;
} /* store_dates */

/***************************************************************************/
/* pull_ikn                                                                */
/*                                                                         */
/* Pull out the issuer key and name, if any, from the tuple's issuer       */
/* field, i, to the .ik and .inam fields of t.                             */
/***************************************************************************/

static int pull_ikn( pTuple t, pSexp i )
{
  int r = 0 ;                   /* return value */
  pSexp w = NULL ;              /* (NO DROP) walker for the issuer */
  pSexp n = NULL ;              /* (NO DROP) walker */
  int rw ;                      /* reserved word of the issuer */

  if ((t == NULL) || (i == NULL)) return ( SPKI_ERR_NULL_PARAM ) ;
  w = i->term ;                 /* go to the first term of the issuer */
  if (w == NULL) return ( SPKI_ERR_MALFORMED_ISSUER ) ;
  if (pull_typ( w ) != S_SEXP) return ( SPKI_ERR_MALFORMED_ISSUER ) ;
  w = w->u.s ;                  /* go to the sub-expression */
  rw = pull_resword( w ) ;
  if (rw == O_NAME) {
    w = w->term ;               /* next term must be the public key */
    if (w == NULL) return ( SPKI_ERR_MALFORMED_ISSUER ) ;
    n = w->term ;               /* and the next term, should be the name */
    if (pull_typ(n) != S_BS) return ( SPKI_ERR_MALFORMED_ISSUER ) ;
    store_bytes( &(t->inam), n->u.b ) ; /* store away the issuer name */
    if (pull_typ( w ) != S_SEXP) return ( SPKI_ERR_MALFORMED_ISSUER ) ;
    w = w->u.s ;                /* go to the sub-expression */
    rw = pull_resword( w ) ;    /* now, enter the key */
    /* fall into the public key case */
  } /* if */

  /* w is an S_KW, through either path and rw matches it. */

  switch (rw) {
  case O_HASH:                   /* all hashes should be removed by now  */
    /*    r = SPKI_ERR_NON_KEY_ISSUER ;  a non-key issuer could not have   */
    /*                           been validated.  Reject it.               */
    /*    break ;                                                          */
    /* Let there be non-key issuers for CDSA AuthCompute.  Revert for SPKI. */
  case O_PUBLIC_KEY:            /* a public key -- as expected */
    r = kh_lookup( w, &(t->ik) ) ; /* plant the Renlists for this key */
    break ;

  default:
    r = SPKI_ERR_MALFORMED_ISSUER ;
    break ;
  } /* switch */
  return ( r ) ;                /* and return */
} /* pull_ikn */

/***************************************************************************/
/* pull_skn_cont                                                           */
/*                                                                         */
/* The working routine for pull_skn -- when you've walked past the         */
/* (subject ...) and are at the definition of the subject.                 */
/***************************************************************************/

static int pull_skn_cont( pTuple t, pSexp a )
{
  int r = 0 ;                   /* return value */

  switch (pull_resword( a )) { /* what kind of subject do we have? */
  case O_K_OF_N: /* handled later */
    break ;

  case O_NAME:                  /* have to pull two fields */
    a = a->term ;               /* go to the first term */
    switch (pull_typ(a)) {
    case S_BS:                  /* name without key */
      store_renlists( &(t->sk), t->ik ) ; /* same key as issuer */
      if (t->ik == NULL) return ( SPKI_ERR_MALFORMED_SUBJECT ) ;
      /* now, store the name away also, in two parts (first, rest) */
      store_bytes( &(t->snam), a->u.b ) ; /* note the name  */
      store_sexp( &(t->rnames), a->term ) ;
      break ;

    case S_SEXP:                /* should be the key or hash */
      r = kh_lookup( a->u.s, &(t->sk) ) ; /* get the subject Renlists block */
      a = a->term ;             /* next term for first name */
      if (pull_typ(a) != S_BS)
        return ( SPKI_ERR_MALFORMED_SUBJECT ) ;
      store_bytes( &(t->snam), a->u.b ) ; /* note the name  */
      store_sexp( &(t->rnames), a->term ) ;
      break ;

    default:
      r = SPKI_ERR_MALFORMED_SUBJECT ;
      break ;
    } /* switch on 1st term of a subject name */
    break ;

  case O_PUBLIC_KEY: case O_HASH:
    r = kh_lookup( a, &(t->sk) ) ; /* get the subject Renlists block */
    break ;

  default:
    r = SPKI_ERR_MALFORMED_SUBJECT ;
  } /* switch */

  return ( r ) ;                /* and return */

} /* pull_skn_cont */

/***************************************************************************/
/* pull_skn                                                                */
/*                                                                         */
/* Pull the subject key and names from the subject in this tuple,          */
/*   and plant them in .sk, .snam and .rnames .                            */
/***************************************************************************/

static int pull_skn( pTuple t, pSexp s )
{
  pSexp w = NULL ;              /* (walker, no drop) for the subject */

  w = s ;                       /* get the subject */
  if (pull_resword(w) != O_SUBJECT) return ( SPKI_ERR_MALFORMED_SUBJECT ) ;
  w = w->term ;                 /* 1st term should be the real subject */
  if (pull_typ(w) != S_SEXP) return ( SPKI_ERR_MALFORMED_SUBJECT ) ;
  return ( pull_skn_cont( t, w->u.s ) ) ;
} /* pull_skn */

/***************************************************************************/
/* splay_fields                                                            */
/*                                                                         */
/* Fill in a tuple from the S-expression                                   */
/***************************************************************************/

static int splay_fields( pTuple t, pSexp x, char is_acl )
{
  int r = 0 ;                   /* return value */

  if (t == NULL) return ( SPKI_ERR_NULL_PARAM ) ;

  /* walk x (using that as the walker), and fill in t */

  while ((x != NULL) && (r == 0)) {
    if (x->typ == S_SEXP) {     /* we care only about sub-expressions */
      pSexp y = x->u.s ;        /* (walker) look at the sub-expression */
      if (y != NULL)            /* must be non-null */
        if (y->typ == S_KW)     /* must be a known keyword */
          switch (y->u.k) {     /* plant this element in the tuple, maybe */
          case O_ISSUER:        /* an issuer block */
            if (t->ik != NULL) r = SPKI_ERR_MPL_ISSUERS ;
            if (is_acl)
              r = SPKI_ERR_ACL_ENTRY_WITH_ISSUER ;
            else {
              t->j = 0 ;
              /* OK, we have an issuer.  Extract information of interest. */
              r = pull_ikn( t, y ) ; /* pull the issuer key and name */
            } /* else */
            break ;
          case O_SUBJECT:       /* a subject block */
            if (t->s != NULL) r = SPKI_ERR_MPL_SUBJECTS ;
            r = pull_skn( t, y ) ; /* pull the subject key and names */
            store_sexp( &(t->s), y ) ; /* plant it */
            break ;
          case O_PROPAGATE:     /* delegation permitted */
            if (t->d != 0) r = SPKI_ERR_MPL_PROPAGATE_FLAGS ;
            t->d = 1 ;          /* note that */
            break ;
          case O_TAG:
            if (t->t != NULL) r = SPKI_ERR_MPL_TAGS ;
            store_sexp( &(t->t), y ) ;
            break ;
          case O_VALID:         /* validity intervals */
          case O_NOT_BEFORE:    /* old style dates at top level */
          case O_NOT_AFTER:
            store_dates( t, y ) ; /* put the dates away */
            break ;
          default:
          ;
          } /* switch, if, if */
    } /* if subexpression */
    x = x->term ;               /* go to the next term */
  } /* while */

  return ( r ) ;                /* and return */
} /* splay_fields */

/***************************************************************************/
/* coerce_dates                                                            */
/*                                                                         */
/* There is a NOW defined.  Reset the validity dates.                      */
/***************************************************************************/

static void coerce_dates( pTuple *pt, pBytes now )
{
  pTuple t = NULL ;             /* the tuple (walker -- no drop) */

  if ((pt == NULL) || (now == NULL)) return ;
  t = *pt ;                     /* get the tuple */
  store_bytes( &(t->nb), bs_max( t->nb, now ) ) ;
  store_bytes( &(t->na), bs_min( t->na, now ) ) ;
  if (bs_compare( t->nb, t->na ) == GREATER ) {
    store_tuple( pt, NULL ) ;   /* wipe it out */
    return ;                    /* and return */
  } /* tuple still valid? */
  return ;
} /* coerce_dates */

/***************************************************************************/
/* int_from_bs                                                             */
/*                                                                         */
/* Fetch a byte string's contents as an integer.  The string is presumed   */
/* to be in Internet standard order (MSB first).                           */
/***************************************************************************/

static int int_from_bs( pBytes b )
{
  int res = 0 ;                 /* result */
  int i ;                       /* loop vbl */

  if (b == NULL) return (-1) ;
  if ((b->d_lth > 0) || (b->v_lth > 3) || (b->v_lth < 1)) return (-1) ;
  for (i=0; i<b->v_lth; i++)
    res = (res << 8) | b->val[i] ;
  return ( res ) ;
} /* int_from_bs */

/***************************************************************************/
/* copy_tuple                                                              */
/*                                                                         */
/* Make a copy of a tuple, using store_xxx to make sure refcnt is all OK.  */
/***************************************************************************/

static int copy_tuple( pTuple *tout, pTuple t )
{
  pTuple x = NULL ;             /* the temp location of *tout */

  if ((tout == NULL) || (t == NULL))
    return ( SPKI_ERR_NULL_PARAM ) ;
  x = new_tuple() ;             /* get the new tuple */
  if (x == NULL) return ( SPKI_ERR_NO_MEM ) ;

  x->active = t->active ;
  x->d = t->d ;
  x->j = t->j ;
  store_renlists( &(x->ik), t->ik ) ;
  store_bytes( &(x->inam), t->inam ) ;
  store_renlists( &(x->sk), t->sk ) ;
  store_bytes( &(x->snam), t->snam ) ;
  store_sexp( &(x->rnames), t->rnames ) ;
  store_thresh( &(x->th), t->th ) ;
  store_sexp( &(x->s), t->s ) ;
  store_sexp( &(x->t), t->t ) ;
  store_bytes( &(x->nb), t->nb ) ;
  store_bytes( &(x->na), t->na ) ;

  store_tuple( tout, x ) ;
  drop_tuple( x ) ;

  return ( 0 ) ;
} /* copy_tuple */

/***************************************************************************/
/* reduce_name                                                             */
/*                                                                         */
/* Have two tuples, issuer and subject, that should be able to reduce      */
/* a name.  Do the name reduction and enter the new tuple                  */
/* using plant_tuple.                                                      */
/*                                                                         */
/* iss defines a name that subj uses.                                      */
/***************************************************************************/

static int reduce_name( pTuple iss, pTuple subj )
{
  pTuple z = NULL ;             /* the resultant tuple */
  int r = 0 ;                   /* return value */

  if ((iss == NULL) || (subj == NULL))
    return ( SPKI_ERR_NULL_PARAM ) ;
  r = copy_tuple( &z, subj ) ;  /* this will be like the subject */
  if (r != 0) {
    drop_tuple( z ) ;
    return ( r ) ;
  } /* if r */
  /* replace the subject in z */
  store_renlists( &(z->sk), iss->sk ) ; /* inherit the iss's subject */
  if (iss->active) z->active = 1 ; /* and iss's activity */
  store_bytes( &(z->nb), bs_max( iss->nb, subj->nb ) ) ;
  store_bytes( &(z->na), bs_min( iss->na, subj->na ) ) ;
  if (z->rnames == NULL)
    store_bytes( &(z->snam), NULL ) ; /* no name -- all reduced */
  else {
    pSexp rn = z->rnames ;      /* (walker - no drop) */
    if (pull_typ( rn ) != S_BS) {
      drop_tuple( z ) ;
      return ( SPKI_ERR_NAME_SYNTAX ) ;
    } /* bad name type */
    store_bytes( &(z->snam), rn->u.b ) ; /* get the next name to reduce */
    store_sexp( &(z->rnames), rn->term ) ; /* and the rest that are waiting */
  } /* else -- peel just one name */
  r = plant_tuple( z, NULL, 1 ) ; /* plant this tuple */
  drop_tuple( z ) ;             /* drop my reference to it */
  return ( r ) ;
} /* reduce_name */
 
/***************************************************************************/
/* reduce_5tuple                                                           */
/*                                                                         */
/* Have two 5-tuples, iss and subj.  The subject of the first matches      */
/* the issuer of the second.  The first is allowed to delegate and the     */
/* second is active.  So, try to make the combined tuple and enter it.     */
/***************************************************************************/

static int reduce_5tuple( pTuple iss, pTuple subj )
{
  pTuple z = NULL ;             /* the generated one */
  int r = 0 ;                   /* return value */

  if ((iss == NULL) || (subj == NULL))
    return ( SPKI_ERR_NULL_PARAM ) ;
  r = copy_tuple( &z, subj ) ;
  if (r != 0) {
    drop_tuple( z ) ;
    return ( r ) ;
  } /* error on copy */
  z->j = iss->j ;               /* ACL, issuer or threshold */
  store_thresh( &(z->th), iss->th ) ; /* give it the issuer's th */
  store_renlists( &(z->ik), iss->ik ) ; /* give it the issuer's ik */
  /* leave the sk alone */
  /* all the names are empty */
  /* the delegation and activity are those of the subj */
  store_sexp( &(z->t), tag_intersect( iss->t, subj->t ) ) ;
  store_bytes( &(z->nb), bs_max( iss->nb, subj->nb ) ) ;
  store_bytes( &(z->na), bs_min( iss->na, subj->na ) ) ;
  if (z->t != NULL)
    if (bs_compare( z->nb, z->na ) != GREATER)
      r = plant_tuple( z, NULL, 1 ) ;
  drop_tuple( z ) ;
  return ( r ) ;
} /* reduce_5tuple */
 
/***************************************************************************/
/* reduce_thresh_pair                                                      */
/*                                                                         */
/* Combine the two tuples, x and y, to form one reduced one.  If they      */
/* reduce, side by side, then emit the reduced tuple as the return value.  */
/***************************************************************************/

static pTuple reduce_thresh_pair( pTuple x, pTuple y )
{
  pTuple z = NULL ;             /* the result */
  int r = 0 ;                   /* return value */

  if ((x == NULL) || (y == NULL)) return ( NULL ) ; /* no reduction */
  r = copy_tuple( &z, x ) ;     /* get a working copy */
  if (r != 0) {
    drop_tuple( z ) ;           /* drop anything carved */
    return ( NULL ) ;           /* and return a failure */
  } /* if */
  store_sexp( &(z->t), tag_intersect( x->t, y->t ) ) ;
  store_bytes( &(z->nb), bs_max( x->nb, y->nb ) ) ;
  store_bytes( &(z->na), bs_min( x->na, y->na ) ) ;
  if ((r != 0) || (z->t == NULL)
      || (bs_compare( z->nb, z->na ) == GREATER) )
    store_tuple( &z, NULL ) ;   /* failure to reduce */
  return ( z ) ;
} /* reduce_thresh_pair */

/***************************************************************************/
/* is_thresh_match                                                         */
/*                                                                         */
/* Return 1 if there is an occurrence of sk in tr, 0 otherwise.            */
/***************************************************************************/

static int is_thresh_match( pTred tr, pRenlists sk )
{
  while (tr != NULL) {
    if (tr->s == sk) return ( 1 ) ;
    tr = tr->nxt ;              /* go to the next element */
  } /* while */
  return ( 0 ) ;                /* none matched */
} /* is_thresh_match */

/***************************************************************************/
/* count_thresh                                                            */
/*                                                                         */
/* Count the # of entries in tarr[] that have subjects sk.                 */
/***************************************************************************/

static int count_thresh( int n, pTred *tarr, pRenlists sk )
{
  int i ;                       /* loop index */
  int tot = 0 ;                 /* count of matches */

  for (i=0; i<=n; i++)
    if (is_thresh_match( tarr[i], sk ))
      tot++ ;
  return ( tot ) ;
} /* count_thresh */
 
/***************************************************************************/
/* try_thresh                                                              */
/*                                                                         */
/* Scan the array of candidates, tarr, skipping tarr[i] (since that's      */
/* forced in the call).  Whenever we find a match to tr->s, build a tuple  */
/* by reducing with what's accumulated in t and resume the process.        */
/* If (nleft + nsucc) > k, also just skip this row that has a match to     */
/* tr->s in it.                                                            */
/*                                                                         */
/* We will end up forming all the possible combinations of k subjects,     */
/* using tr->s as the first one.                                           */
/*                                                                         */
/* If any tuples survive that process, promote their issuers and           */
/* re-plant them.                                                          */
/***************************************************************************/

static int try_thresh( pTred tr, /* the new candidate */
                       int i,   /* its subject number */
                       pTred *tarr, /* array of other candidates */
                       int k,   /* number that have to match */
                       int n,   /* total array size */
                       int j,   /* last scan location */
                       int nsucc, /* # succeeded so far */
                       int nleft, /* # matches left to scan */
                       pTuple t /* accumulated tuple so far */
                     )
{
  pTuple nt = NULL ;            /* next tuple in this process */
  pTred tj ;                    /* walker of the list at tarr[j] */
  int r = 0 ;                   /* return value */

  if (nsucc == k) {             /* have a success.  Emit this tuple. */
    r = copy_tuple( &nt, t ) ;  /* get a copy to work on */
    if (r == 0) {
      pThresh th = t->th ;      /* get its threshold */
      nt->j = th->j ;           /* get its issuer specification */
      store_thresh( &(nt->th), NULL ) ; /* forget its thresh */
      if (nt->j > 0)
        store_thresh( &(nt->th), th->th ) ; /* get the new thresh */
      else
        store_renlists( &(nt->ik), th->ik ) ; /* get the new issuer */
      r = plant_tuple( nt, NULL, 1 ) ; /* plant the tuple, w/ one less layer */
                                /* of threshold.  Leave the active flag */
                                /* alone. */
    } /* if r */
    drop_tuple( nt ) ;
    return ( r ) ;
  } /* if nsucc */

  if (nsucc > k) return ( 0 ) ; /* succeeded already, can't continue */
 
  /** try_thresh ********** continued **/

  /* advance to the next tarr entry that has a matching subject */
  while ((j++) < n)
    if ((j != i) && is_thresh_match( tarr[j], tr->s ))
      break ;
  if (j > n) return ( 0 ) ;     /* all done */
  /* j is the next subject to include */
  tj = tarr[j] ;                /* walk that list */
  while (tj != NULL) {
    if (tj->s == tr->s) {
      pTuple nt = reduce_thresh_pair( t, tj->t ) ;
      if (nt != NULL) {
        r = try_thresh( tr, i, tarr, k, n, j, nsucc+1, nleft-1, nt ) ;
        drop_tuple( nt ) ;
        if (r != 0) return ( r ) ;
      } /* if nt */
    } /* if tj->s */
    tj = tj->nxt ;
  } /* while tj */
  /* Tried all of those.  Now, if there are enough left, try skipping */
  /* this subject and making reductions from the rest. */
  if ((nsucc + nleft) > k)
    r = try_thresh( tr, i, tarr, k, n, j, nsucc, nleft-1, t ) ;
  return ( r ) ;
} /* try_thresh */
 
/***************************************************************************/
/* plant_thresh                                                            */
/*                                                                         */
/* The tuple, t, has a key-only subject and is therefore a candidate for   */
/* threshold reduction.  Plant it (via a Tred) block, listed off its       */
/* threshold block.                                                        */
/*                                                                         */
/* Then, check to see if we can emit any reductions of this threshold.     */
/***************************************************************************/

static int plant_thresh( pTuple t )
{
  pTred tr ;                    /* Tred block for this one */
  pThresh th ;                  /* (walker - no drop) my threshold block */
  int i ;                       /* my index */
  int tot ;                     /* # of subjects matching mine */

  /* start with sanity checks */
  if (t == NULL) return ( 0 ) ; /* none here, no work to do */
  if (t->j < 1) return ( 0 ) ;  /* not a threshold user */
  if ((th = t->th) == NULL) return ( SPKI_ERR_THRESHOLD ) ;
  if (t->snam != NULL) return ( 0 ) ; /* not key-only */
  /* OK.  Try to gather enough others to reduce the threshold. */
  tr = new_tred() ;             /* carve a new Tred for t */
  if (tr == NULL) return ( SPKI_ERR_NO_MEM ) ;
  i = t->j ;                    /* my index into the threshold subject */
  if ((i < 1) || (i > th->n)) return ( SPKI_ERR_THRESHOLD ) ;
  tr->nxt = th->tr[i] ;         /* link this new one in */
  th->tr[i] = tr ;              /* ... */
  tr->th = th ;                 /* initialize it */
  tr->s = t->sk ;               /* t's subject */
  tr->t = t ;                   /* my tuple */
  /* This one is linked in. */

  /* Now, emit any reduced threshold blocks. */
  if ((tot = count_thresh( th->n, th->tr, tr->s )) < th->k )
    return ( 0 ) ;              /* no chance */
  return ( try_thresh( tr, i, th->tr, th->k, th->n, 0, 1, tot-1, t ) ) ;
} /* plant_thresh */
 
/***************************************************************************/
/* expand_thresholds                                                       */
/*                                                                         */
/* We have a threshold subject.                                            */
/*                                                                         */
/* Make one copy of this tuple for each of the n subjects, each with       */
/* a more basic subject and each referring to a Thresh block instead       */
/* of whatever issuer is in the current tuple.  Then, for each of those    */
/* generated tuples, call plant_tuple.                                     */
/***************************************************************************/

static int expand_thresholds( pTuple t, pSexp subj )
{
  pThresh th = NULL ;           /* the threshold block I'll need to make */
  pTuple a = NULL ;             /* the template tuple copy */
  pTuple b = NULL ;             /* a copy of a */
  short k;
  int  n ;                    /* the threshold values */
  short cnt = 0 ;                 /* count the actual number of subjects */
  int r = 0 ;                   /* the return value */
  pSexp s = NULL ;              /* (walker, no drop) to walk the k-of-n */
  pTred *tr = NULL ;            /* array of pTred lists */

  if (t == NULL) return ( SPKI_ERR_NULL_PARAM ) ;
  s = t->s ;                    /* a k-of-n subject */
  r = SPKI_ERR_KOFN_SYNTAX ;    /* default error until the end */
  if (pull_resword( s ) != O_SUBJECT) return ( r ) ;
  if (s->typ != S_KW) return ( r ) ;
  s = s->term ;
  if (s->typ != S_SEXP) return ( r ) ;
  s = s->u.s ;
  if (pull_resword( s ) != O_K_OF_N) return ( r ) ;
  s = s->term ;                 /* get the next term */
  if (pull_typ( s ) != S_BS) return ( r ) ;
  k = (short)int_from_bs( s->u.b ) ;   /* get the integer value for k */
  s = s->term ;                 /* and the next */
  if (pull_typ( s ) != S_BS) return ( r ) ;
  n = int_from_bs( s->u.b ) ;   /* get the integer value for n */
  if ((k < 1) || (n < k)) return ( r ) ;

  /* OK, start the work */
  r = 0 ;                       /* no error assumed any more */
  th = new_thresh() ;           /* get the threshold block */
  if (th == NULL) return ( SPKI_ERR_NO_MEM ) ;
  th->k = k ;                   /* initialze it */
  th->n = n ;
  th->j = t->j ;
  if (th->j == 0)
    store_renlists( &(th->ik), t->ik ) ;
  else if (th->j > 0)
    store_thresh( &(th->th), t->th ) ;
  /* initialize the pTred array */
  th->tr = tr = (pTred *) CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
                            sizeof( pTred ) * (n+1) ,
                            poolptr);
  if (tr != NULL)
    memset( tr, 0, sizeof( pTred ) * (n+1) ) ;
  else r = SPKI_ERR_NO_MEM ;
  /* Thresh block is initialized.  Now we can refer to it. */
 
  /** expand_thresholds ********** continued **/

  /* Make the tuple copy. */
  if (r == 0)
    r = copy_tuple( &a, t ) ;   /* make the copy */
  if (r == 0) {
    /* prepare this tuple to refer to th */
    if (a->j == 0)
      store_renlists( &(a->ik), NULL ) ;
    a->j = 1 ;                  /* actually, to be replaced by cnt */
    store_thresh( &(a->th), th ) ;
  } /* if no errors so far */

  /* The tuple, a, is now a prototype for all the ones we have to */
  /* generate and plant.  Do that in the loop below, iterating over */
  /* the subject S-expressions. */

  while ((r == 0) && (s = s->term) != NULL) { /* walk the whole list */
    if (pull_typ( s ) != S_SEXP) return ( SPKI_ERR_KOFN_SYNTAX ) ;
    cnt++ ;                     /* count the actual # of subjects */
    if (s->term != NULL)
      r = copy_tuple( &b, a ) ; /* make a temp copy */
    a->j = cnt ;                /* identify which subject we're using */
    store_sexp( &(a->s), s->u.s ) ; /* store the new subject away */
    r = pull_skn_cont( a, s->u.s ) ; /* plant that subject into the tuple, a */
    if (r != 0) break ;
    r = plant_tuple( a, subj, 0 ) ; /* and plant that new tuple */
    if (r != 0) break ;
    store_tuple( &a, b ) ;      /* get the temp copy back into a */
  } /* while */

  if (r == 0)
    if (cnt != n) r = SPKI_ERR_KOFN_SYNTAX ;

  drop_thresh( th ) ;           /* drop my use of this threshold block */
  drop_tuple( a ) ;             /* and the temp refs to tuples */
  drop_tuple( b ) ;             /* ... */
  return ( r ) ;                /* and return */
} /* expand_thresholds */
 
/***************************************************************************/
/* ndefreduce                                                              */
/*                                                                         */
/* Name definition reduction.                                              */
/* This tuple defines a name.  See if there are any subject names          */
/* waiting for it, and then plant it in the name definition hash table.    */
/***************************************************************************/

static int ndefreduce( pTuple t )
{
  pRen5 ren = NULL ;            /* (walker - no drop) block for this def */
  SHA_CTX si ;                 /* for hashing (k,n) */
  int wi ;                      /* walker index */
  pRen5 w = NULL ;              /* (walker - no drop) for scanning */
  int r = 0 ;                   /* return value */

  if (t == NULL) return ( 0 ) ; /* no tuple, no work */
  ren = new_ren5() ;            /* get my ren block */
  if (ren == NULL) return ( SPKI_ERR_NO_MEM ) ;
  SHA1_Init( &si ) ;              /* hash the key and name */
  SHA1_Update( &si, t->ik->hash, SHA_DIGEST_LENGTH ) ;
  mpl_bytes_hash( &si, NULL, t->inam ) ;
  SHA1_Final( ren->hash, &si ) ;
  store_tuple( &(ren->t), t ) ; /* note this tuple */
  wi = hindex( ren->hash ) ;    /* find my slot in the hash table */
  /* scan intab looking for the same definition */
  w = intab[ wi ] ;             /* walk this list */
  while (w != NULL) {
    if (memcmp( w->hash, ren->hash, SHA_DIGEST_LENGTH ) == 0)
      if (w->t->sk == ren->t->sk) {
        drop_ren5( ren ) ;      /* hasn't been linked in yet -- drop it */
        return ( 0 ) ;          /* all done -- has been defined */
      } /* definition already there? */
    w = w->nxt ;                /* scan the rest of the list */
  } /* while */
  /* plant this in the definition table first */
  ren->nxt = intab[ wi ] ;      /* link in the current list */
  intab[ wi ] = ren ;           /* and the new one */
                                /* intab inherits the refcnt */
  /* now, scan the tuples waiting to reduce subject names */
  w = sntab[ wi ] ;             /* walk this list */
  while ((w != NULL) && (r == 0)) {
    if (memcmp( w->hash, ren->hash, SHA_DIGEST_LENGTH ) == 0)
      r = reduce_name( ren->t, w->t ) ; /* (issuer, subject) */
    w = w->nxt ;                /* and try the next one in the list */
  } /* while */
  return ( r ) ;                /* and return */
} /* ndefreduce */
 
/***************************************************************************/
/* snreduce                                                                */
/*                                                                         */
/* Subject name reduction.                                                 */
/*                                                                         */
/* This tuple has a name subject.  See if there is a name definition with  */
/* which it can be reduced.  Then plant it in the subject name hash table. */
/***************************************************************************/

static int snreduce( pTuple t )
{
  pRen5 ren = NULL ;            /* (walker - no drop) block for this def */
  SHA_CTX si ;                 /* for hashing (k,n) */
  int wi ;                      /* walker index */
  pRen5 w = NULL ;              /* (walker - no drop) for scanning */
  int r = 0 ;                   /* return value */

  if (t == NULL) return ( 0 ) ; /* no tuple, no work */
  ren = new_ren5() ;            /* get my ren block */
  if (ren == NULL) return ( SPKI_ERR_NO_MEM ) ;
  SHA1_Init( &si ) ;              /* hash the key and name */
  SHA1_Update( &si, t->sk->hash, SHA_DIGEST_LENGTH ) ;
  mpl_bytes_hash( &si, NULL, t->snam ) ;
  SHA1_Final( ren->hash, &si ) ;
  store_tuple( &(ren->t), t ) ; /* note this tuple */
  wi = hindex( ren->hash ) ;    /* find my slot in the hash table */
  /* Scan for a complete match.  If found, don't enter this one. */
  w = sntab[ wi ] ;
  while (w != NULL) {
    if (memcmp( w->hash, ren->hash, SHA_DIGEST_LENGTH ) == 0)
      if (equal_tuple( w->t, ren->t ) ) {
        drop_ren5( ren ) ;      /* hasn't been linked in yet -- drop it */
        return ( 0 ) ;          /* all done -- has been defined */
      } /* definition already there? */
    w = w->nxt ;                /* scan the rest of the list */
  } /* while w */
  /* plant this in the definition table first */
  ren->nxt = sntab[ wi ] ;      /* link in the current list */
  sntab[ wi ] = ren ;           /* and the new one */
                                /* sntab inherits the refcnt */
  /* now, scan the tuples defining this name */
  w = intab[ wi ] ;             /* walk this list */
  while ((w != NULL) && (r == 0)) {
    if (memcmp( w->hash, ren->hash, SHA_DIGEST_LENGTH ) == 0)
      r = reduce_name( w->t, ren->t ) ;
    w = w->nxt ;                /* and try the next one in the list */
  } /* while */
  return ( r ) ;                /* and return */
} /* snreduce */
 
/***************************************************************************/
/* t5reduce                                                                */
/*                                                                         */
/* 5-tuple reduction.                                                      */
/*                                                                         */
/* This tuple is a key to key 5-tuple, ready for 5-tuple reduction,        */
/* assuming it's active.  If it is, try to reduce it.  In any case,        */
/* plant the reference to it under both the issuer and the subject keys.   */
/*                                                                         */
/* If this tuple is part of a threshold, try to reduce that threshold.     */
/*                                                                         */
/*   There are two parts to the 5-tuple reduction process:                 */
/*   1. the subject side                                                   */
/*   2. the issuer side                                                    */
/*                                                                         */
/*   On the subject side, we got here because the subject was a key.       */
/*   It might be active or inactive.  If it's active, try the reduction.   */
/*   If it's inactive, just plant it in the appropriate lists.             */
/*                                                                         */
/*   On the issuer side, there can be:                                     */
/*   a. a normal issuer                                                    */
/*   b. NULL, implying an ACL entry                                        */
/*   c. a threshold join node, implying that the tuple isn't full          */
/*      and needs threshold reduction before it can go anywhere else       */
/*                                                                         */
/***************************************************************************/

static int t5reduce( pTuple t )
{
  pRen5 w = NULL ;              /* (walker - no drop) */
  pRen5 *il = NULL ;            /* issuer list location */
  int r = 0 ;                   /* return value */

  if (t == NULL) return ( 0 ) ; /* no action */
  /* we got here because the subject and issuer were both just keys */
  /* We could be part of a threshold.  Check that first. */
  if (t->j > 0) {
    r = plant_thresh( t ) ;
    if (r != 0) return ( r ) ;
  } else {
    /* OK, it's not a threshold block.  Let's see if we're redundant. */
    if (t->ik == NULL)
      il = &acl_ren ;           /* walk the ACL rendezvous */
    else
      il = &(t->ik->i) ;        /* walk the issuer key rendezvous */
    w = *il ;
    while (w != NULL) {
      if (equal_tuple( t, w->t ))
        return ( 0 ) ;          /* no work to do.  This tuple is there. */
      w = w->nxt ;
    } /* while */
    /* OK, it's a new tuple for the mix. */
    /* Plant it and maybe try to reduce it. */
    w = new_ren5() ;            /* get a block for the issuer side */
    if (w == NULL) return ( SPKI_ERR_NO_MEM ) ;
    store_tuple( &(w->t), t ) ; /* store the tuple away */
    w->nxt = *il ;              /* link it in (and give up refcnt) */
    *il = w ;                   /* ... */
  } /* else -- not a threshold block */
  /* Plant it as a subject */
  w = new_ren5() ;              /* block for the subject side */
  if (w == NULL) return ( SPKI_ERR_NO_MEM ) ;
  store_tuple( &(w->t), t ) ;   /* store the tuple pointer */
  w->nxt = t->sk->s ;           /* in the subject list */
  t->sk->s = w ;                /* and the refcnt goes to that list */
  /* All planted.  Now, see if we should try reduction. */
  if (t->active)
    if (t->ik != NULL) {        /* if active and not ACL, reduce on right */
    /* Look for tuples whose subjects are the same as my issuer */
    /* and that are allowed to delegate. */
    w = t->ik->s ;              /* those whose subject is my issuer */
    while ((w != NULL) && (r == 0)) {
      if (w->t->d)
        r = reduce_5tuple( w->t, t ) ;
      w = w->nxt ;
    } /* while w */
  } /* if t->active */
  if (r != 0) return ( r ) ;

  if (t->d) {                   /* if can delegate, try reduce from the left */
    /* Look for tuples whose issuers are the same as my subject */
    /* and that are active. */
    w = t->sk->i ;              /* those whose issuer is my subject */
    while ((w != NULL) && (r == 0)) {
      if (w->t->active)
        r = reduce_5tuple( t, w->t ) ;
      w = w->nxt ;
    } /* while w */
  } /* if t->d */
  return ( r ) ;
} /* t5reduce */
 
/***************************************************************************/
/* plant_tuple                                                             */
/*                                                                         */
/* Enter the tuple, t, into the reduction process.                         */
/*                                                                         */
/* If this tuple still has a k-of-n subject, then split it first into      */
/* its constituent subjects.  That splitting could be recursive.           */
/***************************************************************************/

static int plant_tuple( pTuple t, pSexp subj, int no_subj )
{
  int r = 0 ;                   /* return value */
  pRenlists srl = NULL ;        /* subj renlist */

  if (subj != NULL) {           /* there's a subject constraint */
    r = kh_lookup( subj, &srl ) ; /* look up this subject */
    if (r != 0) {               /* if any error */
      drop_renlists( srl ) ;    /* drop the result */
      return ( r ) ;            /* and return the error */
    } /* if */
  } /* if subj */

  /* This tuple might still have a threshold subject.  In that case, */
  /* t->sk is NULL.  Otherwise, we can plant it directly. */
  if (t->sk == NULL)
    return ( expand_thresholds( t, subj ) ) ; /* expand threshold subjects */

  /* OK, it's a normal tuple (maybe part of a threshold).  Plant it. */
  /* First, classify it.  Test the subject first. */
  if (t->snam == NULL) {        /* subject is just a key */
    if ((srl == NULL) || (t->sk == srl))
      if (no_subj == 0)         /* if can't spec subject, leave active alone */
        t->active = 1 ;         /* if subject limitation, many tuples */
                                /* aren't active at first. */
    /* Is this a name definition or a 5-tuple? */
    if (t->inam != NULL)
      r = ndefreduce( t ) ;     /* this defines a name, now or later */
    else
      r = t5reduce( t ) ;       /* reduce this as a 5-tuple */
  } else {                      /* subject is a name */
    /* The only operation for this tuple is to reduce the subject. */
    /* Until the subject is just a key, the tuple is otherwise inactive. */
    r = snreduce( t ) ;         /* reduce the subject name */
  } /* else */

  return ( r ) ;
} /* plant_tuple */
 
/***************************************************************************/
/* reduce                                                                  */
/*                                                                         */
/* Walk an S-expression (e.g., an (acl ...) or (sequence ...)) and create  */
/* tuples from their elements.  Enter those tuples in the appropriate      */
/* hash tables, to rendezvous with other tuples, for name reduction,       */
/* 5-tuple reduction or threshold reduction.                               */
/*                                                                         */
/*  x is the S-expression being walked (known to be S_KW).                 */
/*  subj is a subject limitation.  If it is NULL, then all tuples are      */
/*      active.  If it is not NULL, then only tuples matching that subj    */
/*      are active -- and their active state propagates up a chain         */
/*      to an ACL entry.                                                   */
/*  tag is a tag limitation.  If it is non-NULL, then every tuple is       */
/*      intersected with the tag before being entered in the process.      */
/*      If that intersection is NULL, then the tuple is not entered.       */
/*  now is a date&time limitation.  If it is non-NULL, then every tuple    */
/*      has its validity date intersected with now.  If now is not         */
/*      within the dates, then the tuple doesn't enter the process.        */
/*      Otherwise, the tuple's dates are both set to now.                  */
/*                                                                         */
/***************************************************************************/

static int reduce( pSexp x,     /* the entry or cert */
                   pSexp subj,  /* a subject limitation */
                   pSexp tag,   /* a tag limitation */
                   pBytes now ) /* the current date&time */
{
  pTuple t = new_tuple() ;      /* the tuple being made here (NEEDS DROP) */
  char is_acl ;                 /* boolean: is this an ACL entry? */
  int r = 0 ;                   /* return value */

  if (x == NULL) return ( SPKI_ERR_NULL_PARAM ) ;
  if (t == NULL) return ( SPKI_ERR_NO_MEM ) ;
  is_acl = (char)(x->u.k == O_ENTRY) ;
  if (is_acl) t->j = -1 ;       /* note that this is an ACL entry */

  r = splay_fields( t, x, is_acl ) ; /* fill in the tuple */

  if (r != 0) {
    drop_tuple( t ) ;
    return ( r ) ;
  } /* if */

  /* Have the tuple fields extracted. */

  /* First, check dates. */
  if (now != NULL) {            /* intersect them */
    coerce_dates( &t, now ) ;   /* reset the tuple dates */
    if (t == NULL) return ( 0 ) ; /* if tuple was invalid, give up */
  } /* if -- dates */

  /* The tuple has survived any limitation in validity dates. */
  /* Now, intersect the tag, if any. */
  if (t->t != NULL) {           /* if there is a tag in the tuple */
    if (tag != NULL) {          /* and there is a tag limitation */
      store_sexp( &(t->t), tag_intersect( t->t, tag ) ) ;
      if (t->t == NULL) {       /* was there any tag left over? */
        drop_tuple( t ) ;       /* no, so we're done here */
        return ( 0 ) ;
      } /* if */
    } /* if */
  } else                        /* no tag in the tuple */
    if (t->inam == NULL)        /* if not a name definition */
      store_sexp( &(t->t), tag ) ; /* take the default tag */

  /* Now, extract issuer and subject keys and names, to let us decide */
  /* what kind of reduction this tuple needs. */
  if (!is_acl) {                /* then we need an issuer */
    if (t->ik == NULL) {        /* if there is none, it's an error */
      drop_tuple( t ) ;         /* no issuer, but not an ACL */
      return ( SPKI_ERR_NO_ISSUER ) ; /* error return */
    } /* if */
  } /* if */

  /* check for the illegal name+auth cert case. */
  if ((t->inam != NULL) && (t->t != NULL)) {
    drop_tuple( t ) ;
    return ( SPKI_ERR_MIXED_NAME_AND_AUTH ) ;
  } /* if */

  r = plant_tuple( t, subj, 0 ) ; /* enter it into the process */
  drop_tuple( t ) ;             /* drop my temp copy */
  return ( r ) ;                /* and return */
} /* reduce */
 
/***************************************************************************/
/* walk_reduce                                                             */
/*                                                                         */
/* Walk an S-expression (e.g., an (acl ...) or (sequence ...)) and create  */
/* tuples from their elements.  Enter those tuples in the appropriate      */
/* hash tables, to rendezvous with other tuples, for name reduction,       */
/* 5-tuple reduction or threshold reduction.                               */
/*                                                                         */
/*  src is the S-expression being walked.                                  */
/*  subj is a subject limitation.  If it is NULL, then all tuples are      */
/*      active.  If it is not NULL, then only tuples matching that subj    */
/*      are active -- and their active state propagates up a chain         */
/*      to an ACL entry.                                                   */
/*  tag is a tag limitation.  If it is non-NULL, then every tuple is       */
/*      intersected with the tag before being entered in the process.      */
/*      If that intersection is NULL, then the tuple is not entered.       */
/*  now is a date&time limitation.  If it is non-NULL, then every tuple    */
/*      has its validity date intersected with now.  If now is not         */
/*      within the dates, then the tuple doesn't enter the process.        */
/*      Otherwise, the tuple's dates are both set to now.                  */
/*                                                                         */
/***************************************************************************/

static int walk_reduce( pSexp src, /* the source of tuples */
                        pSexp subj, /* a subject limitation */
                        pSexp tag, /* a tag limitation */
                        pBytes now ) /* the current date&time */
{
  pSexp w = src ;               /* walker of the S-expression */
                                /* don't drop this reference */
  int r = 0 ;                   /* return value */

  if (w == NULL)
    return ( 0 ) ;              /* can't walk, but not an error */

  while ((w != NULL) && (r == 0)) {
    if (w->typ == S_SEXP) {     /* look only at sub-expressions */
      switch ( pull_resword( w ) ) {
      case O_ACL: case O_SEQUENCE:
        r = walk_reduce( w->u.s, subj, tag, now ) ;
        break ;
      case O_ENTRY:             /* an ACL entry */
      case O_CERT:              /* a certificate style credential */
        r = reduce( w->u.s, subj, tag, now ) ; /* reduce it */
      default:                  /* otherwise, skip it */
          ;
      } /* switch */
    } /* if */
    w = w->term ;               /* go to the next term in the expression */
  } /* while */
  return ( r ) ;
} /* walk_reduce */
 
/***************************************************************************/
/* append_result                                                           */
/*                                                                         */
/* Have one tuple to report back.  Build an ENTRY for it and append it     */
/* to tail.                                                                */
/***************************************************************************/

static int append_result( pSexp *tail, pTuple t )
{
  pSexp a = NULL ;              /* temps for building the result */
  pSexp b = NULL ;              /* ... */
  pSexp w = NULL ;              /* for walking a */
  pSexp x = NULL ;              /* for walking w */
  pSexp y = NULL ;              /* for walking x */
  char *template =              /* template for the returned item */
"(5:entry(7:subject(10:public-key))(9:propagate)(3:tag)\
(5:valid(10:not_before2:nb)(9:not_after2:na)))" ;
  int r = 0 ;                   /* return value */

  r = lit_sexp( &a, template ) ; /* parse the S-expression */
  if (r == 0) {
    b = new_sexp() ;            /* for the sub-expression node */
    if (b == NULL) r = SPKI_ERR_NO_MEM ;
  } /* if */
  if (r != 0) {
    drop_sexp( a ) ;
    /* b must be NULL */
    return ( r ) ;
  } /* if */
  /* OK, have b & have the template parsed.  Fill it in with specifics. */
  w = a->term ;                 /* go to the first term */
  x = w->u.s ;                  /* the subject node */
  x = x->term ;                 /* the sub-expression for (public-key) */
  store_sexp( &(x->u.s), t->sk->obj ) ; /* the real subject expression */
  /* is delegation set? */
  if (! t->d)                   /* if not, drop the node in the output */
    store_sexp( &(w->term), w->term->term ) ;
  else
    w = w->term ;               /* otherwise walk to it */
  w = w->term ;                 /* go to the tag term */
  store_sexp( &(w->u.s), t->t ) ; /* store the tag away */
  w = w->term ;                 /* go to the valid tag */
  x = w->u.s ;                  /* and go into it */
  if (t->nb != NULL) {
    x = x->term ;               /* go to the not-before sub-expression */
    y = x->u.s ;                /* and go into it */
    y = y->term ;               /* and go to the byte string */
    store_bytes( &(y->u.b), t->nb ) ; /* and save the result */
  } else                        /* none, so drop that node */
    store_sexp( &(x->term), x->term->term ) ;
  if (t->na != NULL) {
    x = x->term ;               /* go to the not-after sub-expression */
    y = x->u.s ;                /* and go into it */
    y = y->term ;               /* and go to the byte string */
    store_bytes( &(y->u.b), t->na ) ; /* and save the datime */
  } else                        /* none, so drop that node */
    store_sexp( &(x->term), x->term->term ) ;
 
  /** append_result  ********  continued **/

  /* Link the result in to the output */
  b->typ = S_SEXP ;             /* sub-expression */
  b->u.s = a ;                  /* and that's the expression */
  (*tail)->term = b ;           /* linked */
  *tail = b ;                   /* and now the new tail */

  /* a and b refcnts have been handed to b and tail, respectively. */

  return ( r ) ;                /* return */
} /* append_result */
 
/***************************************************************************/
/* report_results                                                          */
/*                                                                         */
/* Scan acl_ren, where all ACL entries that result from the reduction      */
/* process live.  For all of those that are marked active build an         */
/* output ACL in aout.                                                     */
/***************************************************************************/

static int report_results( pSexp *aout )
{
  pSexp res = NULL ;            /* current results block */
  pSexp tail = NULL ;           /* (walker - no drop) tail of current list */
  pRen5 w = NULL ;              /* (walker) to scan acl_ren */
  pTuple t = NULL ;             /* (walker) current tuple */
  int r = 0 ;                   /* return value */

  /* construct an ACL S-expression */
  res = new_sexp() ;            /* get a head block for this expression  */
  if (res == NULL) return ( SPKI_ERR_NO_MEM ) ;
  res->typ = S_KW ;
  res->u.k = O_ACL ;
  *aout = res ;                 /* let aout inherit the refcnt */
  res = NULL ;                  /* ... */
  tail = *aout ;                /* walker -- no refcnt */
  /* scan acl_ren, for results */
  w = acl_ren ;                 /* walker - no refcnt */
  while ((w != NULL) && (r == 0)) {
    t = w->t ;                  /* tuple */
    if (t != NULL)
      if (t->active)
        r = append_result( &tail, t ) ; /* something to output */
    w = w->nxt ;
  } /* while */
  return ( r ) ;
} /* report_results */
 
/***************************************************************************/
/* do_reduction                                                            */
/*                                                                         */
/* Reduce a set of credentials.                                            */
/*                                                                         */
/*  Note: this code will edit the parameters, replacing some S-expressions */
/*  with equivalent expressions.  (Some hashes become public-keys and      */
/*  some hashes or keys stay the same, but defined in a common block,      */
/*  so that content equality can be tested by testing pointer equality.)   */
/*                                                                         */
/*                                                                         */
/* Input is:                                                               */
/* acl: the current ACL (acl ...)                                          */
/* cred: certificate bodies and public keys  (sequence ...)                */
/* lim: test limitations                                                   */
/*    subject: a single subject, presumably the one asking for service     */
/*    tag: the thing that subject is asking to do                          */
/*    now: the current date and time                                       */
/* Any or all of the limitations can be missing, in case we want general   */
/*    results rather than a specific answer to one request.                */
/*                                                                         */
/* The result of this process is a list of S-expressions, possibly         */
/* empty.  These will look like an ACL.  Tuples make it into the result    */
/* list only if they  are ACL-type (have no issuer), are active and have   */
/* the desired subj (if any) as their subject.                             */
/*                                                                         */
/***************************************************************************/

int do_reduction( pSexp acl_p,  /* the current ACL */
                  pSexp cred_p, /* the credentials being provided */
                  pSexp subj_p, /* limitation: a subject */
                  pSexp tag_p,  /* limitation: a tag */
                  pBytes now_p, /* limitation: the current date&time */
                  pSexp *aout ) /* output results */
{
  pSexp acl = NULL ;            /* local copy of input params */
  pSexp cred = NULL ;
  pSexp subj = NULL ;           /* all of these need to be dropped on exit */
  pSexp tag = NULL ;
  pBytes now = NULL ;           /* copies to prevent drop mistakes */
  int r = 0 ;                   /* return value */

  hashtab_init() ;              /* init the hash tables */

  if (aout == NULL) return ( SPKI_ERR_NULL_PARAM ) ;
  store_sexp( aout, NULL ) ;    /* no results yet */

  if (acl_p == NULL)
    return ( SPKI_ERR_NO_ACL ) ; /* can't do anything without an ACL */
 
  /* make local copies of the inputs -- each at the S_KW point */
  if (pull_typ( acl_p ) == S_SEXP) store_sexp( &acl, acl_p->u.s ) ;
  else store_sexp( &acl, acl_p ) ;
  if (pull_typ( cred_p ) == S_SEXP) store_sexp( &cred, cred_p->u.s ) ;
  else store_sexp( &cred, cred_p ) ;
  if (pull_typ( subj_p ) == S_SEXP) store_sexp( &subj, subj_p->u.s ) ;
  else store_sexp( &subj, subj_p ) ;
  if (pull_typ( tag_p ) == S_SEXP) store_sexp( &tag, tag_p->u.s ) ;
  else store_sexp( &tag, tag_p ) ;
  store_bytes( &now, now_p ) ;  /* copies made */

  /* 0. Note all defined public keys. */
  pass0( subj ) ;               /* do the subject first */
  pass0( acl ) ;                /* then the ACL */
  pass0( cred ) ;               /* then the issued credentials */

  /* 1. Walk the ACL, entering things in the appropriate hash list. */
  r = walk_reduce( acl, subj, tag, now ) ; /* walk the ACL structure */
                                /* and enter */
                                /* items from it into the appropriate hash */
                                /* tables, for reduction. */

  /* 2. Walk the credentials, entering those. */
  if (r == 0)                   /* if no errors so far */
    r = walk_reduce( cred, subj, tag, now ) ; /* walk credentials */
                                /* and enter those into the hash tables. */

  /* 3. Gather results into aout. */
  if (r == 0)                   /* if no errors in the process */
    r = report_results( aout ) ; /* get the results into aout */

  drop_sexp( acl ) ;            /* drop my copies of the inputs */
  drop_sexp( cred ) ;
  drop_sexp( subj ) ;
  drop_sexp( tag ) ;
  drop_bytes( now ) ;           /* copies dropped */

  hashtab_clear() ;             /* empty the hash tables */
                                /* leaving only those sexp elements */
                                /* reported into aout or in the inputs */
  return ( r ) ; 
} /* do_reduction */
 
/***************************************************************************/
/* clpass0                                                                 */
/*                                                                         */
/* Scan one CSSM_LIST looking for public keys to record in pktab.          */
/***************************************************************************/

static void clpass0( const CSSM_LIST *subj )
{
  pSexp s = cl_to_sexp( subj ) ; /* get the pSexp version */
  pass0( s ) ;                  /* scan it for public keys */
  drop_sexp( s ) ;              /* free the temporary expression */
  return ;                      /* and return */
} /* clpass0 */

/***************************************************************************/
/* lpass0                                                                  */
/*                                                                         */
/* Scan a list of subjects looking for public keys to record in pktab.     */
/***************************************************************************/

void lpass0( uint32 ns, const CSSM_LIST subj[] )
{
  uint32 i ;                    /* loop variable */
  for (i=0; i<ns; i++)
    clpass0( &(subj[i]) ) ;     /* scan each one */
  return ;                      /* and return */
} /* lpass0 */

/***************************************************************************/
/* tgpass0                                                                 */
/*                                                                         */
/* Scan a tuplegroup, looking for public keys to record in pktab.          */
/***************************************************************************/

void tgpass0( const CSSM_TUPLEGROUP *tg )
{
  uint32 i ;                    /* loop variable */
  CSSM_TUPLE *t = NULL ;        /* pointer to one tuple */

  for (i=0; i<tg->NumberOfTuples; i++) {
    t = &(tg->Tuples[i]) ;      /* point to the current tuple */
    clpass0( &(t->Issuer) ) ;   /* scan the issuer */
    clpass0( &(t->Subject) ) ;  /* and the subject */
  } /* for i */
} /* tgpass0 */

/***************************************************************************/
/* lactivate                                                               */
/*                                                                         */
/* Activate a subject list.  If there's none, then no action.              */
/* If there's just one, then use it as the active subject.                 */
/* If there's more, that could be because of a threshold subject to        */
/* biometrics.  So, create a dummy subject, (requestor), and tuples        */
/* delegating to it from each of the subjects.  Activate all of those.     */
/***************************************************************************/

void lactivate( uint32 ns, const CSSM_LIST subj[] )
{
  pRenlists rrl = NULL ;        /* requestor Renlists */
  pSexp req = NULL ;            /* requestor Sexp */
  pSexp tag = NULL ;            /* requestor's tag */
  uint32 i ;                    /* index */
  int r = 0 ;                   /* error return */

  if (ns == 0) return ;         /* no subjects, no work */
  /* create the dummy subject */
  r = lit_sexp( &req, "(9:requestor)" ) ; /* the dummy "subject" */
  if (r == 0)
    r = lit_sexp( &tag, "(3:tag(1:*))" ) ;
  rrl = new_renlists() ;        /* and a Renlists for it */
  if ((r != 0)||(rrl == NULL)) { /* any errors? */
    drop_sexp( req ) ;
    drop_sexp( tag ) ;
    drop_renlists( rrl ) ;
    return ;
  } /* if */
  store_sexp( &(rrl->obj), req ) ; /* note the subject */
  /* this doesn't need a hash -- leave it 0 */
  for (i=0; i<ns; i++) {
    pSexp sbj = cl_to_sexp( &(subj[i]) ) ; /* get the subject */

    if (sbj != NULL) {
      pRenlists srl = NULL ;    /* subject Renlists */
      int r = 0 ;               /* return value */

      r = subj_lookup( sbj, &srl ) ; /* look up this subject */

      if (r == 0) {
        pTuple t = new_tuple() ; /* working tuple (for activation) */

        if (t != NULL) {        /* fill in t and plant it */
          t->active = 1 ;       /* the purpose of tuple is to set activity */
          t->d = 0 ;            /* but it can't delegate */
          t->j = 0 ;            /* normal single issuer */
          t->ik = srl ;         /* take over its refcnt */
          srl = NULL ;          /* ... */
          store_renlists( &(t->sk), rrl ) ; /* the invented subject */
          store_sexp( &(t->s), req ) ; /* record the subject expression */
          store_sexp( &(t->t), tag ) ; /* record the tag */
          plant_tuple( t, NULL, 1 ) ; /* with no change in activity */
        } /* if t */
        store_tuple( &t, NULL ) ; /* drop my reference to it */
      } /* if r */
    } /* if sbj */
  } /* for i */
  drop_sexp( req ) ;
  drop_sexp( tag ) ;
  drop_renlists( rrl ) ;
  return ;
} /* lactivate */
 
/***************************************************************************/
/* tg_reduce1                                                              */
/*                                                                         */
/* Walk the tuple group array, making Tuple entries for each and entering  */
/* them into the reduction process.                                        */
/*                                                                         */
/* This is a combination of walk_reduce and splay_fields.                  */
/***************************************************************************/

CSSM_RETURN tg_reduce1( const CSSM_TUPLEGROUP *tg, /* the tuple group */
                        int active, /* BOOL:  activity of tuples */
                        pSexp sbj, /* the active subject */
                        pSexp tag, /* the tag constraint */
                        pSexp now /* the date constraint */
                        )
{
  uint32 lv ;                    /* loop variable */
  int r = 0 ;                   /* return value */

  if (tg == NULL) return ( CSSMERR_AC_INVALID_INPUT_POINTER ) ;
  for (lv=0; ((lv<tg->NumberOfTuples)&&(r == 0)); lv++) {
    CSSM_TUPLE *tin = &(tg->Tuples[lv]) ; /* current tuple */
    pSexp i = cl_to_sexp( &(tin->Issuer) ) ;
    pSexp s = cl_to_sexp( &(tin->Subject) ) ;
    int d = tin->Delegate ;     /* the delegation permission flag */
    pSexp t = cl_to_sexp( &(tin->AuthorizationTag) ) ;
    pSexp v = cl_to_sexp( &(tin->ValidityPeriod) ) ;
    pTuple ot = new_tuple() ;   /* the new Tuple I'm building */

    if (ot == NULL)
      r = CSSMERR_AC_MEMORY_ERROR ;
    else
      if (pull_resword(s) == O_SUBJECT) {
        /* have all the tuple fields .. make a Tuple and reduce it */
        ot->active = (UC)active ;       /* the tuple activity */
        ot->d = (UC)d ;                 /* the delegate flag */
        if (i != NULL)
          r = pull_ikn( ot, i ) ;   /* pull the issuer fields */
        if (r == 0)
          r = pull_skn( ot, s ) ;   /* pull the subject fields */
        store_sexp( &(ot->s), s ) ; /* and save the raw subject */
        if (t == NULL) {
          if (ot->inam == NULL)
            store_sexp( &(ot->t), tag ) ; /* if no tag & no name, use this */
        } else 
          if (tag != NULL) { /* t is not NULL */
            store_sexp( &(ot->t), tag_intersect( t, tag ) ) ;
            if (ot->t == NULL) {  /* is there any tag left over? */
              drop_tuple( ot ) ;  /* no, so we're done */
              drop_sexp( i ) ;    /* clean up my temp pointers */
              drop_sexp( s ) ;
              drop_sexp( t ) ;
              drop_sexp( v ) ;
              continue ;          /* and resume on this loop */
            }
          } else
            store_sexp( &(ot->t), t ) ; /* give it the TUPLE's tag */
        store_dates( ot, v ) ;      /* and the dates */
        if ((r == 0) && (now != NULL))
          r = intersect_dates( ot, now ) ; /* and intersect with these */

        /* Tuple is initialized.  Reduce it. */
        if (r == 0)
          if ((ot->inam != NULL) && (ot->t != NULL))
            r = SPKI_ERR_MIXED_NAME_AND_AUTH ;

        if (r == 0)
          r = plant_tuple( ot, sbj, (sbj==NULL) ) ; /* activate sbj, if non-NULL */
        r = 0 ;              /* clear the error indication, now that tuple skipped */
      } /* else, if */
    drop_tuple( ot ) ;          /* drop my reference to the Tuple */
    drop_sexp( i ) ;            /* clean up my temp pointers */
    drop_sexp( s ) ;
    drop_sexp( t ) ;
    drop_sexp( v ) ;
  } /* for lv */
  return ( r ) ;                /* and return */
} /* tg_reduce1 */
 
/***************************************************************************/
/* tg_report_results                                                       */
/*                                                                         */
/* Move the results into the output tuple group, aout.                     */
/* We have to carve the array for that output group here.                  */
/***************************************************************************/

CSSM_RETURN tg_report_results( CSSM_AC_HANDLE ach,      /* .. */
                               CSSM_TUPLEGROUP_PTR aout ) /* output param */
{
  pRen5 w = acl_ren ;           /* (walker) of acl_ren */
  int nout = 0 ;                /* # of output values */
  int r = 0 ;                   /* return value */

  while (w != NULL) {           /* count the successes */
    pTuple t = w->t ;           /* the tuple in this list */

    if (t != NULL)
      if (t->active)
        nout++ ;                /* got one */

    w = w->nxt ;                /* the next result */
  } /* while w */

  aout->NumberOfTuples = nout ; /* got it */
  aout->Tuples = (CSSM_TUPLE_PTR) CSSM_MALLOC_WRAPPER(memory_pool_funcs.cssm_memory_funcs.malloc_func,
                                              sizeof( CSSM_TUPLE ) * nout,
                                              poolptr) ;
  if (aout->Tuples == NULL)
    return ( CSSMERR_AC_MEMORY_ERROR ) ;
  memset( aout->Tuples, 0, sizeof( CSSM_TUPLE ) * nout ) ;

  nout = 0 ;                    /* use it as an index now */
  w = acl_ren ;                 /* re-walk the list to fill in the array */
  while ((w != NULL) && (r == 0)) { /* write the output array */
    pTuple t = w->t ;           /* the tuple in this list */

    if (t != NULL)
      if (t->active) {
        /* here's one result */
        CSSM_TUPLE *ot = &(aout->Tuples[ nout++ ]) ; /* the output tuple */

        /* fill it in */
        /* Issuer is NULL -- since this is an ACL entry */
        if (r == 0)
          r = subj_to_cl( ach, &(ot->Subject), t->sk->obj ) ;
        ot->Delegate = t->d ;   /* delegation flag */
        if (r == 0)
          r = sexp_to_cl( ach, &(ot->AuthorizationTag), t->t ) ;
        if (r == 0)
          r = dates_to_cl( ach, &(ot->ValidityPeriod), t->nb, t->na ) ;
      } /* if */
    w = w->nxt ;                /* the next result */
  } /* while w */
  return ( r ) ;                /* return */
} /* tg_report_results */
