/*-----------------------------------------------------------------------
 * File: CCAPI.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#include "precomp.h"


/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_CreateSignatureContext
 *
 *Description:
 *   This function creates a signature cryptographic context for sign and verify given a 
 *   handle of a CSP, an algorithm identification number, a key, and a passphrase structure
 *
 *Parameters: 
 *   CSPHandle (input)         - handle that describes the add-in cryptographic service 
 *                               provider module used to perform this function
 *   AlgorithmID (input)       - algorithm identification number for a signature/verification 
 *                               algorithm
 *   AccessCred (input)        - pointer to the set of one or more credentials required to 
 *                               unlock the private key
 *   Key (input)               - key used to sign
 *   NewContextHandle (output) - cryptographic context handle 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreateSignatureContext(CSSM_CSP_HANDLE CSPHandle,
                                                    CSSM_ALGORITHMS AlgorithmID,
                                                    const CSSM_ACCESS_CREDENTIALS *AccessCred,
                                                    const CSSM_KEY *Key,
                                                    CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CONTEXT_ATTRIBUTE_PTR Attribute = NULL;
    uint32 NumberAttributes = 0;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreateSignatureContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the application here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_SIGNATURE, AlgorithmID, &Context)) == CSSM_OK)
                {
                    if (!port_IsBadReadPtr(AccessCred, (uint32)sizeof(CSSM_ACCESS_CREDENTIALS)))
                        NumberAttributes++;

                    if (!cssm_IsBadCryptoKeyPtr (Key))
                        NumberAttributes++;


                    Context->NumberOfAttributes = NumberAttributes;

                    if ((retValue = cssm_CreateAttribute (&Context->ContextAttributes, NumberAttributes)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    Attribute = Context->ContextAttributes;
                    NumberAttributes--;


                    if (!port_IsBadReadPtr(AccessCred, (uint32)sizeof(CSSM_ACCESS_CREDENTIALS)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_ACCESS_CREDENTIALS, (void *)AccessCred)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (!cssm_IsBadCryptoKeyPtr (Key))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_KEY, (void *)Key)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
                }
            }
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_CreateSymmetricContext
 *
 *Description:
 *   This function creates a symmetric encryption cryptographic context given a handle of a 
 *   CSP, an algorithm identification number, a key, an initial vector, padding, and the 
 *   number of encryption rounds
 *
 *Parameters: 
 *   CSPHandle (input)           - handle that describes the add-in cryptographic service 
 *                                 provider module used to perform this function
 *   AlgorithmID (input)         - algorithm identification number for symmetric encryption
 *   Mode (input)                - mode of the specified algorithm ID
 *   AccessCred (input)          - pointer to the set of one or more credentials required to 
 *                                 unlock the private key
 *   Key (input)                 - key used for symmetric encryption
 *   InitVector (input/optional) - initial vector for symmetric encryption; typically 
 *                                 specified for block ciphers
 *   Padding (input/optional)    - method for padding; typically specified for ciphers 
 *                                 that pad
 *   Reserved (input)            - reserved for future use
 *   NewContextHandle (output)   - cryptographic context handle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreateSymmetricContext(CSSM_CSP_HANDLE CSPHandle,
                                                    CSSM_ALGORITHMS AlgorithmID,
                                                    CSSM_ENCRYPT_MODE Mode,
                                                    const CSSM_ACCESS_CREDENTIALS *AccessCred,
                                                    const CSSM_KEY *Key,
                                                    const CSSM_DATA *InitVector,
                                                    CSSM_PADDING Padding,
                                                    void *Reserved,
                                                    CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CONTEXT_ATTRIBUTE_PTR Attribute = NULL;
    uint32 NumberAttributes = 0;
    CSSM_RETURN retValue = CSSM_OK;


    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreateSymmetricContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the application here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {
                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_SYMMETRIC, AlgorithmID, &Context)) == CSSM_OK)
                {

                    if (!cssm_IsBadCryptoKeyPtr (Key))
                        NumberAttributes++;

                    if (!port_IsBadReadPtr (InitVector, (uint32)sizeof (CSSM_DATA)))
                        NumberAttributes++;

                    if (!port_IsBadReadPtr (AccessCred, (uint32)sizeof (CSSM_ACCESS_CREDENTIALS)))
                        NumberAttributes++;

                    /* for Mode*/
                    NumberAttributes++;
                    /* for padding info */
                    NumberAttributes++;

                    Context->NumberOfAttributes = NumberAttributes;

                    if ((retValue = cssm_CreateAttribute (&Context->ContextAttributes, Context->NumberOfAttributes)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    Attribute = Context->ContextAttributes;
                    NumberAttributes--;


                    if (!cssm_IsBadCryptoKeyPtr (Key))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_KEY, (void *)Key)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (!port_IsBadReadPtr (InitVector, (uint32)sizeof (CSSM_DATA)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_INIT_VECTOR, (void *)InitVector)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_PADDING, (void *)&Padding)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    if (!port_IsBadReadPtr(AccessCred, (uint32)sizeof(CSSM_ACCESS_CREDENTIALS)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_ACCESS_CREDENTIALS, (void *)AccessCred)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_MODE, (void *)&Mode)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                }
            }
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_CreateDigestContext
 *
 *Description:
 *   This function creates a digest cryptographic context, given a handle of a CSP and an 
 *   algorithm identification number
 *
 *Parameters: 
 *   CSPHandle (input)         - handle that describes the add-in cryptographic service 
 *                               provider module used to perform this function
 *   AlgorithmID (input)       - algorithm identification number for message digests   
 *   NewContextHandle (output) - cryptographic context handle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition 
 *  The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreateDigestContext(CSSM_CSP_HANDLE CSPHandle,
                                                 CSSM_ALGORITHMS AlgorithmID,
                                                 CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreateDigestContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_DIGEST, AlgorithmID, &Context)) == CSSM_OK)
                {

                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
                }
            }
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_CreateMacContext
 *
 *Description:
 *   This function creates a message authentication code cryptographic context, given a 
 *   handle of a CSP, algorithm identification number, key, and the length of the key in bits
 *
 *Parameters: 
 *   CSPHandle (input)         - handle that describes the add-in cryptographic service 
 *                               provider module used to perform this function
 *   AlgorithmID (input)       - algorithm identification number for the MAC algorithm
 *   Key (input)               - key used to generate a message authentication code
 *   NewContextHandle (output) - cryptographic context handle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreateMacContext(CSSM_CSP_HANDLE CSPHandle,
                                              CSSM_ALGORITHMS AlgorithmID,
                                              const CSSM_KEY *Key,
                                              CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CONTEXT_ATTRIBUTE_PTR Attribute = NULL;
    uint32 NumberAttributes = 0;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreateMacContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_MAC, AlgorithmID, &Context)) == CSSM_OK)
                {
                    if (!cssm_IsBadCryptoKeyPtr (Key))
                        NumberAttributes++;


                    Context->NumberOfAttributes = NumberAttributes;

                    if ((retValue = cssm_CreateAttribute (&Context->ContextAttributes, NumberAttributes)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    Attribute = Context->ContextAttributes;
                    NumberAttributes--;


                    if (!cssm_IsBadCryptoKeyPtr (Key))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_KEY, (void *)Key)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
                }
            }
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_CreateRandomGenContext
 *
 *Description:
 *   This function creates a random number generation cryptographic context, given a handle 
 *   of a CSP, an algorithm identification number, a seed, and the length of the random 
 *   number in bytes
 *
 *Parameters: 
 *   CSPHandle (input)         - handle that describes the add-in cryptographic service 
 *                               provider module used to perform this function
 *   AlgorithmID (input)       - algorithm identification number for random number generation
 *   Seed (input/optional)     - seed used to generate random number
 *   Length (input)            - length of the random number to be generated 
 *   NewContextHandle (output) - cryptographic context handle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition 
 *  The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreateRandomGenContext(CSSM_CSP_HANDLE CSPHandle,
                                                    CSSM_ALGORITHMS AlgorithmID,
                                                    const CSSM_CRYPTO_DATA *Seed,
                                                    uint32 Length,
                                                    CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CONTEXT_ATTRIBUTE_PTR Attribute = NULL;
    uint32 NumberAttributes = 0;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreateRandomGenContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_RANDOMGEN, AlgorithmID, &Context)) == CSSM_OK)
                {
                    if (!cssm_IsBadCryptoDataPtr (Seed))
                        NumberAttributes++;

                    /* For length attribute */
                    NumberAttributes++;
                    Context->NumberOfAttributes = NumberAttributes;

                    if ((retValue = cssm_CreateAttribute (&Context->ContextAttributes, NumberAttributes)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    Attribute = Context->ContextAttributes;
                    NumberAttributes--;


                    if (!cssm_IsBadCryptoDataPtr (Seed))
                    {
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_SEED, (void *)Seed)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }
                    }

                    if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_OUTPUT_SIZE, (void *)&Length)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
        
                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
                }
            }
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_AsymmetricContext
 *
 *Description:
 *   This function creates an asymmetric encryption cryptographic context , given a handle 
 *   of a CSP, an algorithm identification number, a key,  padding, and the key mode 
 *   (CSSM_ALGMODE_PRIVATE_KEY or CSSM_ ALGMODE_PUBLIC_KEY)
 *
 *Parameters: 
 *   CSPHandle (input)         - handle that describes the add-in cryptographic service 
 *                               provider module used to perform this function
 *   AlgorithmID (input)       - algorithm identification number for the algorithm used for 
 *                               asymmetric encryption
 *   AccessCred (input)        - pointer to the set of one or more credentials required to 
 *                               unlock the private key
 *   Key (input)               - key used for asymmetric encryption
 *   Padding (input/optional)  - method for padding
 *   NewContextHandle (output) - cryptographic context handle
 *   
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition 
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreateAsymmetricContext(CSSM_CSP_HANDLE CSPHandle,
                                                     CSSM_ALGORITHMS AlgorithmID,
                                                     const CSSM_ACCESS_CREDENTIALS *AccessCred,
                                                     const CSSM_KEY *Key,
                                                     CSSM_PADDING Padding,
                                                     CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CONTEXT_ATTRIBUTE_PTR Attribute = NULL;
    uint32 NumberAttributes = 0;
    CSSM_RETURN retValue = CSSM_OK;
    
    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreateAsymmetricContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_ASYMMETRIC, AlgorithmID, &Context)) == CSSM_OK)
                {
                    if (!port_IsBadReadPtr(AccessCred, (uint32)sizeof(CSSM_ACCESS_CREDENTIALS)))
                        NumberAttributes++;

                    if (!cssm_IsBadCryptoKeyPtr (Key))
                        NumberAttributes++;

                    if (Padding != CSSM_PADDING_NONE)
                        NumberAttributes++;


                    Context->NumberOfAttributes = NumberAttributes;
                    if ((retValue = cssm_CreateAttribute (&Context->ContextAttributes, NumberAttributes)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        NewContextHandle = 0;
                        RETURN(retValue)
                    }

                    Attribute = Context->ContextAttributes;
                    NumberAttributes--;


                    if (!cssm_IsBadCryptoKeyPtr (Key))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_KEY, (void *)Key)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            NewContextHandle = 0;
                            RETURN(retValue)
                        }

                    if (!port_IsBadReadPtr(AccessCred, (uint32)sizeof(CSSM_ACCESS_CREDENTIALS)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_ACCESS_CREDENTIALS, (void *)AccessCred)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            NewContextHandle = 0;
                            RETURN(retValue)
                        }
            
                    if (Padding != CSSM_PADDING_NONE)
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_PADDING, (void *)&Padding)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            NewContextHandle = 0;
                            RETURN(retValue)
                        }

                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
                }
            }
        }
    }
   
    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_CreateDeriveKeyContext
 *
 *Description:
 *   This function creates a cryptographic context to derive a symmetric key given a handle 
 *   of a CSP, an algorithm, the type of symmetric key to derive, the length of the derived 
 *   key, and an optional seed or an optional passphrase from which to derive a new key.
 *
 *Parameters: 
 *   CSPHandle (input)                - handle that describes the add-in cryptographic 
 *                                      service provider module used to perform this function
 *   AlgorithmID (input)              - algorithm identification number for a derived key 
 *                                      algorithm
 *   DeriveKeyType (input)            - type of symmetric key to derive
 *   DeriveKeyLengthInBits (input)    - logical length of the key to be derived in bits
 *   AccessCred (input)               - pointer to the set of one or more credentials 
 *                                      required to unlock the private key
 *   BaseKey (input)                  - base key used to derive the new key
 *   IterationCount (input/optional) - number of iterations to be performed during the 
 *                                      derivation process
 *   Salt (input/optional)            - salt used in deriving the key
 *   Seed (input/optional)            - seed used to generate a random number
 *   NewContextHandle (output)        - cryptographic context handle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreateDeriveKeyContext(CSSM_CSP_HANDLE CSPHandle,
                                                    CSSM_ALGORITHMS AlgorithmID,
                                                    CSSM_KEY_TYPE DeriveKeyType,
                                                    uint32 DeriveKeyLengthInBits,
                                                    const CSSM_ACCESS_CREDENTIALS *AccessCred,
                                                    const CSSM_KEY *BaseKey,
                                                    uint32 IterationCount,
                                                    const CSSM_DATA *Salt,
                                                    const CSSM_CRYPTO_DATA *Seed,
                                                    CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CONTEXT_ATTRIBUTE_PTR Attribute = NULL;
    uint32 NumberAttributes = 0;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreateDeriveKeyContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_DERIVEKEY, AlgorithmID, &Context)) == CSSM_OK)
                {
                    if (!port_IsBadReadPtr(AccessCred, (uint32)sizeof(CSSM_ACCESS_CREDENTIALS)))
                        NumberAttributes++;

                    if (!port_IsBadReadPtr (Salt, (uint32)sizeof (CSSM_DATA)))
                        NumberAttributes++;

                    if (!cssm_IsBadCryptoDataPtr (Seed))
                        NumberAttributes++;

                    if (!cssm_IsBadCryptoKeyPtr (BaseKey))
                        NumberAttributes++;

                    if (IterationCount > 0)
                        NumberAttributes++;

                    /* for key size, and type */
                    NumberAttributes += 2;

                    Context->NumberOfAttributes = NumberAttributes;
                    if ((retValue = cssm_CreateAttribute (&Context->ContextAttributes, NumberAttributes)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    Attribute = Context->ContextAttributes;
                    NumberAttributes--;


                    if (!port_IsBadReadPtr(AccessCred, (uint32)sizeof(CSSM_ACCESS_CREDENTIALS)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_ACCESS_CREDENTIALS, (void *)AccessCred)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (!port_IsBadReadPtr (Salt, (uint32)sizeof (CSSM_DATA)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_SALT, (void *)Salt)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN (retValue)
                        }

                    if (!cssm_IsBadCryptoDataPtr (Seed))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_SEED, (void *)Seed)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (IterationCount)
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_ITERATION_COUNT, (void *)&IterationCount)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (!cssm_IsBadCryptoKeyPtr (BaseKey))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_KEY, (void *)BaseKey)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_KEY_TYPE, (void *)&DeriveKeyType)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_KEY_LENGTH, (void *)&DeriveKeyLengthInBits)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
                }
            }
        }
    }
   
    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_CreateKeyGenContext
 *
 *Description:
 *   This function creates a key generation cryptographic context, given a handle of a CSP, 
 *   an algorithm identification number, a pass phrase, a modulus size (for public/private 
 *   keypair generation), a key size (for symmetric key generation), a seed, salt, and a 
 *   label
 *
 *Parameters: 
 *   CSPHandle (input)          - handle that describes the add-in cryptographic service 
 *                                provider module used to perform this function
 *   AlgorithmID (input)        - algorithm identification number of the algorithm used for 
 *                                key generation  
 *   KeySizeInBits (input)      - logical size of the key (specified in bits)
 *   Seed (input/optional)      - seed used to generate the key
 *   Salt (input/optional)      - salt used to generate the key
 *   StartDate (input/optional) - start date for the validity period of the key or key pair 
 *                                being generated
 *   EndDate (input/optional)   - end date for the validity period of the key or key pair 
 *                                being generated
 *   Params (input/optional)    - data buffer containing parameters required to generate a 
 *                                key pair for a specific algorithm     
 *   NewContextHandle (output)  - cryptographic context handle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition 
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreateKeyGenContext(CSSM_CSP_HANDLE CSPHandle,
                                                 CSSM_ALGORITHMS AlgorithmID,
                                                 uint32 KeySizeInBits,
                                                 const CSSM_CRYPTO_DATA *Seed,
                                                 const CSSM_DATA *Salt,
                                                 const CSSM_DATE *StartDate,
                                                 const CSSM_DATE *EndDate,
                                                 const CSSM_DATA *Params,
                                                 CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CONTEXT_ATTRIBUTE_PTR Attribute = NULL;
    uint32 NumberAttributes = 0;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreateKeyGenContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_KEYGEN, AlgorithmID, &Context)) == CSSM_OK)
                {
                    if (!cssm_IsBadCryptoDataPtr (Seed))
                        NumberAttributes++;

                    if (!port_IsBadReadPtr (Salt, (uint32)sizeof (CSSM_DATA)))
                        NumberAttributes++;

                    if (!port_IsBadReadPtr (StartDate, (uint32)sizeof (CSSM_DATE)))
                        NumberAttributes++;

                    if (!port_IsBadReadPtr (EndDate, (uint32)sizeof (CSSM_DATE)))
                        NumberAttributes++;

                    if (!port_IsBadReadPtr (Params, (uint32)sizeof (CSSM_DATA)))
                        NumberAttributes++;

                    /* for key size */
                    NumberAttributes++;

                    Context->NumberOfAttributes = NumberAttributes;
                    if ((retValue = cssm_CreateAttribute (&Context->ContextAttributes, NumberAttributes)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    Attribute = Context->ContextAttributes;
                    NumberAttributes--;


                    if (!cssm_IsBadCryptoDataPtr (Seed))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_SEED, (void *)Seed)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (!port_IsBadReadPtr (Salt, (uint32)sizeof (CSSM_DATA)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_SALT, (void *)Salt)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (!port_IsBadReadPtr (StartDate, (uint32)sizeof (CSSM_DATE)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_START_DATE, (void *)StartDate)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (!port_IsBadReadPtr (EndDate, (uint32)sizeof (CSSM_DATE)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_END_DATE, (void *)EndDate)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }

                    if (!port_IsBadReadPtr (Params, (uint32)sizeof (CSSM_DATA)))
                        if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_ALG_PARAMS, (void *)Params)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }


                    if ((retValue = cssm_AddAttribute (&Attribute[NumberAttributes--], CSSM_ATTRIBUTE_KEY_LENGTH, (void *)&KeySizeInBits)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
                }
            }
        }
    }

    RETURN (retValue)
}


/*---------------------------------------------------------------
 *
 *Name: CSSM_CSP_CreatePassThroughContext
 *
 *Description:
 *   This function creates a custom cryptographic context, given a handle of a CSP and 
 *   pointer to a key to be used for the context.
 *
 *Parameters: 
 *   CSPHandle (input)         - handle that describes the add-in cryptographic service 
 *                               provider module used to perform this function
 *   Key (input)               - key to be used for the context
 *   NewContextHandle (output) - cryptographic context handle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition 
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_CSP_CreatePassThroughContext(CSSM_CSP_HANDLE CSPHandle,
                                                      const CSSM_KEY *Key,
                                                      CSSM_CC_HANDLE *NewContextHandle)
{
    CSSM_CONTEXT_PTR Context = NULL;
    CSSM_CONTEXT_ATTRIBUTE_PTR Attribute = NULL;
    uint32 NumberAttributes = 0, Count = 0;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_CSP_CreatePassThroughContext);
    if( (NewContextHandle == NULL) || 
        (port_IsBadWritePtr(NewContextHandle, (uint32)sizeof(CSSM_CC_HANDLE))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *NewContextHandle = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                if ((retValue = cssm_CreateContext (CSPHandle, CSSM_ALGCLASS_CUSTOM, CSSM_ALGID_NONE, &Context)) == CSSM_OK)
                {
                    if (!port_IsBadReadPtr (Key, (uint32)sizeof (CSSM_KEY)))
                        NumberAttributes++;
        

                    Context->NumberOfAttributes = NumberAttributes;
                    if ((retValue = cssm_CreateAttribute (&Context->ContextAttributes, NumberAttributes)) != CSSM_OK)
                    {
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }

                    Attribute = Context->ContextAttributes;
                    NumberAttributes--;


                    if (!port_IsBadReadPtr (Key, (uint32)sizeof (CSSM_KEY)))
                        if ((retValue = cssm_AddAttribute (&Attribute[Count++], CSSM_ATTRIBUTE_KEY, (void *)Key)) != CSSM_OK)
                        {
                            cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                            CSSM_FREE_WRAPPER(internal_free,Context, 0);
                            RETURN(retValue)
                        }


                    if ((retValue = cssm_InsertContext (Context, CSPHandle, NewContextHandle)) != CSSM_OK)
                    {
                        cssm_FreeAttributes (Context->NumberOfAttributes, Attribute, internal_free, 0, 1);
                        CSSM_FREE_WRAPPER(internal_free,Context, 0);
                        RETURN(retValue)
                    }
                }
            }
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_GetContext
 *
 *Description:
 *   This function retrieves the context information when provided with a context handle
 *   The memory is allocated by CSSM using th ememory functions given at module attach time.
 *Parameters: 
 *   CCHandle (input) - handle to the context information
 *   Context (output) - pointer to the CSSM_CONTEXT structure that describes the context 
 *                      associated with the handle CCHandle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_GetContext(CSSM_CC_HANDLE CCHandle,
                                    CSSM_CONTEXT_PTR *Context)
{
    cssm_INTERNAL_CONTEXT_PTR ContextInfo = NULL;
    CSSM_RETURN retValue = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;

    TRACE(cssm_trace_info, FNCODE_CSSM_GetContext);
    if( (Context == NULL) || (port_IsBadWritePtr(Context, (uint32)sizeof(CSSM_CONTEXT))) )
    {
        retValue = CSSMERR_CSSM_INVALID_POINTER;
    }
    else
    {
        *Context = 0;
        /* make sure that app has done a CSSM_Init */
        if ((retValue = cssm_CheckInit ()) == CSSM_OK)
        {
#ifdef ISL_INCLUDED

            ISL_FUNCTION_PTR retAddress = NULL;
            EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
            if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
            {
                ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL));
                if(CSSM_OK != retValue)
                {
                    ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
                }
            }
#endif
            if( retValue == CSSM_OK)
            {

                /* Get the context record */
                if ((retValue = cssm_FindContextAndMultiLock(CCHandle,
                                                             CSSM_NO_LOCK,
                                                             NULL,
                                                             CSSM_NO_LOCK,
                                                             NULL,
                                                             CSSM_NO_LOCK,
                                                             &AttachRecord,
                                                             CSSM_READER_LOCK,
                                                             &ContextInfo,
                                                             CSSM_READER_LOCK))
                    == CSSM_OK)
                {
                    /* Found the context. Duplicate the context and return the context pointer. */
                    retValue = cssm_DuplicateContext(ContextInfo->Context, 
                                                    AttachRecord->MemoryFunctions->malloc_func,
                                                    AttachRecord->MemoryFunctions->free_func, 
                                                    AttachRecord->MemoryFunctions->AllocRef,
                                                    ContextInfo,
                                                    Context);
                    if( retValue == CSSM_OK )
                    {
                        cssm_ReplaceCssmProxyWithAppCallback(ContextInfo,(*Context)->NumberOfAttributes, (*Context)->ContextAttributes);
                    }
                    /* Release the context record */
                    cssm_ReleaseContextLock(ContextInfo,CSSM_READER_LOCK);
                    cssm_ReleaseAttachLock(AttachRecord,
                                                CSSM_READER_LOCK);

                }
            }
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_FreeContext
 *
 *Description:
 *   This function frees the memory associated with the context structure
 *   The context structure is freed using the memory functions associated with the modulehandle.
 *Parameters: 
 *   Context (input) - pointer to the memory that describes the context structure
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_FreeContext (CSSM_CONTEXT_PTR Context)
{
    uint32 NumberAttributes;
    CSSM_CONTEXT_ATTRIBUTE_PTR ContextAttributes = NULL;
    CSSM_RETURN retValue = CSSM_OK;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;

    TRACE(cssm_trace_info, FNCODE_CSSM_FreeContext);

   /* make sure that app has done a CSSM_Init */
    if ((retValue = cssm_CheckInit ()) == CSSM_OK)
    {
#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != retValue)
            {
                ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }
#endif
        if( retValue == CSSM_OK)
        {

            /* make sure that pointer is good */
            if ( (port_IsBadReadPtr (Context, (uint32)sizeof (CSSM_CONTEXT))) )
            {
                retValue = CSSMERR_CSSM_INVALID_POINTER;
                RETURN(retValue)
            }

            NumberAttributes = Context->NumberOfAttributes;
            ContextAttributes = Context->ContextAttributes;

            /* make sure that pointer is good */
            if(port_IsBadReadPtr(ContextAttributes,
                                 (uint32)sizeof (CSSM_CONTEXT_ATTRIBUTE) * NumberAttributes))
            {
                retValue = CSSMERR_CSSM_INVALID_POINTER;
                RETURN(retValue)
            }

            if((retValue = cssm_FindAttachAndMultiLock(Context->CSPHandle,
                                                        CSSM_NO_LOCK,
                                                        NULL,
                                                        CSSM_NO_LOCK,
                                                        NULL,
                                                        CSSM_NO_LOCK,
                                                        &AttachRecord,
                                                        CSSM_READER_LOCK)) == CSSM_OK)
            {
                if( ( AttachRecord != NULL) && (AttachRecord->MemoryFunctions != NULL) &&
                    (AttachRecord->MemoryFunctions->free_func != NULL) )
                {
                    cssm_FreeContext (&Context, 
                                    AttachRecord->MemoryFunctions->free_func, 
                                    AttachRecord->MemoryFunctions->AllocRef);
                }
                else
                    retValue = CSSMERR_CSSM_INVALID_POINTER;

                cssm_ReleaseAttachLock(AttachRecord,
                                        CSSM_READER_LOCK);

            }
        }
    }

    RETURN(retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_SetContext
 *
 *Description:
 *   This function replaces all of the context information associated with an existing 
 *   context specified by CCHandle
 *
 *Parameters: 
 *   CCHandle (input) - handle to the context
 *   Context (input)  - context data describing the service to replace the current service 
 *                      associated with context handle CCHandle
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_SetContext (CSSM_CC_HANDLE CCHandle,
                                     const CSSM_CONTEXT *Context)
{
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_SetContext);

   /* make sure that app has done a CSSM_Init */
    if ((retValue = cssm_CheckInit ()) == CSSM_OK)
    {
#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the application here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != retValue)
            {
                ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }
#endif
        if( retValue == CSSM_OK)
        {
            /* make sure that pointer is good */
            if(port_IsBadReadPtr(Context, (uint32)sizeof(CSSM_CONTEXT)))
            {
                ERR(retValue = CSSMERR_CSSM_INVALID_POINTER);
            }
            else
            {
                ERR(retValue = cssm_AssignContext(CCHandle, Context));
            }
        }
    
    }
    
    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_DeleteContext
 *
 *Description:
 *   This function frees the context structure allocated by any of the 
 *   CSSM_CreateXXXXXContext functions
 *
 *Parameters: 
 *   CCHandle (input) - The handle that describes a context to be deleted
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_DeleteContext (CSSM_CC_HANDLE CCHandle)
{
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    cssm_INTERNAL_CONTEXT_PTR ContextRecord = NULL;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_DeleteContext);

    /* make sure that app has done a CSSM_Init */
    if((retValue = cssm_CheckInit ()) == CSSM_OK)
    {
#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != retValue)
            {
                ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }
#endif
        if( retValue == CSSM_OK)
        {

            /* Get the internal attach and context records */
            if((retValue = cssm_FindContextAndMultiLock(CCHandle,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   NULL,
                                                   CSSM_NO_LOCK,
                                                   &AttachRecord,
                                                   CSSM_WRITER_LOCK,
                                                   &ContextRecord,
                                                   CSSM_WRITER_LOCK))== CSSM_OK)
            {
                /* Call the internal function to delete the context */
                retValue = cssm_DeleteContext(AttachRecord, &ContextRecord);
                cssm_ReleaseAttachLock(AttachRecord, CSSM_WRITER_LOCK);
            }
            else
                retValue = CSSMERR_CSSM_INVALID_CONTEXT_HANDLE;
        }
    }

    RETURN(retValue)
}


/*---------------------------------------------------------------
 *
 *Name: CSSM_GetContextAttribute
 *
 *Description:
 *   This function returns the value of a context attribute within the supplied context
 *
 *Parameters: 
 *   Context (input)           - pointer to the context
 *   AttributeType (input)     - attribute type of the desired attribute value
 *   ContextAttribute (output) - pointer to the CSSM_CONTEXT_ATTRIBUTE structure that 
 *                               describes the context attributes associated with the handle 
 *                               CCHandle and the attribute type
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *Note :
 *   return pointer is an offset to the context data.  Do not free this pointer.
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI CSSM_GetContextAttribute(const CSSM_CONTEXT *Context,
                                             uint32 AttributeType,
                                             CSSM_CONTEXT_ATTRIBUTE_PTR *ContextAttribute)
{
    uint32 i;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_GetContextAttribute);
    
    /* make sure that app has done a CSSM_Init */
    if(ERR(retValue = cssm_CheckInit ()) == CSSM_OK)
    {
#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

            /* if we have to, we check the PVC on app here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != retValue)
            {
                ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }
#endif
        if( retValue == CSSM_OK)
        {

            if((CSSM_TRUE == port_IsBadReadPtr(Context,
                                               (uint32)sizeof(CSSM_CONTEXT))) ||
               (NULL == ContextAttribute))
            {
                ERR(retValue = CSSMERR_CSSM_INVALID_POINTER);
            }
            else
            {
                *ContextAttribute = NULL;

                for (i=0; i<Context->NumberOfAttributes; i++)
                {
                    if(Context->ContextAttributes[i].AttributeType == 
                       (CSSM_ATTRIBUTE_TYPE)AttributeType)
                    {
                        *ContextAttribute = &Context->ContextAttributes[i];
                    }
                }

                if(NULL == *ContextAttribute)
                {
                    ERR(retValue = CSSMERR_CSSM_ATTRIBUTE_NOT_IN_CONTEXT);
                }
            }
        }
    }

    RETURN (retValue)
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_UpdateContextAttribute
 *
 *Description:
 *   This function updates one or more context attribute values stored as part of an 
 *   existing context specified by CCHandle
 *
 *Parameters: 
 *   CCHandle (input)          - handle to the context
 *   NumberAttributes (input)  - number of CSSM_CONTEXT_ATTRIBUTE structures to allocate
 *   ContextAttributes (input) - pointer to data that describes the attributes to be 
 *                               associated with this context
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_UpdateContextAttributes(CSSM_CC_HANDLE CCHandle,
                             uint32 NumberAttributes,
                             const CSSM_CONTEXT_ATTRIBUTE *ContextAttributes)
{
    cssm_INTERNAL_CONTEXT_PTR InternalContext = NULL;
    CSSM_CONTEXT_PTR tempContext = NULL;
    uint32 i;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_MODULE_HANDLE hAttach = CSSM_UINT64_HIGH(CCHandle);
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_RETURN retValue = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_UpdateContextAttributes);

    /* make sure that app has done a CSSM_Init */
    if ((retValue = cssm_CheckInit ()) == CSSM_OK)
    {
#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress( retAddress);

        /* if we have to, we check the PVC on app here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(retValue = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != retValue)
            {
                ERR(retValue = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }
#endif
        if( retValue == CSSM_OK)
        {
            /* make sure that pointer is good */
            if (port_IsBadReadPtr (ContextAttributes, 
                                    (uint32)sizeof (CSSM_CONTEXT_ATTRIBUTE) * NumberAttributes))
            {
                ERR(retValue = CSSMERR_CSSM_INVALID_POINTER);
            }
            else
            {
                if(ERR(retValue = cssm_FindContextAndMultiLock(CCHandle,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               NULL,
                                                               CSSM_NO_LOCK,
                                                               &AttachRecord,
                                                               CSSM_READER_LOCK,
                                                               &InternalContext,
                                                               CSSM_WRITER_LOCK)) == CSSM_OK)
                {
                    /* duplicate the old context so we can update a copy */
                    ERR(retValue = cssm_DuplicateContext(InternalContext->Context,
                                                         internal_malloc,
                                                         internal_free,
                                                         NULL,
                                                         InternalContext,
                                                         &tempContext));
                    if(CSSM_OK == retValue)
                    {
                        /* Update the data in the duplicate */
                        i = 0;
                        while((i < NumberAttributes) &&
                              (CSSM_OK == retValue))
                        {
                            ERR(retValue = cssm_UpdateContext(tempContext,
                                                              &ContextAttributes[i]));
                            i++;
                        }

                        /* callback to csp to notify context creation */
                        if((CSSM_OK == retValue) && (NULL != AttachRecord->AddInFunctions))
                        {
                            /* get the CSP callback functions */
                            if(ERR(retValue = cssm_GetAddInCallBackByType(AttachRecord,
                                                                          CSSM_SERVICE_CSP,
                                                                          (void *)&CallBack)) == CSSM_OK)/* #ifdef VMS apply cast*/
                            {
                                if(ERR(retValue = cssm_ReplaceAppCallbacksWithCssmProxy(InternalContext,
                                                                                        tempContext->NumberOfAttributes,
                                                                                        tempContext->ContextAttributes)) == CSSM_OK)
                                {
                                    /* this function will lock the mutex if necessary */
                                    ERR(retValue = cssm_LockMultiTaskMutex(AttachRecord));

                                    if(CSSM_OK == retValue)
                                    {
                                        ERR(retValue =  CallBack->EventNotify(hAttach,
                                                                              CSSM_CONTEXT_EVENT_UPDATE,
                                                                              CCHandle,
                                                                              tempContext));

                                        cssm_UnlockMultiTaskMutex(AttachRecord);

                                        if(CSSM_OK == retValue)
                                        {
                                            /* now we free up the old info and insert the new info
                                               since the CSP said it was acceptable
                                            */
                                            cssm_FreeAttributes(InternalContext->Context->NumberOfAttributes, 
                                                                InternalContext->Context->ContextAttributes,
                                                                internal_free,
                                                                0,
                                                                1);
                                            local_free(InternalContext->Context,
                                                       NULL);
                                            InternalContext->Context = tempContext;
                                        }
                                    }

                                    if(CSSM_OK != retValue)
                                    {
                                        /* the update didn't work, so we revert to the old
                                           values
                                        */
                                        ERR(cssm_ReplaceAppCallbacksWithCssmProxy(InternalContext,
                                                                                  InternalContext->Context->NumberOfAttributes,
                                                                                  InternalContext->Context->ContextAttributes));
                                    }
                                }
                            }
                        }

                        if(CSSM_OK != retValue)
                        {
                            ERR(cssm_FreeContext(&tempContext,
                                                 internal_free,
                                                 NULL));
                        }
                    }

                    cssm_ReleaseContextLock(InternalContext,CSSM_WRITER_LOCK);
                    cssm_ReleaseAttachLock(AttachRecord,CSSM_READER_LOCK);
                }
            }
        }
    }

    RETURN (retValue)  
}

/*---------------------------------------------------------------
 *
 *Name: CSSM_DeleteContextAttribute
 *
 *Description:
 *   This function deletes internal data associated with given attribute type of the context 
 *   handle
 *
 *Parameters: 
 *   CCHandle (input)           - handle that describes a context that is to be deleted
 *   NumberOfAttributes (input) - number of attributes to be deletedas specified in the array 
 *                                of context attributes
 *   ContextAttributes (input)  - attribute to be deleted from the context
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition
 *   The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_DeleteContextAttributes(CSSM_CC_HANDLE CCHandle,
                             uint32 NumberAttributes,
                             const CSSM_CONTEXT_ATTRIBUTE *ContextAttributes)
{
    cssm_INTERNAL_CONTEXT_PTR InternalContext = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_SPI_CSP_FUNCS_PTR CallBack = NULL;
    CSSM_CONTEXT_PTR tempContext = NULL;

    uint32 i;
    CSSM_RETURN err = CSSM_OK;
    CSSM_RETURN tempErr = CSSM_OK;

    TRACE(cssm_trace_info, FNCODE_CSSM_DeleteContextAttributes);

    /* make sure that app has done a CSSM_Init */
    if(ERR(err = cssm_CheckInit ()) == CSSM_OK)
    {
#ifdef ISL_INCLUDED

        ISL_FUNCTION_PTR retAddress = NULL;
        EISL_GetReturnAddress(retAddress);

        /* if we have to, we check the PVC on app here */
        if(cssm_GetPVCPolicy() & CSSM_PVC_APP)
        {
            ERR(err = cssm_FindCallerListEntryAndMultiLock((CSSM_PROC_ADDR)retAddress,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL,
                                                           CSSM_NO_LOCK,
                                                           NULL));
            if(CSSM_OK != err)
            {
                ERR(err = CSSMERR_CSSM_PVC_REFERENT_NOT_FOUND);
            }
        }
#endif
        if(CSSM_OK == err)
        {
            /* make sure that pointer is good */
            if(port_IsBadReadPtr(ContextAttributes,
                                 (uint32)sizeof(CSSM_CONTEXT_ATTRIBUTE) * NumberAttributes))
            {
                ERR(err = CSSMERR_CSSM_INVALID_POINTER);
            }
            else
            {
                if(ERR(err = cssm_FindContextAndMultiLock(CCHandle,
                                                          CSSM_NO_LOCK,
                                                          NULL,
                                                          CSSM_NO_LOCK,
                                                          NULL,
                                                          CSSM_NO_LOCK,
                                                          &AttachRecord,
                                                          CSSM_READER_LOCK,
                                                          &InternalContext,
                                                          CSSM_WRITER_LOCK))
                    == CSSM_OK)
                {
			        ERR(err = cssm_GetAddInCallBackByType(AttachRecord,
                                                                      CSSM_SERVICE_CSP,
                                                                      (void *)&CallBack));  /* #ifdef VMS apply cast*/
                    if(CSSM_OK == err)
                    {
                        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->EventNotify))
			            {
                            ERR(err = cssm_DuplicateContext(InternalContext->Context,
                                                            internal_malloc,
                                                            internal_free,
                                                            NULL,
                                                            NULL,
                                                            &tempContext));
                            if(CSSM_OK == err)
                            {
                                /* found and copied the context, now update the data */
                                for(i=0; i<NumberAttributes; i++)
                                {
                                    ERR(err = cssm_DelContextAttr(tempContext,
                                                                  &ContextAttributes[i]));
                                    if(CSSM_OK == tempErr)
                                    {
                                        tempErr = err;
                                    }
                                }

                                /* it doesn't really matter if there was an
                                   an error reported in the above loop
                                */
                                err = CSSM_OK;

                                /* this function will lock the mutex if necessary */
            				    ERR(err = cssm_LockMultiTaskMutex(AttachRecord));

                                if(CSSM_OK == err)
                                {
                                    ERR(err = CallBack->EventNotify(AttachRecord->AttachHandle,
                                                                    CSSM_CONTEXT_EVENT_UPDATE,
                                                                    CCHandle,
                                                                    tempContext));
                                    if(CSSM_OK == err)
                                    {
                                        /* we update the internal context with
                                           the newly accepted information
                                        */
                                        cssm_FreeAttributes(InternalContext->Context->NumberOfAttributes, 
                                                            InternalContext->Context->ContextAttributes,
                                                            internal_free,
                                                            0,
                                                            1);
                                        local_free(InternalContext->Context,
                                                   NULL);
                                        InternalContext->Context = tempContext;

                                    }

                                    cssm_UnlockMultiTaskMutex(AttachRecord);
                                }

                                /* if there's been a problem, we destroy the
                                   new context and revert to the old attributes
                                */
                                if(CSSM_OK != err)
                                {
                                    ERR(cssm_FreeContext(&tempContext,
                                                         internal_free,
                                                         NULL));
                                }

                                /* if there was a problem updating any
                                   attribute, we note that here.  This isn't
                                   a fatal error, and may occur in correct
                                   operations, so we keep the updated attribute
                                   list.
                                */
                                if(CSSM_OK == err)
                                {
                                    err = tempErr;
                                }
                            }
                        }
                        else
                        {
                            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
                        }
                    }


                    cssm_ReleaseAttachLock(AttachRecord,
                                           CSSM_READER_LOCK);
                    cssm_ReleaseContextLock(InternalContext,
                                            CSSM_WRITER_LOCK);
                }
            }
        }
    }

    return err;
}
