/*-----------------------------------------------------------------------
 * File: CLAPI.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */
 
/*
 * This file contains the functions that are contained in the CLI portion
 * of the CSSM exported functions.
 */

#include "precomp.h"


extern CSSM_RETURN CSSMAPI CSSM_CRYPTO_DATA_API_CALLBACK (CSSM_DATA_PTR OutData, void *CallbackCtx );
/* Proposed Certificate Library API */
/* Certificate Library Operations */


/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertCreateTemplate
 *
 *Description:
 *   This function allocates and initializes memory for an encoded 
 *   certificate template returned in CertTemplate
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the certificate 
 *                     library module used to perform this function.
 *
 *  CertFields(input) - A pointer to an array of OID/value pairs that 
 *                      identify the field values to initialize a new 
 *                      certificate.
 *  
 *  NumberOfFields (input) - The number of certificate field values specified 
 *                           in the CertFields. 
 *
 *  CertTemplate (output) - A pointer to a CSSM_DATA structure that will 
 *                          contain the unsigned certificate template as a 
 *                          result of this function.3
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. All other 
 *   values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertCreateTemplate(CSSM_CL_HANDLE CLHandle,
                           uint32 NumberOfFields,
                           const CSSM_FIELD *CertFields,
                           CSSM_DATA_PTR CertTemplate)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertCreateTemplate);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertCreateTemplate))
        {
            ERR(err = CallBack->CertCreateTemplate(CLHandle,
                                                   NumberOfFields,
                                                   CertFields,
                                                   CertTemplate));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGetAllTemplateFields
 *
 *Description:
 *   Extracts and returns a copy of the values stored in the encoded CertTemplate
 *
 *Parameters: 
 *   CLHandle (input)        - handle of CL module
 *   CertTemplate (input)    - pointer to the CSSM_DATA structure containing the packed, 
 *                             encoded certificate template
 *   CertFields(output)      - pointer to an array of CSSM_FIELD structures which 
 *                             contains the names and values of the fields of the input 
 *                             certificate template
 *   NumberOfFields (output) - length of the output array of fields
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertGetAllTemplateFields(CSSM_CL_HANDLE CLHandle,
                                 const CSSM_DATA *CertTemplate,
                                 uint32 *NumberOfFields,
                                 CSSM_FIELD_PTR *CertFields)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGetAllTemplateFields);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGetAllTemplateFields))
        {
            ERR(err = CallBack->CertGetAllTemplateFields(CLHandle,
                                                         CertTemplate,
                                                         NumberOfFields,
                                                         CertFields));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertSign
 *
 *Description:
 *   This function signs a certificate and returns the result as a signed, encoded 
 *   certificate in SignedCert
 *
 *Parameters: 
 *   CLHandle (input)           - handle of CL module
 *   CCHandle (input/optional)  - handle that describes the context of this 
 *                                cryptographic operation
 *   CertTemplate(input)        - pointer to a CSSM_DATA structure containing a 
 *                                certificate template in the default format supported 
 *                                by this CL
 *   SignerCert (input)         - pointer to the CSSM_DATA structure containing the 
 *                                certificate identifying the entity that is performing 
 *                                the signing operations
 *   SignScope (input/optional) - pointer to the CSSM_FIELD array containing the 
 *                                OID/value pairs of the fields to be signed
 *   ScopeSize (input)          - number of entries in the SignScope list
 *   SignedCert (output)        - pointer to the CSSM_DATA structure containing the 
 *                                signed certificate
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertSign(CSSM_CL_HANDLE CLHandle,
                 CSSM_CC_HANDLE CCHandle,
                 const CSSM_DATA *CertTemplate,
                 const CSSM_FIELD *SignScope,
                 uint32 ScopeSize,
                 CSSM_DATA_PTR SignedCert)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertSign);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertSign))
        {
            ERR(err = CallBack->CertSign(CLHandle,
                                         CCHandle,
                                         CertTemplate,
                                         SignScope,
                                         ScopeSize,
                                         SignedCert));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}





/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertVerify
 *
 *Description:
 *   Function for CL module to verify the signature of a signed certificate
 *
 *Parameters: 
 *   CLHandle (input)             - handle of CL module
 *   CCHandle (input)             - handle of cryptographic context
 *   CertToBeVerified (input)     - pointer to the signed certificate
 *   SignerCert (input)           - pointer to certificate used in signing
 *   VerifyScope (input/optional) - pointer to the scope of fields of signature
 *   ScopeSize (input)            - number of fields in VerifyScope
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertVerify(CSSM_CL_HANDLE CLHandle,
                   CSSM_CC_HANDLE CCHandle,
                   const CSSM_DATA *CertToBeVerified,
                   const CSSM_DATA *SignerCert,
                   const CSSM_FIELD *VerifyScope,
                   uint32 ScopeSize)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertVerify);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertVerify))
        {
            ERR(err = CallBack->CertVerify(CLHandle,
                                           CCHandle,
                                           CertToBeVerified,
                                           SignerCert,
                                           VerifyScope,
                                           ScopeSize));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertVerifyWithKey
 *
 *Description:
 *   This function verifies that the CertToBeVerified was signed using a specific 
 *   private key and that the certificate has not been altered since it was signed 
 *   using that private key
 *
 *Parameters: 
 *   CLHandle (input)         - handle that describes the certificate library service 
 *                              module used to perform this function
 *   CCHandle (input)         - signature verification context defining the CSP, 
 *                              verification algorithm, and public key that must be used 
 *                              to perform the operation.
 *   CertToBeVerified (input) - A signed certificate whose signature is to be verified
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertVerifyWithKey(CSSM_CL_HANDLE CLHandle,
                          CSSM_CC_HANDLE CCHandle,
                          const CSSM_DATA *CertToBeVerified)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertVerifyWithKey);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertVerifyWithKey))
        {
            ERR(err = CallBack->CertVerifyWithKey(CLHandle,
                                                  CCHandle,
                                                  CertToBeVerified));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGetFirstFieldValue
 *
 *Description:
 *   This function returns the value of the certificate field designated by the 
 *   CSSM_OID CertField
 *
 *Parameters: 
 *   CLHandle (input)               - handle that describes the add-in certificate 
 *                                    library module used to perform this function
 *   Cert (input)                   - pointer to the CSSM_DATA structure containing the 
 *                                    certificate
 *   CertField (input)              - pointer to an object identifier which identifies 
 *                                    the field value to be extracted from the Cert
 *   ResultsHandle (output)         - pointer to the CSSM_HANDLE which should be used 
 *                                    to obtain any additional matching fields
 *   NumberOfMatchedFields (output) - total number of fields that match the CertField OID
 *   Value (output)                 - pointer to the CSSM_DATA structure containing the 
 *                                    value of the requested field
 *   
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertGetFirstFieldValue(CSSM_CL_HANDLE CLHandle,
                               const CSSM_DATA *Cert,
                               const CSSM_OID *CertField,
                               CSSM_HANDLE_PTR ResultsHandle,
                               uint32 *NumberOfMatchedFields,
                               CSSM_DATA_PTR *Value)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGetFirstFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGetFirstFieldValue))
        {
            ERR(err = CallBack->CertGetFirstFieldValue(CLHandle,
                                                       Cert,
                                                       CertField,
                                                       ResultsHandle,
                                                       NumberOfMatchedFields,
                                                       Value));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGetNextFieldValue
 *
 *Description:
 *   This function returns the value of a certificate field, when that field occurs 
 *   multiple times in a certificate
 *
 *Parameters: 
 *   CLHandle (input)      - handle that describes the add-in certificate library module 
 *                           used to perform this function
 *   ResultsHandle (input) - handle which identifies the results of a certificate query
 *   Value (output)        - pointer to the CSSM_DATA structure containing the value of 
 *                           the requested field
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertGetNextFieldValue(CSSM_CL_HANDLE CLHandle,
                              CSSM_HANDLE ResultsHandle,
                              CSSM_DATA_PTR *Value)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGetNextFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGetNextFieldValue))
        {
            ERR(err = CallBack->CertGetNextFieldValue(CLHandle,
                                                      ResultsHandle,
                                                      Value));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertAbortQuery
 *
 *Description:
 *   This function terminates a results handle used to access multiple certificate 
 *   fields identified by a single OID
 *
 *Parameters: 
 *   CLHandle (input)      - handle that describes the add-in certificate library module 
 *                           used to perform this function.
 *   ResultsHandle (input) - pointer to the handle which identifies the results of a 
 *                           CSSM_CL_GetFieldValue request. 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertAbortQuery(CSSM_CL_HANDLE CLHandle,
                       CSSM_HANDLE ResultsHandle)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertAbortQuery);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertAbortQuery))
        {
            ERR(err = CallBack->CertAbortQuery(CLHandle,
                                               ResultsHandle));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGetKeyInfo
 *
 *Description:
 *   This function returns the public key and integral information 
 *   about the key from the specified certificate
 *
 *Parameters: 
 *   CLHandle (input) - The handle that describes the add-in 
 *                      certificate library module used to perform 
 *                      this function.
 *
 *   Cert (input) - A pointer to the CSSM_DATA structure containing 
 *                  the certificate from which to extract the public 
 *                  key information.         
 *
 *   Key (output) - A pointer to the CSSM_KEY structure containing the 
 *                  public key and possibly other key information. 
 *                  The CSSM_KEY structure is allocated by the service 
 *                  provider and must be deallocated by the application.
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertGetKeyInfo(CSSM_CL_HANDLE CLHandle,
                       const CSSM_DATA *Cert,
                       CSSM_KEY_PTR *Key)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGetKeyInfo);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGetKeyInfo))
        {
            ERR(err = CallBack->CertGetKeyInfo(CLHandle,
                                               Cert,
                                               Key));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGetAllFields
 *
 *Description:
 *   This function returns a list of the values stored in the input certificate
 *
 *Parameters: 
 *   CLHandle (input)        - handle that describes the add-in certificate library 
 *                             module used to perform this function
 *   Cert (input)            - pointer to the CSSM_DATA structure containing the 
 *                             certificate whose fields will be returned
 *   NumberOfFields (output) - length of the returned array of fields
 *   FieldList (output)      - pointer to an array of CSSM_FIELD structures that contain 
 *                             the values of all of the fields of the input certificate
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertGetAllFields(CSSM_CL_HANDLE CLHandle,
                         const CSSM_DATA *Cert,
                         uint32 *NumberOfFields,
                         CSSM_FIELD_PTR *FieldList)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGetAllFields);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGetAllFields))
        {
            ERR(err = CallBack->CertGetAllFields(CLHandle,
                                                 Cert,
                                                 NumberOfFields,
                                                 FieldList));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_FreeFieldValue
 *
 *Description:
 *   This function frees the data specified by CertOrCrlField->FieldValue 
 *
 *Parameters: 
 *   CLHandle - handle of CL module
 *   CertOrCrlField - pointer to a CSSM_FIELD structure
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *   The value CSSM_OK indicates success. All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_FreeFieldValue(CSSM_CL_HANDLE CLHandle,
                       const CSSM_OID *CertOrCrlOid,
                       CSSM_DATA_PTR Value)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_FreeFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->FreeFieldValue))
        {
            ERR(err = CallBack->FreeFieldValue(CLHandle,
                                               CertOrCrlOid,
                                               Value));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertCache
 *
 *Description:
 *   This function caches a copy of a certificate for subsequent 
 *   accesses using the functions CSSM_CL_CertGetFirstCachedFieldValue 
 *   and CSSM_CL_CertGetNextCachedFieldValue
 *
 *Parameters: 
 *   CLHandle  (input) - The handle that describes the certificate 
 *                       library module used to perform this function.
 *
 *   Cert  (input) - A pointer to the CSSM_DATA structure containing 
 *                   the encoded certificate.
 *
 *   CertHandle  (output) - A pointer to the CSSM_HANDLE that should be 
 *                          used in all future references to the cached 
 *                          certificate.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CertCache(CSSM_CL_HANDLE CLHandle,
                  const CSSM_DATA *Cert,
                  CSSM_HANDLE_PTR CertHandle)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertCache);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertCache))
        {
            ERR(err = CallBack->CertCache(CLHandle,
                                          Cert,
                                          CertHandle));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGetFirstCachedFieldValue
 *
 *Description:
 *   This function returns a single structure containing a set of 
 *   field values from the cached certificate identified by CertHandle
 *
 *Parameters: 
 *   CLHandle  (input) - The handle that describes the add-in certificate 
 *                       library module used to perform this function.
 *
 *   CertHandle  (input) - A handle identifying a certificate that the 
 *                         application has temporarily cached with the 
 *                         Certificate Library module. The referenced 
 *                         certificate is searched for the field value named 
 *                         by CertField.
 *
 *   CertField  (input) - A pointer to an object identifier that 
 *                        identifies the field value to be extracted 
 *                        from the Cert.
 *
 *   ResultsHandle  (output) - A pointer to the CSSM_HANDLE that 
 *                             should be used to obtain any additional 
 *                             matching fields.
 *
 *   NumberOfMatchedFields  (output) - The total number of fields 
 *                                     that match the CertField OID. This 
 *                                     count includes the first 
 *                                     match, which was returned by this 
 *                                     function. 
 *
 *   FieldValue  (output) - A pointer to the CSSM_DATA structure 
 *                          containing the value of the requested field.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CertGetFirstCachedFieldValue(CSSM_CL_HANDLE CLHandle,
                                     CSSM_HANDLE CertHandle,
                                     const CSSM_OID *CertField,
                                     CSSM_HANDLE_PTR ResultsHandle,
                                     uint32 *NumberOfMatchedFields,
                                     CSSM_DATA_PTR *FieldValue)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGetFirstCachedFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGetFirstCachedFieldValue))
        {
            ERR(err = CallBack->CertGetFirstCachedFieldValue(CLHandle,
                                                             CertHandle,
                                                             CertField,
                                                             ResultsHandle,
                                                             NumberOfMatchedFields,
                                                             FieldValue));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGetNextCachedFieldValue
 *
 *Description:
 *   This function returns the value of a certificate field, 
 *   when that field occurs multiple times in a certificate
 *
 *Parameters: 
 *  CLHandle  (input) - The handle that describes the certificate 
 *                      library module used to perform this function. 
 *
 *  ResultsHandle  (input) - The handle that identifies the 
 *                           results of a certificate query. 
 *
 *  FieldValue  (output) - A pointer to the CSSM_DATA structure 
 *                         containing the value of the requested field.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertGetNextCachedFieldValue(CSSM_CL_HANDLE CLHandle,
                                    CSSM_HANDLE ResultsHandle,
                                    CSSM_DATA_PTR *FieldValue)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGetNextCachedFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGetNextCachedFieldValue))
        {
            ERR(err = CallBack->CertGetNextCachedFieldValue(CLHandle,
                                                            ResultsHandle,
                                                            FieldValue));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertAbortCache
 *
 *Description:
 *   This function terminates a certificate cache handle created 
 *   and returned by the function CSSM_CL_CertCache
 *
 *Parameters: 
 *  CLHandle  (input) - The handle that describes the certificate 
 *                      library module used to perform this function.
 *
 *  CertHandle  (input) - The handle that identifies the cached certificate. 
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CertAbortCache(CSSM_CL_HANDLE CLHandle,
                       CSSM_HANDLE CertHandle)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertAbortCache);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertAbortCache))
        {
            ERR(err = CallBack->CertAbortCache(CLHandle,
                                               CertHandle));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGroupToSignedBundle
 *
 *Description:
 *   This function accepts as input a certificate group 
 *   (as an array of individual certificates) and returns a certificate 
 *   bundle (a codified and signed aggregation of the certificates in the 
 *   group)
 *
 *Parameters: 
 *  CLHandle (input) - The handle of the add-in module to perform 
 *  this operation.
 *
 *  CCHandle (input/optional) - The handle of the cryptographic 
 *                              context to control the signing operation.  
 *                              The operation will fail if a signature 
 *                              is required for this type of bundle and the 
 *                              cryptographic context is not valid.
 *
 *  CertGroupToBundle (input) - An array of individual, encoded 
 *                              certificates. All of the certificates 
 *                              in this list will be included in the 
 *                              resulting certificate bundle.
 * 
 *  SignerCert (input/optional) - If signing is required for this 
 *                                type of certificate bundle, this is 
 *                                the certificate to be used to sign 
 *                                the bundle. If a signing certificate 
 *                                is required but not specified, then 
 *                                the module will assume a default certificate.  
 *                                If a signature is not required for this 
 *                                certificate bundle type, this parameter 
 *                                will be ignored. 
 *
 *  BundleInfo (input/optional) - A structure containing the 
 *                                type and encoding of the bundle to be created. 
 *                                If the type and the encoding are not specified, 
 *                                then the module will assume a default bundle 
 *                                type and bundle encoding.
 *
 *  SignedBundle (output) - The function returns a pointer to a signed 
 *                          certificate bundle containing all of the 
 *                          certificates in the certificate group. The 
 *                          bundle is of the type and encoding requested by 
 *                          the caller or is the default type defined by the 
 *                          library module if the BundleInfo was not specified 
 *                          by the caller. The certificate group is allocated by 
 *                          the service provider and must be deallocated by the 
 *                          application.
 *
 *
 *Returns:
 * A CSSM_RETURN value indicating success or specifying a particular error condition. 
 * The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertGroupToSignedBundle(CSSM_CL_HANDLE CLHandle,
                                CSSM_CC_HANDLE CCHandle,
                                const CSSM_CERTGROUP *CertGroupToBundle,
                                const CSSM_CERT_BUNDLE_HEADER *BundleInfo,
                                CSSM_DATA_PTR SignedBundle)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGroupToSignedBundle);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGroupToSignedBundle))
        {
            ERR(err = CallBack->CertGroupToSignedBundle(CLHandle,
                                                        CCHandle,
                                                        CertGroupToBundle,
                                                        BundleInfo,
                                                        SignedBundle));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGroupFromVerifiedBundle
 *
 *Description:
 *   This function accepts as input a certificate bundle (a codified 
 *    and signed aggregation of the certificates in the group)
 *
 *
 *Parameters: 
 *  CLHandle (input) - The handle of the add-in module to perform 
 *                     this operation.
 *
 *  CCHandle (input) - The handle of the cryptographic context to 
 *                     control the verification operation.
 *
 *  CertBundle (input) - A structure containing a reference to a signed, 
 *                       encoded bundle of certificates, and to descriptors 
 *                       of the type and encoding of the bundle. The bundled 
 *                       certificates are to be separated into a certificate 
 *                       group (list of individual encoded certificates).  
 *                       If the bundle type and bundle encoding are not specified, 
 *                       the add-in module may either attempt to decode the bundle 
 *                       assuming a default type and encoding or may immediately fail.
 *
 *  SignerCert (input/optional) - The certificate to be used to verify the 
 *                                signature on the certificate bundle. If the 
 *                                bundle is signed but this field is not specified, 
 *                                then the module will assume a default certificate 
 *                                for verification. 
 *
 *  CertGroup (output) - A pointer to the certificate group, represented as 
 *                       an array of individual, encoded certificates. 
 *                       The group contains all of the certificates contained 
 *                       in the certificate bundle.
 *
 *Returns:
 *  A CSSM_RETURN value indicating success or specifying a particular error condition. 
 *  The value CSSM_OK indicates success. All other values represent an error condition
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertGroupFromVerifiedBundle(CSSM_CL_HANDLE CLHandle,
                                    CSSM_CC_HANDLE CCHandle,
                                    const CSSM_CERT_BUNDLE *CertBundle,
                                    const CSSM_DATA *SignerCert,
                                    CSSM_CERTGROUP_PTR *CertGroup)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertGroupFromVerifiedBundle);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertGroupFromVerifiedBundle))
        {
            ERR(err = CallBack->CertGroupFromVerifiedBundle(CLHandle,
                                                            CCHandle,
                                                            CertBundle,
                                                            SignerCert,
                                                            CertGroup));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CertGetKeyInfo
 *
 *Description:
 *   This function returns the public key and integral information 
 *   about the key from the specified certificate
 *
 *Parameters: 
 *   CLHandle (input) - The handle that describes the add-in 
 *                      certificate library module used to perform 
 *                      this function.
 *
 *   Cert (input) - A pointer to the CSSM_DATA structure containing 
 *                  the certificate from which to extract the public 
 *                  key information.         
 *
 *   Key (output) - A pointer to the CSSM_KEY structure containing the 
 *                  public key and possibly other key information. 
 *                  The CSSM_KEY structure is allocated by the service 
 *                  provider and must be deallocated by the application.
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CertDescribeFormat(CSSM_CL_HANDLE CLHandle,
                           uint32 *NumberOfOids,
                           CSSM_OID_PTR *OidList)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CertDescribeFormat);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CertDescribeFormat))
        {
            ERR(err = CallBack->CertDescribeFormat(CLHandle,
                                                   NumberOfOids,
                                                   OidList));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlCreateTemplate
 *
 *Description:
 *   This function creates an unsigned, memory-resident CRL. 
 *   Fields in the CRL are initialized with the descriptive 
 *   data specified by the OID/value input pairs. 
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in 
 *                     certificate library module used to perform 
 *                     this function.
 *
 *  CrlTemplate (input) - An array of OID/value pairs specifying 
 *                        the initial values for descriptive data 
 *                        fields of the new CRL.
 *
 *  NumberOfFields (input) - The number of OID/value pairs specified 
 *                           in the CrlTemplate input parameter.         
 *
 *  NewCrl (output) - A pointer to the CSSM_DATA structure containing 
 *                    the new CRL. The CRL is allocated by the service 
 *                    provider and must be deallocated by the application.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlCreateTemplate(CSSM_CL_HANDLE CLHandle,
                          uint32 NumberOfFields,
                          const CSSM_FIELD *CrlTemplate,
                          CSSM_DATA_PTR NewCrl)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlCreateTemplate);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlCreateTemplate))
        {
            ERR(err = CallBack->CrlCreateTemplate(CLHandle,
                                                  NumberOfFields,
                                                  CrlTemplate,
                                                  NewCrl));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlSetFields
 *
 *Description:
 *   This function will set the fields of the input CRL to the new 
 *   values, specified by the input OID/value pairs
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  CrlTemplate (input) - Any array of field OID/value pairs containing 
 *                        the values to initialize the CRL attribute fields.
 *
 *  NumberOfFields (input) - The number of OID/value pairs specified in 
 *                           the CrlTemplate input parameter.
 *
 *  OldCrl (input) - The CRL to be updated with the new attribute values. 
 *                   The CRL must be unsigned and available for update.         
 *
 *  ModifiedCrl (output) - A pointer to the modified, unsigned CRL. The 
 *                         CRL is allocated by the service provider and must 
 *                         be deallocated by the application.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlSetFields(CSSM_CL_HANDLE CLHandle,
                     uint32 NumberOfFields,
                     const CSSM_FIELD *CrlTemplate,
                     const CSSM_DATA *OldCrl,
                     CSSM_DATA_PTR ModifiedCrl)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlSetFields);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlSetFields))
        {
            ERR(err = CallBack->CrlSetFields(CLHandle,
                                             NumberOfFields,
                                             CrlTemplate,
                                             OldCrl,
                                             ModifiedCrl));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlAddCert
 *
 *Description:
 *   This function revokes the input certificate by adding a record 
 *   representing the certificate to the CRL
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  CCHandle (input) - The handle that describes the context of this 
 *                     cryptographic operation.
 *
 *  Cert (input) - A pointer to the CSSM_DATA structure containing the 
 *                 certificate to be revoked.
 *
 *  RevokerCert (input) - A pointer to the CSSM_DATA structure containing 
 *                        the revoker's certificate.
 *
 *  CrlEntryFields (input) - An array of OID/value pairs specifying the 
 *                           initial values for descriptive data fields 
 *                           of the new CRL entry.
 *
 *  NumberOfFields (input) - The number of OID/value pairs specified in 
 *                           the CrlEntryFields input parameter.
 *
 *  OldCrl (input) - A pointer to the CSSM_DATA structure containing the 
 *                   CRL to which the newly-revoked certificate will be added.         
 *
 *  NewCrl (output) - A pointer to the CSSM_DATA structure containing the updated 
 *                    CRL. The CRL is allocated by the service provider and 
 *                    must be deallocated by the application
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CrlAddCert(CSSM_CL_HANDLE CLHandle,
                   CSSM_CC_HANDLE CCHandle,
                   const CSSM_DATA *Cert,
                   uint32 NumberOfFields,
                   const CSSM_FIELD *CrlEntryFields,
                   const CSSM_DATA *OldCrl,
                   CSSM_DATA_PTR NewCrl)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlAddCert);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlAddCert))
        {
            ERR(err = CallBack->CrlAddCert(CLHandle,
                                           CCHandle,
                                           Cert,
                                           NumberOfFields,
                                           CrlEntryFields,
                                           OldCrl,
                                           NewCrl));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlRemoveCert
 *
 *Description:
 *   This function reinstates a certificate by removing it from 
 *   the specified CRL
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  Cert (input) - A pointer to the CSSM_DATA structure containing the 
 *                 certificate to be reinstated.
 *
 *  OldCrl (input) - A pointer to the CSSM_DATA structure containing 
 *                   the CRL from which the certificate is to be removed.         
 *
 *  NewCrl (output) - A pointer to the CSSM_DATA structure containing 
 *                    the updated CRL. The CRL is allocated by the service 
 *                    provider and must be deallocated by the application.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlRemoveCert(CSSM_CL_HANDLE CLHandle,
                      const CSSM_DATA *Cert,
                      const CSSM_DATA *OldCrl,
                      CSSM_DATA_PTR NewCrl)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlRemoveCert);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlRemoveCert))
        {
            ERR(err = CallBack->CrlRemoveCert(CLHandle,
                                              Cert,
                                              OldCrl,
                                              NewCrl));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlSign
 *
 *Description:
 *   This function signs, in accordance with the specified cryptographic 
 *   context, the fields of the CRL indicated in the SignScope parameter
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  CCHandle (input) - The handle that describes the context of this 
 *                     cryptographic operation.
 *
 *  UnsignedCrl (input) - A pointer to the CSSM_DATA structure containing 
 *                        the CRL to be signed.
 *
 *  SignerCert (input) - A pointer to the CSSM_DATA structure containing 
 *                       the certificate to be used to sign the CRL.
 *
 *  SignScope (input/optional) - A pointer to the CSSM_FIELD array containing 
 *                               the tag/value pairs of the fields to be signed.  
 *                               If the signing scope is null, the certificate 
 *                               library module includes a default set of CRL 
 *                               fields in the signing process.
 *
 *  ScopeSize (input) - The number of entries in the sign scope list. 
 *                      If the signing scope is not specified, the input 
 *                      scope size must be zero.         
 *
 *  SignedCrl (output) - A pointer to the CSSM_DATA structure containing 
 *                       the signed CRL. The CRL is allocated by the service 
 *                       provider and must be deallocated by the application
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlSign(CSSM_CL_HANDLE CLHandle,
                CSSM_CC_HANDLE CCHandle,
                const CSSM_DATA *UnsignedCrl,
                const CSSM_FIELD *SignScope,
                uint32 ScopeSize,
                CSSM_DATA_PTR SignedCrl)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlSign);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlSign))
        {
            ERR(err = CallBack->CrlSign(CLHandle,
                                        CCHandle,
                                        UnsignedCrl,
                                        SignScope,
                                        ScopeSize,
                                        SignedCrl));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}








/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlVerify
 *
 *Description:
 *   This function verifies that the signed CRL has not been altered 
 *   since it was signed by the designated signer
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  CCHandle (input) - The handle that describes the context of this 
 *                     cryptographic operation.
 *
 *  CrlToBeVerified (input) - A pointer to the CSSM_DATA structure containing 
 *                            the CRL to be verified.
 *
 *  SignerCert (input) - A pointer to the CSSM_DATA structure containing the 
 *                       certificate used to sign the CRL.
 *
 *  VerifyScope (input/optional) - A pointer to the CSSM_FIELD array containing 
 *                                 the tag/value pairs of the fields to be verified.  
 *                                 If the verification scope is null, the certificate 
 *                                 library module assumes that a default set of 
 *                                 fields were used in the signing process and 
 *                                 those same fields are used in the verification 
 *                                 process.
 *
 *  ScopeSize (input) - The number of entries in the verify scope list. 
 *                      If the verification scope is not specified, the 
 *                      input value for scope size must be zero. 
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlVerify(CSSM_CL_HANDLE CLHandle,
                  CSSM_CC_HANDLE CCHandle,
                  const CSSM_DATA *CrlToBeVerified,
                  const CSSM_DATA *SignerCert,
                  const CSSM_FIELD *VerifyScope,
                  uint32 ScopeSize)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlVerify);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlVerify))
        {
            ERR(err = CallBack->CrlVerify(CLHandle,
                                          CCHandle,
                                          CrlToBeVerified,
                                          SignerCert,
                                          VerifyScope,
                                          ScopeSize));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlVerifyWithKey
 *
 *Description:
 *  This function verifies that the CrlToBeVerified was signed using 
 *  a specific private key and that the certificate revocation list 
 *  has not been altered since it was signed using that private key
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the certificate library 
 *                     service module used to perform this function.
 *
 *  CCHandle (input) - A signature verification context defining the CSP, 
 *                     verification algorithm, and public key that must be 
 *                     used to perform the operation.
 *
 *  CrlToBeVerified (input) - A signed certificate revocation list whose 
 *                            signature is to be verified
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlVerifyWithKey(CSSM_CL_HANDLE CLHandle,
                         CSSM_CC_HANDLE CCHandle,
                         const CSSM_DATA *CrlToBeVerified)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlVerifyWithKey);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlVerifyWithKey))
        {
            ERR(err = CallBack->CrlVerifyWithKey(CLHandle,
                                                 CCHandle,
                                                 CrlToBeVerified));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_IsCertInCrl
 *
 *Description:
 *   This function searches the CRL for a record corresponding 
 *   to the certificate
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in 
 *                     certificate library module used to perform 
 *                     this function.
 *
 *  Cert (input) - A pointer to the CSSM_DATA structure containing 
 *                 the certificate to be located.
 *
 *
 *  Crl (input) - A pointer to the CSSM_DATA structure containing 
 *                the CRL to be searched. 
 *
 *  CertFound - A pointer indicating yes or no status of the certificate 
 *              to be found
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_IsCertInCrl(CSSM_CL_HANDLE CLHandle,
                    const CSSM_DATA *Cert,
                    const CSSM_DATA *Crl,
                    CSSM_BOOL *CertFound)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_IsCertInCrl);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->IsCertInCrl))
        {
            ERR(err = CallBack->IsCertInCrl(CLHandle,
                                            Cert,
                                            Crl,
                                            CertFound));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlGetFirstFieldValue
 *
 *Description:
 *   This function returns the value of the CRL field designated 
 *   by the CSSM_OID CrlField
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in 
 *                     certificate library module used to 
 *                     perform this function.
 *
 *  Crl (input) - A pointer to the CSSM_DATA structure which contains 
 *                the CRL from which the field is to be retrieved.
 *
 *  CrlField (input) - An object identifier which identifies the 
 *                     field value to be extracted from the Crl.
 *
 *  ResultsHandle (output) - A pointer to the CSSM_HANDLE which 
 *                           should be used to obtain any additional 
 *                           matching fields.
 *
 *  NumberOfMatchedFields (output) - The total number of fields that 
 *                                   match the CrlField OID. This count 
 *                                   includes the first match, which was 
 *                                   returned by this function.
 *
 *  Value (output) - Returns a pointer to a CSSM_DATA structure containing 
 *                   the first field that matched the CrlField. The 
 *                   field value is allocated by the service provider 
 *                   and must be deallocated by the application.
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlGetFirstFieldValue(CSSM_CL_HANDLE CLHandle,
                              const CSSM_DATA *Crl,
                              const CSSM_OID *CrlField,
                              CSSM_HANDLE_PTR ResultsHandle,
                              uint32 *NumberOfMatchedFields,
                              CSSM_DATA_PTR *Value)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlGetFirstFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlGetFirstFieldValue))
        {
            ERR(err = CallBack->CrlGetFirstFieldValue(CLHandle,
                                                      Crl,
                                                      CrlField,
                                                      ResultsHandle,
                                                      NumberOfMatchedFields,
                                                      Value));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlGetNextFieldValue
 *
 *Description:
 *   This function returns the value of a CRL field, when that 
 *   field occurs multiple times in a CRL
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  ResultsHandle (input) - The handle that identifies the results 
 *                          of a CRL query.         
 *
 *  Value (output) - Returns a pointer to a CSSM_DATA structure 
 *                   containing the first field that matched the 
 *                   CrlField specified to the CSSM_CL_CrlGetFirstFieldValue 
 *                   function. The field value is allocated by the 
 *                   service provider and must be deallocated by 
 *                   the application.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlGetNextFieldValue(CSSM_CL_HANDLE CLHandle,
                             CSSM_HANDLE ResultsHandle,
                             CSSM_DATA_PTR *Value)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlGetNextFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlGetNextFieldValue))
        {
            ERR(err = CallBack->CrlGetNextFieldValue(CLHandle,
                                                     ResultsHandle,
                                                     Value));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlAbortQuery
 *
 *Description:
 *  This function terminates the query initiated by 
 *  CSSM_CL_CrlGetFirstFieldValue or CSSM_CL_CrlGetFirstCachedFieldValue 
 *  and allows the CL to release all intermediate state information associated 
 *  with the repeating-value get operationl.  Once this function has been 
 *  invoked the results handle is invalid.
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  ResultsHandle (input) - The handle which identifies the results of 
 *                          a CRL query. 
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlAbortQuery(CSSM_CL_HANDLE CLHandle,
                      CSSM_HANDLE ResultsHandle)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlAbortQuery);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlAbortQuery))
        {
            ERR(err = CallBack->CrlAbortQuery(CLHandle,
                                              ResultsHandle));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlGetAllFields
 *
 *Description:
 *   This function returns one or more structures each containing 
 *   a set of  field values from the encoded, packed CRL contained 
 *   in Crl
 *
 *Parameters: 
 *  CLHandle  (input) - The handle that describes the add-in certificate 
 *                      library module used to perform this function.
 *
 *  Crl  (input) - A pointer to the CSSM_DATA structure that contains 
 *                 the encoded, packed, CRL from which field values 
 *                 are to be extracted.
 *
 *  CrlFields  (output) - A pointer to an array of OID-value pairs 
 *                        that describe the contents of the CRL. 
 *                        The extracted CRL fields are returned as 
 *                        the value portion of each OID-value pair.
 *
 *  NumberOfCrlFields (output) - The number of entries in the array 
 *                               CrlFields
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlGetAllFields(CSSM_CL_HANDLE CLHandle,
                        const CSSM_DATA *Crl,
                        uint32 *NumberOfCrlFields,
                        CSSM_FIELD_PTR *CrlFields)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlGetAllFields);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlGetAllFields))
        {
            ERR(err = CallBack->CrlGetAllFields(CLHandle,
                                                Crl,
                                                NumberOfCrlFields,
                                                CrlFields));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*
/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlCache
 *
 *Description:
 *   This function caches a copy of a CertificateRevocationList 
 *   (CRL) for subsequent accesses using the functions 
 *   CSSM_CL_CrlGetFirstFieldValue and CSSM_CL_CrlGetNextFieldValue
 *
 *Parameters: 
 *  CLHandle  (input) - The handle that describes the certificate 
 *                      library module used to perform this function.
 *
 *  Crl  (input) - A pointer to the CSSM_DATA structure containing 
 *                 the encoded CRL.
 *
 *  CrlHandle  (output) - A pointer to the CSSM_HANDLE that should 
 *                        be used in all future references to the 
 *                        cached CRL.
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlCache(CSSM_CL_HANDLE CLHandle,
                 const CSSM_DATA *Crl,
                 CSSM_HANDLE_PTR CrlHandle)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlCache);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlCache))
        {
            ERR(err = CallBack->CrlCache(CLHandle,
                                         Crl,
                                         CrlHandle));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_IsCertInCachedCrl
 *
 *Description:
 *   This function searches the cached CRL for a record corresponding 
 *   to the certificate
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  Cert (input) - A pointer to the CSSM_DATA structure containing 
 *                 an encoded, packed certificate.
 *
 *  CrlHandle (input) - A handle identifying a CRL that the application 
 *                      has temporarily cached with the Certificate Library 
 *                      module. The referenced CRL is searched for a 
 *                      revocation record matching the specified Cert. 
 *
 *  CertFound (output) - A pointer to a CSSM_BOOL indicating success 
 *                       or failure in finding the specified certificate 
 *                       in the CRL. CSSM_TRUE signifies that the certificate 
 *                       was found in the CRL. CSSM_FALSE indicates that the 
 *                       certificate was not found in the CRL.
 *
 *  CrlRecordIndex (output/optional) - A pointer to a CSSM_DATA structure 
 *                                     containing an index descriptor for 
 *                                     direct access to the located CRL record. 
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_IsCertInCachedCrl(CSSM_CL_HANDLE CLHandle,
                          const CSSM_DATA *Cert,
                          CSSM_HANDLE CrlHandle,
                          CSSM_BOOL *CertFound,
                          CSSM_DATA_PTR CrlRecordIndex)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_IsCertInCachedCrl);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->IsCertInCachedCrl))
        {
            ERR(err = CallBack->IsCertInCachedCrl(CLHandle,
                                                  Cert,
                                                  CrlHandle,
                                                  CertFound,
                                                  CrlRecordIndex));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlGetFirstCachedFieldValue
 *
 *Description:
 *  This function returns a single structure containing a set of 
 *  field values from the cached CRL identified by CrlHandle
 *
 *Parameters: 
 *  CLHandle  (input) - The handle that describes the add-in 
 *                      certificate library module used to perform 
 *                      this function.
 *
 *  CrlHandle (input) -  A handle identifying a CRL that the 
 *                       application has temporarily cached with 
 *                       the Certificate Library module. The 
 *                       referenced CRL is searched for the field 
 *                       values identified by CrlField. 
 *
 *  CrlRecordIndex (input/optional) -  An index value identifying 
 *                                     a particular revocation record 
 *                                     in a cached CRL. If an index value 
 *                                     is supplied, the scan for the field 
 *                                     values identified by CrlField is 
 *                                     limited to the pre-selected revocation 
 *                                     record.
 *
 *  CrlField  (input) - A pointer to an object identifier that identifies 
 *                      the field value to be extracted from the Crl.
 *
 *  ResultsHandle  (output) - A pointer to the CSSM_HANDLE, which should 
 *                            be used to obtain any additional matching fields.
 *
 *  NumberOfMatchedFields  (output) - The total number of fields that match 
 *                                    the CrlField OID. This count includes 
 *                                    the first match, which was returned 
 *                                    by this function.
 *
 *  FieldValue  (output) - A pointer to the CSSM_DATA structure containing 
 *                         the first field value that matched the requested 
 *                         field.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CrlGetFirstCachedFieldValue(CSSM_CL_HANDLE CLHandle,
                                    CSSM_HANDLE CrlHandle,
                                    const CSSM_DATA *CrlRecordIndex,
                                    const CSSM_OID *CrlField,
                                    CSSM_HANDLE_PTR ResultsHandle,
                                    uint32 *NumberOfMatchedFields,
                                    CSSM_DATA_PTR *FieldValue)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlGetFirstCachedFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlGetFirstCachedFieldValue))
        {
            ERR(err = CallBack->CrlGetFirstCachedFieldValue(CLHandle,
                                                            CrlHandle,
                                                            CrlRecordIndex,
                                                            CrlField,
                                                            ResultsHandle,
                                                            NumberOfMatchedFields,
                                                            FieldValue));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlGetNextCachedFieldValue
 *
 *Description:
 *  This function returns the value of a CRL field, when that 
 *  field occurs multiple times in a CRL
 *
 *Parameters: 
 *  CLHandle  (input) - The handle that describes the add-in 
 *                      certificate library module used to perform 
 *                      this function.
 *
 *  ResultsHandle  (input) - The handle that identifies the results 
 *                           of a CRL query. 
 *
 *  FieldValue  (output) - A pointer to the CSSM_DATA structure 
 *                         containing the next field value that 
 *                         matched the requested field specified 
 *                         in the CL_CrlGetFirstCachedFieldValue 
 *                         function. 
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlGetNextCachedFieldValue(CSSM_CL_HANDLE CLHandle,
                                   CSSM_HANDLE ResultsHandle,
                                   CSSM_DATA_PTR *FieldValue)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlGetNextCachedFieldValue);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlGetNextCachedFieldValue))
        {
            ERR(err = CallBack->CrlGetNextCachedFieldValue(CLHandle,
                                                           ResultsHandle,
                                                           FieldValue));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlGetAllCachedRecordFields
 *
 *Description:
 *   This function returns all field values from a pre-located, 
 *   cached CRL record. The scanned CRL record is identified by 
 *   CrlRecordIndex, which is returned by the function 
 *   CSSM_CL_IsCertInCachedCrl.
 *
 *Parameters: 
 *  CLHandle  (input) - The handle that describes the add-in 
 *                      certificate library module used to perform 
 *                      this function.
 *
 *  CrlHandle (input) -  A handle identifying a CRL that the 
 *                       application has temporarily cached with the 
 *                       Certificate Library module. The referenced 
 *                       CRL must contain the CRL record identified 
 *                       by CrlRecordIndex. 
 *
 *  CrlRecordIndex (input) -  An index value identifying a particular 
 *                            revocation record in a cached CRL.
 *
 *  NumberOfFields  (output) - The number of OID-value pairs returned 
 *                             by this function. 
 *
 *  Fields  (output) - A pointer to an array of CSSM_FIELD structures, 
 *                     describing the contents of the pre-selected CRL 
 *                     record using OID-value pairs.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/

CSSM_RETURN CSSMAPI
CSSM_CL_CrlGetAllCachedRecordFields(CSSM_CL_HANDLE CLHandle,
                                    CSSM_HANDLE CrlHandle,
                                    const CSSM_DATA *CrlRecordIndex,
                                    uint32 *NumberOfFields,
                                    CSSM_FIELD_PTR *Fields)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlGetAllCachedRecordFields);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlGetAllCachedRecordFields))
        {
            ERR(err = CallBack->CrlGetAllCachedRecordFields(CLHandle,
                                                            CrlHandle,
                                                            CrlRecordIndex,
                                                            NumberOfFields,
                                                            Fields));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}




/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlAbortCache
 *
 *Description:
 *   This function terminates a CRL cache handle created and returned 
 *   by the function CSSM_CL_CrlCache
 *
 *Parameters: 
 *  CLHandle  (input) - The handle that describes the certificate 
 *                      library module used to perform this function.
 *
 *  CrlHandle  (input) - The handle that identifies the cached CRL. 
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.  
 *
 *----------------------------------------------------------------*/


CSSM_RETURN CSSMAPI
CSSM_CL_CrlAbortCache(CSSM_CL_HANDLE CLHandle,
                      CSSM_HANDLE CrlHandle)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlAbortCache);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlAbortCache))
        {
            ERR(err = CallBack->CrlAbortCache(CLHandle,
                                              CrlHandle));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_CrlDescribeFormat
 *
 *Description:
 *  This function returns a list of the CSSM_OID values this 
 *  certificate library module uses to name and reference fields 
 *  of a CRL
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 *
 *  NumberOfFields (output) - The length of the returned array of OIDs.         
 *
 *  OidList (output) - A pointer to the array of CSSM_OIDs which 
 *                     represent the supported CRL format. 
 *                     The OID list is allocated by the service provider 
 *                     and must be deallocated by the application.
 *
 *
 *Returns:
 *   A CSSM_RETURN value indicating success or specifying a particular 
 *   error condition. The value CSSM_OK indicates success. 
 *   All other values represent an error condition.
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_CrlDescribeFormat(CSSM_CL_HANDLE CLHandle,
                          uint32 *NumberOfOids,
                          CSSM_OID_PTR *OidList)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_CrlDescribeFormat);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->CrlDescribeFormat))
        {
            ERR(err = CallBack->CrlDescribeFormat(CLHandle,
                                                  NumberOfOids,
                                                  OidList));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



/*---------------------------------------------------------------
 *
 *Name: CSSM_CL_PassThrough
 *
 *Description:
 *   This function allows applications to call certificate library 
 *   module-specific operations
 *
 *Parameters: 
 *  CLHandle (input) - The handle that describes the add-in certificate 
 *                     library module used to perform this function.
 * 
 *  CCHandle (input/optional) - The handle that describes the context 
 *                              of the cryptographic operation. If 
 *                              the module-specific operation does not 
 *                              perform any cryptographic operations a 
 *                              cryptographic context is not required.
 *
 *  PassThroughId (input) - An identifier assigned by the CL module to 
 *                          indicate the exported function to perform.
 *
 *  InputParams (input) - A pointer to a module implementation-specific 
 *                        structure containing parameters to be interpreted 
 *                        in a function-specific manner by the requested CL
 *                        module.         
 *
 *  OutputParams (output) - A pointer to a module, implementation-specific 
 *                          structure containing the output data. The 
 *                          service provider will allocate the memory for 
 *                          this structure. The application should free
 *                          the memory for the structure.
 *
 *
 *Returns:
 *  NULL - unable to complete pass through call
 *  non NULL - result from pass through call
 *
 *----------------------------------------------------------------*/
CSSM_RETURN CSSMAPI
CSSM_CL_PassThrough(CSSM_CL_HANDLE CLHandle,
                    CSSM_CC_HANDLE CCHandle,
                    uint32 PassThroughId,
                    const void *InputParams,
                    void **OutputParams)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_PassThrough);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->PassThrough))
        {
            ERR(err = CallBack->PassThrough(CLHandle,
                                            CCHandle,
                                            PassThroughId,
                                            InputParams,
                                            OutputParams));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}



CSSM_RETURN CSSMAPI
CSSM_CL_FreeFields(CSSM_CL_HANDLE CLHandle,
                   uint32 NumberOfFields,
                   CSSM_FIELD_PTR FieldArray)
{
    CSSM_SPI_CL_FUNCS_PTR CallBack = NULL;
    cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord = NULL;
    CSSM_RETURN err = CSSM_OK;

#ifdef ISL_INCLUDED
    /* now we get the return address */
    ISL_FUNCTION_PTR retAddress = NULL;
    EISL_GetReturnAddress(retAddress);
#else
    CSSM_PROC_ADDR retAddress = NULL;   /* dummy for parameter passing */
#endif

    TRACE(cssm_trace_info, FNCODE_CSSM_CL_FreeFields);

    /* check the caller and do other setup for this SPI call */
    ERR(err = cssm_SetupStandardSPICall((CSSM_PROC_ADDR)retAddress,
                                        CLHandle,
                                        CSSM_SERVICE_CL,
                                        (void**)(&CallBack),
                                        &AttachRecord));
    if(CSSM_OK == err)
    {
        /*
         * Call the callback function. 
         * Make sure that function pointer is valid.
         */
        if(CSSM_FALSE == port_IsBadCodePtr((CSSM_PROC_ADDR)CallBack->FreeFields))
        {
            ERR(err = CallBack->FreeFields(CLHandle,
                                           NumberOfFields,
                                           FieldArray));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_FUNCTION_NOT_IMPLEMENTED);
        }

        /* cleanup from the SPI call */
        ERR(cssm_CleanupStandardSPICall(AttachRecord));
    }

    RETURN (err)
}
