/*-----------------------------------------------------------------------
 * File: KEYMGMT.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

#ifdef VMS
#include <string.h>
#endif
#include "precomp.h"

#ifdef ISL_INCLUDED


/*---------------------------------------------------------------
 *
 *Name: cssm_VerifyCaller
 *
 *Description:Verifies the Caller's manifest based on the keytype and returns
 * the verifiedmoduleptr
 *
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/


CSSM_RETURN
cssm_VerifyCaller(uint32 KeyType,
                  uint32* NumAppSections, 
                  ISL_VERIFIED_MODULE_PTR** VerifiedAppPtrArr,
                  const CSSM_GUID* CallerGuid)
{
    ISL_MANIFEST_SECTION_PTR* AppManifestSectionPtrArr = NULL;
    CSSM_API_MEMORY_FUNCS fx;
    CSSM_RETURN err = CSSM_OK;

    *NumAppSections = 0;
    *VerifiedAppPtrArr = NULL;

    fx.malloc_func = internal_malloc;
    fx.calloc_func = internal_calloc;
    fx.free_func = internal_free;
    fx.realloc_func = internal_realloc;
    fx.AllocRef = NULL;


    if(NULL != CallerGuid)
    {
        CSSM_DATA tmpCredential;
        CSSM_DATA tmpPath;
        ISL_CONST_DATA AppCredential;
        ISL_CONST_DATA AppSearchPath;

        ERR(err = mdsutil_GetModuleCredentialInfo(CallerGuid,
                                                  &tmpCredential,
                                                  NULL,
                                                  &tmpPath));
        if(CSSM_OK == err)
        {
            AppCredential.Length = tmpCredential.Length;
            AppCredential.Data = tmpCredential.Data;
            AppSearchPath.Length = tmpPath.Length;
            AppSearchPath.Data =  tmpPath.Data;

            AppManifestSectionPtrArr = cssm_VerifyCredentials(AppCredential,
                                                              AppSearchPath,
                                                              KeyType,
                                                              NumAppSections,
                                                              &fx);

            local_free ((char *)AppCredential.Data, NULL);   
            local_free ((char *)AppSearchPath.Data, NULL);

            if(NULL != AppManifestSectionPtrArr)
            {
                *VerifiedAppPtrArr = cssm_VerifyLoadedOrLoadAllModules(AppManifestSectionPtrArr,
                                                                       NumAppSections,
                                                                       CSSM_FALSE,
                                                                       &fx);
                if( *VerifiedAppPtrArr == NULL)
                {
                    cssm_FreeManifestSections(AppManifestSectionPtrArr,
                                              *NumAppSections);
                    ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
                }

                local_free( AppManifestSectionPtrArr, NULL);

            }
            else
            {
                ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
            }
        }
        else
        {
            ERR(err = CSSMERR_CSSM_MDS_ERROR);
        }
    }
    else
    {
        ERR(err = CSSMERR_CSSM_INVALID_POINTER);
    }

    return err;
}
/*---------------------------------------------------------------
 *
 *Name: cssm_UseeMatchForCallerAndSP
 *
 *Description:Checks if the requested usee is in both caller's and SP's manifest
 *
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/


CSSM_BOOL cssm_UseeMatchForCallerAndSP( uint32 usee,
                                       cssm_CALLER_LIST_ENTRY_PTR CallerListEntry,
                                       cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord)
{
    CSSM_BOOL bSPUseeFound = CSSM_FALSE;
    CSSM_BOOL bCallerUseeFound = CSSM_FALSE;
    uint32 i = 0;
    if( usee != 0 )
    {
        while ( ( !bSPUseeFound ) && ( i < ServiceRecord->NumSPTags ) )
        {
            if( usee != ServiceRecord->SPTags[i] )
                i++;
            else
                bSPUseeFound = CSSM_TRUE;
        }
        i = 0;
        while ( ( !bCallerUseeFound ) && ( i < CallerListEntry->NumUSEETags ) )
        {
            if( usee != CallerListEntry->USEETags[i] )
                i++;
            else
                bCallerUseeFound = CSSM_TRUE;
        }
        if( bSPUseeFound && bCallerUseeFound )
            RETURN (CSSM_TRUE)
        else
            RETURN (CSSM_FALSE)
    }
    RETURN (CSSM_TRUE)

}
/*---------------------------------------------------------------
 *
 *Name: cssm_UseeMatchForCssmAndSP
 *
 *Description:Checks if the requested usee is in both cssm's and SP's manifest
 *
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/


CSSM_BOOL cssm_UseeMatchForCssmAndSP( uint32 usee,
                                   cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord)
{
    CSSM_BOOL bSPUseeFound = CSSM_FALSE;
    CSSM_BOOL bCssmUseeFound = CSSM_FALSE;
    uint32 i = 0;
    if( usee != 0 )
    {
        while ( ( !bSPUseeFound ) && ( i < ServiceRecord->NumSPTags ) )
        {
            if( usee != ServiceRecord->SPTags[i] )
                i++;
            else
                bSPUseeFound = CSSM_TRUE;
        }
        i = 0;
        while ( ( !bCssmUseeFound ) && ( i < cssm_NumUseetags ) )
        {
            if( usee != cssm_UseeTagList[i] )
                i++;
            else
                bCssmUseeFound = CSSM_TRUE;
        }
        if( bSPUseeFound && bCssmUseeFound )
            RETURN (CSSM_TRUE)
        else
            RETURN (CSSM_FALSE)
    }
    RETURN (CSSM_TRUE)

}

/*---------------------------------------------------------------
 *
 *Name: cssm_VendorPrivMatchForCallerAndSP
 *
 *Description:Checks if the requested vendor privilege is in both caller's and SP's manifest
 *
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/


CSSM_BOOL cssm_VendorPrivMatchForCallerAndSP( uint32 VendorPriv,
                                       cssm_CALLER_LIST_ENTRY_PTR CallerListEntry,
                                       cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord)
{
    CSSM_BOOL bSPVendorFound = CSSM_FALSE;
    CSSM_BOOL bCallerVendorFound = CSSM_FALSE;
    uint32 i = 0;
    while ( ( !bSPVendorFound ) && ( i < AttachRecord->NumVendorPrivs ) )
    {
        if( VendorPriv != AttachRecord->VendorPrivs[i] )
            i++;
        else
            bSPVendorFound = CSSM_TRUE;
    }
    i = 0;
    if( CallerListEntry->KeyHierarchy & AttachRecord->KeyHierarchy )
    {
        cssm_VENDOR_PRIVILEGES_PTR tempPtr = CallerListEntry->VendorPrivs;
        CSSM_BOOL bKHFound = CSSM_FALSE;
        while ( tempPtr != NULL )
        {
            if( tempPtr->KeyHierarchy & AttachRecord->KeyHierarchy )
            {
                bKHFound = CSSM_TRUE;
                break;
            }
            else
                tempPtr = tempPtr->Next;
        }
        if( bKHFound )
        {
            while ( ( !bCallerVendorFound ) && ( i < tempPtr->NumVendorPrivs ) )
            {
                if( VendorPriv != tempPtr->VendorPrivs[i] )
                    i++;
                else
                    bCallerVendorFound = CSSM_TRUE;
            }
        }
    }
    if( bSPVendorFound && bCallerVendorFound )
        RETURN (CSSM_TRUE)
    else
        RETURN (CSSM_FALSE)

}

/*---------------------------------------------------------------
 *
 *Name: cssm_VendorPrivMatchForCssmAndSP
 *
 *Description:Checks if the requested vendor privilege is in both cssm's and SP's manifest
 *
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_BOOL cssm_VendorPrivMatchForCssmAndSP( uint32 VendorPriv,
                                       cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord)
{
    CSSM_BOOL bSPVendorFound = CSSM_FALSE;
    CSSM_BOOL bCssmVendorFound = CSSM_FALSE;
    uint32 i = 0;
    cssm_VENDOR_PRIVILEGES_PTR tempptr = NULL;
    while ( ( !bSPVendorFound ) && ( i < AttachRecord->NumVendorPrivs ) )
    {
        if( VendorPriv != AttachRecord->VendorPrivs[i] )
            i++;
        else
            bSPVendorFound = CSSM_TRUE;
    }
    /* cssm manifest's vendor privilege is in the global variable cssm_CssmVendorPrivs,
       which gets filled up in CSSM_INit*/
    tempptr = cssm_CssmVendorPrivs;
    while( tempptr != NULL )
    {
        if( tempptr->KeyHierarchy & AttachRecord->KeyHierarchy )
            break;
        else
            tempptr = tempptr->Next;
    }
    if( tempptr != NULL)
    {
        i = 0;
        while ( ( !bCssmVendorFound ) && i < (tempptr->NumVendorPrivs ) )
        {
            if( VendorPriv != tempptr->VendorPrivs[i] )
                i++;
            else
                bCssmVendorFound = CSSM_TRUE;
        }
    }
    if( bSPVendorFound && bCssmVendorFound )
        RETURN (CSSM_TRUE)
    else
        RETURN (CSSM_FALSE)

}
/*---------------------------------------------------------------
 *
 *Name: cssm_TOGPrivMatchForCssmAndSP
 *
 *Description:Checks if the requested TOG privilege is in both cssm's and SP's manifest
 *
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_BOOL cssm_TOGPrivMatchForCssmAndSP( uint32 TOGPriv,
                                       cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord)
{
    CSSM_BOOL bSPTOGFound = CSSM_FALSE;
    CSSM_BOOL bCssmTOGFound = CSSM_FALSE;
    uint32 i = 0;
    if( TOGPriv < CSSM_TOGMASK )
    {
        while ( ( !bSPTOGFound ) && ( i < ServiceRecord->NumTOGPrivs ) )
        {
            if( TOGPriv != ServiceRecord->TOGPrivs[i] )
                i++;
            else
                bSPTOGFound = CSSM_TRUE;
        }
        i = 0;
        while ( ( !bCssmTOGFound ) && ( i < cssm_NumCssmTOGPrivs ) )
        {
            if( TOGPriv != cssm_CssmTOGPrivs[i] )
                i++;
            else
                bCssmTOGFound = CSSM_TRUE;
        }
    }
    if( bSPTOGFound && bCssmTOGFound )
        RETURN (CSSM_TRUE)
    else
        RETURN (CSSM_FALSE)

}

/*---------------------------------------------------------------
 *
 *Name: cssm_TOGPrivMatchForCallerAndSP
 *
 *Description:Checks if the requested TOG privilege is in both caller's and SP's manifest
 *
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/


CSSM_BOOL cssm_TOGPrivMatchForCallerAndSP( uint32 TOGPriv,
                                       cssm_CALLER_LIST_ENTRY_PTR CallerListEntry,
                                       cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord)
{
    CSSM_BOOL bSPTOGFound = CSSM_FALSE;
    CSSM_BOOL bCallerTOGFound = CSSM_FALSE;
    uint32 i = 0;
    if( TOGPriv < CSSM_TOGMASK )
    {
        while ( ( !bSPTOGFound ) && ( i < ServiceRecord->NumTOGPrivs ) )
        {
            if( TOGPriv != ServiceRecord->TOGPrivs[i] )
                i++;
            else
                bSPTOGFound = CSSM_TRUE;
        }
        i = 0;
        while ( ( !bCallerTOGFound ) && ( i < CallerListEntry->NumTOGPrivs ) )
        {
            if( TOGPriv != CallerListEntry->TOGPrivs[i] )
                i++;
            else
                bCallerTOGFound = CSSM_TRUE;
        }
    }
    if( bSPTOGFound && bCallerTOGFound )
        RETURN (CSSM_TRUE)
    else
        RETURN (CSSM_FALSE)

}
/*---------------------------------------------------------------
 *
 *Name: cssm_PrivilegeMatchForCallerAndSP
 *
 *Description:Checks if the requested privilege is in both caller's and SP's manifest
 *
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_BOOL cssm_PrivilegeMatchForCallerAndSP( CSSM_PRIVILEGE Privilege,
                                             cssm_CALLER_LIST_ENTRY_PTR CallerListEntry,
                                             cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
                                             cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord)
{
    /* Split the CSSM_PRIVILEGE into USEE, TOG and Vendor Privileges*
    /* check if each of the above privileges match for the caller and SP*/
    /*Vendor and TOG privilege applies to  Integrity Keyhierarchy*/

    uint32 reqUseeTag = (CSSM_UINT64_HIGH(Privilege)) & CSSM_USEE_MASK;
    uint32 reqTOGPriv = (CSSM_UINT64_LOW(Privilege)) & CSSM_TOGVENDOR_MASK;
    CSSM_BOOL bMatch = CSSM_FALSE;

    if( AttachRecord->KeyHierarchy & CSSM_KEY_HIERARCHY_EXPORT )
    {
        if( CallerListEntry != NULL)
        {
        if( cssm_UseeMatchForCallerAndSP( reqUseeTag, CallerListEntry, ServiceRecord ) )
            bMatch = CSSM_TRUE;
        }
        else
        {
            if( cssm_UseeMatchForCssmAndSP ( reqUseeTag, ServiceRecord ) )
                bMatch = CSSM_TRUE;
        }
    }
    else if( ( AttachRecord->KeyHierarchy & CSSM_KEY_HIERARCHY_INTEG ) ||
           ( AttachRecord->KeyHierarchy == CSSM_KEY_HIERARCHY_NONE ) )
    {
        if( CallerListEntry != NULL)
        {
            if( ( cssm_TOGPrivMatchForCallerAndSP( reqTOGPriv, CallerListEntry,ServiceRecord)) &&
                ( cssm_VendorPrivMatchForCallerAndSP( reqTOGPriv, CallerListEntry, AttachRecord)) )
                bMatch = CSSM_TRUE;
        }
        else
        {
            if( ( cssm_TOGPrivMatchForCssmAndSP( reqTOGPriv, ServiceRecord)) &&
                ( cssm_VendorPrivMatchForCssmAndSP( reqTOGPriv, AttachRecord)) )
                bMatch = CSSM_TRUE;
        }
    }

    RETURN (bMatch)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_ExtractPrivilege
 *
 *Description: From the manifest signer info attribute header, retrieves the  privilege
               by name and returns in the Privilege pointer
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_RETURN cssm_ExtractPrivilege( ISL_MANIFEST_SECTION_PTR ManifestSectionptr,
                                 uint32 *NumPriv,
                                 uint32** Privilege,
                                 ISL_CONST_DATA AttribName)
{
    CSSM_RETURN ret = CSSM_OK;
    ISL_CONST_DATA AttribValue;
    ISL_STATUS Status = ISL_OK;
     /*
     * Signer info header attribute API's
     */
     Status = EISL_FindSignerInfoAttribute(EISL_GetManifestSignatureRoot(ManifestSectionptr),
                                                                        AttribName,
                                                                        &AttribValue);
     if(Status == ISL_OK)
     {
         ERR( ret = GetPrivTagInCredentialAttribute( &AttribValue,
                                                    NumPriv,
                                                    Privilege));
     }
     else 
     {
        *NumPriv = 0;
        *Privilege = NULL;
     }

     RETURN (ret)
}
/*---------------------------------------------------------------
 *
 *Name: cssm_VerifyManifestSections
 *
 *Description: Given the module guid, manifest in buffer and search path,
               verifies the manifest sections based on the key hierarchy and returns
               the verified manifest sections and number of sections.
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/


CSSM_RETURN
cssm_VerifyManifestSections(CSSM_GUID_PTR Guid,
                            ISL_MANIFEST_SECTION_PTR** ManifestSectionPtrArr,
                            uint32* NumSections,
                            CSSM_KEY_HIERARCHY KeyHierarchy,
                            CSSM_DATA_PTR Manifest,
                            CSSM_DATA_PTR Path)
{
    ISL_CONST_DATA Credential;
    ISL_CONST_DATA ModuleSearchPath;
    CSSM_API_MEMORY_FUNCS fx;

    fx.malloc_func = internal_malloc;
    fx.calloc_func = internal_calloc;
    fx.free_func = internal_free;
    fx.realloc_func = internal_realloc;
    fx.AllocRef = NULL;

    
    Credential.Length = Manifest->Length;
    Credential.Data = Manifest->Data;
    ModuleSearchPath.Length = Path->Length;
    ModuleSearchPath.Data =  Path->Data;
    if( (KeyHierarchy & CSSM_KEY_HIERARCHY_INTEG) || (KeyHierarchy == CSSM_KEY_HIERARCHY_NONE ) )
    {
        *ManifestSectionPtrArr = cssm_VerifyCredentials(Credential,
                                                        ModuleSearchPath,
                                                        CSSM_KEY_HIERARCHY_INTEG,
                                                        NumSections,
                                                        &fx);
    }
    else if(KeyHierarchy & CSSM_KEY_HIERARCHY_EXPORT)
    {
        *ManifestSectionPtrArr = cssm_VerifyCredentials(Credential, 
            ModuleSearchPath, 
            CSSM_KEY_HIERARCHY_EXPORT,
            NumSections, 
            &fx);
    }
    if(*ManifestSectionPtrArr == NULL)
    {
        RETURN (CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED)
    }
    RETURN (CSSM_OK)
}


/*---------------------------------------------------------------
 *
 *Name: cssm_FilterTOGAndVendorPrivs
 *
 *Description: Given an array of privileges, seperates into tog and vendor privileges
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_RETURN cssm_FilterTOGAndVendorPrivs( uint32 NumPrivs, uint32* Privs, 
                              uint32* NumVendorPrivs, uint32** VendorPrivs,
                              uint32* NumTOGPrivs, uint32** TOGPrivs)
{
    CSSM_RETURN err = CSSM_OK;
    uint32* temppriv = Privs;
    uint32 i =0 ;

    *VendorPrivs = NULL;
    *NumVendorPrivs = 0;
    *NumTOGPrivs = 0;
    *TOGPrivs = NULL;
    if( NumPrivs != 0 && Privs != NULL)
    {
        while( i< NumPrivs)
        {
            if( *temppriv & CSSM_TOGMASK )
            {
                if(0 == *NumVendorPrivs)
                {
                    *VendorPrivs = local_malloc(sizeof(uint32),
                                          NULL);
                }
                else
                {
                    *VendorPrivs = local_realloc(*VendorPrivs,
                                           sizeof(uint32) * (*NumVendorPrivs + 1),
                                           NULL);
                }
                if( *VendorPrivs == NULL)
                    RETURN ( CSSMERR_CSSM_MEMORY_ERROR )
                memcpy(&((*VendorPrivs)[*NumVendorPrivs]),
                       temppriv,
                       sizeof(uint32));
                (*NumVendorPrivs)++;
            }
            else
            {
                if(0 == *NumTOGPrivs)
                {
                    *TOGPrivs = local_malloc(sizeof(uint32),
                                              NULL);
                }
                else
                {
                    *TOGPrivs = local_realloc(*TOGPrivs,
                                               sizeof(uint32) * (*NumTOGPrivs + 1),
                                               NULL);
                }
                if( *TOGPrivs == NULL)
                    RETURN ( CSSMERR_CSSM_MEMORY_ERROR )

                memcpy(&((*TOGPrivs)[*NumTOGPrivs]),
                           temppriv,
                           sizeof(uint32));
                (*NumTOGPrivs)++;
            }
            temppriv++;
            i++;
        }
    }
    RETURN (err)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_GetVendorPrivs
 *
 *Description: Given an array of cssm_vendor_privileges, 
               extracts the vendor privileges based on key hierarchy.
               csmm_vendor_privileges have vendor privileges for multiple key hierarchies
 *Parameters: 
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_RETURN cssm_GetVendorPrivs( CSSM_KEY_HIERARCHY KeyHierarchy, 
                                cssm_VENDOR_PRIVILEGES_PTR VendorPrivilegeptr,
                                uint32* NumVendorPrivs,
                                uint32** VendorPrivs)
{
    CSSM_RETURN ret = CSSM_OK;
    cssm_VENDOR_PRIVILEGES_PTR pTemppriv = NULL;
    CSSM_BOOL bFound = CSSM_FALSE;
    uint32* tempvendorpriv = NULL;

    if( (VendorPrivilegeptr == NULL) || 
        ( NumVendorPrivs == NULL) || 
        (VendorPrivs == NULL) )
        ret = CSSMERR_CSSM_INVALID_POINTER;
    else
    {
        pTemppriv = VendorPrivilegeptr;

        while( pTemppriv != NULL )
        {
            if( KeyHierarchy == pTemppriv->KeyHierarchy )
            {
                bFound = CSSM_TRUE;
                break;
            }
            pTemppriv = pTemppriv->Next;
        }
        if( bFound )
        {
            *NumVendorPrivs = pTemppriv->NumVendorPrivs;
            
            if(0 < pTemppriv->NumVendorPrivs)
            {
                tempvendorpriv = local_malloc(sizeof( uint32) * pTemppriv->NumVendorPrivs,
                                              NULL);
                if( tempvendorpriv == NULL )
                {
                    *NumVendorPrivs = 0;
                    ERR(ret = CSSMERR_CSSM_MEMORY_ERROR);
                }
                else
                {
                    memcpy(tempvendorpriv,
                           pTemppriv->VendorPrivs,
                           sizeof(uint32) * pTemppriv->NumVendorPrivs );
                    *VendorPrivs = tempvendorpriv;
                }
            }
            else
            {
                *VendorPrivs = NULL;
            }
        }
        else
        {
            ERR(ret = CSSMERR_CSSM_INTERNAL_ERROR);
        }
    }
    RETURN ( ret )
}


/*---------------------------------------------------------------
 *
 *Name: cssm_AddPrivs
 *
 *Description: Adds the vendor privilege into a service record's vendore privilege
               list based on key hierarchy.
 *
 *Parameters: 
 *
 *Returns:
 *
 *-------------------------------------------------------------*/

CSSM_RETURN cssm_AddPrivs(
                      cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord,
                      uint32 NumVendorPrivs,
                      uint32* VendorPrivs,
                      CSSM_KEY_HIERARCHY KeyHierarchy)
{
    cssm_VENDOR_PRIVILEGES_PTR pTempPriv = NULL;
    CSSM_RETURN ret = CSSM_OK;
    cssm_VENDOR_PRIVILEGES_PTR pTempInsert = NULL;
    CSSM_BOOL bFound = CSSM_FALSE;
    uint32* newVendorprivs = NULL;
    /* Validate input */
    if(NULL == VendorPrivs) 
    {
        RETURN (CSSMERR_CSSM_INVALID_POINTER)
    }
    pTempInsert = ServiceRecord->VendorPrivs;

    while( pTempInsert != NULL)
    {
        if( pTempInsert->KeyHierarchy == KeyHierarchy )
        {
            bFound = CSSM_TRUE;
            break;
        }
        else if( pTempInsert->Next == NULL)
            break;
        else
            pTempInsert = pTempInsert->Next;
    }
    if( bFound )
        RETURN ( ret ) 

    
    /* Allocate a CSSM_VENDOR_PRIVILEGE record */
    pTempPriv = (cssm_VENDOR_PRIVILEGES_PTR)
                        local_calloc((uint32)sizeof(cssm_VENDOR_PRIVILEGES),
                                     1,
                                     NULL);
    if ( pTempPriv == NULL )
    {
        ret = CSSMERR_CSSM_MEMORY_ERROR;
    }
    else
    {
        port_memset( pTempPriv, 0, sizeof(cssm_VENDOR_PRIVILEGES) );

        pTempPriv->NumVendorPrivs = NumVendorPrivs;

        if(0 != NumVendorPrivs)
        {
            newVendorprivs = local_malloc((uint32)sizeof(uint32) * NumVendorPrivs, NULL);

            if ( newVendorprivs == NULL )
            {
                ret = CSSMERR_CSSM_MEMORY_ERROR;
            }

            memcpy( newVendorprivs, VendorPrivs, sizeof(uint32)*NumVendorPrivs);
        }

        pTempPriv->VendorPrivs = newVendorprivs;

        pTempPriv->KeyHierarchy = KeyHierarchy;
        
        if( pTempInsert == NULL )
        {
            ServiceRecord->VendorPrivs = pTempPriv;
        }
        else
        {
            pTempInsert->Next = pTempPriv;
        }
    }
    RETURN (ret)
}


/*---------------------------------------------------------------
 *
 *Name: cssm_AddInLoadUsingKeyHierarchy
 *
 *Description:This function gets called when the ServiceRecord->SPCredential is NULL 
              or when there is an attach with a key hierarchy which is not done before. 
              If this is the first attach then we do a verification of the module using
              the key hierarchy and save the result in the ServiceRecord. 
              But if the ServiceRecord->SPCredential is not NULL, then we do a verification
              using the new key hierarchy and recycle the verified module ptr without saving it.
 *
 *Parameters: 
 *
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_RETURN cssm_AddInLoadUsingKeyHierarchy(const CSSM_GUID* ModuleGuid,
                              CSSM_KEY_HIERARCHY KeyHierarchy,
                              cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord,
                              cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
                              CSSM_DATA_PTR Manifest,
                              CSSM_DATA_PTR ModulePath ) 
{
    ISL_MANIFEST_SECTION_PTR* ManifestSectionPtrArr = NULL;
    uint32 NumManifestSections = 0;
    CSSM_RETURN ret = CSSM_OK; 
    CSSM_API_MEMORY_FUNCS fx;
    ISL_CONST_DATA AttributeName;
    const char *szUSEEAttribName = "CDSA_USEE";
    const char *szTOGVENDORPrivAttribName = "CDSA_PRIV";
    uint32 NumPrivs = 0;
    uint32* Privs = NULL;
    uint32 NumVendorPrivs = 0;
    uint32* VendorPrivs = NULL;
    uint32 NumTOGPrivs = 0;
    uint32* TOGPrivs = NULL;
    ISL_VERIFIED_MODULE_PTR* tempVerifiedModulePtrArr = NULL;
    uint32 tempNumVerifiedSections = 0;

    
    fx.malloc_func = internal_malloc;
    fx.calloc_func = internal_calloc;
    fx.free_func = internal_free;
    fx.realloc_func = internal_realloc;
    fx.AllocRef = NULL;

    ret  = cssm_VerifyManifestSections(
                             (CSSM_GUID*)ModuleGuid,
                             &ManifestSectionPtrArr,
                             &NumManifestSections,
                             KeyHierarchy,
                             Manifest,
                             ModulePath);
    if( (ret == CSSM_OK) && ( ManifestSectionPtrArr != NULL)) 
    {
        AttributeName.Length = (uint32)strlen(szUSEEAttribName);
        AttributeName.Data = (uint8*)szUSEEAttribName;
        if( ServiceRecord->NumSPTags == 0 && ServiceRecord->SPTags == NULL)
        {
            ERR( ret = cssm_ExtractPrivilege( *ManifestSectionPtrArr,
                                   &ServiceRecord->NumSPTags,
                                   &ServiceRecord->SPTags,
                                   AttributeName));
        }
        if( ret ==CSSM_OK )
        {
            AttributeName.Length = (uint32)strlen(szTOGVENDORPrivAttribName);
            AttributeName.Data = (uint8*)szTOGVENDORPrivAttribName;
            /* Extract the CDSA_PRIV and filter into TOG and Vendor Privs.*/
            ERR( ret = cssm_ExtractPrivilege( *ManifestSectionPtrArr,
                                   &NumPrivs,
                                   &Privs,
                                   AttributeName));
            if( ret == CSSM_OK )
            {
                ERR( ret = cssm_FilterTOGAndVendorPrivs( NumPrivs, Privs, 
                                                         &NumVendorPrivs, &VendorPrivs,
                                                         &NumTOGPrivs, &TOGPrivs));

                if( NumTOGPrivs != 0 && TOGPrivs != NULL)
                {
                    ServiceRecord->NumTOGPrivs = NumTOGPrivs;
                    ServiceRecord->TOGPrivs = TOGPrivs;
                }
                if( NumVendorPrivs != 0 && VendorPrivs != NULL)
                {
                    ret = cssm_AddPrivs( ServiceRecord, 
                                         NumVendorPrivs,
                                         VendorPrivs,
                                         KeyHierarchy);
                    AttachRecord->NumVendorPrivs = NumVendorPrivs;
                    AttachRecord->VendorPrivs = VendorPrivs;
                }
                if( ServiceRecord->bDiffPrimaryManifest )
                    tempVerifiedModulePtrArr = cssm_VerifyLoadedOrLoadAllModules( ManifestSectionPtrArr,
                                                                        &NumManifestSections,CSSM_TRUE,
                                                                        &fx);
                else
                    tempVerifiedModulePtrArr = cssm_VerifyLoadedOrLoadAllModules( ManifestSectionPtrArr,
                                                                        &NumManifestSections,CSSM_FALSE,
                                                                        &fx);
                if(tempVerifiedModulePtrArr == NULL)
                {
                    cssm_FreeManifestSections( ManifestSectionPtrArr,  NumManifestSections);
                    RETURN ( CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED)
                }


                ServiceRecord->KeyHierarchy |= KeyHierarchy;
                tempNumVerifiedSections = NumManifestSections;
                if( ServiceRecord->SPCredentialPtr == NULL )
                {
                    ServiceRecord->SPCredentialPtr = tempVerifiedModulePtrArr;
                    ServiceRecord->NumSPCreds = tempNumVerifiedSections;
                }
                else
                {
                    cssm_FreeVerifiedCredentials( tempVerifiedModulePtrArr, tempNumVerifiedSections );
                    local_free( tempVerifiedModulePtrArr, NULL);
                }
            }
        }

        local_free(ManifestSectionPtrArr, NULL);
    }
    
    RETURN ( ret )
}
/*---------------------------------------------------------------
 *
 *Name: cssm_AttachModuleUsingExportOrIntegrityKeys
 *
 *Description:
 *Parameters: 
 *
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_RETURN cssm_AttachModuleUsingExportOrIntegrityKeys(const CSSM_GUID *ModuleGuid,
                                            const CSSM_VERSION *Version,
                                            uint32 SubServiceId,
                                            CSSM_SERVICE_TYPE SubServiceType,
                                            CSSM_ATTACH_FLAGS AttachFlags,
                                            CSSM_MODULE_HANDLE ModuleHandle,
                                            CSSM_KEY_HIERARCHY KeyHierarchy,
                                            CSSM_MODULE_FUNCS_PTR *FuncTbl,
                                            cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord,
                                            cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
                                            const CSSM_GUID *CallerGuid,
                                            CSSM_DATA_PTR Manifest,
                                            CSSM_DATA_PTR ModulePath)
{
    SPI_ATTACH_FUNC_PTR pfnModuleAttach;
    char szModuleAttach[] = "CSSM_SPI_ModuleAttach";
    ISL_CONST_DATA ProcedureName;
    CSSM_MODULE_FUNCS_PTR TempFuncTable;
    CSSM_RETURN ret = CSSM_OK;
    uint32 NumFuncs;
    uint32 i;
    CSSM_GUID EMMGuid;
    CSSM_BOOL bFound = CSSM_FALSE;

    ERR( ret = cssm_AddInLoadUsingKeyHierarchy( ModuleGuid,
                                                KeyHierarchy,
                                                ServiceRecord,
                                                AttachRecord,
                                                Manifest,
                                                ModulePath));
    if( ret == CSSM_OK)
    {
        /* Find the address of the AddinAuthenticate function in the addin module*/
        ProcedureName.Length = (uint32)strlen(szModuleAttach);
        ProcedureName.Data = (const unsigned char*)szModuleAttach;

        i = 0;
        pfnModuleAttach = NULL;

        while((NULL == pfnModuleAttach) &&
              (i < ServiceRecord->NumSPCreds))
        {
            pfnModuleAttach = (SPI_ATTACH_FUNC_PTR)EISL_LocateProcedureAddress(ServiceRecord->SPCredentialPtr[i],
                                                                               ProcedureName);
            i++;
        }

        if(pfnModuleAttach == NULL)
        {
            cssm_FreeVerifiedCredentials( ServiceRecord->SPCredentialPtr, ServiceRecord->NumSPCreds);
            local_free( ServiceRecord->SPCredentialPtr, NULL);
            RETURN(CSSMERR_CSSM_ADDIN_AUTHENTICATE_FAILED)
        }
        i=0;
        bFound = CSSM_FALSE;
        while( i< ServiceRecord->NumSPCreds )
        {
            /* Verifies the AddInAuthenticate function is within the AddinModule*/
            if (EISL_CheckAddressWithinModule(ServiceRecord->SPCredentialPtr[i],
                                             (ISL_FUNCTION_PTR)pfnModuleAttach) == ISL_OK)
            {
                bFound = CSSM_TRUE;
                break;
            }
            i++;
        }
        if( !bFound )
        {
            cssm_FreeVerifiedCredentials( ServiceRecord->SPCredentialPtr, ServiceRecord->NumSPCreds);
            local_free( ServiceRecord->SPCredentialPtr, NULL);

            RETURN(CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL)
        }
        if( AttachRecord->pEMMRecord != NULL)
        {
            memcpy( &EMMGuid, &(AttachRecord->pEMMRecord->EMMGuid), sizeof(CSSM_GUID));
        }
        else
        {
            memcpy(&EMMGuid, &cssm30_GUID, sizeof(CSSM_GUID));
        }
        /* Calls AddinAuthenticate function with cssm credential path and cssm section name*/
        if( (ret = pfnModuleAttach(ModuleGuid,
                           Version,
                           SubServiceId,
                           SubServiceType,
                           AttachFlags,
                           ModuleHandle,
                           KeyHierarchy,
                           &cssm30_GUID,
                           &EMMGuid,  
                           CallerGuid,
                           &InternalUpcalls,
                           &TempFuncTable)) != CSSM_OK)
        {
            RETURN(ret)
        }
        else
        {
            if((CSSM_FALSE == port_IsBadReadPtr(TempFuncTable,
                                                sizeof(CSSM_MODULE_FUNCS))) &&
               (CSSM_FALSE == port_IsBadReadPtr(TempFuncTable->ServiceFuncs,
                                                (TempFuncTable->NumberOfServiceFuncs * sizeof(CSSM_PROC_ADDR)))))
            {
                NumFuncs = TempFuncTable->NumberOfServiceFuncs;


                *FuncTbl = (CSSM_MODULE_FUNCS_PTR)local_calloc(sizeof(CSSM_MODULE_FUNCS),
                                                              1,
                                                              NULL);
                if(NULL != *FuncTbl)
                {
                    (*FuncTbl)->ServiceType = TempFuncTable->ServiceType;
                    (*FuncTbl)->NumberOfServiceFuncs = NumFuncs;

                    if(0 != NumFuncs)
                    {
                        (*FuncTbl)->ServiceFuncs = local_calloc((sizeof(CSSM_PROC_ADDR) * NumFuncs),
                                                               1,
                                                               NULL);
                        if(NULL != (*FuncTbl)->ServiceFuncs)
                        {
                            memcpy((void *)(*FuncTbl)->ServiceFuncs,	/* ifdef VMS apply cast */
                                   TempFuncTable->ServiceFuncs,
                                   (sizeof(CSSM_PROC_ADDR) * NumFuncs));
                        }
                        else
                        {
                            ERR(ret = CSSMERR_CSSM_MEMORY_ERROR);
                            local_free((*FuncTbl),
                                       NULL);
                            *FuncTbl = NULL;
                        }
                    }
                }
                else
                {
                    ERR(ret = CSSMERR_CSSM_MEMORY_ERROR);
                }
            }
            else
            {
                ERR(ret = CSSMERR_CSSM_INVALID_ADDIN_FUNCTION_TABLE);
            }
        }
    }
    RETURN(ret) 
}



/*---------------------------------------------------------------
 *
 *Name: cssm_EMMLoadAndCheckFromGUID
 *
 *Description:
 *   Verify and load the EMM using keytype based on guid of the emm
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_RETURN
cssm_EMMLoadAndCheckFromGUID(const CSSM_GUID *ModuleGuid,
                             CSSM_KEY_HIERARCHY KeyType,
                             CSSM_BOOL LoadModule,
                             uint32 *NumEmmSections,
                             ISL_VERIFIED_MODULE_PTR** pEmmVerifiedModule)
{
    CSSM_RETURN err = CSSM_OK;
    ISL_CONST_DATA Credential;
    ISL_CONST_DATA ModuleSearchPath;
    CSSM_API_MEMORY_FUNCS fx;
    CSSM_DATA tmpCredential;
    CSSM_DATA tmpPath;
    ISL_MANIFEST_SECTION_PTR* ManifestSectionPtrArr = NULL;
    *NumEmmSections = 0;
    fx.malloc_func = internal_malloc;
    fx.calloc_func = internal_calloc;
    fx.free_func = internal_free;
    fx.realloc_func = internal_realloc;
    fx.AllocRef = NULL;
    
    
    if(ERR(err = mdsutil_GetModuleCredentialInfo(ModuleGuid,
                                                 &tmpCredential,
                                                 NULL,
                                                 &tmpPath)) != CSSM_OK)
    {
        ERR(err = CSSMERR_CSSM_MDS_ERROR);
    }
    else
    {
        Credential.Length = tmpCredential.Length;
        Credential.Data = tmpCredential.Data;
        ModuleSearchPath.Length = tmpPath.Length;
        ModuleSearchPath.Data =  tmpPath.Data;
    
        ManifestSectionPtrArr = cssm_VerifyCredentials(Credential,
                                                       ModuleSearchPath,
                                                       KeyType,
                                                       NumEmmSections,
                                                       &fx);
        local_free ((char *)Credential.Data, NULL);
        local_free ((char *)ModuleSearchPath.Data, NULL);

        if(ManifestSectionPtrArr == NULL)
        {
            ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
        }
        else
        {
            *pEmmVerifiedModule = cssm_VerifyLoadedOrLoadAllModules(ManifestSectionPtrArr,
                                                                    NumEmmSections,
                                                                    LoadModule,
                                                                    &fx);
            if(*pEmmVerifiedModule == NULL)
            {
                cssm_FreeManifestSections(ManifestSectionPtrArr,
                                          *NumEmmSections);
                local_free( ManifestSectionPtrArr, NULL );
                *pEmmVerifiedModule = NULL;
                *NumEmmSections = 0;
                ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
            }

            local_free( ManifestSectionPtrArr, NULL );

        }
    }

    RETURN (err)
}

/*---------------------------------------------------------------
 *
 *Name: cssm_EMMLoadAndCheckFromGUID
 *
 *Description:
 *   Verify and load the EMM using keytype
 *Returns:
 *
 *----------------------------------------------------------------*/

CSSM_RETURN
cssm_LoadEMMUsingExportOrIntegrityKeys(CSSM_DATA EMMCredential,
                                       CSSM_DATA EMMSearchPath,
                                       CSSM_GUID_PTR EmmGuid,
                                       const CSSM_GUID *CallerGuid,
                                       CSSM_KEY_HIERARCHY KeyHierarchy,
                                       uint32* NumEmmSections,
                                       ISL_VERIFIED_MODULE_PTR** VerifiedModulePtrArr,
                                       CSSM_MANAGER_REGISTRATION_INFO_PTR ManagerInfo)
{
    CSSM_RETURN err = CSSM_OK;
    uint32 i;


    ERR(err = cssm_EMMLoadAndCheckFromCredAndPath(EMMCredential,
                                                  EMMSearchPath,
                                                  KeyHierarchy,
                                                  CSSM_TRUE,
                                                  NumEmmSections,
                                                  VerifiedModulePtrArr));

    if(CSSM_OK == err)
    {
        ERR(err = cssm_IdentifyVerifiedModuleByGUID(*NumEmmSections,
                                                    *VerifiedModulePtrArr,
                                                    EmmGuid,
                                                    &i));


        if(CSSM_OK == err)
        {
            ERR(err = cssm_EmmManagerAuthenticate(KeyHierarchy,
                                                  CallerGuid,
                                                  (*VerifiedModulePtrArr)[i],
                                                  ManagerInfo));
        }
        else
        {
            ERR(err = CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED);
        }

        if(CSSM_OK != err)
        {
            if( *VerifiedModulePtrArr )
            {
                for( i =0; i< *NumEmmSections; i++)
                {
                    port_FreeLibrary( EISL_GetLibHandle( (*VerifiedModulePtrArr)[i]));
                }
                ERR(cssm_FreeVerifiedCredentials(*VerifiedModulePtrArr,
                                                 *NumEmmSections));
                local_free( *VerifiedModulePtrArr, NULL);
                *NumEmmSections = 0;
                *VerifiedModulePtrArr = NULL;
            }
        }
    }

    return err;
}

#endif

CSSM_RETURN
cssm_AttachModuleNoKeys(const CSSM_GUID *ModuleGuid,
                        const CSSM_VERSION *Version,
                        uint32 SubServiceId,
                        CSSM_SERVICE_TYPE SubServiceType,
                        CSSM_ATTACH_FLAGS AttachFlags,
                        CSSM_MODULE_HANDLE ModuleHandle,
                        CSSM_KEY_HIERARCHY KeyHierarchy,
                        const CSSM_GUID *CallerGuid,                
                        CSSM_MODULE_FUNCS_PTR *AttachFuncTbl,
                        cssm_INTERNAL_ATTACH_LIST_PTR AttachRecord,
                        cssm_INTERNAL_SERVICE_LIST_PTR ServiceRecord)
{
    CSSM_RETURN ret = CSSM_OK; 
    SPI_ATTACH_FUNC_PTR pfnModuleAttach = NULL;
    uint32 NumFuncs;
    char szModuleAttach[] = "CSSM_SPI_ModuleAttach";
    CSSM_MODULE_FUNCS_PTR TempFuncTable;
    uint32 NativeServices = 0;
    CSSM_GUID ManagerGuid;
#ifdef ISL_INCLUDED
    ISL_CONST_DATA ProcedureName;
    uint32 i;
#else
    cssm_INTERNAL_MODULE_LIST_PTR ModuleRecord;
    CSSM_LIB_HANDLE LibHandle;
#endif

#ifdef ISL_INCLUDED
    if( ServiceRecord->SPCredentialPtr != NULL)
    {
        /* Find the address of the AddinAuthenticate function in the addin module*/
        ProcedureName.Length = (uint32)strlen(szModuleAttach);
        ProcedureName.Data = (const unsigned char*)szModuleAttach;
        for( i=0; i < ServiceRecord->NumSPCreds; i++)
        {
            pfnModuleAttach = (SPI_ATTACH_FUNC_PTR)EISL_LocateProcedureAddress(ServiceRecord->SPCredentialPtr[i],
                                                                         ProcedureName);
            if( pfnModuleAttach != NULL )
            {
                /* Verifies the AddInAuthenticate function is within the AddinModule*/
                if (EISL_CheckAddressWithinModule(ServiceRecord->SPCredentialPtr[i],
                                                 (ISL_FUNCTION_PTR)pfnModuleAttach) != ISL_OK)
                {
                    RETURN(CSSMERR_CSSM_FUNCTION_INTEGRITY_FAIL)
                }
                break;
            }
            
        }

    }
    else
    {
        RETURN(CSSMERR_CSSM_ADDIN_AUTHENTICATE_FAILED)
    }

#else

    if((ret = cssm_FindModuleAndMultiLock(ModuleHandle,
                                         CSSM_NO_LOCK,
                                         &ModuleRecord,
                                         CSSM_READER_LOCK)) == CSSM_OK)
    {
        LibHandle = ModuleRecord->LibHandle;
        ret = port_GetProcAddress(LibHandle, szModuleAttach, (CSSM_PROC_ADDR*)&pfnModuleAttach);

        /* Release the module record now - we still have the attach record
           locked, so we know no one can unload this on us.
        */
        cssm_ReleaseModuleLock(ModuleRecord,
                               CSSM_READER_LOCK);
    }
    else
    {
        RETURN(ret)
    }
#endif  /* !ISL_INCLUDED */

    if(pfnModuleAttach == NULL)
    {
        RETURN(CSSMERR_CSSM_ADDIN_AUTHENTICATE_FAILED)
    }
    if( (ret = cssm_GetCSSMNativeServices(&NativeServices)) != CSSM_OK )
    {
        RETURN ( CSSMERR_CSSM_MDS_ERROR );
    }
    if( NativeServices & SubServiceType )
        memcpy( &ManagerGuid, &cssm30_GUID, sizeof(CSSM_GUID) );
    else
        memcpy( &ManagerGuid, &AttachRecord->pEMMRecord->EMMGuid, sizeof( CSSM_GUID ));
    /* Calls AddinAuthenticate function */
    ERR( ret = pfnModuleAttach(ModuleGuid,
                       Version,
                       SubServiceId,
                       SubServiceType,
                       AttachFlags,
                       ModuleHandle,
                       KeyHierarchy,
                       &cssm30_GUID,
                       &ManagerGuid, 
                       CallerGuid,
                       &InternalUpcalls,
                       &TempFuncTable)); 
    if( ret != CSSM_OK )
    {
        RETURN(ret)
    }
    else
    {
        if((CSSM_FALSE == port_IsBadReadPtr(TempFuncTable,
                                            sizeof(CSSM_MODULE_FUNCS))) &&
           (CSSM_FALSE == port_IsBadReadPtr(TempFuncTable->ServiceFuncs,
                                            (TempFuncTable->NumberOfServiceFuncs * sizeof(CSSM_PROC_ADDR)))))
        {
            NumFuncs = TempFuncTable->NumberOfServiceFuncs;

            if(0 != NumFuncs)
            {

                *AttachFuncTbl = (CSSM_MODULE_FUNCS_PTR)local_calloc(sizeof(CSSM_MODULE_FUNCS),
                                                                    1,
                                                                    NULL);
                if(NULL != *AttachFuncTbl)
                {
                    (*AttachFuncTbl)->ServiceType = TempFuncTable->ServiceType;
                    (*AttachFuncTbl)->NumberOfServiceFuncs = NumFuncs;

                    (*AttachFuncTbl)->ServiceFuncs = local_calloc((sizeof(CSSM_PROC_ADDR) * NumFuncs),
                                                                  1,
                                                                  NULL);
                    if(NULL != (*AttachFuncTbl)->ServiceFuncs)
                    {
                        memcpy((void *)(*AttachFuncTbl)->ServiceFuncs,  /* #ifdef VMS apply cast */
                               TempFuncTable->ServiceFuncs,
                               (sizeof(CSSM_PROC_ADDR) * NumFuncs));
                    }
                    else
                    {
                        ERR(ret = CSSMERR_CSSM_MEMORY_ERROR);
                        local_free((*AttachFuncTbl),
                                       NULL);
                        *AttachFuncTbl = NULL;
                    }
                }
                else
                {
                    ERR(ret = CSSMERR_CSSM_MEMORY_ERROR);
                }
            }
        }
        else
        {
            RETURN(CSSMERR_CSSM_INVALID_ADDIN_FUNCTION_TABLE)
        }
    }

    RETURN(CSSM_OK)
}

#ifdef ISL_INCLUDED
/*---------------------------------------------------------------
 *
 *Name: cssm_VerifyLoadedCSSMUsingExportRoot
 *
 *Description:
 *   Verifies the loaded cssm module using the export root
 *Parameters: 
 *
 *----------------------------------------------------------------*/

CSSM_RETURN cssm_VerifyLoadedCSSMUsingExportRoot(CSSM_DATA_PTR ModuleCredential, 
                                                CSSM_DATA_PTR SearchPath,
                                                ISL_VERIFIED_MODULE_PTR** VerifiedModulePtrArr,
                                                uint32* NumSections)
{
    ISL_MANIFEST_SECTION_PTR* ManifestSectionPtrArr = NULL;
    CSSM_API_MEMORY_FUNCS fx;
    ISL_CONST_DATA Credential;
    ISL_CONST_DATA ModuleSearchPath;
    
    *NumSections = 0;
    *VerifiedModulePtrArr = NULL,
    fx.malloc_func = internal_malloc;
    fx.calloc_func = internal_calloc;
    fx.free_func = internal_free;
    fx.realloc_func = internal_realloc;
    
    Credential.Length = ModuleCredential->Length;
    Credential.Data = ModuleCredential->Data;
    ModuleSearchPath.Length = SearchPath->Length;
    ModuleSearchPath.Data =  SearchPath->Data;

    ManifestSectionPtrArr = cssm_VerifyCredentials(Credential,
                                                   ModuleSearchPath,
                                                   CSSM_KEY_HIERARCHY_EXPORT,
                                                   NumSections,
                                                   &fx);

    if(ManifestSectionPtrArr != NULL) 
    {
        *VerifiedModulePtrArr = cssm_VerifyLoadedOrLoadAllModules( ManifestSectionPtrArr,
                                                            NumSections,CSSM_FALSE,
                                                            &fx);
        if(*VerifiedModulePtrArr == NULL)
        {
            cssm_FreeManifestSections( ManifestSectionPtrArr,  *NumSections);
            local_free( ManifestSectionPtrArr, NULL);
            RETURN ( CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED )
        }

        local_free(ManifestSectionPtrArr, NULL);

        RETURN ( CSSM_OK )
    }
    RETURN ( CSSMERR_CSSM_MODULE_MANIFEST_VERIFY_FAILED )
}
#endif
