/*-----------------------------------------------------------------------
 * File: CDSACSSM.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */


#include <assert.h>
#include <string.h>

#include "cssmtype.h"
#include "cssmerr.h"
#ifdef VMS
#include "cssmport.h"
#else
#include "port/cssmport.h"
#endif
#include "mds_schema.h"

#include "mds.h"
#include "mds_util.h"

#include "convert.h"
#include "internal.h"

#define __ModuleID          ( 0 )
#define __CdsaVersion       ( 1 )
#define __CssmVendor        ( 2 )
#define __Desc              ( 3 )
#define __NativeServices    ( 4 )

typedef struct _mdsu_cdsa_cssm_attr_state
{
    CSSM_DB_ATTRIBUTE_DATA DbRecordAttributes[MDS_CDSADIR_CSSM_NUM_ATTRIBUTES];
    CSSM_DATA AttributeValues[MDS_CDSADIR_CSSM_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintableCDSAVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    /*uint8 PrintableEMMVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintableEMMSpecVersion[_MDSU_PRINTABLE_VERSION_LENGTH];*/
}
_MDSU_CDSA_CSSM_ATTR_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSACSSMAttributes(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaData,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFillValues,
                            void **ppAttrState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_CSSM_SCHEMA *pCSSM = 
                            (const MDSU_CDSA_CSSM_SCHEMA*)pSchemaData;
    _MDSU_CDSA_CSSM_ATTR_STATE *pState = NULL;
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && ppAttrState );
    assert( ( bFillValues == CSSM_FALSE ) ||
            ( ( bFillValues == CSSM_TRUE ) && pSchemaData ) );

    /* Allocate the memory for the attributes */
    pState = MDSU_malloc( pContext, sizeof(_MDSU_CDSA_CSSM_ATTR_STATE) );
    if ( pState == NULL )
    {
        ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
    }
    else
    {
        /* Clear the structure */
        memset( pState, 0, sizeof(_MDSU_CDSA_CSSM_ATTR_STATE) );
        DbRecordAttributes = pState->DbRecordAttributes;

        /* Fill in the record information */
        DbRecord->DataRecordType = MDS_CDSADIR_CSSM_RECORDTYPE;
        DbRecord->SemanticInformation = 0;
        DbRecord->NumberOfAttributes = MDS_CDSADIR_CSSM_NUM_ATTRIBUTES;
        DbRecord->AttributeData = DbRecordAttributes;

        /* Fill in attribute info */
        DbRecordAttributes[__ModuleID].Info = s_MdsAttrInfo_ModuleId;
        DbRecordAttributes[__CdsaVersion].Info = s_MdsAttrInfo_CDSAVersion;
        DbRecordAttributes[__CssmVendor].Info = s_MdsAttrInfo_Vendor;
        DbRecordAttributes[__Desc].Info = s_MdsAttrInfo_Desc;
        DbRecordAttributes[__NativeServices].Info = s_MdsAttrInfo_NativeServices;

        if ( bFillValues == CSSM_TRUE )
        {
            /* Connect the attribute value array to the attribute array */
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = &pState->AttributeValues[i];
                DbRecordAttributes[i].NumberOfValues = 1;
            }

            /* Fill in the record contents */
            /* ModuleID */
            _MDSU_GetPrintableGUID( &pCSSM->ModuleID,
                                    (char*)pState->PrintableGuid );
            DbRecordAttributes[__ModuleID].Value->Data = 
                                                    pState->PrintableGuid;
            DbRecordAttributes[__ModuleID].Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;

            /* Manifest */
            /*(*DbRecordAttributes[__Manifest].Value) = pEMM->Manifest;*/

            /* ModuleName */
            /*DbRecordAttributes[__ModuleName].Value->Length = 
                                        strlen( pEMM->ModuleName ) + 1 ;
            DbRecordAttributes[__ModuleName].Value->Data =
                                                (uint8*)pEMM->ModuleName;*/

            /* Path */
            /*DbRecordAttributes[__Path].Value->Length = 
                                        strlen( pEMM->Path ) + 1 ;
            DbRecordAttributes[__Path].Value->Data =
                                                (uint8*)pEMM->Path;*/

            /* CDSA Version */
            _MDSU_GetPrintableVersion( &pCSSM->CDSAVersion,
                                       (char*)pState->PrintableCDSAVersion );
            DbRecordAttributes[__CdsaVersion].Value->Data = 
                                                    pState->PrintableCDSAVersion;
            DbRecordAttributes[__CdsaVersion].Value->Length = 
                                strlen( (char*)pState->PrintableCDSAVersion ) + 1;

            /* EMM Spec Version */
            /*_MDSU_GetPrintableVersion( &pEMM->EmmSpecVersion,
                                       (char*)pState->PrintableEMMSpecVersion );
            DbRecordAttributes[__EmmSpecVersion].Value->Data = 
                                                    pState->PrintableEMMSpecVersion;
            DbRecordAttributes[__EmmSpecVersion].Value->Length = 
                                strlen( (char*)pState->PrintableEMMSpecVersion ) + 1;*/

            /* CssmVendor */
            DbRecordAttributes[__CssmVendor].Value->Length =
                                            strlen( pCSSM->CssmVendor ) + 1;
            DbRecordAttributes[__CssmVendor].Value->Data =
                                            (uint8*)pCSSM->CssmVendor;

            /* Desc */
            DbRecordAttributes[__Desc].Value->Length = 
                                        strlen( pCSSM->Desc ) + 1 ;
            DbRecordAttributes[__Desc].Value->Data =
                                                (uint8*)pCSSM->Desc;

            /* PolicyStmt */
            /*(*DbRecordAttributes[__PolicyStmt].Value) = pEMM->PolicyStmt;*/

            /* EmmVersion */
            /*_MDSU_GetPrintableVersion( &pEMM->EmmVersion,
                                       (char*)pState->PrintableEMMVersion );
            DbRecordAttributes[__EmmVersion].Value->Data = 
                                                    pState->PrintableEMMVersion;
            DbRecordAttributes[__EmmVersion].Value->Length = 
                                strlen( (char*)pState->PrintableEMMVersion ) + 1;*/



            /* NativeServices */
            DbRecordAttributes[__NativeServices].Value->Length = sizeof(uint32);
            DbRecordAttributes[__NativeServices].Value->Data = (uint8*)&pCSSM->NativeServices;

        } /* bFillValues */
        else
        {
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = NULL;
                DbRecordAttributes[i].NumberOfValues = 0;
            }
        }

        if ( rv != CSSM_OK )
        {
            /* Deallocate memory if there was an error */
            MDSU_free( pContext,
                       pState );
            pState = NULL;
        }
    } /* allocate DbRecordAttributes */

    *ppAttrState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSACSSMAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFreeAll,
                            void *pAttrState )
{
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && pAttrState );

    /* Free the buffers allocated for the attribute values */
    DbRecordAttributes = DbRecord->AttributeData;
    if ( bFreeAll == CSSM_TRUE )
    {
        for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
        {
            if ( DbRecordAttributes[i].Value )
            {
                if ( DbRecordAttributes[i].Value->Data )
                {
                    MDSU_free( pContext,
                               DbRecordAttributes[i].Value->Data );
                }
                MDSU_free( pContext,
                           DbRecordAttributes[i].Value );
            }
        }
    }

    /* Free the state structure */
    MDSU_free( pContext,
               pAttrState );
    DbRecord->AttributeData = NULL;

    return CSSM_OK;
}

typedef struct _mdsu_cdsa_cssm_pred_state
{
    CSSM_SELECTION_PREDICATE Predicates[MDS_CDSADIR_CSSM_NUM_ATTRIBUTES];
    CSSM_DATA PredicateValues[MDS_CDSADIR_CSSM_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintableCDSAVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    //uint8 PrintableEMMVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    //uint8 PrintableEMMSpecVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_CSSM_PRED_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSACSSMPredicate(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaTemplate,
                            MDSU_TEMPLATE_MASK ValidTemplates,
                            CSSM_QUERY *pQuery,
                            void **ppPredState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_CSSM_SCHEMA *pCSSM = 
                            (const MDSU_CDSA_CSSM_SCHEMA*)pSchemaTemplate;
    _MDSU_CDSA_CSSM_PRED_STATE *pState = NULL;
    CSSM_SELECTION_PREDICATE *pPredicates = NULL;
    uint32 uNumPredicates;
    uint32 uIndex = 0;
    uint32 i;

    assert( pContext && 
            pQuery &&
            ppPredState &&
            ( ( pSchemaTemplate && ValidTemplates ) ||
              ( !pSchemaTemplate && !ValidTemplates ) ) );

    /* Count the predicates */
    /*
    // Clear the manifest bit. Searching on manifest is not supported 
    ValidTemplates &= (~MDSU_EMM_MANIFEST);*/
    _MDSU_CountBits( ValidTemplates, uNumPredicates );

    /* Allocate the predicate state and link the value structures */
    if ( uNumPredicates != 0 )
    {
        pState = MDSU_malloc( pContext, sizeof(_MDSU_CDSA_CSSM_PRED_STATE) );
        if ( pState == NULL )
        {
            ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
        }
        else
        {
            memset( pState,
                    0,
                    sizeof(_MDSU_CDSA_CSSM_PRED_STATE) );
            pPredicates = pState->Predicates;

            /* Allocate a single attribute value struct per attribute */
            for ( i = 0; i < uNumPredicates; i++ )
            {
                pPredicates[i].Attribute.Value = &pState->PredicateValues[i];
                pPredicates[i].Attribute.NumberOfValues = 1;
            }
        } /* Allocate and link predicate values */
    } /* uNumPredicates != 0 */

    if ( rv == CSSM_OK )
    {
        /* Fill in the query parameters */
        pQuery->RecordType = MDS_CDSADIR_CSSM_RECORDTYPE;
        pQuery->Conjunctive = CSSM_DB_AND;
        pQuery->NumSelectionPredicates = uNumPredicates;
        pQuery->QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
        pQuery->QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
        pQuery->QueryFlags = 0;
        pQuery->SelectionPredicate = pPredicates;
        for ( uIndex = 0; uIndex < uNumPredicates; uIndex++ )
        {
            pPredicates[uIndex].DbOperator = CSSM_DB_EQUAL;
        }

        /* Fill in the predicate data */
        uIndex = 0;
        if ( ValidTemplates & MDSU_CSSM_MODULE_ID )
        {
            /* ModuleID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleId;
            _MDSU_GetPrintableGUID( &pCSSM->ModuleID,
                                    (char*)pState->PrintableGuid );
            pPredicates[uIndex].Attribute.Value->Data = pState->PrintableGuid;
            pPredicates[uIndex].Attribute.Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;
            uIndex++;
        }
        /*
        if ( ValidTemplates & MDSU_EMM_MODULE_NAME )
        {
            // ModuleName 
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleName;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pEMM->ModuleName ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pEMM->ModuleName;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_PATH )
        {
            // Path 
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Path;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pEMM->Path ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pEMM->Path;
            uIndex++;
        }
        */
        if ( ValidTemplates & MDSU_CSSM_CDSA_VERSION )
        {
            /* CDSA Version */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_CDSAVersion;
            _MDSU_GetPrintableVersion( &pCSSM->CDSAVersion,
                                       (char*)pState->PrintableCDSAVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableCDSAVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableCDSAVersion ) + 1;
            uIndex++;
        }
        /*
        if ( ValidTemplates & MDSU_EMM_EMM_SPEC_VERSION )
        {
            // EMM Spec Version 
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_EMMSpecVersion;
            _MDSU_GetPrintableVersion( &pEMM->EmmSpecVersion,
                                       (char*)pState->PrintableEMMSpecVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableEMMSpecVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableEMMSpecVersion ) + 1;
            uIndex++;
        }
        */
        if ( ValidTemplates & MDSU_CSSM_DESC )
        {
            /* Desc */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Desc;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pCSSM->Desc ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pCSSM->Desc;
            uIndex++;
        }
        /*
        if ( ValidTemplates & MDSU_EMM_POLICY_STMT )
        {
            // PolicyStmt 
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_PolicyStmt;
            (*pPredicates[uIndex].Attribute.Value) = pEMM->PolicyStmt;

            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMM_EMM_VERSION )
        {
            // EmmVersion 
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_EmmVersion;
            _MDSU_GetPrintableVersion( &pEMM->EmmVersion,
                                       (char*)pState->PrintableEMMVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableEMMVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableEMMVersion ) + 1;
            uIndex++;
        }
        */
        if ( ValidTemplates & MDSU_CSSM_VENDOR )
        {
            /* CssmVendor */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Vendor;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pCSSM->CssmVendor ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pCSSM->CssmVendor;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_CSSM_NATIVE_SERVICES )
        {
            /* NativeServices - a ServiceMask */
            pPredicates[uIndex].Attribute.Info = 
                                                s_MdsAttrInfo_NativeServices;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pCSSM->NativeServices;
            uIndex++;
        }

    } /* Initial allocation successful */

    *ppPredState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSACSSMPredicate(
                            MDSU_CONTEXT *pContext,
                            CSSM_QUERY *pQuery,
                            void *pPredState )
{
    assert( pContext && pQuery );

    /* Free the attribute array */
    if ( pPredState != NULL )
    {
        MDSU_free( pContext, pPredState );
    }

    return CSSM_OK;
}

static CSSM_RETURN CSSMAPI __MDSU_ConvertCDSACSSMAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            void *pSchemaData )
{
    CSSM_RETURN crReturn;
    MDSU_CDSA_CSSM_SCHEMA *pCSSM = (MDSU_CDSA_CSSM_SCHEMA*)pSchemaData;

    const CSSM_DB_ATTRIBUTE_DATA *DbAttr = NULL;

    assert( pSchemaData && DbRecord && pContext );

    DbAttr = DbRecord->AttributeData;

    /* ModuleID */
    crReturn = _MDSU_GetStructuredGUID( 
                                (char*)DbAttr[__ModuleID].Value->Data,
                                &pCSSM->ModuleID );
    if ( crReturn != CSSM_OK )
    {
        return crReturn;
    }

    // Manifest 
    /*
    if ( DbAttr[__Manifest].Value->Data != NULL )
    {
        pEMM->Manifest.Data = MDSU_malloc( 
                                        pContext,
                                        DbAttr[__Manifest].Value->Length );
        if ( pEMM->Manifest.Data == NULL )
        {
            return CSSMERR_CSSM_MEMORY_ERROR;
        }
        pEMM->Manifest.Length = DbAttr[__Manifest].Value->Length;
        memcpy( pEMM->Manifest.Data,
                DbAttr[__Manifest].Value->Data,
                DbAttr[__Manifest].Value->Length );
    }
    else
    {
        pEMM->Manifest.Data = NULL;
        pEMM->Manifest.Length = 0;
    }
    */
    /* ModuleName */
    /*strcpy( pCSSM->ModuleName,
            (char*)DbAttr[__ModuleName].Value->Data );*/

    /* Path */
    /*strcpy( pEMM->Path,
            (char*)DbAttr[__Path].Value->Data );*/

    /* CDSA Version */
    crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__CdsaVersion].Value->Data,
                                    &pCSSM->CDSAVersion );
    /*if ( crReturn != CSSM_OK )
    {
        if(NULL != pEMM->Manifest.Data)
        {
            MDSU_free( pContext, pEMM->Manifest.Data );
            pEMM->Manifest.Data = NULL;
        }
    }*/

    /* EMM Spec Version */
    /*crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__EmmSpecVersion].Value->Data,
                                    &pEMM->EmmSpecVersion );
    if ( crReturn != CSSM_OK )
    {
        if(NULL != pEMM->Manifest.Data)
        {
            MDSU_free( pContext, pEMM->Manifest.Data );
            pEMM->Manifest.Data = NULL;
        }
    }
    */

    /* CssmVendor */
	if ( DbAttr[__CssmVendor].Value->Data != NULL )
        strncpy( pCSSM->CssmVendor,
                 (char*)DbAttr[__CssmVendor].Value->Data,
				 DbAttr[__CssmVendor].Value->Length );
	else
		strcpy( pCSSM->CssmVendor, "" );

    /* Desc */
	if ( DbAttr[__Desc].Value->Data != NULL )
        strncpy( pCSSM->Desc,
                 (char*)DbAttr[__Desc].Value->Data,
				 DbAttr[__Desc].Value->Length );
	else
		strcpy( pCSSM->Desc, "" );

    /* PolicyStmt */
    /*if ( DbAttr[__PolicyStmt].Value->Data != NULL )
    {
        pEMM->PolicyStmt.Data = MDSU_malloc( 
                                        pContext,
                                        DbAttr[__PolicyStmt].Value->Length );
        if ( pEMM->PolicyStmt.Data == NULL )
        {
            return CSSMERR_CSSM_MEMORY_ERROR;
        }
        pEMM->PolicyStmt.Length = DbAttr[__PolicyStmt].Value->Length;
        memcpy( pEMM->PolicyStmt.Data,
                DbAttr[__PolicyStmt].Value->Data,
                DbAttr[__PolicyStmt].Value->Length );
    }
    else
    {
        pEMM->PolicyStmt.Data = NULL;
        pEMM->PolicyStmt.Length = 0;
    }*/

    /* EmmVersion */
    /*crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__EmmVersion].Value->Data,
                                    &pEMM->EmmVersion );
    if ( crReturn != CSSM_OK )
    {
        if(NULL != pEMM->PolicyStmt.Data)
        {
            MDSU_free( pContext, pEMM->PolicyStmt.Data );
            pEMM->PolicyStmt.Data = NULL;
        }
        if(NULL != pEMM->Manifest.Data)
        {
            MDSU_free( pContext, pEMM->Manifest.Data );
            pEMM->Manifest.Data = NULL;
        }
    }
    */

    /* NativeServices - a ServiceMask */
    pCSSM->NativeServices = *(uint32*)DbAttr[__NativeServices].Value->Data;

    return CSSM_OK;
}

const IfMdsuSchemaManipulator IfiMdsuCdsadirCSSMSchema =
{
    __MDSU_ConstructCDSACSSMAttributes,
    __MDSU_DestructCDSACSSMAttributes,
    __MDSU_ConstructCDSACSSMPredicate,
    __MDSU_DestructCDSACSSMPredicate,
    __MDSU_ConvertCDSACSSMAttributes,
    MDS_CDSADIR_CSSM_RECORDTYPE
};
