/*-----------------------------------------------------------------------
 * File: CDSAEMMPRIMARY.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */


#include <assert.h>
#include <string.h>

#include "cssmtype.h"
#include "cssmerr.h"
#ifdef VMS
#include "cssmport.h"
#else
#include "port/cssmport.h"
#endif
#include "mds_schema.h"

#include "mds.h"
#include "mds_util.h"

#include "convert.h"
#include "internal.h"

#define MDSU_EMMP_MODULE_ID         ( 0x00000001 )
#define MDSU_EMMP_SSID              ( 0x00000002 )
#define MDSU_EMMP_SERVICE_TYPE      ( 0x00000004 )
#define MDSU_EMMP_MANIFEST          ( 0x00000008 )
#define MDSU_EMMP_MODULE_NAME       ( 0x00000010 )
#define MDSU_EMMP_VERSION           ( 0x00000020 )
#define MDSU_EMMP_VENDOR            ( 0x00000040 )
#define MDSU_EMMP_SAMPLE_TYPES      ( 0x00000080 )
#define MDSU_EMMP_ACL_SUBJECT_TYPES ( 0x00000100 )
#define MDSU_EMMP_AUTH_TAGS         ( 0x00000200 )
#define MDSU_EMMP_EMM_SPEC_VERSION  ( 0x00000400 )
#define __ModuleID          ( 0 )
#define __SSID              ( 1 )
#define __ServiceType       ( 2 )
#define __Manifest          ( 3 )
#define __ModuleName        ( 4 )
#define __Version           ( 5 )
#define __Vendor            ( 6 )
#define __SampleTypes       ( 7 )
#define __AclSubjectTypes   ( 8 )
#define __AuthTags          ( 9 )
#define __EmmSpecVersion    ( 10 )


typedef struct _mdsu_cdsa_emm_primary_attr_state
{
    CSSM_DB_ATTRIBUTE_DATA DbRecordAttributes[MDS_CDSADIR_EMM_PRIMARY_NUM_ATTRIBUTES];
    CSSM_DATA AttributeValues[MDS_CDSADIR_EMM_PRIMARY_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintableVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintableEmmSpecVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_EMM_PRIMARY_ATTR_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSAEMMPrimaryAttributes(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaData,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFillValues,
                            void **ppAttrState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_EMM_PRIMARY_SCHEMA *pEMMPrimary = 
                            (const MDSU_CDSA_EMM_PRIMARY_SCHEMA*)pSchemaData;
    _MDSU_CDSA_EMM_PRIMARY_ATTR_STATE *pState = NULL;
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && ppAttrState );
    assert( ( bFillValues == CSSM_FALSE ) ||
            ( ( bFillValues == CSSM_TRUE ) && pSchemaData ) );

    /* Allocate the memory for the attributes */
    pState = MDSU_malloc( pContext, sizeof(_MDSU_CDSA_EMM_PRIMARY_ATTR_STATE) );
    if ( pState == NULL )
    {
        ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
    }
    else
    {
        /* Clear the structure */
        memset( pState, 0, sizeof(_MDSU_CDSA_EMM_PRIMARY_ATTR_STATE) );
        DbRecordAttributes = pState->DbRecordAttributes;

        /* Fill in the record information */
        DbRecord->DataRecordType = MDS_CDSADIR_EMM_PRIMARY_RECORDTYPE;
        DbRecord->SemanticInformation = 0;
        DbRecord->NumberOfAttributes = MDS_CDSADIR_EMM_PRIMARY_NUM_ATTRIBUTES;
        DbRecord->AttributeData = DbRecordAttributes;

        /* Fill in attribute info */
        DbRecordAttributes[__ModuleID].Info = s_MdsAttrInfo_ModuleId;
        DbRecordAttributes[__SSID].Info = s_MdsAttrInfo_SSID;
        DbRecordAttributes[__ServiceType].Info = s_MdsAttrInfo_ServiceType;
        DbRecordAttributes[__Manifest].Info = s_MdsAttrInfo_Manifest;
        DbRecordAttributes[__ModuleName].Info = s_MdsAttrInfo_ModuleName;
        DbRecordAttributes[__Version].Info = s_MdsAttrInfo_Version;
        DbRecordAttributes[__Vendor].Info = s_MdsAttrInfo_Vendor;
        DbRecordAttributes[__SampleTypes].Info = s_MdsAttrInfo_SampleTypes;
        DbRecordAttributes[__AclSubjectTypes].Info = s_MdsAttrInfo_AclSubjectTypes;
        DbRecordAttributes[__AuthTags].Info = s_MdsAttrInfo_AuthTags;
        DbRecordAttributes[__EmmSpecVersion].Info = s_MdsAttrInfo_EMMSpecVersion;

        if ( bFillValues == CSSM_TRUE )
        {
            /* Connect the attribute value array to the attribute array */
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = &pState->AttributeValues[i];
                DbRecordAttributes[i].NumberOfValues = 1;
            }

            /* Fill in the record contents */
            /* ModuleID */
            _MDSU_GetPrintableGUID( &pEMMPrimary->ModuleID,
                                    (char*)pState->PrintableGuid );
            DbRecordAttributes[__ModuleID].Value->Data = 
                                                    pState->PrintableGuid;
            DbRecordAttributes[__ModuleID].Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;

            /* SSID */
            DbRecordAttributes[__SSID].Value->Length = sizeof(uint32);
            DbRecordAttributes[__SSID].Value->Data = (uint8*)&pEMMPrimary->SSID;

            /* ServiceType */
            DbRecordAttributes[__ServiceType].Value->Length = sizeof(uint32);
            DbRecordAttributes[__ServiceType].Value->Data = (uint8*)&pEMMPrimary->ServiceType;

            /* Manifest */
            (*DbRecordAttributes[__Manifest].Value) = pEMMPrimary->Manifest;

            /* ModuleName */
            DbRecordAttributes[__ModuleName].Value->Length = 
                                        strlen( pEMMPrimary->ModuleName ) + 1 ;
            DbRecordAttributes[__ModuleName].Value->Data =
                                                (uint8*)pEMMPrimary->ModuleName;

            /* Version */
            _MDSU_GetPrintableVersion( &pEMMPrimary->Version,
                                       (char*)pState->PrintableVersion );
            DbRecordAttributes[__Version].Value->Data = 
                                                    pState->PrintableVersion;
            DbRecordAttributes[__Version].Value->Length = 
                                strlen( (char*)pState->PrintableVersion ) + 1;

            /* Vendor */
            DbRecordAttributes[__Vendor].Value->Length = 
                                        strlen( pEMMPrimary->Vendor ) + 1 ;
            DbRecordAttributes[__Vendor].Value->Data =
                                                (uint8*)pEMMPrimary->Vendor;


            /* SampleTypes */
            DbRecordAttributes[__SampleTypes].Value->Length =
                                            sizeof(uint32) * pEMMPrimary->NumSampleTypes;
            DbRecordAttributes[__SampleTypes].Value->Data = (uint8*)&pEMMPrimary->SampleTypes;

            /* AclSubjectTypes */
            DbRecordAttributes[__AclSubjectTypes].Value->Length =
                                            sizeof(uint32) * pEMMPrimary->NumAclSubjectTypes;
            DbRecordAttributes[__AclSubjectTypes].Value->Data = (uint8*)&pEMMPrimary->AclSubjectTypes;

            /* AuthTags */
            DbRecordAttributes[__AuthTags].Value->Length =
                                            sizeof(uint32) * pEMMPrimary->NumAuthTags;
            DbRecordAttributes[__AuthTags].Value->Data = (uint8*)&pEMMPrimary->AuthTags;

            /* EMM Spec Version */
            _MDSU_GetPrintableVersion( &pEMMPrimary->EmmSpecVersion,
                                       (char*)pState->PrintableEmmSpecVersion );
            DbRecordAttributes[__EmmSpecVersion].Value->Data = 
                                                    pState->PrintableEmmSpecVersion;
            DbRecordAttributes[__EmmSpecVersion].Value->Length = 
                                strlen( (char*)pState->PrintableVersion ) + 1;


        } /* bFillValues */
        else
        {
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = NULL;
                DbRecordAttributes[i].NumberOfValues = 0;
            }
        }

        if ( rv != CSSM_OK )
        {
            /* Deallocate memory if there was an error */
            MDSU_free( pContext,
                       pState );
            pState = NULL;
        }
    } /* allocate DbRecordAttributes */

    *ppAttrState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSAEMMPrimaryAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFreeAll,
                            void *pAttrState )
{
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord && pAttrState );

    /* Free the buffers allocated for the attribute values */
    DbRecordAttributes = DbRecord->AttributeData;
    if ( bFreeAll == CSSM_TRUE )
    {
        for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
        {
            if ( DbRecordAttributes[i].Value )
            {
                if ( DbRecordAttributes[i].Value->Data )
                {
                    MDSU_free( pContext,
                               DbRecordAttributes[i].Value->Data );
                }
                MDSU_free( pContext,
                           DbRecordAttributes[i].Value );
            }
        }
    }

    /* Free the state structure */
    MDSU_free( pContext,
               pAttrState );
    DbRecord->AttributeData = NULL;

    return CSSM_OK;
}

typedef struct _mdsu_cdsa_emm_primary_pred_state
{
    CSSM_SELECTION_PREDICATE Predicates[MDS_CDSADIR_EMM_PRIMARY_NUM_ATTRIBUTES];
    CSSM_DATA PredicateValues[MDS_CDSADIR_EMM_PRIMARY_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintableVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintableEmmSpecVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_EMM_PRIMARY_PRED_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSAEMMPrimaryPredicate(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaTemplate,
                            MDSU_TEMPLATE_MASK ValidTemplates,
                            CSSM_QUERY *pQuery,
                            void **ppPredState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_EMM_PRIMARY_SCHEMA *pEMMPrimary = 
                            (const MDSU_CDSA_EMM_PRIMARY_SCHEMA*)pSchemaTemplate;
    _MDSU_CDSA_EMM_PRIMARY_PRED_STATE *pState = NULL;
    CSSM_SELECTION_PREDICATE *pPredicates = NULL;
    uint32 uNumPredicates;
    uint32 uIndex = 0;
    uint32 i;

    assert( pContext && 
            pQuery &&
            ppPredState &&
            ( ( pSchemaTemplate && ValidTemplates ) ||
              ( !pSchemaTemplate && !ValidTemplates ) ) );

    /* Count the predicates */
    /* Clear the manifest bit. Searching on manifest is not supported */
    ValidTemplates &= (~MDSU_EMMP_MANIFEST);
    _MDSU_CountBits( ValidTemplates, uNumPredicates );

    /* Allocate the predicate state and link the value structures */
    if ( uNumPredicates != 0 )
    {
        pState = MDSU_malloc( pContext, sizeof(_MDSU_CDSA_EMM_PRIMARY_PRED_STATE) );
        if ( pState == NULL )
        {
            ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
        }
        else
        {
            memset( pState,
                    0,
                    sizeof(_MDSU_CDSA_EMM_PRIMARY_PRED_STATE) );
            pPredicates = pState->Predicates;

            /* Allocate a single attribute value struct per attribute */
            for ( i = 0; i < uNumPredicates; i++ )
            {
                pPredicates[i].Attribute.Value = &pState->PredicateValues[i];
                pPredicates[i].Attribute.NumberOfValues = 1;
            }
        } /* Allocate and link predicate values */
    } /* uNumPredicates != 0 */

    if ( rv == CSSM_OK )
    {
        /* Fill in the query parameters */
        pQuery->RecordType = MDS_CDSADIR_EMM_PRIMARY_RECORDTYPE;
        pQuery->Conjunctive = CSSM_DB_AND;
        pQuery->NumSelectionPredicates = uNumPredicates;
        pQuery->QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
        pQuery->QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
        pQuery->QueryFlags = 0;
        pQuery->SelectionPredicate = pPredicates;
        for ( uIndex = 0; uIndex < uNumPredicates; uIndex++ )
        {
            pPredicates[uIndex].DbOperator = CSSM_DB_EQUAL;
        }


        /* Fill in the predicate data */
        uIndex = 0;
        if ( ValidTemplates & MDSU_EMMP_MODULE_ID )
        {
            /* ModuleID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleId;
            _MDSU_GetPrintableGUID( &pEMMPrimary->ModuleID,
                                    (char*)pState->PrintableGuid );
            pPredicates[uIndex].Attribute.Value->Data = pState->PrintableGuid;
            pPredicates[uIndex].Attribute.Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_SSID )
        {
            /* SSID */
            pPredicates[uIndex].Attribute.Info = 
                                                s_MdsAttrInfo_SSID;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pEMMPrimary->SSID;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_SERVICE_TYPE )
        {
            /* ServiceType */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ServiceType;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pEMMPrimary->ServiceType;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_MODULE_NAME )
        {
            /* ModuleName */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleName;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pEMMPrimary->ModuleName ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pEMMPrimary->ModuleName;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_VERSION )
        {
            /* Version */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Version;
            _MDSU_GetPrintableVersion( &pEMMPrimary->Version,
                                       (char*)pState->PrintableVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_VENDOR )
        {
            /* Vendor */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_Vendor;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pEMMPrimary->Vendor ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pEMMPrimary->Vendor;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_SAMPLE_TYPES )
        {
            /* SampleTypes */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SampleTypes;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        sizeof(uint32) * pEMMPrimary->NumSampleTypes;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pEMMPrimary->SampleTypes;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_ACL_SUBJECT_TYPES )
        {
            /* AclSubjectTypes */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_AclSubjectTypes;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        sizeof(uint32) * pEMMPrimary->NumAclSubjectTypes;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pEMMPrimary->AclSubjectTypes;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_AUTH_TAGS )
        {
            /* AuthTags */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_AuthTags;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        sizeof(uint32) * pEMMPrimary->NumAuthTags;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)&pEMMPrimary->AuthTags;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_EMMP_EMM_SPEC_VERSION )
        {
            /* EMM Spec Version */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_EMMSpecVersion;
            _MDSU_GetPrintableVersion( &pEMMPrimary->EmmSpecVersion,
                                       (char*)pState->PrintableEmmSpecVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintableEmmSpecVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintableEmmSpecVersion ) + 1;
            uIndex++;
        }

    } /* Initial allocation successful */

    *ppPredState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSAEMMPrimaryPredicate(
                            MDSU_CONTEXT *pContext,
                            CSSM_QUERY *pQuery,
                            void *pPredState )
{
    assert( pContext && pQuery );

    /* Free the attribute array */
    if ( pPredState != NULL )
    {
        MDSU_free( pContext, pPredState );
    }

    return CSSM_OK;
}

static CSSM_RETURN CSSMAPI __MDSU_ConvertCDSAEMMPrimaryAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            void *pSchemaData )
{
    CSSM_RETURN crReturn;
    MDSU_CDSA_EMM_PRIMARY_SCHEMA *pEMMPrimary = (MDSU_CDSA_EMM_PRIMARY_SCHEMA*)pSchemaData;

    const CSSM_DB_ATTRIBUTE_DATA *DbAttr = NULL;

    assert( pSchemaData && DbRecord && pContext );

    DbAttr = DbRecord->AttributeData;


    /* ModuleID */
    crReturn = _MDSU_GetStructuredGUID( 
                                (char*)DbAttr[__ModuleID].Value->Data,
                                &pEMMPrimary->ModuleID );
    if ( crReturn != CSSM_OK )
    {
        return crReturn;
    }

    /* SSID */
    pEMMPrimary->SSID = *(uint32*)DbAttr[__SSID].Value->Data;

    /* ServiceType */
    pEMMPrimary->ServiceType = *(uint32*)DbAttr[__ServiceType].Value->Data;

    /* Manifest */
    if ( DbAttr[__Manifest].Value->Data != NULL )
    {
        pEMMPrimary->Manifest.Data = MDSU_malloc( 
                                        pContext,
                                        DbAttr[__Manifest].Value->Length );
        if ( pEMMPrimary->Manifest.Data == NULL )
        {
            return CSSMERR_CSSM_MEMORY_ERROR;
        }
        pEMMPrimary->Manifest.Length = DbAttr[__Manifest].Value->Length;
        memcpy( pEMMPrimary->Manifest.Data,
                DbAttr[__Manifest].Value->Data,
                DbAttr[__Manifest].Value->Length );
    }
    else
    {
        pEMMPrimary->Manifest.Data = NULL;
        pEMMPrimary->Manifest.Length = 0;
    }
    
    /* ModuleName */
	if ( DbAttr[__ModuleName].Value->Data != NULL )
        strncpy( pEMMPrimary->ModuleName,
                 (char*)DbAttr[__ModuleName].Value->Data,
				 DbAttr[__ModuleName].Value->Length );
	else
		strcpy( pEMMPrimary->ModuleName, "" );

    /* Version */
    crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__Version].Value->Data,
                                    &pEMMPrimary->Version );
    if ( crReturn != CSSM_OK )
    {
        if(NULL != pEMMPrimary->Manifest.Data)
        {
            MDSU_free( pContext, pEMMPrimary->Manifest.Data );
            pEMMPrimary->Manifest.Data = NULL;
        }
    }

    /* Vendor */
	if ( DbAttr[__Vendor].Value->Data != NULL )
        strncpy( pEMMPrimary->Vendor,
                 (char*)DbAttr[__Vendor].Value->Data,
				 DbAttr[__Vendor].Value->Length );
	else
		strcpy( pEMMPrimary->Vendor, "" );


    /* SampleTypes */
    pEMMPrimary->NumSampleTypes = (DbAttr[__SampleTypes].Value->Length / sizeof(uint32));
    if((0 != pEMMPrimary->NumSampleTypes) &&
       (16 >= pEMMPrimary->NumSampleTypes))
    {
        memcpy(pEMMPrimary->SampleTypes,
               DbAttr[__SampleTypes].Value->Data,
               (sizeof(uint32) * pEMMPrimary->NumSampleTypes));
    }
    else
    {
        /* we can't handle more than 16 sample types */
        if(16 < pEMMPrimary->NumSampleTypes)
        {
            pEMMPrimary->NumSampleTypes = 0;
            return CSSMERR_CSSM_INTERNAL_ERROR;
        }
    }

    /* AclSubjectTypes */
    pEMMPrimary->NumAclSubjectTypes = (DbAttr[__AclSubjectTypes].Value->Length / sizeof(uint32));
    if((0 != pEMMPrimary->NumAclSubjectTypes) &&
       (16 >= pEMMPrimary->NumAclSubjectTypes))
    {
        memcpy(pEMMPrimary->AclSubjectTypes,
               DbAttr[__AclSubjectTypes].Value->Data,
               (sizeof(uint32) * pEMMPrimary->NumAclSubjectTypes));
    }
    else
    {
        /* we can't handle more than 16 ACL subject types */
        if(16 < pEMMPrimary->NumAclSubjectTypes)
        {
            pEMMPrimary->NumAclSubjectTypes = 0;
            return CSSMERR_CSSM_INTERNAL_ERROR;
        }
    }

    /* AuthTags */
    pEMMPrimary->NumAuthTags = (DbAttr[__AuthTags].Value->Length / sizeof(uint32));
    if((0 != pEMMPrimary->NumAuthTags) &&
       (16 >= pEMMPrimary->NumAuthTags))
    {
        memcpy(pEMMPrimary->AuthTags,
               DbAttr[__AuthTags].Value->Data,
               (sizeof(uint32) * pEMMPrimary->NumAuthTags));
    }
    else
    {
        /* we can't handle more than 16 sample types */
        if(16 < pEMMPrimary->NumAuthTags)
        {
            pEMMPrimary->NumAuthTags = 0;
            return CSSMERR_CSSM_INTERNAL_ERROR;
        }
    }

    /* EMM Spec Version */
    crReturn = _MDSU_GetStructuredVersion( 
                                    (char*)DbAttr[__EmmSpecVersion].Value->Data,
                                    &pEMMPrimary->EmmSpecVersion );
    if ( crReturn != CSSM_OK )
    {
        if(NULL != pEMMPrimary->Manifest.Data)
        {
            MDSU_free( pContext, pEMMPrimary->Manifest.Data );
            pEMMPrimary->Manifest.Data = NULL;
        }
    }

    return CSSM_OK;
}

const IfMdsuSchemaManipulator IfiMdsuCdsadirEMMPrimarySchema =
{
    __MDSU_ConstructCDSAEMMPrimaryAttributes,
    __MDSU_DestructCDSAEMMPrimaryAttributes,
    __MDSU_ConstructCDSAEMMPrimaryPredicate,
    __MDSU_DestructCDSAEMMPrimaryPredicate,
    __MDSU_ConvertCDSAEMMPrimaryAttributes,
    MDS_CDSADIR_EMM_PRIMARY_RECORDTYPE
};
