/*-----------------------------------------------------------------------
 * File: CDSATPENCPRODUCT.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */


#include <assert.h>
#include <string.h>
#ifdef VMS
#include "cssmport.h"
#else
#include "port/cssmport.h"
#endif
#include "cssmtype.h"
#include "cssmerr.h"
#include "mds_schema.h"

#include "mds.h"
#include "mds_util.h"
#include "convert.h"
#include "internal.h"

#define __ModuleID              ( 0 )
#define __SSID                  ( 1 )
#define __ProductDesc           ( 2 )
#define __ProductVendor         ( 3 )
#define __ProductVersion        ( 4 )
#define __ProductFlags          ( 5 )
#define __AuthorityRequestType  ( 6 )
#define __StandardDesc          ( 7 )
#define __StandardVersion       ( 8 )
#define __ProtocolDesc          ( 9 )
#define __ProtocolFlags         ( 10 )
#define __CertClassName         ( 11 )
#define __RootCertificate       ( 12 )
#define __RootCertTypeFormat    ( 13 )

typedef struct _mdsu_cdsa_tp_enc_prod_attr_state
{
    CSSM_DB_ATTRIBUTE_DATA DbRecordAttributes[MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    CSSM_DATA AttributeValues[MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintProductVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintStandardVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_TP_ENC_PROD_ATTR_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSATPEncProductAttributes(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaData,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFillValues,
                            void **ppAttrState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_TP_ENC_PRODUCT_SCHEMA *pProduct = 
                        (const MDSU_CDSA_TP_ENC_PRODUCT_SCHEMA*)pSchemaData;
    _MDSU_CDSA_TP_ENC_PROD_ATTR_STATE *pState = NULL;
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( ( bFillValues == CSSM_FALSE ) ||
            ( ( bFillValues == CSSM_TRUE ) && pContext && pSchemaData && DbRecord ) );

    /* Allocate the memory for the attributes */
    pState = MDSU_malloc( pContext, sizeof(_MDSU_CDSA_TP_ENC_PROD_ATTR_STATE) );
    if ( pState == NULL )
    {
        ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
    }
    else
    {
        /* Clear the structure */
        memset( pState,
                0,
                sizeof(*pState) );
        DbRecordAttributes = pState->DbRecordAttributes;

        /* Fill in the record information */
        DbRecord->DataRecordType = 
                            MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_RECORDTYPE;
        DbRecord->SemanticInformation = 0;
        DbRecord->NumberOfAttributes = 
                            MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES;
        DbRecord->AttributeData = DbRecordAttributes;

        /* Fill in the attribute info */
        DbRecordAttributes[__ModuleID].Info = s_MdsAttrInfo_ModuleId;
        DbRecordAttributes[__SSID].Info = s_MdsAttrInfo_SSID;
        DbRecordAttributes[__ProductDesc].Info = s_MdsAttrInfo_ProductDesc;
        DbRecordAttributes[__ProductVendor].Info = s_MdsAttrInfo_ProductVendor;
        DbRecordAttributes[__ProductVersion].Info = s_MdsAttrInfo_ProductVersion;
        DbRecordAttributes[__ProductFlags].Info = s_MdsAttrInfo_ProductFlags;
        DbRecordAttributes[__AuthorityRequestType].Info = s_MdsAttrInfo_AuthorityRequestType;
        DbRecordAttributes[__StandardDesc].Info = s_MdsAttrInfo_StandardDesc;
        DbRecordAttributes[__StandardVersion].Info = s_MdsAttrInfo_StandardVersion;
        DbRecordAttributes[__ProtocolDesc].Info = s_MdsAttrInfo_ProtocolDesc;
        DbRecordAttributes[__ProtocolFlags].Info = s_MdsAttrInfo_ProtocolFlags;
        DbRecordAttributes[__CertClassName].Info = s_MdsAttrInfo_CertClassName;
        DbRecordAttributes[__RootCertificate].Info = s_MdsAttrInfo_RootCertificate;
        DbRecordAttributes[__RootCertTypeFormat].Info = s_MdsAttrInfo_RootCertTypeFormat;

        if ( bFillValues == CSSM_TRUE )
        {
            /* Connect the attribute value array to the attribute array */
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = &pState->AttributeValues[i];
                DbRecordAttributes[i].NumberOfValues = 1;
            }

            /* Fill in the record contents */
            /* ModuleID */
            _MDSU_GetPrintableGUID( &pProduct->ModuleID,
                                    (char*)pState->PrintableGuid );
            DbRecordAttributes[__ModuleID].Value->Data =
                                                    pState->PrintableGuid;
            DbRecordAttributes[__ModuleID].Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;

            /* SSID */
            DbRecordAttributes[__SSID].Value->Length = sizeof(uint32);
            DbRecordAttributes[__SSID].Value->Data = (uint8*)&pProduct->SSID;

            /* ProductDesc */
            DbRecordAttributes[__ProductDesc].Value->Data = 
                                                    (uint8*)pProduct->ProductDesc;
            DbRecordAttributes[__ProductDesc].Value->Length =
                                            strlen( pProduct->ProductDesc ) + 1;

            /* ProductVendor */
            DbRecordAttributes[__ProductVendor].Value->Length = 
                                            strlen( pProduct->ProductVendor ) + 1;
            DbRecordAttributes[__ProductVendor].Value->Data =
                                                (uint8*)pProduct->ProductVendor;
            
            /* ProductVersion */
            _MDSU_GetPrintableVersion(
                                &pProduct->ProductVersion,
                                (char*)pState->PrintProductVersion );
            DbRecordAttributes[__ProductVersion].Value->Data =
                                                pState->PrintProductVersion;
            DbRecordAttributes[__ProductVersion].Value->Length = 
                            strlen( (char*)pState->PrintProductVersion ) + 1;

            /* ProductFlags */
            DbRecordAttributes[__ProductFlags].Value->Length = sizeof(uint32);
            DbRecordAttributes[__ProductFlags].Value->Data =
                                                (uint8*)&pProduct->ProductFlags;

            /* AuthorityRequestType */
            DbRecordAttributes[__AuthorityRequestType].Value->Length = 
                                                sizeof(uint32) * pProduct->NumAuthorityRequestType;
            DbRecordAttributes[__AuthorityRequestType].Value->Data =
                                                (uint8*)&pProduct->AuthorityRequestType;

            /* StandardDesc */
            DbRecordAttributes[__StandardDesc].Value->Length = 
                                            strlen( pProduct->StandardDesc ) + 1;
            DbRecordAttributes[__StandardDesc].Value->Data =
                                                (uint8*)pProduct->StandardDesc;

            /* StandardVersion */
            _MDSU_GetPrintableVersion(
                                &pProduct->StandardVersion,
                                (char*)pState->PrintStandardVersion );
            DbRecordAttributes[__StandardVersion].Value->Data =
                                            pState->PrintStandardVersion;
            DbRecordAttributes[__StandardVersion].Value->Length = 
                        strlen( (char*)pState->PrintStandardVersion ) + 1;

            /* ProtocolDesc */
            DbRecordAttributes[__ProtocolDesc].Value->Length = strlen( pProduct->ProtocolDesc ) + 1;
            DbRecordAttributes[__ProtocolDesc].Value->Data = (uint8*)pProduct->ProtocolDesc;

            /* ProtocolFlags */
            DbRecordAttributes[__ProtocolFlags].Value->Length = sizeof(uint32);
            DbRecordAttributes[__ProtocolFlags].Value->Data = (uint8*)&pProduct->ProtocolFlags;

            /* CertClassName */
            DbRecordAttributes[__CertClassName].Value->Length = strlen( pProduct->CertClassName ) + 1;
            DbRecordAttributes[__CertClassName].Value->Data = (uint8*)pProduct->CertClassName;

            /* RootCertificate */
            (*DbRecordAttributes[__RootCertificate].Value) = pProduct->RootCertificate;

            /* RootCertTypeFormat */
            DbRecordAttributes[__RootCertTypeFormat].Value->Length = sizeof(uint32);
            DbRecordAttributes[__RootCertTypeFormat].Value->Data = (uint8*)&pProduct->RootCertTypeFormat;
        }
        else
        {
            for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
            {
                DbRecordAttributes[i].Value = NULL;
                DbRecordAttributes[i].NumberOfValues = 0;
            }
        }
    } /* Allocate attributes array */

    *ppAttrState = pState;

    return rv;

}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSATPEncProductAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            CSSM_BOOL bFreeAll,
                            void *pAttrState )
{
    CSSM_DB_ATTRIBUTE_DATA *DbRecordAttributes;
    uint32 i;

    assert( pContext && DbRecord );

    /* Free the buffers allocated for the attribute values */
    DbRecordAttributes = DbRecord->AttributeData;
    if ( bFreeAll == CSSM_TRUE )
    {
        for ( i = 0; i < DbRecord->NumberOfAttributes; i++ )
        {
            if ( DbRecordAttributes[i].Value )
            {
                if ( DbRecordAttributes[i].Value->Data )
                {
                    MDSU_free( pContext,
                               DbRecordAttributes[i].Value->Data );
                }
                MDSU_free( pContext,
                           DbRecordAttributes[i].Value );
            }
        }
    }

    /* Free the attribute array */
    MDSU_free( pContext,
               pAttrState );

    return CSSM_OK;
}

typedef struct _mdsu_cdsa_tp_enc_prod_pred_state
{
    CSSM_SELECTION_PREDICATE Predicates[MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    CSSM_DATA PredicateValues[MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_NUM_ATTRIBUTES];
    uint8 PrintableGuid[_MDSU_PRINTABLE_GUID_LENGTH];
    uint8 PrintProductVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
    uint8 PrintStandardVersion[_MDSU_PRINTABLE_VERSION_LENGTH];
}
_MDSU_CDSA_TP_ENC_PROD_PRED_STATE;

static CSSM_RETURN CSSMAPI __MDSU_ConstructCDSATPEncProductPredicate(
                            MDSU_CONTEXT *pContext,
                            const void *pSchemaTemplate,
                            MDSU_TEMPLATE_MASK ValidTemplates,
                            CSSM_QUERY *pQuery,
                            void **ppPredState )
{
    CSSM_RETURN rv = CSSM_OK;
    const MDSU_CDSA_TP_ENC_PRODUCT_SCHEMA *pProduct = 
                    (const MDSU_CDSA_TP_ENC_PRODUCT_SCHEMA*)pSchemaTemplate;
    _MDSU_CDSA_TP_ENC_PROD_PRED_STATE *pState = NULL;
    CSSM_SELECTION_PREDICATE *pPredicates = NULL;
    uint32 uNumPredicates;
    uint32 uIndex = 0;
    uint32 i;

    assert( pContext && 
            pQuery &&
            ppPredState &&
            ( ( pSchemaTemplate && ValidTemplates ) ||
              ( !pSchemaTemplate && !ValidTemplates ) ) );

    /* Clear certificate and multi_uint32 bits. Searching on them are not supported. */
    ValidTemplates &= ~( MDSU_TPEP_AUTH_REQUEST_TYPE |
                         MDSU_TPEP_ROOT_CERTIFICATE );

    /* Count the predicates */
    _MDSU_CountBits( ValidTemplates, uNumPredicates );

    /* Allocate the predicate state and link the value structures */
    if ( uNumPredicates != 0 )
    {
        pState = MDSU_malloc( pContext, sizeof(*pState) );
        if ( pState == NULL )
        {
            ERR( rv = CSSMERR_CSSM_MEMORY_ERROR );
        }
        else
        {
            memset( pState,
                    0,
                    sizeof(*pState) );
            pPredicates = pState->Predicates;

            /* Allocate a single attribute value struct per attribute */
            for ( i = 0; i < uNumPredicates; i++ )
            {
                pPredicates[i].Attribute.Value =
                                            &pState->PredicateValues[i];
                pPredicates[i].Attribute.NumberOfValues = 1;
            }
        } /* Allocate and link predicate values */
    } /* uNumPredicates != 0 */

    if ( rv == CSSM_OK )
    {
        /* Fill in the query parameters */
        pQuery->RecordType = MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_RECORDTYPE;
        pQuery->Conjunctive = CSSM_DB_AND;
        pQuery->NumSelectionPredicates = uNumPredicates;
        pQuery->QueryLimits.TimeLimit = CSSM_QUERY_TIMELIMIT_NONE;
        pQuery->QueryLimits.SizeLimit = CSSM_QUERY_SIZELIMIT_NONE;
        pQuery->QueryFlags = 0;
        pQuery->SelectionPredicate = pPredicates;
        for ( uIndex = 0; uIndex < uNumPredicates; uIndex++ )
        {
            pPredicates[uIndex].DbOperator = CSSM_DB_EQUAL;
        }

        /* Fill in the predicate data */
        uIndex = 0;
        if ( ValidTemplates & MDSU_TPEP_MODULE_ID )
        {
            /* ModuleID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ModuleId;
            _MDSU_GetPrintableGUID( &pProduct->ModuleID,
                                    (char*)pState->PrintableGuid );
            pPredicates[uIndex].Attribute.Value->Data = pState->PrintableGuid;
            pPredicates[uIndex].Attribute.Value->Length = 
                                strlen( (char*)pState->PrintableGuid ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_SSID )
        {
            /* SSID */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_SSID;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        sizeof(pProduct->SSID) ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pProduct->SSID;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_PRODUCT_DESC )
        {
            /* ProductDesc */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ProductDesc;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pProduct->ProductDesc ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pProduct->ProductDesc;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_PRODUCT_VENDOR )
        {
            /* ProductVendor */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ProductVendor;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pProduct->ProductVendor ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pProduct->ProductVendor;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_PRODUCT_VERSION )
        {
            /* ProductVersion */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ProductVersion;
            _MDSU_GetPrintableVersion( &pProduct->ProductVersion,
                                       (char*)pState->PrintProductVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintProductVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintProductVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_PRODUCT_FLAGS )
        {
            /* ProductFlags */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ProductFlags;
            pPredicates[uIndex].Attribute.Value->Length = sizeof(uint32);
            pPredicates[uIndex].Attribute.Value->Data = 
                                                    (uint8*)&pProduct->ProductFlags;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_STANDARD_DESC )
        {
            /* StandardDesc */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_StandardDesc;
            pPredicates[uIndex].Attribute.Value->Length = 
                                        strlen( pProduct->StandardDesc ) + 1 ;
            pPredicates[uIndex].Attribute.Value->Data =
                                                (uint8*)pProduct->StandardDesc;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_STANDARD_VERSION )
        {
            /* StandardVersion */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_StandardVersion;
            _MDSU_GetPrintableVersion( &pProduct->StandardVersion,
                                       (char*)pState->PrintStandardVersion );
            pPredicates[uIndex].Attribute.Value->Data =
                                                pState->PrintStandardVersion;
            pPredicates[uIndex].Attribute.Value->Length = 
                            strlen( (char*)pState->PrintStandardVersion ) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_PROTOCOL_DESC )
        {
            /* ProtocolDesc */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ProtocolDesc;
            pPredicates[uIndex].Attribute.Value->Length = 
                                                    strlen(pProduct->ProtocolDesc) + 1;
            pPredicates[uIndex].Attribute.Value->Data = 
                                                    (uint8*)pProduct->ProtocolDesc;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_PROTOCOL_DESC )
        {
            /* ProtocolFlags */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_ProtocolFlags;
            pPredicates[uIndex].Attribute.Value->Length = 
                                                    sizeof(pProduct->ProtocolFlags);
            pPredicates[uIndex].Attribute.Value->Data = 
                                                    (uint8*)&pProduct->ProtocolFlags;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_CERT_CLASS_NAME )
        {
            /* CertClassName */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_CertClassName;
            pPredicates[uIndex].Attribute.Value->Data = 
                                                    (uint8*)pProduct->CertClassName;
            pPredicates[uIndex].Attribute.Value->Length = 
                                                    strlen(pProduct->CertClassName) + 1;
            uIndex++;
        }
        if ( ValidTemplates & MDSU_TPEP_ROOT_CERT_TYPE_FORMAT )
        {
            /* RootCertTypeFormat */
            pPredicates[uIndex].Attribute.Info = s_MdsAttrInfo_RootCertTypeFormat;
            pPredicates[uIndex].Attribute.Value->Length = 
                                                    sizeof(pProduct->RootCertTypeFormat);
            pPredicates[uIndex].Attribute.Value->Data = 
                                                    (uint8*)&pProduct->RootCertTypeFormat;
            uIndex++;
        }
    } /* Initial allocation successful */

    *ppPredState = (void*)pState;

    return rv;
}

static CSSM_RETURN CSSMAPI __MDSU_DestructCDSATPEncProductPredicate(
                            MDSU_CONTEXT *pContext,
                            CSSM_QUERY *pQuery,
                            void *pPredState )
{
    assert( pContext && pQuery );

    /* Free the attribute array */
    if ( pPredState != NULL )
    {
        MDSU_free( pContext, pPredState );
    }

    return CSSM_OK;
}

static CSSM_RETURN CSSMAPI __MDSU_ConvertCDSATPEncProductAttributes(
                            MDSU_CONTEXT *pContext,
                            CSSM_DB_RECORD_ATTRIBUTE_DATA *DbRecord,
                            void *pSchemaData )
{
    CSSM_RETURN crReturn;
    MDSU_CDSA_TP_ENC_PRODUCT_SCHEMA *pProduct = 
                            (MDSU_CDSA_TP_ENC_PRODUCT_SCHEMA*)pSchemaData;
    const CSSM_DB_ATTRIBUTE_DATA *DbAttr = NULL;

    assert( pSchemaData && DbRecord && pContext );

    DbAttr = DbRecord->AttributeData;

    /* ModuleID */
    crReturn = _MDSU_GetStructuredGUID( 
                                (char*)DbAttr[__ModuleID].Value->Data,
                                &pProduct->ModuleID );

    /* SSID */
    pProduct->SSID = *(uint32*)DbAttr[__SSID].Value->Data;

    /* ProductDesc */
	if ( DbAttr[__ProductDesc].Value->Data != NULL )
        strncpy( pProduct->ProductDesc,
                 (char*)DbAttr[__ProductDesc].Value->Data,
				 DbAttr[__ProductDesc].Value->Length );
	else
		strcpy( pProduct->ProductDesc, "" );

    /* ProductVendor */
 	if ( DbAttr[__ProductVendor].Value->Data != NULL )
        strncpy( pProduct->ProductVendor,
                 (char*)DbAttr[__ProductVendor].Value->Data,
				 DbAttr[__ProductVendor].Value->Length );
	else
		strcpy( pProduct->ProductVendor, "" );

    /* ProductVersion */
    _MDSU_GetStructuredVersion( (char*)DbAttr[__ProductVersion].Value->Data,
                                &pProduct->ProductVersion );

    /* ProductFlags */
    pProduct->ProductFlags = *(uint32*)DbAttr[__ProductFlags].Value->Data;

    /* AuthorityRequestType */
    memcpy( pProduct->AuthorityRequestType,
            DbAttr[__AuthorityRequestType].Value->Data,
            DbAttr[__AuthorityRequestType].Value->Length );
    pProduct->NumAuthorityRequestType = DbAttr[__AuthorityRequestType].Value->Length / 
                                sizeof(pProduct->AuthorityRequestType[0]);

    /* StandardDesc */
   	if ( DbAttr[__StandardDesc].Value->Data != NULL )
        strncpy( pProduct->StandardDesc,
                 (char*)DbAttr[__StandardDesc].Value->Data,
				 DbAttr[__StandardDesc].Value->Length );
	else
		strcpy( pProduct->StandardDesc, "" );

    /* StandardVersion */
    _MDSU_GetStructuredVersion( (char*)DbAttr[__StandardVersion].Value->Data,
                                &pProduct->StandardVersion );

    /* ProtocolDesc */
   	if ( DbAttr[__ProtocolDesc].Value->Data != NULL )
        strncpy( pProduct->ProtocolDesc,
                 (char*)DbAttr[__ProtocolDesc].Value->Data,
				 DbAttr[__ProtocolDesc].Value->Length );
	else
		strcpy( pProduct->ProtocolDesc, "" );

    /* ProtocolFlags */
    pProduct->ProtocolFlags = *(uint32*)DbAttr[__ProtocolFlags].Value->Data;

    /* CertClassName */
   	if ( DbAttr[__CertClassName].Value->Data != NULL )
        strncpy( pProduct->CertClassName,
                 (char*)DbAttr[__CertClassName].Value->Data,
				 DbAttr[__CertClassName].Value->Length );
	else
		strcpy( pProduct->CertClassName, "" );

    /* RootCertTypeFormat */
    pProduct->RootCertTypeFormat = *(uint32*)DbAttr[__RootCertTypeFormat].Value->Data;

    /* RootCertificate */
    if ( DbAttr[__RootCertificate].Value->Data != NULL )
    {
        pProduct->RootCertificate.Data = MDSU_malloc( pContext,
                                        DbAttr[__RootCertificate].Value->Length );
        if ( pProduct->RootCertificate.Data == NULL )
        {
            return CSSMERR_CSSM_MEMORY_ERROR;
        }
        pProduct->RootCertificate.Length = DbAttr[__RootCertificate].Value->Length;
        memcpy( pProduct->RootCertificate.Data,
                DbAttr[__RootCertificate].Value->Data,
                DbAttr[__RootCertificate].Value->Length );
    }
    else
    {
        pProduct->RootCertificate.Data = NULL;
        pProduct->RootCertificate.Length = 0;
    }

    return CSSM_OK;
}

const IfMdsuSchemaManipulator IfiMdsuCdsadirTPEncProductSchema =
{
    __MDSU_ConstructCDSATPEncProductAttributes,
    __MDSU_DestructCDSATPEncProductAttributes,
    __MDSU_ConstructCDSATPEncProductPredicate,
    __MDSU_DestructCDSATPEncProductPredicate,
    __MDSU_ConvertCDSATPEncProductAttributes,
    MDS_CDSADIR_TP_ENCAPSULATED_PRODUCT_RECORDTYPE
};

