/*-----------------------------------------------------------------------
 * File: CSSMCSPF.H
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

/* Function declarations for the Cryptographic Service Provider module. The
 * CSSM_DEFINE_FUNC, __CSSM_PASTE2 and __CSSM_PASTE3 macros must be defined
 * by the file that includes this header.
 */

/* Check to make sure that all macros were defined */
#ifndef CSSM_DEFINE_FUNCTION
#error "The CSSM_DEFINE_FUNCTION(name,params) macro must be defined!"
#endif
#ifndef __CSSM_PASTE2
#error "The __CSSM_PASTE2(x,y) macro must be defined!"
#endif
#ifndef __CSSM_PASTE3
#error "The __CSSM_PASTE3(x,y,z) macro must be defined!"
#endif
#if defined(_CSSM_CUSTOM_INITIAL_PARAMS)
# ifndef _CSSM_CSPSPI_FULL_PARAMS
# error "The _CSSM_CSPSPI_FULL_PARAMS macro must be defined!"
# endif
# ifndef _CSSM_CSPSPI_SHORT_PARAMS
# error "The _CSSM_CSPSPI_SHORT_PARAMS macro must be defined!"
# endif
#endif

/* Use this macro to define the name of CSP functions that have and explicit
 * CSP indication in the API name.
 */
#ifdef _CSSM_API_DEFINITIONS
# define _CSSM_EXPLICIT_NAME( name ) __CSSM_PASTE2( CSP_, name )
#else
# define _CSSM_EXPLICIT_NAME( name ) name
#endif

/* Use this macro to define parameter data types that are a pointer for the
 * API and a value parameter for the SPI.
 */
#ifdef _CSSM_API_DEFINITIONS
# define _CSSM_PAPI_VSPI( name ) name *
#else
# define _CSSM_PAPI_VSPI( name ) name
#endif

/* Use these macros to define the leading parameters on CSP functions that
 * do not have symmetric parameters on the API and SPI.
 * Define _CSSM_CUSTOM_INITIAL_PARAMS to redefine the initial parameters
 * for common that use a context or context handle.
 */
#ifndef _CSSM_CUSTOM_INITIAL_PARAMS
# ifdef _CSSM_API_DEFINITIONS
#  define _CSSM_CSPSPI_FULL_PARAMS \
					CSSM_CC_HANDLE CCHandle
#  define _CSSM_CSPSPI_SHORT_PARAMS \
					CSSM_CC_HANDLE CCHandle
# else
#  define _CSSM_CSPSPI_FULL_PARAMS \
					CSSM_CSP_HANDLE CSPHandle, \
					CSSM_CC_HANDLE CCHandle, \
					const CSSM_CONTEXT *Context
#  define _CSSM_CSPSPI_SHORT_PARAMS \
					CSSM_CSP_HANDLE CSPHandle, \
					CSSM_CC_HANDLE CCHandle
# endif
#endif /* ifndef _CSSM_CUSTOM_INITIAL_PARAMS */

/* Use this macro to generate the 'P' variant of the privileged API names.
 */
#ifdef _CSSM_API_DEFINITIONS
# define _CSSM_CSPAPI_PNAME( name ) __CSSM_PASTE2( name, P )
#else
# define _CSSM_CSPAPI_PNAME( name ) name
#endif

/* Use this macro to wrap parameters that appear in the SPI functions only.
 */
#ifdef _CSSM_API_DEFINITIONS
# define _CSSM_CSPSPI_ONLY1ST(type,name)
# define _CSSM_CSPSPI_ONLY(type,name)
#else
# define _CSSM_CSPSPI_ONLY1ST(type,name) type name,
# define _CSSM_CSPSPI_ONLY(type,name) ,type name
#endif

/*************************************************************************/
/*** Function declarations ***********************************************/
/*************************************************************************/

#ifndef _CSSM_API_DEFINITIONS
/* This function only appears in the SPI table */
CSSM_DEFINE_FUNCTION( EventNotify, (
						CSSM_MODULE_HANDLE CSPHandle,
						CSSM_CONTEXT_EVENT Event,
						CSSM_CC_HANDLE CCHandle,
						const CSSM_CONTEXT* Context ) );
#endif /* ifndef _CSSM_API_DEFINITIONS */

/* Information Functions */
CSSM_DEFINE_FUNCTION( QuerySize, (
						_CSSM_CSPSPI_FULL_PARAMS,
						CSSM_BOOL Encrypt,
						uint32 QuerySizeCount,
						CSSM_QUERY_SIZE_DATA_PTR DataBlock ) );

/* Crypto Functions */
CSSM_DEFINE_FUNCTION( SignData, (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount,
						CSSM_ALGORITHMS DigestAlgorithm,
						CSSM_DATA_PTR Signature ) );

CSSM_DEFINE_FUNCTION( SignDataInit, (
						_CSSM_CSPSPI_FULL_PARAMS ) );

CSSM_DEFINE_FUNCTION( SignDataUpdate, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount ) );

CSSM_DEFINE_FUNCTION( SignDataFinal, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						CSSM_DATA_PTR Signature ) );

CSSM_DEFINE_FUNCTION( VerifyData, (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount,
						CSSM_ALGORITHMS DigestAlgorithm,
						const CSSM_DATA *Signature ) );

CSSM_DEFINE_FUNCTION( VerifyDataInit, (
						_CSSM_CSPSPI_FULL_PARAMS ) );

CSSM_DEFINE_FUNCTION( VerifyDataUpdate, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount ) );

CSSM_DEFINE_FUNCTION( VerifyDataFinal, (
					 	_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *Signature ) );

CSSM_DEFINE_FUNCTION( DigestData, (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount,
						CSSM_DATA_PTR Digest ) );

CSSM_DEFINE_FUNCTION( DigestDataInit, (
						_CSSM_CSPSPI_FULL_PARAMS ) );

CSSM_DEFINE_FUNCTION( DigestDataUpdate, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount ) );

CSSM_DEFINE_FUNCTION( DigestDataClone, (
						_CSSM_CSPSPI_ONLY1ST( CSSM_CSP_HANDLE, CSPHandle )
						CSSM_CC_HANDLE CCHandle,
						_CSSM_PAPI_VSPI( CSSM_CC_HANDLE ) ClonednewCCHandle ) );

CSSM_DEFINE_FUNCTION( DigestDataFinal, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						CSSM_DATA_PTR Digest ) );

CSSM_DEFINE_FUNCTION( GenerateMac, (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount,
						CSSM_DATA_PTR Mac ) );

CSSM_DEFINE_FUNCTION( GenerateMacInit, (
						_CSSM_CSPSPI_FULL_PARAMS ) );

CSSM_DEFINE_FUNCTION( GenerateMacUpdate, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount ) );

CSSM_DEFINE_FUNCTION( GenerateMacFinal, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						CSSM_DATA_PTR Mac ) );

CSSM_DEFINE_FUNCTION( VerifyMac, (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount,
						const CSSM_DATA *Mac ) );

CSSM_DEFINE_FUNCTION( VerifyMacInit, (
						_CSSM_CSPSPI_FULL_PARAMS ) );
    
CSSM_DEFINE_FUNCTION( VerifyMacUpdate, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *DataBufs,
						uint32 DataBufCount ) );

CSSM_DEFINE_FUNCTION( VerifyMacFinal, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *Mac ) );

#ifdef _CSSM_API_DEFINITIONS
/* This function maps to the normal variant of EncryptData */
CSSM_DEFINE_FUNCTION( EncryptData, (
						_CSSM_CSPSPI_FULL_PARAMS, 
						const CSSM_DATA *ClearBufs, 
						uint32 ClearBufCount, 
						CSSM_DATA_PTR CipherBufs, 
						uint32 CipherBufCount, 
						uint32 *bytesEncrypted, 
						CSSM_DATA_PTR RemData ) );
#endif /* ifdef _CSSM_API_DEFINITIONS */

/* This function maps to the 'P' variant of EncryptData and the SPI */
CSSM_DEFINE_FUNCTION( _CSSM_CSPAPI_PNAME( EncryptData ), (
						_CSSM_CSPSPI_FULL_PARAMS, 
						const CSSM_DATA *ClearBufs, 
						uint32 ClearBufCount, 
						CSSM_DATA_PTR CipherBufs, 
						uint32 CipherBufCount, 
						uint32 *bytesEncrypted, 
						CSSM_DATA_PTR RemData,
						CSSM_PRIVILEGE Privilege ) ); 

#ifdef _CSSM_API_DEFINITIONS
/* This function maps to the normal variant of EncryptDataInit */
CSSM_DEFINE_FUNCTION( EncryptDataInit, (
						_CSSM_CSPSPI_FULL_PARAMS ) );
#endif /* ifdef _CSSM_API_DEFINITIONS */

/* This function maps to the 'P' variant of EncryptDataInit and the SPI */
CSSM_DEFINE_FUNCTION( _CSSM_CSPAPI_PNAME( EncryptDataInit ), (
						_CSSM_CSPSPI_FULL_PARAMS,
						CSSM_PRIVILEGE Privilege ) );

CSSM_DEFINE_FUNCTION( EncryptDataUpdate, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *ClearBufs,
						uint32 ClearBufCount,
						CSSM_DATA_PTR CipherBufs,
						uint32 CipherBufCount,
						uint32 *bytesEncrypted ) );

CSSM_DEFINE_FUNCTION( EncryptDataFinal, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						CSSM_DATA_PTR RemData ) );

#ifdef _CSSM_API_DEFINITIONS
/* This function maps to the normal variant of DecryptData */
CSSM_DEFINE_FUNCTION( DecryptData, (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_DATA *CipherBufs,
						uint32 CipherBufCount,
						CSSM_DATA_PTR ClearBufs,
						uint32 ClearBufCount,
						uint32 *bytesDecrypted,
						CSSM_DATA_PTR RemData ) );
#endif /* ifdef _CSSM_API_DEFINITIONS */

/* This function maps to the 'P' variant of DecryptData and the SPI */
CSSM_DEFINE_FUNCTION( _CSSM_CSPAPI_PNAME( DecryptData ), (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_DATA *CipherBufs,
						uint32 CipherBufCount,
						CSSM_DATA_PTR ClearBufs,
						uint32 ClearBufCount,
						uint32 *bytesDecrypted,
						CSSM_DATA_PTR RemData,
						CSSM_PRIVILEGE Privilege ) );

#ifdef _CSSM_API_DEFINITIONS
/* This function maps to the normal variant of DecryptDataInit */
CSSM_DEFINE_FUNCTION( DecryptDataInit, (
						_CSSM_CSPSPI_FULL_PARAMS ) );
#endif /* ifdef _CSSM_API_DEFINITIONS */

/* This function maps to the 'P' variant of DecryptDataInit and the SPI */
CSSM_DEFINE_FUNCTION( _CSSM_CSPAPI_PNAME( DecryptDataInit ), (
						_CSSM_CSPSPI_FULL_PARAMS,
						CSSM_PRIVILEGE Privilege ) );

CSSM_DEFINE_FUNCTION( DecryptDataUpdate, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						const CSSM_DATA *CipherBufs,
						uint32 CipherBufCount,
						CSSM_DATA_PTR ClearBufs,
						uint32 ClearBufCount,
						uint32 *bytesDecrypted ) );

CSSM_DEFINE_FUNCTION( DecryptDataFinal, (
						_CSSM_CSPSPI_SHORT_PARAMS,
						CSSM_DATA_PTR RemData ) );

CSSM_DEFINE_FUNCTION( QueryKeySizeInBits, (
						CSSM_CSP_HANDLE CSPHandle,
						CSSM_CC_HANDLE CCHandle
						_CSSM_CSPSPI_ONLY( const CSSM_CONTEXT *, Context ),
						const CSSM_KEY *Key,
						CSSM_KEY_SIZE_PTR KeySize ) );

#ifdef _CSSM_API_DEFINITIONS
/* This function maps to the normal variant of GenerateKey */
CSSM_DEFINE_FUNCTION( GenerateKey, (
						_CSSM_CSPSPI_FULL_PARAMS,
						uint32 KeyUsage,
						uint32 KeyAttr,
						const CSSM_DATA *KeyLabel,
						const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
						CSSM_KEY_PTR Key ) );
#endif /* ifdef _CSSM_API_DEFINITIONS */

/* This function maps to the 'P' variant of GenerateKey and the SPI */
CSSM_DEFINE_FUNCTION( _CSSM_CSPAPI_PNAME( GenerateKey ), (
						_CSSM_CSPSPI_FULL_PARAMS,
						uint32 KeyUsage,
						uint32 KeyAttr,
						const CSSM_DATA *KeyLabel,
						const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
						CSSM_KEY_PTR Key,
						CSSM_PRIVILEGE Privilege ) );

#ifdef _CSSM_API_DEFINITIONS
/* This function maps to the normal variant of GenerateKeyPair */
CSSM_DEFINE_FUNCTION( GenerateKeyPair, (
						_CSSM_CSPSPI_FULL_PARAMS,
						uint32 PublicKeyUsage,
						uint32 PublicKeyAttr,
						const CSSM_DATA *PublicKeyLabel,
						CSSM_KEY_PTR PublicKey,
						uint32 PrivateKeyUsage,
						uint32 PrivateKeyAttr,
						const CSSM_DATA *PrivateKeyLabel,
						const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
						CSSM_KEY_PTR PrivateKey ) );
#endif /* ifdef _CSSM_API_DEFINITIONS */

/* This function maps to the 'P' variant of GenerateKeyPair and the SPI */
CSSM_DEFINE_FUNCTION( _CSSM_CSPAPI_PNAME( GenerateKeyPair ), (
						_CSSM_CSPSPI_FULL_PARAMS,
						uint32 PublicKeyUsage,
						uint32 PublicKeyAttr,
						const CSSM_DATA *PublicKeyLabel,
						CSSM_KEY_PTR PublicKey,
						uint32 PrivateKeyUsage,
						uint32 PrivateKeyAttr,
						const CSSM_DATA *PrivateKeyLabel,
						const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
						CSSM_KEY_PTR PrivateKey,
						CSSM_PRIVILEGE Privilege ) );

CSSM_DEFINE_FUNCTION( GenerateRandom, (
						_CSSM_CSPSPI_FULL_PARAMS,
						CSSM_DATA_PTR RandomNumber ) );

CSSM_DEFINE_FUNCTION( GenerateAlgorithmParams, (
						_CSSM_CSPSPI_FULL_PARAMS,
						uint32 ParamBits,
						CSSM_DATA_PTR Param
						_CSSM_CSPSPI_ONLY( uint32* , NumberOfUpdatedAttributes )
						_CSSM_CSPSPI_ONLY( CSSM_CONTEXT_ATTRIBUTE_PTR*, UpdatedAttributes ) ) );

#ifdef _CSSM_API_DEFINITIONS
/* This function maps to the normal variant of WrapKey */
CSSM_DEFINE_FUNCTION( WrapKey, (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_ACCESS_CREDENTIALS *AccessCred,
						const CSSM_KEY *Key,
						const CSSM_DATA *DescriptiveData,
						CSSM_WRAP_KEY_PTR WrappedKey ) );
#endif /* ifdef _CSSM_API_DEFINITIONS */

/********************************************************
CSSM_RETURN CSSMAPI CSSM_WrapKeyP (	CSSM_CC_HANDLE CCHandle,
									const CSSM_CRYPTO_DATA *PassPhrase, 
									const CSSM_KEY *Key,
									const CSSM_DATA *DescriptiveData,
									CSSM_WRAP_KEY_PTR WrappedKey,
									CSSM_PRIVILEGE Privilege);
*/
/* This function maps to the 'P' variant of WrapKey and the SPI */
CSSM_DEFINE_FUNCTION( _CSSM_CSPAPI_PNAME( WrapKey ), (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_ACCESS_CREDENTIALS *Credentials,
						const CSSM_KEY *Key,
						const CSSM_DATA *DescriptiveData,
						CSSM_WRAP_KEY_PTR WrappedKey,
						CSSM_PRIVILEGE Privilege ) );

#ifdef _CSSM_API_DEFINITIONS
/* This function maps to the normal variant of UnwrapKey */
CSSM_DEFINE_FUNCTION( UnwrapKey, (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_KEY *PublicKey,
						const CSSM_WRAP_KEY *WrappedKey,
						uint32 KeyUsage,
						uint32 KeyAttr,
						const CSSM_DATA *KeyLabel,
						const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
						CSSM_KEY_PTR UnwrappedKey,
						CSSM_DATA_PTR DescriptiveData ) );
#endif /* ifdef _CSSM_API_DEFINITIONS */

/******************************************
CSSM_RETURN CSSMAPI CSSM_UnwrapKeyP ( CSSM_CC_HANDLE CCHandle,
									const CSSM_CRYPTO_DATA *NewPassPhrase,
									const CSSM_KEY *PublicKey,
									const CSSM_WRAP_KEY *WrappedKey,
									uint32 KeyUsage,
									uint32 KeyAttr,
									const CSSM_DATA *KeyLabel,
									CSSM_KEY_PTR UnwrappedKey,
									CSSM_DATA_PTR DescriptiveData,
									CSSM_PRIVILEGE Privilege);
*/
/* This function maps to the 'P' variant of UnwrapKey and the SPI */
CSSM_DEFINE_FUNCTION( _CSSM_CSPAPI_PNAME( UnwrapKey ), (
						_CSSM_CSPSPI_FULL_PARAMS,
						const CSSM_KEY *PublicKey,
						const CSSM_WRAP_KEY *WrappedKey,
						uint32 KeyUsage,
						uint32 KeyAttr,
						const CSSM_DATA *KeyLabel,
						const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
						CSSM_KEY_PTR UnwrappedKey,
						CSSM_DATA_PTR DescriptiveData,
						CSSM_PRIVILEGE Privilege ) );

CSSM_DEFINE_FUNCTION( DeriveKey, (
						_CSSM_CSPSPI_FULL_PARAMS,
                        CSSM_DATA_PTR Param,
                        uint32 KeyUsage,
                        uint32 KeyAttr,
                        const CSSM_DATA *KeyLabel,
						const CSSM_RESOURCE_CONTROL_CONTEXT *CredAndAclEntry,
                        CSSM_KEY_PTR DerivedKey ) );

CSSM_DEFINE_FUNCTION( FreeKey, (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_ACCESS_CREDENTIALS *AccessCred,
						CSSM_KEY_PTR KeyPtr,
						CSSM_BOOL Delete ) );

/* Expandability Functions */
CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME( PassThrough ), (
						_CSSM_CSPSPI_FULL_PARAMS,
						uint32 PassThroughId,
						const void *InData,
						void **OutData ) );

/* User Login Functions */
CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME( Login ), (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_ACCESS_CREDENTIALS *AccessCred,
						const CSSM_DATA *LoginName,
						const void *Reserved ) );
    
CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME( Logout ), (
						CSSM_CSP_HANDLE CSPHandle ) );

CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME( ChangeLoginAcl ), (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_ACCESS_CREDENTIALS *AccessCred,
						const CSSM_ACL_EDIT *AclEdit ) );

CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME( ObtainPrivateKeyFromPublicKey ), (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_KEY *PublicKey,
						CSSM_KEY_PTR PrivateKey ) );

CSSM_DEFINE_FUNCTION( RetrieveUniqueId, (
						CSSM_CSP_HANDLE CSPHandle,
						CSSM_DATA_PTR UniqueID ) );

CSSM_DEFINE_FUNCTION( RetrieveCounter, (
						CSSM_CSP_HANDLE CSPHandle,
						CSSM_DATA_PTR Counter ) );

CSSM_DEFINE_FUNCTION( VerifyDevice, (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_DATA *DeviceCert ) );

CSSM_DEFINE_FUNCTION( GetTimeValue, (
						CSSM_CSP_HANDLE CSPHandle,
						CSSM_ALGORITHMS TimeAlgorithm,
						CSSM_DATA *TimeData ) );

CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME( GetOperationalStatistics ), (
						CSSM_CSP_HANDLE CSPHandle,
						CSSM_CSP_OPERATIONAL_STATISTICS *Statistics ) );

CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME(GetLoginAcl), (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_STRING *SelectionTag,
						uint32 *NumberOfAclInfos,
						CSSM_ACL_ENTRY_INFO_PTR *AclInfos ) );

CSSM_DEFINE_FUNCTION( GetKeyAcl, (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_KEY *Key,
						const CSSM_STRING *SelectionTag,
						uint32 *NumberOfAclInfos,
						CSSM_ACL_ENTRY_INFO_PTR *AclInfos ) );


CSSM_DEFINE_FUNCTION( ChangeKeyAcl, (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_ACCESS_CREDENTIALS *AccessCred,
						const CSSM_ACL_EDIT *AclEdit,
						const CSSM_KEY *Key ) );

CSSM_DEFINE_FUNCTION( GetKeyOwner, (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_KEY *Key,
						CSSM_ACL_OWNER_PROTOTYPE_PTR Owner ) );

CSSM_DEFINE_FUNCTION( ChangeKeyOwner, (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_ACCESS_CREDENTIALS *AccessCred,
						const CSSM_KEY *Key,
						const CSSM_ACL_OWNER_PROTOTYPE *NewOwner ) );

CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME(GetLoginOwner), (
						CSSM_CSP_HANDLE CSPHandle,
						CSSM_ACL_OWNER_PROTOTYPE_PTR Owner ) );

CSSM_DEFINE_FUNCTION( _CSSM_EXPLICIT_NAME(ChangeLoginOwner), (
						CSSM_CSP_HANDLE CSPHandle,
						const CSSM_ACCESS_CREDENTIALS *AccessCred,
						const CSSM_ACL_OWNER_PROTOTYPE *NewOwner ) );


/* Undefine macros from the pre-processor namespace */
#undef _CSSM_CSPSPI_FULL_PARAMS
#undef _CSSM_CSPSPI_SHORT_PARAMS
#undef _CSSM_EXPLICIT_NAME
#undef _CSSM_PAPI_VSPI
#undef _CSSM_CSPSPI_ONLY1ST
#undef _CSSM_CSPSPI_ONLY
#undef _CSSM_CSPAPI_PNAME
