/*-----------------------------------------------------------------------
 * File: BER_BN.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
 * Basic Encoding Rules and Distinguished Encoding Rules (BER/DER) routines.
 * This file is part of the Token Adaptation Layer (TAL) source code base.
 * The TAL code makes it easier for CSP venders to develop CSPs that plug
 * into the Intel CDSA infrastructure.
 */

#include "ber_bn.h"
#include "isl_internal.h"
#include <isl_memtrack.h>

/* 
 * decode large BER integer into an array of unsigned long values, 
 * return count of unsigned long, or zero if the number is too large (ignoring sign).
 * (NOTE: zero has a length of one).
 */
int 
BER_BerToVeryUnLong(const unsigned char *octets, void *bignum, int maxCount)
{
	int i, Length, noctets;
	unsigned char *target, *array;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x00);

	array = (unsigned char *) bignum;
	/* bypass Tag, assume integer with primitive definite Length */
	if ((*octets++ & 31) == 31)	{ /* long Tag */
		while ((*octets++ & 128) != 0);
	}
	/* compute Length of Contents */
	noctets = *octets & 127;
	if ((*octets++ & 128) != 0) { 
		Length = 0;	 		 /* long definite */
		while (noctets-- > 0) {
			NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x0b);			
			Length = 256 * Length + *octets++;
		}
	} else Length = noctets; /* short definite */
	/* get value */
	target = (unsigned char *) array;
	memset(target, 0, maxCount);
	/* skip leading zeroes (should be at most one if DER) */
	while (Length > 0 && *octets == 0) {
		NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x01);
		Length--;
		octets++;
	}
	if (Length > maxCount) return 0;	/* error: number does not fit */
	// test endian value dynamically
	i = 0x1234;
	if (((unsigned char*) &i)[0] == 0x34) {	/* LITTLE ENDIAN -- intel */
		for (i = 0; i < Length; i++) 
		{
			NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x02);
			target[Length-i-1] = octets[i];
		}
		while (Length > 1 && target[Length-1] == 0) {
			NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x03);
			Length--;
		}
	} else {
		memcpy(target + (maxCount - Length), octets, Length);   /* BIG ENDIAN */
	}
	if (Length == 0) Length = 1;	/* special case for ICL */
	return 	Length;
}

/*	
 * encode "Count" bytes of a bignum into a DER integer at the 
 * given memory, return length in octets 
 */
size_t 
DER_VeryUnLongToDer(
	void *bignum,		/* the number in native byte sex format */
	int Count,		/* size of array in bytes */
	unsigned char *octets)	/* target for DER output */
{
	unsigned char *array, *start;
	size_t result;
	int Length, size, i;

	NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x04);

	array = (unsigned char *) bignum;
	start = octets;
	*octets++ = BER_INTEGER;	/* UNIVERSAL, primitive */
	i = 0x1234;		/* check ENDIAN property */
	if (((unsigned char*) &i)[0] == 0x34) {	 /* LITTLE ENDIAN */
		for (Length = Count; Length > 1; Length--)
		{
			NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x05);
			if (array[Length-1]) break;
		}
		if (array[Length-1] & 0x80) Length++;
		size = BER_OutputLength(Length, octets);
		octets += size;
		if (Length > Count) {
			*octets++ = 0;
			Length--;
		}
		for (Length--; Length >= 0; Length--)
		{
			NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x06);
			*octets++ = array[Length];
		}
	} else {  /* BIG ENDIAN */
		for (i = 0, Length = Count; Length > 1;
				i++, Length--) 
			if (array[i]) break;
		if (array[i] & 0x80) Length++;
		size = BER_OutputLength(Length, octets);
		octets += size;
		if (array[i] & 0x80) {
			*octets++ = 0;
			Length--;
		}
		memcpy(octets, array + i, Length);
		octets += Length;
	}
	result = octets - start;
	return result;
}

uint32
DER_Encode(const EICLData *DataBufs, sint16 DataBufCount, uint8 *Octets)
{

	int		i,
			encodeLen,
			LengthLen,
			disp,
			sumOfEncodeLen = 0;
	unsigned char	*octet_ptr; 
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x07);

	octet_ptr = &Octets[8];
	for (i = 0; i < DataBufCount; i++) {
		NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x08);
		encodeLen = DER_VeryUnLongToDer(DataBufs[i].value, DataBufs[i].length,
					octet_ptr);
		sumOfEncodeLen += encodeLen;
		octet_ptr += encodeLen;
	}

	
	Octets[0] = BER_CONSTRUCTED_SEQUENCE;
	octet_ptr = &Octets[1];
	LengthLen = BER_OutputLength(sumOfEncodeLen, octet_ptr);
	octet_ptr = &Octets[LengthLen+1];
	if (LengthLen+1 < 8) {
		disp = 8 - (LengthLen + 1);
		for (i = 0; i < sumOfEncodeLen; i++) {
		NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x09);
			Octets[8+i-disp] = Octets[8+i];
		}
	}
	return (1+LengthLen+sumOfEncodeLen);
}

sint16
BER_Decode(uint8 *Octets, uint32 OctetCount, EICLData *DataBufs,
		sint16 DataBufCount)
{
	int		octetLen;
	sint16	i;
	const unsigned char	*octet_ptr;
	unsigned char	Tag[8], Content[256];
	BER_PARSED_ITEM parsedBERItem;

    NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x0a);
	octetLen = OctetCount;
	octet_ptr = &Octets[0];
	parsedBERItem.Tag = Tag;
	parsedBERItem.Content = Content;

	octet_ptr = BER_ExpandItem(octet_ptr, octetLen, &parsedBERItem);
	if ((octet_ptr == NULL) || 
			(parsedBERItem.Tag[0] != BER_CONSTRUCTED_SEQUENCE))
		return(0);
	octet_ptr = parsedBERItem.Content;

	for (i = 0; i < DataBufCount; i++) {
		NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_BER_BN_TAG,0x0b);
		if (octet_ptr >= 
			(parsedBERItem.Content + parsedBERItem.ContentLength))
			break;
		DataBufs[i].length = BER_BerToVeryUnLong(octet_ptr, DataBufs[i].value,
						DataBufs[i].length);
		if (DataBufs[i].length == 0) 
			return 0;	/* output buffer was too small */
		octet_ptr += BER_SizeofObject(octet_ptr);
	}
	return (i);
}
