/*-----------------------------------------------------------------------
 * File: PKCS1RSA.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */
 
#include "integapi.h"
#include "isl_cssm.h"
#include <isl_memtrack.h>

extern ISL_SIGN_VERIFY_METHODS PKCS1RSAMD2Methods, PKCS1RSASHA1Methods, PKCS1RSAMD5Methods;
extern ISL_DIGEST_METHODS PKCS7MD2Methods, PKCS7SHA1Methods, PKCS7MD5Methods;

static unsigned char pkcs1rsaoid[] = { 
	0x06, 0x09, 0x2a, 0x86, 0x48, 0x86, 0xf7, 0x0d, 0x01, 0x01, 0x01 };
/*
**	CSSM 1.1a versions of PKCS#1 RSA sign and verify
*/
static void id(
	ISL_SERVICE_CLASS *algClass,				
	ISL_CONST_DATA *algID,				
	ISL_CONST_DATA *serviceName)
{
	*algClass = ISL_ServiceSignVerify;
	algID->Data = pkcs1rsaoid;
	algID->Length = sizeof(pkcs1rsaoid);
	serviceName->Data = (const unsigned char *)"PKCS1 RSA";               /* ifdef VMS apply cast */
	serviceName->Length = strlen((const char *)serviceName->Data);        /* ifdef VMS apply cast */
}

/*
**	The RSA sign is not implemented in 1.1 (especially the staged version), so we use
**	the encrypt API after computing the hash function and padding.
**	In the future, if RSA encrypt adds PCKS#1 padding, we may have to disable the padding
**	(or perhaps even use the sign API) 
*/
#ifdef VMS
static ISL_STATUS sign(
#else
static ISL_ERROR_CODE sign(
#endif
	void *appContext,
	ISL_CONST_DATA parameters,
	const ISL_KEY_BLOCK *key,
	const ISL_DIGEST_METHODS	 *digestMethods,
	ISL_CONST_DATA digestParameters,
	const ISL_GET_DATA_METHODS *getDataMethods,
	ISL_CONST_DATA getDataParameters,
	ISL_DATA *encryptedDigest)
{
	if (digestMethods == &PKCS7MD5Methods) {
		return (*PKCS1RSAMD5Methods.Sign)(appContext, 
			parameters, key, 
			digestMethods, digestParameters, 
			getDataMethods, getDataParameters, 
			encryptedDigest);
	} else if (digestMethods == &PKCS7SHA1Methods) {
		return (*PKCS1RSASHA1Methods.Sign)(appContext, 
			parameters, key, 
			digestMethods, digestParameters, 
			getDataMethods, getDataParameters, 
			encryptedDigest);
	} else if (digestMethods == &PKCS7MD2Methods) {
		return (*PKCS1RSAMD2Methods.Sign)(appContext, 
			parameters, key, 
			digestMethods, digestParameters, 
			getDataMethods, getDataParameters, 
			encryptedDigest);
	}
	return ISL_FAIL;
}

static ISL_ERROR_CODE verify(
	ISL_CONST_DATA parameters,
	const ISL_KEY_BLOCK *key,
	ISL_CONST_DATA encryptedDigest,
	const ISL_DIGEST_METHODS *digestMethods,
	ISL_CONST_DATA digestParameters,
	const ISL_GET_DATA_METHODS *getDataMethods,
	ISL_CONST_DATA getDataParameters)
{
	if (digestMethods == &PKCS7MD5Methods) {
		return (*PKCS1RSAMD5Methods.Verify)(
			parameters, key, encryptedDigest, 
			digestMethods, digestParameters, 
			getDataMethods, getDataParameters);
	} else if (digestMethods == &PKCS7SHA1Methods) {
		return (*PKCS1RSASHA1Methods.Verify)( 
			parameters, key, encryptedDigest, 
			digestMethods, digestParameters, 
			getDataMethods, getDataParameters);
	} else if (digestMethods == &PKCS7MD2Methods) {
		return (*PKCS1RSAMD2Methods.Verify)(
			parameters, key, encryptedDigest, 
			digestMethods, digestParameters, 
			getDataMethods, getDataParameters);
	}
	return INTEGRITY_UNIMPLEMENTED;
}
static ISL_STATUS split(					   /* separate algorithms for PKCS#7 */
		const ISL_DIGEST_METHODS **DigestMethods,	   /* digest algorithm */
		const ISL_SIGN_VERIFY_METHODS **SignMethods)   /* signature algorithm */
{
	return ISL_FAIL;		/* not a combination algorithm */
}

static ISL_STATUS SignWithClass(
		void *AppContext,
		ISL_CONST_DATA Parameters,
		const ISL_KEY_BLOCK *Key,
		const ISL_DIGEST_METHODS *DigestMethods,
		ISL_CONST_DATA DigestParameters,
		const ISL_GET_DATA_METHODS *GetDataMethods,
		ISL_CONST_DATA GetDataParameters,
		ISL_DATA_PTR EncryptedDigest,
		ISL_CLASS_PTR Class){
	
	uint32 i, count;
	ISL_SIGN_VERIFY_METHODS *signMethods;

	
	count = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.NumberOfMethods;

	for (i=0; i<count; i++) {
		if (DigestMethods == ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[i].DigestMethods)
		{
			signMethods = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[i].SignMethods;
			return (*signMethods->SignWithClass)(AppContext, 
			Parameters, Key, 
			DigestMethods, DigestParameters, 
			GetDataMethods, GetDataParameters, 
			EncryptedDigest, signMethods->ServiceMethods.Class);
		}
	}
	return ISL_FAIL;
}

static ISL_ERROR_CODE VerifyWithClass(
		ISL_CONST_DATA Parameters,
		const ISL_KEY_BLOCK *Key,
		ISL_CONST_DATA Signature,
		const ISL_DIGEST_METHODS *DigestMethods,
		ISL_CONST_DATA DigestParameters,
		const ISL_GET_DATA_METHODS *GetDataMethods,
		ISL_CONST_DATA GetDataParameters,
		ISL_CLASS_PTR Class){

	uint32 i, count;
	ISL_SIGN_VERIFY_METHODS *signMethods;

	
	count = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.NumberOfMethods;

	for (i=0; i<count; i++) {
		if (DigestMethods == ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[i].DigestMethods)
		{
			signMethods = ((ISL_SIGN_VERIFY_CLASS_CONTEXT *)Class->ClassContext)->SignatureMethodList.Methods[i].SignMethods;
			return (*signMethods->VerifyWithClass)(
			Parameters, Key, Signature, 
			DigestMethods, DigestParameters, 
			GetDataMethods, GetDataParameters,
			signMethods->ServiceMethods.Class);
		}
	}
	return ISL_FAIL;
}

static ISL_STATUS SplitCombinationWithClass(			/* separate algorithms for PKCS#7 */
		const ISL_DIGEST_METHODS **DigestMethods,	   /* digest algorithm */
		const ISL_SIGN_VERIFY_METHODS **SignMethods,  /* signature algorithm */
		ISL_CLASS_PTR Class){
	 return split(DigestMethods,SignMethods);
}
ISL_SIGN_VERIFY_METHODS PKCS1RSAMethods = {
	id,
	0, 
	sign,
	verify,
	split,
	SignWithClass, 
	VerifyWithClass, 
	SplitCombinationWithClass
};
