/*-----------------------------------------------------------------------
 * File: ATTR.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

/*
 * This file contains the routines for extracting the signer_info and manifest
 * attributes.
 */
#include "isl_internal.h"
#include <isl_memtrack.h>

static ISL_STATUS 
isl_FindAttribute(
	ISL_HEADER_INFO_PTR pHeaderInfo,
	ISL_CONST_DATA Name,
	ISL_CONST_DATA_PTR Value)
{
	ISL_NAME_VALUE_PTR pAttribute = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x01);
	if (NULL == pHeaderInfo)
		return ISL_FAIL;

	for(pAttribute = pHeaderInfo->Attribute;
		pAttribute;
		pAttribute = pAttribute->Next)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x02);
		if (IS_EQUAL(Name, pAttribute->Name))
		{
			Value->Length = pAttribute->Value.Length;
			Value->Data   = pAttribute->Value.Data;
			return ISL_OK;
		}
	}
	return ISL_FAIL;
}


static ISL_ITERATOR_PTR 
isl_CreateAttributeEnumerator(
	const ISL_ARCHIVE_CONTEXT_PTR pArchive,
	const ISL_HEADER_INFO_PTR pHeaderInfo)
{
	ISL_ITERATOR_PTR pItr = NULL;
	ISL_DATA				AttributePointerArray = {0, NULL};
	ISL_NAME_VALUE_PTR		*Attributes = NULL;
	ISL_NAME_VALUE_PTR		pAttribute = NULL;
	uint32					i = 0;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x03);
	if (NULL == pArchive || NULL == pHeaderInfo)
		return NULL;

	AttributePointerArray.Length = sizeof(ISL_NAME_VALUE_PTR) * pHeaderInfo->AttributeCount;
	Attributes = isl_AllocateMemory(pArchive, AttributePointerArray.Length);
	if (!Attributes) {
#if (!defined(EISL) || defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_COPY_MEMORY);
		sprintf(gErrBuf, gErrStr, pArchive, AttributePointerArray.Length);
		isl_ReportError(pArchive, ISL_NO_MEMORY, gErrBuf);
#endif
		return NULL;
	}

	AttributePointerArray.Data = (void*)Attributes;

	for(pAttribute = pHeaderInfo->Attribute, i = 0;
		pAttribute;
		pAttribute = pAttribute->Next, i++)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x04);
		Attributes[i] = pAttribute;
	}

	pItr = isl_CreateIterator(pArchive, &AttributePointerArray);
	isl_FreeMemory(pArchive, Attributes);
	if (!pItr) {
		REPORT_CREATE_ITERATOR_ERROR(pArchive, &AttributePointerArray);
		return NULL;
	}
	return pItr;
}


ISL_STATUS 
EISL_GetNextAttribute(
	const ISL_ITERATOR_PTR Iterator,
	ISL_CONST_DATA_PTR Name,
	ISL_CONST_DATA_PTR Value)
{
	ISL_NAME_VALUE_PTR node = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x05);
	node = (ISL_NAME_VALUE_PTR) isl_GetNextNode(Iterator);
	if (!node)
		return ISL_FAIL;

	Name->Length  = node->Name.Length;
	Name->Data	  = node->Name.Data;
	Value->Length = node->Value.Length;
	Value->Data   = node->Value.Data;
	return ISL_OK;
}

ISL_STATUS 
EISL_RecycleAttributeEnumerator(
	const ISL_ITERATOR_PTR Iterator)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x06);
	return isl_RecycleIterator(Iterator);
}

/*
 * EISL_FindSignerInfoAttribute
 *
 * This function updates the length and pointer to refer to the 
 * Signer info  Attribute (or metadata) Value corresponding to the given name, 
 * or returns ISL_FAIL if there is no such attribute.
 */
ISL_STATUS 
EISL_FindSignerInfoAttribute(
	ISL_VERIFIED_SIGNATURE_ROOT_PTR Context,
	ISL_CONST_DATA Name,
	ISL_CONST_DATA_PTR Value)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x01);
	if (NULL == Context || NULL == Context->Parent->SigSections)
		return ISL_FAIL;
	
	return isl_FindAttribute (&Context->Parent->SigSections->HeaderInfo, Name, Value);

}

/*
 * EISL_CreateSignerInfoAttributeEnumerator
 */
ISL_ITERATOR_PTR 
EISL_CreateSignerInfoAttributeEnumerator(
	const ISL_VERIFIED_SIGNATURE_ROOT_PTR Context)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x03);
	if (NULL == Context || NULL == Context->Parent || NULL == Context->Parent->SigSections)
		return NULL;

	return isl_CreateAttributeEnumerator (Context->Parent->Archive, &Context->Parent->SigSections->HeaderInfo);
}


/*
 * EISL_FindManifestAttribute
 *
 * This function updates the length and pointer to refer to the 
 * Manifest  Attribute (or metadata) Value corresponding to the given name, 
 * or returns ISL_FAIL if there is no such attribute.
 */
ISL_STATUS 
EISL_FindManifestAttribute(
	ISL_VERIFIED_SIGNATURE_ROOT_PTR Context,
	ISL_CONST_DATA Name,
	ISL_CONST_DATA_PTR Value)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x01);
	if (NULL == Context || NULL == Context->Parent || NULL == Context->Parent->Archive)
		return ISL_FAIL;
	
	return isl_FindAttribute (&Context->Parent->Archive->Manifest.Header, Name, Value);

}

/*
 * EISL_CreateManifestAttributeEnumerator
 */
ISL_ITERATOR_PTR 
EISL_CreateManifestAttributeEnumerator(
	const ISL_VERIFIED_SIGNATURE_ROOT_PTR Context)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MANIFEST_TAG,0x03);
	if (NULL == Context || NULL == Context->Parent || NULL == Context->Parent->Archive)
		return NULL;

	return isl_CreateAttributeEnumerator (Context->Parent->Archive,
												&Context->Parent->Archive->Manifest.Header);
}
