/*-----------------------------------------------------------------------
 * File: M_PARSE.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 

#include <stdio.h>
#include <sys/stat.h>
#include "isl_internal.h"
#include "islutil.h"
#include "isl_parse.h"
#include <isl_memtrack.h>

#ifndef EISL
extern ISL_ARCHIVE_CONFIG_METHODS ArchiveConfigMethods;
#endif

/*-----------------------------------------------------------------------------
 * Name: AllocAlgList
 *
 * Description:
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static ISL_ALG_INFO_PTR
AllocAlgList(	
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr,
	unsigned long count)				/* length of list */
{
	unsigned long			i = 0;			/* loop counter */
	ISL_ALG_INFO_PTR		list = NULL;	/* list to be allocated */

	ASSERT(ArchivePtr);
	ASSERT(count > 0);	/* length of list should be positive */

	/* allocating list of ISL_ALG_INFO */
	list = (ISL_ALG_INFO_PTR) isl_AllocateMemory(ArchivePtr, count*sizeof(ISL_ALG_INFO));
	if (!list) {
#if (!defined(EISL) || defined(DEBUG))
		isl_ReportError(ArchivePtr, ISL_NO_MEMORY, gErrBuf);
#endif
		return NULL;
	}
    memset(list, 0, count*sizeof(ISL_ALG_INFO));

	/* initializing each item in the list */
	for(i=0; i < count; i++)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x1f);
		list[i].Parent = NULL;
		list[i].Next = list + (i+1);
	}

    list[count-1].Next = NULL;
	return list;
}


/*-----------------------------------------------------------------------------
 * Name: CountItemsInList
 *
 * Description:
 * Counts the number of items in a space separated list
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static unsigned long
CountItemsInList(
	const unsigned char *ptr, 
	size_t len)
{
	unsigned long count = 0;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x20);
	while (len > 0)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x21);

		while (len > 0 && *ptr == ' ' && !IS_NEWLINE(*ptr))
		{
            NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x22);

			ptr++; len--;
		}

		if (len == 0 || IS_NEWLINE(*ptr))
			return count;

		count++;

		while (len > 0 && IS_HEADERCHAR(*ptr))
		{
            NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x23);

			ptr++; len--;
		}
	}

	return count;
}


/*-----------------------------------------------------------------------------
 * Name: InitAlgList
 *
 * Description:
 * This function will initialize algorithm information list using the value
 * of an attribute called Digest-Algorithms
 *
 * Parameters:
 * list (input/output)			: list of algorithm infos
 * Value (input)				: space separated list of digest algorithms
 * pArchive (input)				: used for list of supported algorithms
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static ISL_STATUS
InitAlgList(
	ISL_ALG_INFO_PTR list, 
	ISL_CONST_DATA Value,
	ISL_ARCHIVE_CONTEXT_PTR	pArchive)
{
	const uint8	        *oldptr = NULL;
    const uint8         *ptr = NULL;
	uint32	            oldlen = 0;
    uint32              len = 0;

    NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x24);

	for(ptr = Value.Data, len = Value.Length;
		list != NULL;
		list = list->Next)
	{
#ifndef EISL
        ISL_SERVICE_CLASS_METHODS *methods;
#endif

        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x25);
		if (!ptr)
			return ISL_FAIL;

		/* looking for beginning of token */
		while (len > 0 && !IS_HEADERCHAR(*ptr) )
		{
            NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x26);
			ptr++;
			len--;
		}

		/* testing for valid length */
		if (!(len > 0))
			return ISL_FAIL;

		/* looking for end of token */
		oldptr = ptr;
		ptr = memchr(oldptr, ' ', len);
	
		/* calculate the length of token */
		oldlen = (ptr) ? (ptr-oldptr) : len;

		/* initialize AlgName */
        list->AlgName.Length = oldlen;
        list->AlgName.Data = isl_CopyMemory(pArchive, oldptr, oldlen);
        if (!list->AlgName.Data) {
			REPORT_COPY_MEMORY_ERROR(pArchive, oldptr, oldlen);
			return ISL_FAIL;
        }
#ifndef EISL
        methods = ArchiveConfigMethods.FindAlgorithm(pArchive->Config, list->AlgName);
		if (!methods) {
			REPORT_ARCHIVE_CONFIG_FIND_ALG_ERROR(pArchive, pArchive->Config, list->AlgName);
		}
        list->Methods = (ISL_DIGEST_METHODS *) methods;
#endif
		/* update length of ptr*/
		len -= oldlen;
	}

	return ISL_OK;
}


/*-----------------------------------------------------------------------------
 * Name: InitializeSectionAlgList
 *
 * Description:
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static ISL_ALG_INFO_PTR
InitializeSectionAlgList(
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr,
	ISL_SECTION_PTR SectionPtr, 
	ISL_CONST_DATA Value)
{
	ISL_ALG_INFO_PTR		list = NULL;
	unsigned long			count = 0;

	count = CountItemsInList(Value.Data,Value.Length);
	list = AllocAlgList(ArchivePtr, count);
	if (!list)
		return NULL;

	if (InitAlgList(list, Value, ArchivePtr) != ISL_OK)
		return NULL;

	SectionPtr->Algorithm = list;
	SectionPtr->AlgCount = count;
	return list;
}

/*-----------------------------------------------------------------------------
 * Name: AddAttribute
 *
 * Description:
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static ISL_STATUS 
AddAttribute(
	ISL_CONST_DATA Name, 
	ISL_CONST_DATA Value, 
	ISL_SECTION_PTR SectionPtr,
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr)
{
	ISL_NAME_VALUE_PTR pAttribute = NULL;
	ISL_NAME_VALUE_PTR pCurr = NULL;	

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x02);
	ASSERT(pManSec);
	pAttribute = (ISL_NAME_VALUE_PTR) isl_AllocateMemory(ArchivePtr,sizeof(ISL_NAME_VALUE));
	if (!pAttribute) {
		REPORT_ALLOCATE_MEMORY_ERROR(ArchivePtr, sizeof(ISL_NAME_VALUE));
		return ISL_FAIL;
	}
	pAttribute->Name = Name;
	pAttribute->Value = Value;
	pAttribute->Next = NULL;

	SectionPtr->AttributeCount++;
	pCurr = SectionPtr->Attribute;

	if (!pCurr)
	{
		SectionPtr->Attribute = pAttribute;
		return ISL_OK;
	}
	
	while (pCurr->Next != NULL) {
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x03);

		pCurr = pCurr->Next;
    }

	pCurr->Next = pAttribute;

	return ISL_OK;
}

/*-----------------------------------------------------------------------------
 * Name: AddAttributeToHeader
 *
 * Description: This function adds attribute (name/value pair) to 
 * header section
 *
 * Parameters: 
 *	ISL_CONST_DATA			(input)			Name, 
 *	ISL_CONST_DATA			(input)			Value, 
 *	ISL_HEADER_PTR	(input/output)	HeaderSectionPtr,
 *	ISL_ARCHIVE_CONTEXT_PTR (input)			ArchivePtr
 *
 * Return value:
 *	ISL_OK
 * 
 * Error Codes:
 *	ISL_FAIL;
 *
 *---------------------------------------------------------------------------*/
static ISL_STATUS 
AddAttributeToHeader(
	ISL_CONST_DATA Name, 
	ISL_CONST_DATA Value, 
	ISL_HEADER_INFO_PTR HeaderSectionPtr,
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr)
{
	ISL_NAME_VALUE_PTR pAttribute = NULL;
	ISL_NAME_VALUE_PTR pCurr = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x02);
	pAttribute = (ISL_NAME_VALUE_PTR) isl_AllocateMemory(ArchivePtr,sizeof(ISL_NAME_VALUE));
	if (!pAttribute) {
		REPORT_ALLOCATE_MEMORY_ERROR(ArchivePtr, sizeof(ISL_NAME_VALUE));
		return ISL_FAIL;
	}
	pAttribute->Name = Name;
	pAttribute->Value = Value;
	pAttribute->Next = NULL;

	HeaderSectionPtr->AttributeCount++;
	pCurr = HeaderSectionPtr->Attribute;
	/* 
	 * Put the new attribute at the end of the list 
	 */
	if (!pCurr)
	{
		HeaderSectionPtr->Attribute = pAttribute;
		return ISL_OK;
	}
	while (pCurr->Next != NULL) {
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x03);
		pCurr = pCurr->Next;
    }
	pCurr->Next = pAttribute;
	
	return ISL_OK;
}

/*-----------------------------------------------------------------------------
 * Name: AddDigest
 *
 * Description:
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static ISL_STATUS
AddDigest(
	ISL_CONST_DATA Name, 
	ISL_CONST_DATA Value, 
	ISL_ALG_INFO_PTR pAlgNode,
	ISL_ARCHIVE_CONTEXT_PTR pArchive)
{
	ISL_HASH_INFO_PTR			pHashNode = NULL;
	ISL_DATA_PTR				pDecHash = NULL;
	uint8 *tmpPtr = NULL;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x12);
	ASSERT(pAlgNode);
	/* Walking Tree to find ISL_ARCHIVE_CONTEXT_PTR */
	ASSERT(!pArchive);

	/* Allocating a HashNode */
	pHashNode = (ISL_HASH_INFO_PTR) isl_AllocateMemory(pArchive,sizeof(ISL_HASH_INFO));
	if(!pHashNode) {
		REPORT_ALLOCATE_MEMORY_ERROR(pArchive,sizeof(ISL_HASH_INFO));
		return ISL_FAIL;
	}

	/* Initializing HashNode */
	pHashNode->Name = Name;
	pDecHash = IslUtil_Base64Decode((ISL_DATA *)&Value);
	if (!pDecHash) 
		return ISL_FAIL;
	pHashNode->Value.Length = pDecHash->Length;

	if ((tmpPtr = isl_AllocateMemory(pArchive, pHashNode->Value.Length)) == NULL) {
		IslUtil_FreeData(pDecHash);
		REPORT_ALLOCATE_MEMORY_ERROR(pArchive, pHashNode->Value.Length);
		return ISL_FAIL;
	}
	memcpy(tmpPtr, pDecHash->Data, pHashNode->Value.Length);
	pHashNode->Value.Data = tmpPtr;

	IslUtil_FreeData(pDecHash);
	pHashNode->Next = pAlgNode->Hash;
	
	/* Inserting at the head of the list */
	pAlgNode->Hash = pHashNode;
	pAlgNode->HashCount++;
	return ISL_OK;
}



/*-----------------------------------------------------------------------------
 * Name: isl_GetProtocolAndParameters
 *
 * Description:
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static ISL_STATUS
isl_GetProtocolAndParameters(
	ISL_CONST_DATA Name, 
	ISL_CONST_DATA_PTR Protocol, 
	ISL_CONST_DATA_PTR Parameters)
{
	if (NULL == Protocol || NULL == Parameters) 
		return ISL_FAIL;

	Parameters->Data = memchr(Name.Data,':',Name.Length);
	if (NULL == Parameters->Data)
		return ISL_FAIL;
	Parameters->Data++;    /* skip over ':' */
	Parameters->Length = Name.Length - (Parameters->Data - Name.Data);
	if (Parameters->Length > Name.Length) 
		return ISL_FAIL;
	Protocol->Data = Name.Data;
	Protocol->Length = Name.Length - (Parameters->Length + 1);
	if (Protocol->Length > Name.Length) 
		return ISL_FAIL;
	return ISL_OK;
}






/*-----------------------------------------------------------------------------
 * Name: isl_ParseMemoryWithTemplate
 *
 * Description:
 * This function parses the given contents for the manifest using the 
 * defines in the template. This function succeeds in case of a valid manifest
 * and returns pointers to various manifest sections/fields in Results. It fails
 * in case of an invalid manifest.
 *
 * Parameters:
 * ArchivePtr (input)			: Archive context
 * Content (input)				: Manifest contents to be parsed.
 * Template (input)				: template that specifies the content type 
 *									(e.g manifest/signer_info)
 * Results (output)				: List of pointers to various/sections fields in parsed content.
 *
 * Return value:
 *	ISL_OK - in case of a valid manifest
 *  ISL_FAIL - otherwise.
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
ISL_STATUS
isl_ParseMemoryWithTemplate(
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr,
	ISL_CONST_DATA Content,
	ISL_CONST_DATA Template,
	ISL_DATA_CHAIN_PTR *Results)
{
	ISL_DATA defines;
	unsigned char defines_buf[TEMPLATE_BUFFER_SIZE];

	/*
	 * Get the definitions for a valid manifest.
	 */
	defines.Length = TEMPLATE_BUFFER_SIZE;
	defines.Data = defines_buf;
	if (isl_get_template (&defines) != ISL_OK)
		return ISL_FAIL;

	/*
	 * parse the contents.
	 */
	if (parseMain(ArchivePtr, (ISL_CONST_DATA_PTR)&defines, &Template, &Content,
#ifndef NODBG
		  0,
#endif
		  Results, MANIFEST_LAST_INDEX) == 0)
		return ISL_FAIL;

	return ISL_OK;

}

ISL_STATUS
BuildSectionTemplate(
	ISL_ARCHIVE_CONTEXT_PTR	Context,
	ISL_CONST_DATA_PTR	Template)
{
    uint8   *ptr = NULL;
    uint8   len = 10;
    
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_ARCHIVE_TAG,0x0f);
	
	/* allocate memory */
	Template->Data = isl_AllocateMemory(Context, len);
    if (!Template->Data) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, len);
	}

    ptr = (uint8 *)Template->Data;

	/* build up template */
    *(ptr++) = '<';
    *(ptr++) = 's';
    *(ptr++) = 'e';
    *(ptr++) = 'c';
    *(ptr++) = 't';
    *(ptr++) = 'i';
    *(ptr++) = 'o';
    *(ptr++) = 'n';
    *(ptr++) = '>';

    Template->Length = ptr - Template->Data;
    *(ptr++) = '\0';
    return ISL_OK;
}

/*-----------------------------------------------------------------------------
 * Name: StripAttributeValue
 *
 * Description:
 * This function will strip the attribute value of continuation characters.
 * continutation characters.  This routine will copy the buffer in case the
 * Attribute value contains a continuation character.
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static ISL_STATUS 
StripAttributeValue(
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr,
	ISL_DATA_PTR AttrValue)
{
	uint32 oldlen;
	uint8 *oldptr;
	uint8 *endptr;
	uint8 *DataPtr;
	uint8 *newPtr;
	uint8 *CR;
	uint8 *LF;
	uint8 *ptr;

	if (NULL == AttrValue)
		return ISL_FAIL;

	oldptr = AttrValue->Data;
	oldlen = AttrValue->Length;

	CR = memchr(oldptr, '\r', oldlen);
	LF = memchr(oldptr, '\n', oldlen);
	ptr = (CR) ? CR : LF;

	/*
	 * If no continuation, simply return ISL_OK
	 */
    if (ptr == NULL)
		return ISL_OK;

	if ((newPtr = isl_AllocateMemory(ArchivePtr, oldlen)) == NULL)
		return ISL_FAIL;

	DataPtr = newPtr;
	endptr = oldptr + oldlen;

	while (oldptr < endptr) 
	{
		CR = memchr(oldptr, '\r', oldlen);
		LF = memchr(oldptr, '\n', oldlen);
		ptr = (CR) ? CR : LF;
        if (ptr == NULL)							/* tesing for newline */
		{
			memcpy (DataPtr, oldptr, oldlen);			
			DataPtr += oldlen;						/* increment counter */
			break;									/* exit */
		}
		else
		{	
			oldlen = ptr - oldptr;					/* calculating length */
			memcpy (DataPtr, oldptr, oldlen);
			DataPtr += oldlen;						/* incrementing length */
		}
		ptr = (LF) ? LF : CR;						/* finding newline */
		ptr++;										/* skipping newline */
        if (*ptr == ' ')							/* testing for continuation */
            ptr++;									/* skipping space */
        oldptr = ptr;
		oldlen = endptr - oldptr;
	}
	AttrValue->Data = newPtr;
	AttrValue->Length = DataPtr - newPtr;
	return ISL_OK;
}

ISL_STATUS
isl_BuildHeaderInfo(
	ISL_HEADER_INFO_PTR HeaderInfoPtr,
	ISL_DATA_CHAIN_PTR *resultsPtr,
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr)
{
	ISL_DATA_CHAIN_PTR attrName;
	ISL_DATA_CHAIN_PTR attrValue;

	HeaderInfoPtr->MajorVersion = MANIFEST_MAJOR;
	HeaderInfoPtr->MinorVersion = MANIFEST_MINOR;

	attrName = resultsPtr[HEADER_ATTR_NAME_INDEX];
	attrValue = resultsPtr[HEADER_ATTR_VALUE_INDEX];
	
	while (attrName && attrValue) {
		if (StripAttributeValue (ArchivePtr, attrValue->ptr) == ISL_FAIL)
			return ISL_FAIL;
	
		if (AddAttributeToHeader(*(ISL_CONST_DATA_PTR)attrName->ptr, *(ISL_CONST_DATA_PTR)attrValue->ptr,
												HeaderInfoPtr, ArchivePtr) != ISL_OK)
				return ISL_FAIL;
		attrName = attrName->next;
		attrValue = attrValue->next;
	}
	return ISL_OK;
}


/*-----------------------------------------------------------------------------
 * Name: MatchDigest
 *
 * Description:
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
static ISL_ALG_INFO_PTR 
MatchDigest(
	ISL_CONST_DATA Name, 
	ISL_ALG_INFO_PTR AlgList)
{
	ISL_CONST_DATA AlgName;		/* digest name e.g. md5, sha-1 */	
	int len = 0;			/* length of the digest name */

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x1c);
	ASSERT(AlgList);

	len = (int)(Name.Length - HASH_STR_LENGTH);	/* length of the digest name */
	if (len <= 0)
		return NULL;

	/* loop through AlgList for AlgName */
	for(AlgName.Data = Name.Data, 	AlgName.Length = len;
		AlgList != NULL;
		AlgList = AlgList->Next)
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x1d);
		if (IS_EQUAL(AlgName, AlgList->AlgName))
			return (AlgList);
	}

	return NULL;	/* no match found */

}


static ISL_STATUS
InitializeSection(
	ISL_SECTION_PTR	SectionPtr,
	ISL_DATA_CHAIN_PTR *SectionInfoPtr,
	ISL_ARCHIVE_CONTEXT_PTR ArchivePtr)

{
	ISL_ALG_INFO_PTR pDigest = NULL;
	ISL_ALG_INFO_PTR pHashInfo = NULL;
	ISL_DATA_CHAIN_PTR AttrNamePtr;
	ISL_DATA_CHAIN_PTR AttrValuePtr;
	ISL_DATA_CHAIN_PTR digestVal, digestName;


	/*
	 * Assign the well defined section attributes.
	 */
	SectionPtr->Image = *(ISL_CONST_DATA_PTR)SectionInfoPtr[SECTION_INDEX]->ptr;

	if (SectionInfoPtr[NAME_INDEX]) {
		if (StripAttributeValue (ArchivePtr, SectionInfoPtr[NAME_INDEX]->ptr) == ISL_FAIL)
			return ISL_FAIL;
		SectionPtr->Name = *(ISL_CONST_DATA_PTR)SectionInfoPtr[NAME_INDEX]->ptr;
	}

	if (SectionInfoPtr[SECTION_NAME_INDEX]) {
		if (StripAttributeValue (ArchivePtr, SectionInfoPtr[SECTION_NAME_INDEX]->ptr) == ISL_FAIL)
			return ISL_FAIL;
		SectionPtr->SectionName = *(ISL_CONST_DATA_PTR)SectionInfoPtr[SECTION_NAME_INDEX]->ptr;
	}

	if (SectionInfoPtr[DIGEST_ALG_INDEX]) {
		if (StripAttributeValue (ArchivePtr, SectionInfoPtr[DIGEST_ALG_INDEX]->ptr) == ISL_FAIL)
			return ISL_FAIL;
		pDigest = InitializeSectionAlgList(ArchivePtr, SectionPtr, 
												*(ISL_CONST_DATA_PTR)SectionInfoPtr[DIGEST_ALG_INDEX]->ptr);
	}

	digestName = SectionInfoPtr[DIGEST_NAME_INDEX];
	digestVal = SectionInfoPtr[DIGEST_VAL_INDEX];
	while (digestName && digestVal) {
		pHashInfo = MatchDigest(*(ISL_CONST_DATA_PTR)digestName->ptr, pDigest);
		if (pHashInfo) {
			if (StripAttributeValue (ArchivePtr, digestVal->ptr) == ISL_FAIL)
				return ISL_FAIL;
			if (AddDigest(*(ISL_CONST_DATA_PTR)digestName->ptr,
							*(ISL_CONST_DATA_PTR)digestVal->ptr,
							pHashInfo, ArchivePtr) != ISL_OK)
				return ISL_FAIL;
			digestName = digestName->next;
			digestVal = digestVal->next;
		}
	}

	/*
	 * Assign the section attributes.
	 */
	AttrNamePtr = SectionInfoPtr[SECTION_ATTR_NAME_INDEX];
	AttrValuePtr = SectionInfoPtr[SECTION_ATTR_VALUE_INDEX];
	while (AttrNamePtr) {
		ISL_CONST_DATA Name;
		ISL_CONST_DATA Value;

		Name = *(ISL_CONST_DATA_PTR)AttrNamePtr->ptr;
		Value = *(ISL_CONST_DATA_PTR)AttrValuePtr->ptr;
		if (StripAttributeValue (ArchivePtr, (ISL_DATA_PTR)&Value) == ISL_FAIL)
			return ISL_FAIL;
	
		if (AddAttribute(Name, Value, SectionPtr, ArchivePtr) != ISL_OK)
				return ISL_FAIL;

		AttrNamePtr = AttrNamePtr->next;
		AttrValuePtr = AttrValuePtr->next;
	}
	
	return ISL_OK;

}


ISL_STATUS
isl_BuildManifestSection(
	ISL_MANIFEST_SECTION_PTR ManifestSectionPtr,
	ISL_DATA_CHAIN_PTR *SectionInfoPtr)
{
#ifndef EISL
    ISL_SERVICE_CLASS_METHODS *methods;
#endif
	ISL_ARCHIVE_CONTEXT_PTR pArchive;
	ISL_SECTION_PTR SectionPtr;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x00);

	if (NULL == SectionInfoPtr || NULL == ManifestSectionPtr) 
		return ISL_FAIL;

	pArchive = ManifestSectionPtr->Parent;
	if (pArchive == NULL) 
		return ISL_FAIL;

	SectionPtr = &(ManifestSectionPtr->SectionInfo);
	
	if (InitializeSection(SectionPtr, SectionInfoPtr, pArchive) != ISL_OK)
		return ISL_FAIL;

	
	if (isl_GetProtocolAndParameters (ManifestSectionPtr->SectionInfo.Name,
									&ManifestSectionPtr->Protocol,
									&ManifestSectionPtr->Parameters) == ISL_FAIL)
		return ISL_FAIL;
	
#ifndef EISL
	methods = ArchiveConfigMethods.FindAlgorithm(
		pArchive->Config, 
		ManifestSectionPtr->Protocol);
	
	ManifestSectionPtr->GetDataMethods = (ISL_GET_DATA_METHODS *)methods;
#endif

	return ISL_OK;
}

/*-----------------------------------------------------------------------------
 * Name: isl_BuildSigSection
 *
 * Description:
 *
 * Parameters: 
 *
 * Return value:
 * 
 * Error Codes:
 *---------------------------------------------------------------------------*/
ISL_STATUS
isl_BuildSigSection( 
	ISL_SIG_SECTION_PTR SigSectionPtr,
	ISL_DATA_CHAIN_PTR *SectionInfoPtr)
{	
	ISL_ARCHIVE_CONTEXT_PTR		ArchivePtr;
	ISL_MANIFEST_SECTION_PTR	ManifestSectionPtr;
	ISL_SECTION_PTR				SectionPtr;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x0f);
	if (NULL == SigSectionPtr || NULL == SigSectionPtr->Parent ||
								NULL == SigSectionPtr->Parent->Archive)
		return ISL_FAIL;

	ArchivePtr = SigSectionPtr->Parent->Archive;
 	SectionPtr = &(SigSectionPtr->SectionInfo);
	
	if (InitializeSection(SectionPtr, SectionInfoPtr, ArchivePtr) != ISL_OK)
		return ISL_FAIL;
 
	for (ManifestSectionPtr = ArchivePtr->Manifest.Section;
								 ManifestSectionPtr != NULL;
								 ManifestSectionPtr = ManifestSectionPtr->Next) {
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MPARSE_TAG,0x0c);
		if (IS_EQUAL(ManifestSectionPtr->SectionInfo.Name, SigSectionPtr->SectionInfo.Name))
			SigSectionPtr->ManSect = ManifestSectionPtr;
	}
	
	if (SigSectionPtr->ManSect == NULL) 
		return ISL_FAIL;

	return ISL_OK;
}
