/*-----------------------------------------------------------------------
 * File: ARCHIVE.C
 *
 * Copyright (c) 1995-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

 
/*
**	Implementation of ISL_ARCHIVE_CONTEXT class 
*/
#ifndef EISL
#include <fcntl.h>
#ifdef WIN32
#include <io.h>
#include <direct.h>
#endif
#include <stdio.h>
#include <sys/stat.h>
#include "integapi.h"
#endif

#include "isl_internal.h"
#include <isl_memtrack.h>

extern ISL_STATUS Archive_InitializeNew(ISL_ARCHIVE_CONTEXT *result,
								ISL_ARCHIVE_CONFIG_PTR configContext, 	
								ISL_CONST_DATA Path);


#ifndef EISL
extern struct isl_archive_methods ArchiveMethods;
extern struct isl_manifest_section_methods ManifestSectionMethods;
extern struct isl_archive_config_methods ArchiveConfigMethods;
extern struct isl_signature_methods	PKCS7SignatureMethods;
extern struct isl_certificate_methods X509CertMethods; 
#endif

/*
 * EISL-ISL Common functions.
 */




/*
 * ISL-ONLY_FUNCTIONS below this line.
 */

#ifndef EISL
/* internal routines */
/* shallow memory copy--add new item to archive's free list */
void *isl_CopyMemory(ISL_ARCHIVE_CONTEXT *Context, const void *Buffer, ISL_SIZE Size)
{
	struct isl_memory_buffer *item;

	if (!Context) return NULL;

	if (!Buffer) {
#if (defined(DEBUG)) 
		gErrStr = isl_GetErrorStr(ISL_COPY_MEMORY);
		sprintf(gErrBuf, gErrStr, Context, Buffer, Size);
		isl_ReportError(Context, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return NULL;
	}

	item = ISL_MALLOC_WRAPPER(Context->Config->MemoryMethods->Malloc, 
		Context->Config->MemoryContext, sizeof(struct isl_memory_buffer));
	if (!item) return 0;
	item->Buffer = ISL_MALLOC_WRAPPER(Context->Config->MemoryMethods->Malloc, 
		Context->Config->MemoryContext, Size);
	if (!item->Buffer) {
		ISL_FREE_WRAPPER(
			Context->Config->MemoryMethods->Free,
			Context->Config->MemoryContext,
			item);
		return 0;
	}
	item->Next = Context->Buffers;
	Context->Buffers = item;
	memcpy(item->Buffer, Buffer, Size);
	return item->Buffer;
}



void *isl_AllocateMemory(ISL_ARCHIVE_CONTEXT *archive, ISL_SIZE size)
{
	struct isl_memory_buffer *item;

	if (!archive) return NULL;

	item = ISL_MALLOC_WRAPPER(
		archive->Config->MemoryMethods->Malloc,
		archive->Config->MemoryContext, sizeof(struct isl_memory_buffer));
	if (!item) 
		return 0;
	item->Buffer = ISL_MALLOC_WRAPPER(
		archive->Config->MemoryMethods->Malloc,
		archive->Config->MemoryContext, size);
	if (!item->Buffer) {
		ISL_FREE_WRAPPER(
			archive->Config->MemoryMethods->Free,
			archive->Config->MemoryContext,
			item);
		return 0;
	}
	item->Next = archive->Buffers;
	archive->Buffers = item;
	memset(item->Buffer, 0, size);
	return item->Buffer;
}


void isl_FreeMemory(ISL_ARCHIVE_CONTEXT *Context, void *Buffer)
{
	ISL_MEMORY_BUFFER_PTR	Curr = NULL;
	ISL_MEMORY_BUFFER_PTR	Prev = NULL;

	if (!Context) return;
	if (!Buffer) return;

	for(Curr = Context->Buffers;											/* looping through all buffers */
		Curr;																/* exit condition */
		Curr = Curr->Next)													/* getting next buffer */
	{
        NAMELOOP( ISL_DSCM_TAG, ISL_EISL_TAG, ISL_MEMORY_TAG,0x05);
		if (Buffer == Curr->Buffer)											/* match pointers */
		{
			if (!Prev)														/* testing for head of list */
				Context->Buffers = Curr->Next;								/* head of list is next buffer */
			else
				Prev->Next = Curr->Next;									/* pointer previous bufffer to next buffer */
			
			ISL_FREE_WRAPPER(
				Context->Config->MemoryMethods->Free, 						/* free buffer */
				Context->Config->MemoryContext, 
				(void*) Buffer);

			ISL_FREE_WRAPPER(
				Context->Config->MemoryMethods->Free, 						/* free book keeping */
				Context->Config->MemoryContext, 
				(void*) Curr);																			
			return;
		}
		Prev = Curr;														/* save current pointer */
	}
	return;
	
	
}




ISL_STATUS isl_CertificateExternalize(ISL_ARCHIVE_CONTEXT_PTR pContext,
									  ISL_CERTIFICATE_LIST_PTR pCertList)
{
	uint8 metaPath[256];
	PK_CONST_DATA filename;
	PK_CONST_DATA filedata;
#define METAINFDIRSLASH "META-INF/"

	if(!pContext)
		return ISL_FAIL;

	if ((!pCertList) || 
		(!pCertList->Cert) || 
		(!pCertList->Name.Data) ||
		(!pCertList->Name.Length))
		return ISL_FAIL;

	filename.Data = metaPath;
	filename.Length = strlen(METAINFDIRSLASH);
	memcpy(metaPath, METAINFDIRSLASH, filename.Length);
	memcpy(metaPath + filename.Length, 
		pCertList->Name.Data, 
		pCertList->Name.Length);
	filename.Length += pCertList->Name.Length;
	memcpy(metaPath + filename.Length, ".CER", 4);
	filename.Length += 4;
	filedata.Data = pCertList->Cert->theCert.Data;
	filedata.Length = pCertList->Cert->theCert.Length;
	if (pk_AddFileToNewArchive(pContext->NewZipArchive, filename, filedata) != PK_OK) {
		REPORT_PK_ADD_FILE_TO_NEW_ARCHIVE(pContext, filename, filedata);
		return ISL_FAIL;
	}
	return ISL_OK;
}

/* class methods */
static	ISL_SIZE SizeofObject()			/* returns sizeof object */
{	
	return sizeof(ISL_ARCHIVE_CONTEXT);
}

/*
const char NewManifestHeader[] = "Manifest-Version: 2.0\r\n\r\n";
*/
const char NewManifestHeader[] = "Manifest-Version: ";

static	ISL_STATUS Externalize(			/* create exteral rep. of archive */
		ISL_ARCHIVE_CONTEXT *context)		/* context */
{
	uint32 ManifestSize;
	PK_STATUS pkstatus;
	PK_CONST_DATA filename, filedata;
	ISL_CONST_DATA result;
	ISL_CONST_DATA headerResult;
	ISL_MANIFEST_SECTION_PTR section;
	ISL_SIGNATURE_LIST_PTR sigCtx;
	ISL_STATUS status;
	ISL_CERTIFICATE_LIST_PTR pCertList;

	if (!context) return ISL_FAIL;

	/*
	 * Flatten the manifest header
	 */
	if (isl_HeaderFlatten(context, &context->Manifest.Header, NewManifestHeader, &headerResult) != ISL_OK)
		return ISL_FAIL;

	ManifestSize = headerResult.Length;

	/* 
	 * for each manifest section, create the external representation for it 
	 */ 
	for (section = context->Manifest.Section; section; section = section->Next) {
		status = isl_SectionFlatten(context, &section->SectionInfo, &result);
		if (status != ISL_OK){
#if (defined(DEBUG)) 
			gErrStr = isl_GetErrorStr(ISL_MANIFEST_SECTION_FLATTEN);
			sprintf(gErrBuf, gErrStr, section, &result);
			isl_ReportError(context, ISL_IO_ERROR, gErrBuf);
#endif
			return status;
		}
		ManifestSize += result.Length;
	}
	/* initialize new zip archive */
	filedata.Data = isl_AllocateMemory(context, ManifestSize);
	if (!filedata.Data) {
		REPORT_ALLOCATE_MEMORY_ERROR(context, ManifestSize);
	}
	filedata.Length = 0;
	context->NewZipArchive = pk_CreateNewArchive(context);
	if (!context->NewZipArchive) {
#if (defined(DEBUG)) 
		gErrCode = ISL_IO_ERROR;
		gErrStr = isl_GetErrorStr(PK_CREATE_NEW_ARCHIVE);
		sprintf(gErrBuf, gErrStr, context);
#endif
		goto PK_FAIL;
	}
	/* create manifest.mf file */
	/*
	 * Copy the manifest header data.
	 */
	memcpy((void*)filedata.Data, headerResult.Data, headerResult.Length);
	filedata.Length = headerResult.Length;
	/*
	 * Copy the data for each of the sections.
	 */
	for (section = context->Manifest.Section; section; section = section->Next) {
		memcpy((void*)(filedata.Data + filedata.Length), section->SectionInfo.Image.Data, 
			section->SectionInfo.Image.Length);
		filedata.Length += section->SectionInfo.Image.Length;
	}
	filename.Data = (uint8 *)"META-INF/MANIFEST.MF";
	filename.Length = strlen((char *)filename.Data);
	pkstatus = pk_AddFileToNewArchive(context->NewZipArchive, filename, filedata);
	if (pkstatus != PK_OK){
#if (defined(DEBUG)) 
		gErrCode = ISL_IO_ERROR;
		gErrStr = isl_GetErrorStr(PK_ADD_FILE_TO_NEW_ARCHIVE);
		sprintf(gErrBuf, gErrStr, context->NewZipArchive, filename, filedata);
#endif
		goto PK_FAIL;
	}
	/* create the signature info and signature block files */
	for (sigCtx = context->Signatures; sigCtx; sigCtx = sigCtx->Next) {
		status = (*PKCS7SignatureMethods.Externalize)(sigCtx->Signature);
		if (status != ISL_OK) {
#if (defined(DEBUG)) 
			gErrCode = ISL_IO_ERROR;
			gErrStr = isl_GetErrorStr(ISL_SIGNATURE_EXTERNALIZE);
			sprintf(gErrBuf, gErrStr, sigCtx);
#endif
			goto PK_FAIL;
		}
	}

	
	/* create certificate file for each certificate in the archive */
	for (pCertList=context->Certs; pCertList; pCertList=pCertList->Next)
	{
		status = isl_CertificateExternalize(context, pCertList);
		if (status != ISL_OK) {
#if (defined(DEBUG)) 
			gErrCode = ISL_IO_ERROR;
			gErrStr = isl_GetErrorStr(ISL_CERTIFICATE_EXTERNALIZE);
			sprintf(gErrBuf, gErrStr, sigCtx);
#endif
		goto PK_FAIL;
		}
	}

	filename.Data = context->PathName.Data;
	filename.Length = context->PathName.Length;
	pkstatus = pk_WriteNewArchive(context->NewZipArchive, filename);
	if (pkstatus != PK_OK) {
#if (defined(DEBUG)) 
		gErrCode = ISL_IO_ERROR;
		gErrStr = isl_GetErrorStr(PK_WRITE_NEW_ARCHIVE);
		sprintf(gErrBuf, gErrStr, context->NewZipArchive, filename);
#endif
		goto PK_FAIL;
	}
	pkstatus = pk_RecycleNewArchive(context->NewZipArchive);
	context->NewZipArchive = 0;
	if (pkstatus != PK_OK) {
#if (defined(DEBUG)) 
		gErrCode = ISL_IO_ERROR;
		gErrStr = isl_GetErrorStr(PK_RECYCLE_NEW_ARCHIVE);
		sprintf(gErrBuf, gErrStr, context->NewZipArchive);
#endif
		goto PK_FAIL;
	}
	return ISL_OK;

PK_FAIL:
	(void) pk_RecycleNewArchive(context->NewZipArchive);
	context->NewZipArchive = 0;
#if (!defined(DEBUG)) 
	isl_ReportError(context, gErrCode, gErrBuf);
#endif
	return ISL_FAIL;
}

static	ISL_MANIFEST_SECTION_PTR CreateNewSignableObject(	/* add a new object that can be signed */
		ISL_ARCHIVE_CONTEXT *Context,		/* context */
		ISL_CONST_DATA Name,
		ISL_CONST_DATA GetServiceName,
		ISL_CONST_DATA GetParameters)
{
	ISL_MANIFEST_SECTION *item, *tmpSection, *prevSection = NULL;
	uint8 *temp;
	ISL_SECTION *pSectionInfo;

	item = NULL;

	if (!Context) return NULL;

	if ( (!GetServiceName.Data) ||
		 (!GetParameters.Data) )
	{
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_CREATE_NEW_SIGNABLE_OBJECT);
		sprintf(gErrBuf, gErrStr, Context, Name.Data, GetServiceName.Data, GetParameters.Data);
		isl_ReportError(Context, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return NULL;
	}
	/* allocate memory for item */
	item = (ISL_MANIFEST_SECTION *)isl_AllocateMemory(Context, sizeof(ISL_MANIFEST_SECTION));				
	if (!item) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, sizeof(ISL_MANIFEST_SECTION));
		return NULL;
	}
	pSectionInfo = &item->SectionInfo;

	memset(item, 0, sizeof(ISL_MANIFEST_SECTION));
#ifndef EISL
	item->Methods = &ManifestSectionMethods;
#endif
	item->Parent = Context;
	
	if (Name.Data || (Name.Length > 0)) {
		pSectionInfo->SectionName.Length  = Name.Length;	/* copy the unique section name */
		pSectionInfo->SectionName.Data =  
			isl_CopyMemory(Context, (void *)Name.Data, Name.Length);
		if (!pSectionInfo->SectionName.Data){ 
			REPORT_COPY_MEMORY_ERROR(Context, (void *)Name.Data, Name.Length);
			return NULL;
		}
	}

	item->Protocol.Length  = GetServiceName.Length;	/* copy the protocol */
	item->Protocol.Data =  
		isl_CopyMemory(Context, (void *)GetServiceName.Data, GetServiceName.Length);
	if (!item->Protocol.Data) {
		REPORT_COPY_MEMORY_ERROR(Context, (void *)GetServiceName.Data, GetServiceName.Length);
		return NULL;
	}

	item->Parameters.Length  = GetParameters.Length;	/* copy the parmeter */
	item->Parameters.Data =  
		isl_CopyMemory(Context, (void *)GetParameters.Data, GetParameters.Length);
	if (!item->Parameters.Data) { 
		REPORT_COPY_MEMORY_ERROR(Context, (void *)GetParameters.Data, GetParameters.Length);
		return NULL;
	}
	/* the format of archive specific name is:
		protocol:parameters (i.e. file:foo.txt)
	*/
	pSectionInfo->Name.Length = item->Protocol.Length + item->Parameters.Length + 1;
	pSectionInfo->Name.Data = (uint8 *)isl_AllocateMemory(Context, pSectionInfo->Name.Length);
	if (!pSectionInfo->Name.Data) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, pSectionInfo->Name.Length);
		return NULL;
	}
	temp = (uint8 *) pSectionInfo->Name.Data;
	memcpy(temp, item->Protocol.Data, item->Protocol.Length);
	temp += item->Protocol.Length;
	memcpy(temp, ":", 1);
	temp += 1;
	memcpy(temp, item->Parameters.Data, item->Parameters.Length);

	/* copy get data methods */
#ifndef EISL
	item->GetDataMethods = (ISL_GET_DATA_METHODS_PTR) (*ArchiveConfigMethods.FindAlgorithm)(
								item->Parent->Config, 
								GetServiceName);
	if (!item->GetDataMethods) {
		REPORT_ARCHIVE_CONFIG_FIND_ALG_ERROR(
			item->Parent, 
			item->Parent->Config, 
			GetServiceName);
		return NULL;
	}
#endif	
	item->Next = NULL;

	/* add the new signable object to the archive */
	if (Context->Manifest.Section == NULL)
	{
		Context->Manifest.Section = item;
	}
	else
	{
		tmpSection = Context->Manifest.Section;
		while(tmpSection != NULL) {
			prevSection = tmpSection;
			tmpSection  = tmpSection->Next;
		}
		prevSection->Next = item;
	}

	Context->Manifest.SectionCount++;

	return item;
}

static	ISL_ITERATOR_PTR CreateSignableObjectEnumerator(
		ISL_ARCHIVE_CONTEXT *Context)		/* pointer to configuration context */
{
	ISL_ITERATOR_PTR		pItr = NULL;
	ISL_DATA				ManifestSectionPointerArray = {0, NULL};
	ISL_MANIFEST_SECTION_PTR pManSect = NULL;
	ISL_MANIFEST_SECTION_PTR *ManifestSections = NULL;
	unsigned long i = 0;

	if (!Context)
		return NULL;

	if (!Context->Manifest.SectionCount)
		return NULL;

	ManifestSectionPointerArray.Length = sizeof(ISL_MANIFEST_SECTION_PTR) *
										 Context->Manifest.SectionCount;

	ManifestSections = isl_AllocateMemory(Context,
										  ManifestSectionPointerArray.Length);
	if (!ManifestSections){
		REPORT_ALLOCATE_MEMORY_ERROR(
			Context,
			ManifestSectionPointerArray.Length);
		return NULL;
	}

	for(pManSect = Context->Manifest.Section, i = Context->Manifest.SectionCount - 1;
		pManSect;
		pManSect = pManSect->Next, i--)
	{
		ManifestSections[i] = pManSect;
	}

	ManifestSectionPointerArray.Data = (void*)ManifestSections;

	pItr = isl_CreateIterator(Context, &ManifestSectionPointerArray);
	if (!pItr) {
		REPORT_CREATE_ITERATOR_ERROR(Context, &ManifestSectionPointerArray);
		return NULL;
	}

	isl_FreeMemory(Context, ManifestSections);

	return pItr;
}


static 	ISL_MANIFEST_SECTION_PTR GetNextSignableObject(
		ISL_ITERATOR_PTR Iterator)
{
	void * node = isl_GetNextNode(Iterator);
	return (ISL_MANIFEST_SECTION_PTR) node;
}

static	ISL_STATUS RecycleSignableObjectEnumerator(
		ISL_ITERATOR_PTR Iterator)
{
	return isl_RecycleIterator(Iterator);
}

static 	
ISL_SIGNATURE_CONTEXT *
CreateNewSignature(	                                /* add PGP/PKCS#7 signature to archive */
		ISL_ARCHIVE_CONTEXT *ArchiveContext,		/* archive context */
		ISL_SIGNATURE_CONFIG *SigConfigContext,     /* signature config context */
		ISL_CONST_DATA Name)				        /* file name within archive for signature */
{
	ISL_SIGNATURE_CONTEXT *SigContext=NULL;
	ISL_STATUS status;

	if (!ArchiveContext) {
		return NULL;
	}

	if ( (!SigConfigContext) ||
		 (!Name.Data) )
	{
#if (defined(DEBUG))
			gErrStr = isl_GetErrorStr(ISL_ARCHIVE_CREATE_NEW_SIGNATURE);
			sprintf(gErrBuf, gErrStr, ArchiveContext, SigConfigContext, Name.Data);
			isl_ReportError(ArchiveContext, ISL_INVALID_PARAMETER, gErrBuf);
#endif
			return NULL;
	}
	/* allocate memory for the new signature context */
	SigContext = (ISL_SIGNATURE_CONTEXT *)isl_AllocateMemory(ArchiveContext, sizeof(ISL_SIGNATURE_CONTEXT));				
	if (!SigContext) {
		REPORT_ALLOCATE_MEMORY_ERROR(ArchiveContext, sizeof(ISL_SIGNATURE_CONTEXT));
		return NULL;
	}

	memset(SigContext, 0, sizeof(ISL_SIGNATURE_CONTEXT));

	/* initialize new signature */
#ifdef EISL
	status = isl_InitializeNewSignature(SigContext,
										SigConfigContext, 
										ArchiveContext, 
										Name);
#else
	status = PKCS7SignatureMethods.InitializeNew(SigContext,
										         SigConfigContext, 
										         ArchiveContext, 
										         Name);
#endif
	if (status != ISL_OK) {
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_SIGNATURE_INITIALIZE_NEW);
		sprintf(gErrBuf, gErrStr, SigContext, SigConfigContext, ArchiveContext, Name);
		isl_ReportError(ArchiveContext, ISL_IO_ERROR, gErrBuf);
#endif
		return NULL;
	}
	
	return SigContext;
}

static	ISL_STATUS RemoveSignature(				/* finish component list and enclosing signature */
		ISL_ARCHIVE_CONTEXT *Context,		/* archive context */
		ISL_SIGNATURE_CONTEXT *Signature)	/* signature within the archive */
{
	ISL_SIGNATURE_LIST *current, *prev;
	
	if (!Context) return ISL_FAIL;
	
	current = Context->Signatures;
	prev = current;

	/*
	 * search through the signature list.  delete the siganture context
	 * from the  the list if found 
	 */
	while (current) {
		if (current->Signature == Signature)  {
			/* head of the signature list */
			if (prev == current) 
				Context->Signatures = current->Next;
			else 
				prev->Next = current->Next;
			break;
		}
		prev = current;
		current = current->Next;
	}

	if (!current) {
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_REMOVE_SIGNATURE);
		sprintf(gErrBuf, gErrStr, Context, Signature);
		isl_ReportError(Context, ISL_SIGNATURE_NOT_FOUND, gErrBuf);
#endif
		return ISL_FAIL;
	}
	return ISL_OK;
}

static 	ISL_SIGNATURE_CONTEXT *FindSignature(/* return found object or NULL */
		ISL_ARCHIVE_CONTEXT *Context,		/* context */
		const ISL_CONST_DATA Name)				/* archive-unique name */
{
    ISL_SIGNATURE_LIST   *sigContext = NULL;

	if (!Context) return NULL;

	if (!Name.Data) {
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_FIND_SIGNATURE);
		sprintf(gErrBuf, gErrStr, Context, Name.Data);
		isl_ReportError(Context, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return NULL;
	}

    for(sigContext = Context->Signatures;
        sigContext != NULL;
        sigContext = sigContext->Next)
    {
        if ( (Name.Length == sigContext->Name.Length) &&
             (memcmp(Name.Data, sigContext->Name.Data, Name.Length) == 0) )
             break;
    }
	if (!sigContext) {
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_FIND_SIGNATURE);
		sprintf(gErrBuf, gErrStr, Context, Name);
		isl_ReportError(Context, ISL_SIGNATURE_NOT_FOUND, gErrBuf);
#endif
		return NULL;
	}
	
	return sigContext->Signature;
}

static	ISL_ITERATOR_PTR CreateSignatureEnumerator(
		ISL_ARCHIVE_CONTEXT *Context)		/* pointer to configuration context */
{
	ISL_ITERATOR_PTR		pItr = NULL;
	ISL_DATA				SignaturePointerArray = {0, NULL};
	ISL_SIGNATURE_LIST_PTR pSignatureCtx = NULL;
	ISL_SIGNATURE_LIST_PTR *Signatures = NULL;
	unsigned long i = 0;
	int signatureCount;

	if (!Context) return NULL;

	if (NULL == Context->Signatures)
		return NULL;

	//calculte the total number of signatures
	pSignatureCtx = Context->Signatures;
	signatureCount = 0;
	while (pSignatureCtx) {
		signatureCount++;
		pSignatureCtx = pSignatureCtx->Next;
	}

	SignaturePointerArray.Length = sizeof(ISL_SIGNATURE_LIST_PTR) *
										 signatureCount;

	Signatures = isl_AllocateMemory(Context,
									SignaturePointerArray.Length);
	if (!Signatures) {
		REPORT_ALLOCATE_MEMORY_ERROR(
			Context,
			SignaturePointerArray.Length);
		return NULL;
	}

	for(pSignatureCtx = Context->Signatures, i = 0;
		pSignatureCtx;
		pSignatureCtx = pSignatureCtx->Next, i++)
	{
		Signatures[i] = pSignatureCtx;
	}

	SignaturePointerArray.Data = (void*)Signatures;

	pItr = isl_CreateIterator(Context, &SignaturePointerArray);
	isl_FreeMemory(Context, Signatures);
	if (!pItr) {
		REPORT_CREATE_ITERATOR_ERROR(Context, &SignaturePointerArray);
		return NULL;
	}

	return pItr;
}

static	ISL_SIGNATURE_CONTEXT *GetNextSignature(
		ISL_ITERATOR_PTR Iterator)
{
	void * node = isl_GetNextNode(Iterator);
	if (node)
		return (ISL_SIGNATURE_CONTEXT_PTR) ((ISL_SIGNATURE_LIST_PTR)node)->Signature;
	else
		return NULL;
}

static	ISL_STATUS RecycleSignatureEnumerator(
		ISL_ITERATOR_PTR Iterator)
{
	return isl_RecycleIterator(Iterator);
}

static	ISL_STATUS AddCertificate(			/* add certificate file to archive */
		ISL_ARCHIVE_CONTEXT *Context,		/* archive context */
		ISL_CONST_DATA Name,				/* archive-unique name */
		ISL_CERTIFICATE *Certificate)		/* certificate */
{
	ISL_CERTIFICATE_LIST_PTR	pNode;
	ISL_CERTIFICATE_PTR			pCert;
	ISL_STATUS					status;
	ISL_CONST_DATA				theCert;

	if (!Context) return ISL_FAIL;

	if ( (!Name.Data) ||
		 (!Certificate) )
	{
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_ADD_CERTIFICATE);
		sprintf(gErrBuf, gErrStr, Context, Name.Data, Certificate);
		isl_ReportError(Context, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return ISL_FAIL;
	}

	pNode = isl_AllocateMemory(Context, sizeof(ISL_CERTIFICATE_LIST));
	if (!pNode) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, sizeof(ISL_CERTIFICATE_LIST));
		return ISL_FAIL;
	}
	memset(pNode, 0, sizeof(ISL_CERTIFICATE_LIST));

	pCert = isl_AllocateMemory(Context, X509CertMethods.SizeofObject()); 
	if (!pCert) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, X509CertMethods.SizeofObject());
		return ISL_FAIL;
	}
	memset(pCert, 0, X509CertMethods.SizeofObject());
	
	status = X509CertMethods.Flatten(Certificate, &theCert);
	if (status != ISL_OK) {
		return ISL_FAIL;
	}
	status = X509CertMethods.Initialize(
		pCert,
		Context->Config->MemoryContext,
		Context->Config->MemoryMethods,
		theCert);
	if (status != ISL_OK) {
		return ISL_FAIL;
	}
	pNode->Cert = pCert;

	pNode->Name.Length = Name.Length;
	pNode->Name.Data   = isl_CopyMemory(Context, Name.Data, Name.Length);
	if (!pNode->Name.Data) {
		REPORT_COPY_MEMORY_ERROR(Context, Name.Data, Name.Length);
		return ISL_FAIL;
	}

	pNode->Next = Context->Certs;
	Context->Certs = pNode;
	return ISL_OK;
}

static	ISL_STATUS RemoveCertificate(		/* remove certificate file from archive */
		ISL_ARCHIVE_CONTEXT *Context,		/* archive context */
		ISL_CERTIFICATE *Certificate)		/* certificate */
{
	ISL_CERTIFICATE_LIST_PTR	curr, prev;
	ISL_CONST_DATA				theCert;
	ISL_STATUS					status;

	if (!Context) return ISL_FAIL;

	if (!Certificate) {
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_REMOVE_CERTIFICATE);
		sprintf(gErrBuf, gErrStr, Context, Certificate);
		isl_ReportError(Context, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return ISL_FAIL;
	}
	status = X509CertMethods.Flatten(Certificate, &theCert);
	if (status != ISL_OK) {
		return ISL_FAIL;
	}

	for(curr = Context->Certs, prev = Context->Certs;
		curr;
		curr = curr->Next) 
	{
		ISL_CONST_DATA		theCurrCert;

		status = X509CertMethods.Flatten(Certificate, &theCurrCert);
		if (status != ISL_OK) {
			continue;
		}

		if (IS_EQUAL(theCert, theCurrCert))
		{
			if (curr != prev) {
				prev->Next = curr->Next;
			} else {
				Context->Certs = curr->Next;
			}
			return ISL_OK;
		}
		prev = curr;
	}

#if (defined(DEBUG))
	gErrStr = isl_GetErrorStr(ISL_ARCHIVE_REMOVE_CERTIFICATE);
	sprintf(gErrBuf, gErrStr, Context, Certificate);
	isl_ReportError(Context, ISL_OBJECT_NOT_FOUND, gErrBuf);
#endif
	return ISL_FAIL;
}

static	ISL_CERTIFICATE *FindCertificate(	/* return found object or NULL */
		ISL_ARCHIVE_CONTEXT *Context,		/* context */
		const ISL_CONST_DATA Name)			/* archive-unique name */
{
	ISL_CERTIFICATE_LIST_PTR	curr;

	if (!Context) return NULL;

	if (!Name.Data) {
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_FIND_CERTIFICATE);
		sprintf(gErrBuf, gErrStr, Context, Name.Data);
		isl_ReportError(Context, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return NULL;
	}

	for(curr = Context->Certs;
		curr;
		curr = curr->Next) 
	{
		if (IS_EQUAL(curr->Name, Name))
		{
			return curr->Cert;
		}
	}

#if (defined(DEBUG))
	gErrStr = isl_GetErrorStr(ISL_ARCHIVE_FIND_CERTIFICATE);
	sprintf(gErrBuf, gErrStr, Context, Name.Data);
	isl_ReportError(Context, ISL_OBJECT_NOT_FOUND, gErrBuf);
#endif
	return NULL;
}

static	ISL_ITERATOR_PTR CreateCertificateEnumerator(
		ISL_ARCHIVE_CONTEXT_PTR Context)		/* pointer to archive context */
{
	ISL_ITERATOR_PTR			pItr = NULL;
	ISL_DATA					CertificatePointerArray = {0, NULL};
	ISL_CERTIFICATE_LIST_PTR	pList = NULL;
	ISL_CERTIFICATE_PTR			*Certificates;

	unsigned long i = 0;
	int Count;

	if (!Context) return NULL;

	//calculte the total number of signatures
	pList = Context->Certs;
	Count = 0;
	while (pList) {
		Count++;
		pList = pList->Next;
	}

	CertificatePointerArray.Length = sizeof(ISL_CERTIFICATE_PTR) *
										 Count;

	Certificates = isl_AllocateMemory(Context,
									CertificatePointerArray.Length);
	if (!Certificates) {
		REPORT_ALLOCATE_MEMORY_ERROR(
			Context,
			CertificatePointerArray.Length);
		return NULL;
	}

	for(pList = Context->Certs, i = 0;
		pList;
		pList = pList->Next, i++)
	{
		Certificates[i] = pList->Cert;
	}

	CertificatePointerArray.Data = (void*)Certificates;

	pItr = isl_CreateIterator(Context, &CertificatePointerArray);
	isl_FreeMemory(Context, Certificates);
	if (!pItr) {
		REPORT_CREATE_ITERATOR_ERROR(Context, &CertificatePointerArray);
		return NULL;
	}

	return pItr;
}

static	ISL_CERTIFICATE *GetNextCertificate(
		ISL_ITERATOR_PTR Iterator)
{
	void * node;

	if (!Iterator) return NULL;
	node = isl_GetNextNode(Iterator);
	return (ISL_CERTIFICATE *) node;

}

static	ISL_STATUS RecycleCertificateEnumerator(
		ISL_ITERATOR_PTR Iterator)
{
	if (!Iterator) return ISL_FAIL;
	return isl_RecycleIterator(Iterator);
}

static	ISL_STATUS RemoveSignableObject(		/* remove an object that can be signed */
		ISL_ARCHIVE_CONTEXT *Context,		/* arhive context */
		ISL_MANIFEST_SECTION_PTR Section)	/* manifest section */
{
	ISL_MANIFEST_SECTION_PTR prev, current;

	if (!Context) return ISL_FAIL;

	//search through manifest list for the specified section.
	//If found, delete it from the list
	prev = Context->Manifest.Section;
	current = prev;
	while (current) {
		if (current == Section)  {
			/* head of the signature list */
			if (prev == current) 
				Context->Manifest.Section = current->Next;
			else 
				prev->Next = current->Next;
			Context->Manifest.SectionCount--;
			return ISL_OK;
		}
		prev = current;
		current = current->Next;
	}

#if (defined(DEBUG))
	gErrStr = isl_GetErrorStr(ISL_ARCHIVE_REMOVE_SIGNABLE_OBJECT);
	sprintf(gErrBuf, gErrStr, Context, Section);
	isl_ReportError(Context, ISL_OBJECT_NOT_FOUND, gErrBuf);
#endif
	return ISL_FAIL;

}

static	ISL_MANIFEST_SECTION_PTR FindSignableObject(	/* return found object or NULL */
		ISL_ARCHIVE_CONTEXT *Context,		/* context */
		ISL_CONST_DATA Name)				/* manifest section name */
{
	ISL_MANIFEST_SECTION_PTR current;

	if (!Context) return NULL;

	if (!Name.Data) {
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_FIND_SIGNABLE_OBJECT);
		sprintf(gErrBuf, gErrStr, Context, Name.Data);
		isl_ReportError(Context, ISL_OBJECT_NOT_FOUND, gErrBuf);
#endif
		return NULL;
	}

	//search through manifest list for the specified section.
	//If found, delete it from the list
	for(current = Context->Manifest.Section;
		current;
		current = current->Next) 
	{
		if (IS_EQUAL(current->SectionInfo.Name, Name))  {
			return current;
		}
		
	}

#if (defined(DEBUG))
	gErrStr = isl_GetErrorStr(ISL_ARCHIVE_FIND_SIGNABLE_OBJECT);
	sprintf(gErrBuf, gErrStr, Context, Name.Data);
	isl_ReportError(Context, ISL_OBJECT_NOT_FOUND, gErrBuf);
#endif
	return NULL;
}

static ISL_STATUS GetArchivePath(
		ISL_ARCHIVE_CONTEXT *Context,	/*context */
		ISL_CONST_DATA_PTR PathName)	
{
	if (!Context) return ISL_FAIL;

	if (!Context->PathName.Data || Context->PathName.Length == 0) {	
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_GET_ARCHIVE_PATH);
		sprintf(gErrBuf, gErrStr, Context, PathName);
		isl_ReportError(Context, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return ISL_FAIL;
	}

	PathName->Length = Context->PathName.Length;
	PathName->Data =  isl_CopyMemory(Context, (void*) Context->PathName.Data, PathName->Length);
	if (!PathName->Data) {	
		REPORT_COPY_MEMORY_ERROR(Context, (void*) Context->PathName.Data, PathName->Length);
		return ISL_FAIL;
	}
	return ISL_OK;
}

static ISL_STATUS GetArchiveDirectory(
		ISL_ARCHIVE_CONTEXT *Context,	/*context */
		ISL_CONST_DATA_PTR DirName)	
{
	long i;

	if (!Context) return ISL_FAIL;

	if (!Context->PathName.Data || Context->PathName.Length == 0) {	
#if (defined(DEBUG))
		gErrStr = isl_GetErrorStr(ISL_ARCHIVE_GET_ARCHIVE_DIRECTORY);
		sprintf(gErrBuf, gErrStr, Context, Context->PathName);
		isl_ReportError(Context, ISL_INVALID_PARAMETER, gErrBuf);
#endif
		return ISL_FAIL;
	}

	for(i = (long) (Context->PathName.Length-1); i>0; i--) {
		if (Context->PathName.Data[i] == '\\')
			break;
	}
	DirName->Length = (uint32) i;
	DirName->Data =  isl_CopyMemory(Context, (void*) Context->PathName.Data, DirName->Length);
	if (!DirName->Data) {	
		REPORT_COPY_MEMORY_ERROR(Context, (void*) Context->PathName.Data, DirName->Length);
		return ISL_FAIL;
	}
	return ISL_OK;
}


static	ISL_STATUS AddAttribute(			/* add metadata to manifest header */
		ISL_ARCHIVE_CONTEXT_PTR Context,	/* archive context */
		ISL_CONST_DATA Attribute,			/* attribute "name" */
		ISL_CONST_DATA Value) 				/* attribute value */
{

	ISL_NAME_VALUE *item;
	ISL_NAME_VALUE *temp, *prev = NULL;
	ISL_HEADER_INFO_PTR pHeaderInfo;

	if (NULL == Context || NULL == Attribute.Data || NULL == Value.Data)
		return ISL_FAIL;

	if ((item = isl_CreateAttribute(Context, Attribute, Value)) == NULL)
		return ISL_FAIL;


	pHeaderInfo = &Context->Manifest.Header;

	/* 
	 * add the attribute to the end of attribute list 
	 */
	temp = pHeaderInfo->Attribute;
	while (temp!= NULL)	{
		prev = temp;
		temp = temp->Next;
	}
	if (pHeaderInfo->Attribute == NULL) /* head of the attribute list */
		pHeaderInfo->Attribute = item;
	else
		prev->Next = item;

	pHeaderInfo->AttributeCount++;
	return ISL_OK;
}


static ISL_STATUS RemoveAttribute(				/* remove metadata from Manifest header */
		ISL_ARCHIVE_CONTEXT_PTR Context,		/* Archive context */
		ISL_CONST_DATA Attribute)				/* attribute "name" */
{
	ISL_NAME_VALUE *curr, *prev;
	ISL_HEADER_INFO_PTR pHeaderInfo;

    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG,ISL_PKCS7SIG_TAG,0x1a);
	
	if (NULL == Context || NULL == Attribute.Data)
		return ISL_FAIL;

	pHeaderInfo = &Context->Manifest.Header;

	prev = pHeaderInfo->Attribute;
	for (curr = prev; curr; curr = curr->Next)
	{
		if (IS_EQUAL(curr->Name, Attribute))
		{
			if (prev != curr) {
				prev->Next = curr->Next;
			} else {
				pHeaderInfo->Attribute = curr->Next;
			}

			pHeaderInfo->AttributeCount--;
			return ISL_OK;
		}
		prev = curr;
	}

	return ISL_FAIL;

}

static ISL_STATUS FindAttributeValue(			/* find metadata in manifest header */
		ISL_ARCHIVE_CONTEXT_PTR Context,			/* signer context */
		ISL_CONST_DATA Attribute,				/* attribute "name" */
		ISL_CONST_DATA_PTR Value)				/* fill in Value and Length if found */
{
	ISL_NAME_VALUE *curr;
	ISL_HEADER_INFO_PTR pHeaderInfo;


    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG,ISL_PKCS7SIG_TAG,0x1b);
	
	if (NULL == Context || NULL == Attribute.Data || NULL == Value)
		return ISL_FAIL;

	pHeaderInfo = &Context->Manifest.Header;

	for (curr = pHeaderInfo->Attribute; curr; curr = curr->Next)
	{
		if (IS_EQUAL(curr->Name, Attribute))
		{
			Value->Length  = curr->Value.Length;	/* copy the unique section name */
			Value->Data =  isl_CopyMemory(Context, curr->Value.Data, Value->Length);
			if (!Value->Data) {
				REPORT_COPY_MEMORY_ERROR(Context, curr->Value.Data, Value->Length);
			}
			return ISL_OK;
		}
	}

	return ISL_FAIL;
}


static ISL_ITERATOR_PTR CreateAttributeEnumerator(
		ISL_ARCHIVE_CONTEXT_PTR Context)		/* pointer to Archive context */
{
	ISL_ITERATOR_PTR pItr = NULL;
	ISL_DATA				AttributePointerArray = {0, NULL};
	ISL_NAME_VALUE_PTR		*Attributes = NULL;
	ISL_NAME_VALUE_PTR		pAttribute = NULL;
	uint32					i = 0;
	ISL_HEADER_INFO_PTR pHeaderInfo;

    
	NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG,ISL_PKCS7SIG_TAG,0x1c);


	if (NULL == Context)
		return NULL;
	
	pHeaderInfo = &Context->Manifest.Header;

	AttributePointerArray.Length = sizeof(ISL_NAME_VALUE_PTR) * pHeaderInfo->AttributeCount;
	Attributes = isl_AllocateMemory(Context, AttributePointerArray.Length);
	if (!Attributes) {
		REPORT_ALLOCATE_MEMORY_ERROR(Context, AttributePointerArray.Length);
		return NULL;
	}

	AttributePointerArray.Data = (void*)Attributes;

	for(pAttribute = pHeaderInfo->Attribute, i = 0;
		pAttribute;
		pAttribute = pAttribute->Next, i++)
	{
		Attributes[i] = pAttribute;
	}

	pItr = isl_CreateIterator(Context, &AttributePointerArray);
	isl_FreeMemory(Context, Attributes);
	if (!pItr) {
		REPORT_CREATE_ITERATOR_ERROR(Context, &AttributePointerArray);
		return NULL;
	}
	return pItr;

}

static ISL_STATUS GetNextAttribute(
		ISL_ITERATOR_PTR Context,
		ISL_CONST_DATA_PTR Name,
		ISL_CONST_DATA_PTR Value)
{
	ISL_NAME_VALUE_PTR node = NULL;

	NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG,ISL_PKCS7SIG_TAG,0x1d);

	if (NULL == Context || NULL == Name || NULL == Value)
		return ISL_FAIL;

	node = (ISL_NAME_VALUE_PTR) isl_GetNextNode(Context);
	if (!node)
		return ISL_FAIL;

	Name->Length  = node->Name.Length;
	Name->Data	  = node->Name.Data;
	Value->Length = node->Value.Length;
	Value->Data   = node->Value.Data;
	return ISL_OK;

}

#ifndef EISL
static ISL_STATUS RecycleAttributeEnumerator(
		ISL_ITERATOR_PTR Iterator)
{
    NAMEPROC( ISL_DSCM_TAG, ISL_EISL_TAG,ISL_PKCS7SIG_TAG,0x1e);

	return isl_RecycleIterator(Iterator);
}
#endif



ISL_ARCHIVE_METHODS ArchiveMethods = {
		SizeofObject,
		Archive_InitializeNew,
		Archive_InitializeFromPath,
		Archive_InitializeFromMemory,
		Externalize,
		isl_RecycleArchiveContext,
		CreateNewSignableObject,
		RemoveSignableObject,
		FindSignableObject,
		CreateSignableObjectEnumerator,
		GetNextSignableObject,
		RecycleSignableObjectEnumerator,
		CreateNewSignature,
		RemoveSignature,
		FindSignature,
		CreateSignatureEnumerator,
		GetNextSignature,
		RecycleSignatureEnumerator,
		AddCertificate,
		RemoveCertificate,
		FindCertificate,
		CreateCertificateEnumerator,
		GetNextCertificate,
		RecycleCertificateEnumerator,
		GetArchivePath,
		GetArchiveDirectory,
		AddAttribute,
		RemoveAttribute,
		FindAttributeValue,
		CreateAttributeEnumerator,
		GetNextAttribute,
		RecycleAttributeEnumerator,
		0,
		0
};

#endif
