/*-----------------------------------------------------------------------
 * File: OBJECTREF.CPP
 *
 * Copyright (c) 1998-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

// ObjectRef.cpp
//
// This file contains the class implementation for the CObjectRef
//  class, a class which implements an object reference based on 
//  the CDSA 2.0 Signed Manifest Specification (Open Group standard)
//  and its definition of a "manifest section".
//////////////////////////////////////////////////////////////////////

#ifdef CSSM_MEMTRACK
#undef new
#include <crtdbg.h>
static char SZ_FILE_NAME[] = __FILE__;
#define new new(_NORMAL_BLOCK, SZ_FILE_NAME, __LINE__)  /*need space*/
#endif /* ifdef VMS comment this out - CSSM_MEMTRACK */

#include "ObjectRef.h"
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <sys/stat.h>
#include <scMemTrack.h>
#include "cssmport.h"

//////////////////////////////////////////////////////////////////////
// CObjectRef::CObjectRef()
//  constructs an empty CObjectRef object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CObjectRef::CObjectRef()
{
    mName = 0;
    mType = RefNone;
    mReferencePath = 0;
    mAttributes = 0;
}


//////////////////////////////////////////////////////////////////////
// CObjectRef::~CObjectRef()
//  destructs a CObjectRef object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CObjectRef::~CObjectRef()
{
    if( mName )
        free( mName );
    if( mReferencePath )
        free( mReferencePath );
    if( mAttributes )
        delete mAttributes;
}


//////////////////////////////////////////////////////////////////////
// CObjectRef::CObjectRef(char* Name, 
//                        REF_TYPE Type,
//                        char* ManifestPath)
//  constructs a CObjectRef object with  the provided name, path, and
//   type
//
// Parameters:
//  Name(input)         name to be attached to the object in the manifest
//                      (must be unique WRT the manifest)
//  Type(input)         type of reference (tells how to retrieve)
//  ManifestPath        path to the object
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CObjectRef::CObjectRef(char* Name, 
                       REF_TYPE Type,
                       char* ReferencePath)
{
	mName = (char*)malloc(strlen(Name)+1);
    memset( mName, 0, strlen(Name)+1);
    memcpy(mName, Name, strlen(Name));
	mType = Type;
	mReferencePath = (char*)malloc(strlen(ReferencePath)+1);
    memset( mReferencePath, 0, strlen(ReferencePath)+1);
    memcpy(mReferencePath, ReferencePath, strlen(ReferencePath));
    mAttributes = 0;
}


//////////////////////////////////////////////////////////////////////
// CObjectRef::CObjectRef(const CObjectRef *object)
//  constructs a CObjectRef object which is a copy of the one passed 
//
// Parameters:
//  object(input)       pointer to the CObjectRef object to copy
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CObjectRef::CObjectRef(const CObjectRef *object)
{
	mName = (char*)calloc(strlen(object->mName)+1, 1);
    memcpy(mName, object->mName, strlen(object->mName));
	mType = object->mType;
	mReferencePath = (char*)calloc(strlen(object->mReferencePath)+1, 1);
    memcpy( mReferencePath, object->mReferencePath, strlen(object->mReferencePath));
    if(object->mAttributes)
        mAttributes = new CAttributeList(object->mAttributes);
    else
        mAttributes = 0;
}


//////////////////////////////////////////////////////////////////////
// CObjectRef::CObjectRef(char* Name, 
//                        REF_TYPE Type,
//                        char* SourcePath, 
//                        char* ManifestPath, 
//                        CAttributeList *Attributes )
//  constructs a CObjectRef object with the provided name, path, type,
//   and attributes.  The source path is used to find where the original
//   of the referant object is
//
// Parameters:
//  Name(input)         name to be attached to the object in the manifest
//  Type(input)         type of reference (tells how to retrieve)
//  SourcePath(input)   path to the original object
//  ManifestPath(input) path to the object, as it is to be embedded in
//                      the manifest
//  Attributes(input)   pointer to a list of attributes to be associated
//                      with this reference
// Returns:
//  none
// Side effects:
//  Will create a copy of the original object in a scratch directory,
//   so that the embedded ManifestPath is correct
//////////////////////////////////////////////////////////////////////
CObjectRef::CObjectRef( char* Name, 
                        REF_TYPE Type, 
                        char* SourcePath, 
                        char* ReferencePath, 
                        CAttributeList *Attributes )
{
    FILE* fp;
    char scratch_path[WF_MAX_PATH];
    struct _stat st;
    unsigned char* buffer = 0;

    mName = (char*)calloc(strlen(Name)+1, 1);
    memcpy(mName, Name, strlen(Name));
	mType = Type;
	mReferencePath = (char*)calloc(strlen(ReferencePath)+1, 1);
    memcpy(mReferencePath, ReferencePath, strlen(ReferencePath));
    int RefFlag = 0;
    if( Attributes )
    {
        mAttributes = new CAttributeList( Attributes );
    
        CAttributeListNode *node = mAttributes->mHead;
        int NodeCount = mAttributes->GetNumberOfNodes();

        for( int i=0; i<NodeCount; i++ )
        {
            // check for meta data 
            char name[WF_MAX_STRING], value[WF_MAX_STRING];
            memset( name, 0, WF_MAX_STRING );
            memset( value, 0, WF_MAX_STRING );
        
            if(node->mAttribute.Length < WF_MAX_STRING)
                memcpy( name, node->mAttribute.Data, node->mAttribute.Length );
            else // truncate because strings are too long
                memcpy( name, node->mAttribute.Data, WF_MAX_STRING-1);
        
            if(node->mValue.Length < WF_MAX_STRING)
                memcpy( value, node->mValue.Data, node->mValue.Length );
            else // truncate because strings are too long
                memcpy( value, node->mValue.Data, WF_MAX_STRING-1);

            if( !strcmp( name, "Integrity-VerifyData") || !strcmp( value, "reference" ) )
                RefFlag = 1;
        
            node = node->mNext;
        }
    }
    else
        mAttributes = 0;

    if(!RefFlag)
    {
        // copy file from source to scratch dir
#ifdef WIN32
        sprintf( scratch_path, "%s\\%s", MRM_SCRATCH_DIR, ReferencePath );
#else
        sprintf( scratch_path, "%s/%s", MRM_SCRATCH_DIR, ReferencePath );
#endif
        // load source file into buffer
        fp = fopen( SourcePath, "rb" );
        if( !fp )
            return;
        // stat the file and find out how much memory we need
	    if ((_stat(SourcePath,&st)) != 0)
		    return;
        buffer = (unsigned char *)malloc(st.st_size);
        if( !buffer )
            return;
	    if ((fread(buffer,st.st_size,1,fp)) != 1)
        {
            free(buffer);
            return;
        }
        fclose(fp);
        // write source to scratch location
        fp = fopen( scratch_path, "wb" );
        if( !fp )
        {
            free(buffer);
            return;
        }
	    if ((fwrite(buffer,st.st_size,1,fp)) != 1)
        {
            free(buffer);
            return;
        }
        fclose(fp);
        free(buffer);
    }
}
