/*-----------------------------------------------------------------------
 * File: SIGNATUREGRAPH.CPP
 *
 * Copyright (c) 1998-2000 Intel Corporation. All rights reserved.
 *-----------------------------------------------------------------------
 */

// SignatureGraph.cpp
//
// This file contains the class implementation for the CSignatureGraph
//  and CSignatureGraphNode classes. These classes implement a graph
//  of signers indexed by objects signed.
//
// Copyright(c) 1998, Intel Corporation.  All rights reserved.
//
//////////////////////////////////////////////////////////////////////
#ifdef CSSM_MEMTRACK
#undef new
#include <crtdbg.h>
static char SZ_FILE_NAME[] = __FILE__;
#define new new(_NORMAL_BLOCK, SZ_FILE_NAME, __LINE__)  /*need space*/
#endif /* ifdef VMS comment this out - CSSM_MEMTRACK */

#include "SignatureGraph.h"
#include <scMemTrack.h>

//=======================//
//                       //
//  CSignatureGraphNode  //
//                       //
//=======================//

//////////////////////////////////////////////////////////////////////
// CSignatureGraphNode::~CSignatureGraphNode()
//  destructs a CSignatureGraphNode object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CSignatureGraphNode::~CSignatureGraphNode()
{
    if(mSigners)
        delete mSigners;
    if(mObjectReference)
        delete mObjectReference;
}

//////////////////////////////////////////////////////////////////////
// CSignatureGraphNode::CSignatureGraphNode(const CSignatureGraphNode *node)
//  constructs a CSignatureGraphNode object which is a copy of the one 
//  passed 
//
// Parameters:
//  node(input)         
//      pointer to the CSignatureGraphNode object to copy
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CSignatureGraphNode::CSignatureGraphNode(const CSignatureGraphNode *node)
{
	mNext = node->mNext;
	if(node->mSigners)
        mSigners = new CSignerList(node->mSigners);
    else
        mSigners = NULL;
	mObjectReference = new CObjectRef(node->mObjectReference);
}

//////////////////////////////////////////////////////////////////////
// CSignatureGraphNode::CSignatureGraphNode(CObjectRef* ObjectRef,
//                                          CSignerList* Signers)
//  constructs a CSignatureGraphNode object with the reference 
//  field filled with the passed ref
//
// Parameters:
//  ObjectRef(input)
//      reference to be associated with this node
//  Signers(input)
//      pointer to the list of signers which signed the object
// Returns:
//  none
// Side effects:
//  ObjectRef and Signers objects are copied before construction
//////////////////////////////////////////////////////////////////////
CSignatureGraphNode::CSignatureGraphNode(CObjectRef* ObjectRef,
                                         CSignerList* Signers)
{
    mNext = NULL;
    if( Signers )
        mSigners = new CSignerList( Signers );
    else
        mSigners = NULL;
    if( ObjectRef )
        mObjectReference = new CObjectRef( ObjectRef );
    else 
        mObjectReference = NULL;

}

//////////////////////////////////////////////////////////////////////
// CSignatureGraphNode::AddSigner(CSignerListNode* SignerToAdd)
//  appends a signer to the end of the mSigners list in the node
//
// Parameters:
//  SignerToAdd(input)         
//      pointer to the CSignerListNode object to add to the list
// Returns:
//  none
// Side effects:
//  node will be copied prior to addition
//////////////////////////////////////////////////////////////////////
WF_RETURN CSignatureGraphNode::AddSigner(CSignerListNode* SignerToAdd)
{
    if(mSigners)
        return(mSigners->AddNode( SignerToAdd ));
    else
    {
        mSigners = new CSignerList( SignerToAdd );
        return WF_OK;
    }

}


//==================//
//                  //
//  CSignatureGraph //
//                  //
//==================//

//////////////////////////////////////////////////////////////////////
// CSignatureGraph::CSignatureGraph()
//  constructs an empty CSignatureGraph object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CSignatureGraph::CSignatureGraph()
{
    mHead = 0;
	mTail = 0;
	mNumberOfNodes = 0;
}

//////////////////////////////////////////////////////////////////////
// CSignatureGraph::~CSignatureGraph()
//  destructs a CSignatureGraph object
//
// Parameters:
//  none
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CSignatureGraph::~CSignatureGraph()
{
    CSignatureGraphNode *current_node = mHead, *next_node;

	// delete all nodes
	for( int i=0; i<mNumberOfNodes; i++ )
	{
		next_node = current_node->mNext;
		delete current_node;
		current_node = next_node;
	}
}

//////////////////////////////////////////////////////////////////////
// CSignatureGraph::CSignatureGraph(const CSignatureGraph *Graph)
//  constructs a CSignatureGraph object which is a copy of the one 
//  passed 
//
// Parameters:
//  Graph(input)         
//      pointer to the CSignatureGraph object to copy
// Returns:
//  none
// Side effects:
//  none
//////////////////////////////////////////////////////////////////////
CSignatureGraph::CSignatureGraph(const CSignatureGraph *Graph)
{
	CSignatureGraphNode *new_node, *old_node;
	
	// setup head ptr
	if(Graph->mHead)
    {
        mHead = new CSignatureGraphNode(Graph->mHead);
	    mTail = mHead;
	    new_node = mHead;
	    old_node = Graph->mHead;
        // copy all nodes
	    for( int i=1; i<Graph->mNumberOfNodes; i++ )
	    {
		    new_node->mNext = new CSignatureGraphNode( old_node->mNext );
		    mTail = new_node;
		    new_node = new_node->mNext;
		    old_node = old_node->mNext;
	    }
    }
    else
    {
        mHead = 0;
        mTail = 0;
    }
    // intialize number of nodes
    mNumberOfNodes = Graph->mNumberOfNodes;
}

//////////////////////////////////////////////////////////////////////
// CSignatureGraph::AddNode( CSignatureGraphNode *NodeToAdd )
//  appends NodeToAdd to the end of the list
//
// Parameters:
//  NodeToAdd(input)         
//      pointer to the CObjectListNode object to add to the list
// Returns:
//  WF_OK       Success
//  WF_FAIL     Failure
// Side effects:
//  node will be copied prior to addition
//////////////////////////////////////////////////////////////////////
WF_RETURN CSignatureGraph::AddNode( CSignatureGraphNode *NodeToAdd )
{
    if( mHead )
    {
        mTail->mNext = new CSignatureGraphNode( NodeToAdd );
	    mTail = mTail->mNext;
	    mTail->mNext = 0;
    }
    else
    {
        mHead = new CSignatureGraphNode( NodeToAdd );
        mTail = mHead;
    }
	mNumberOfNodes++;
	return WF_OK;
}
